<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * This class is an alternative to csv files.
 *
 * Each line represents a row of data.
 * Tags are used to mark up the fields.
 *
 * The use of tag delimiters inside field values is not allowed.
 * If a tag is found in a value it automatically gets removed.
 *
 * The structure is not as strict as you might know from CSV files,
 * which means the number and names of fields inside a row may
 * vary from one row to another.
 *
 * This provides you with more flexibility. For example:
 * You may decide not to use a field for future entries
 * or just add a new one and you don't have to change your previous
 * entries inside the file.
 *
 * This is also a good way around if you just don't know the
 * structure of future entries yet.
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 *
 */
class DatFile extends SecureFileStream
{

    /**
    * retrieve a line of data from the file
    *
    * @uses    $datfile->get(0)
    * @access  public
    * @param   int  $line_nr  line to retrieve
    * @return  array
    */
    function get($line_nr)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_INT, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        $data      = $this->content[$line_nr];
        $REF       = array();
        while (preg_match("/<(.*)>(.*)<\/.*>/Ui", $data, $matches))
        {
            $matches[2]       = preg_replace("/(\S{80})(\S*)/i", "\\1<wbr>\\2", $matches[2]);
            $REF[strtoupper($matches[1])] = $matches[2];
            $data             = preg_replace("/<".$matches[1].">.*<\/.*>/U", "", $data);
            $matches          = array();
        }
        assert('is_array($REF); /* unexcpected result */');
        return $REF;
    }

    /**
    * insert (append) an entry to the file
    *
    * @uses    $datfile->insert($new_entry)
    * @access  public
    * @param   array  $new_entry a ssociative array containing the new entry
    * @return  bool
    */
    function insert($REF)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_ARRAY, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        $txt       = "";
        assert('!isset($key); /* cannot redeclare variable $key */');
        assert('!isset($element); /* cannot redeclare variable $element */');
        foreach ($REF as $key => $element)
        {
            $key     = strtoupper($key);
            $element = strip_tags($element);
            $txt    .= "<$key>$element</$key>";
        }
        unset($key, $element);
        $txt       = nl2br($txt);
        $txt       = preg_replace("/\s/"," ",$txt);
        $txt      .= "\n";
        if (!$this->isEmpty()) {
            array_unshift($this->content,$txt);
        } else {
            $this->content[0] = $txt;
        }
        return true;
    }

    /**
    * remove an entry from the file
    *
    * @uses    $datfile->insert($new_entry)
    * @access  public
    * @param   array  $line_nr  line to retrieve
    * @return  bool
    */
    function remove($line_nr)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_INT, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        if (isset($this->content[$line_nr])) {
            unset($this->content[$line_nr]);
            return true;
        } else {
            return false;
        }
    }

    /**
    * return file content as string (if available)
    *
    * @uses    $datfile->toString()
    * @access  public
    * @return  string
    */
     function toString()
     {
        if (!$this->exists()) {
            $result = "Dat-file ".$this->getPath()." does not exist";
        } else if ($this->isEmpty()) {
            $result = "Dat-file ".$this->getPath()." is wether empty or not opened";
        } else {
            $result = implode("\n", $this->content);
        }
        assert('is_array($result); /* unexcpected result */');
        return $result;
     }


    /**
    * return the number of rows in the file
    *
    * @uses    $datfile->length()
    * @access  public
    * @return  int
    */
    function length()
    {
        $result = count($this->content);
        assert('is_int($result); /* unexcpected result */');
        return $result;
    }

}

?>