<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * decorator  database Creator
 *
 * This decorator class is intended to create SQL DDL from YANA Framework -
 * database structure files.
 *
 * For this task it provides functions which create specific
 * DDL for various DBS.
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 * @since       2.8.6
 */
class DbCreator extends Object
{

    /**#@+
     * @ignore
     * @access  private
     */

    /** @var StructureFile */ var $structure = null;

    /**#@-*/

    /**
     * create a new instance
     *
     * This class requires an object of class StructureFile
     * as input.
     *
     * Example of Usage:
     * <code>
     * $dbc = new DbCreator( new StructureFile('guestbook'));
     * $array_of_stmts = $dbc->createMySQL();
     * // This will output the result as text
     * print implode("\n",$array_of_stmts);
     * // This will send the statements to the database
     * $db = new DbStream('guestbook');
     * foreach ($array_of_stmts as $stmt)
     * {
     *    $db->query($stmt);
     * }
     * </code>
     *
     * @todo   sort table definitions by foreign keys, so referenced tables are created before the references (scheduled for: v3.0)
     *
     * @param  string  &$structure  a structure resource
     */
    function DbCreator(&$structure)
    {
        /* this is a simple work-around to use PHP5-like type-checking in PHP4 */
        assert('is_object($structure) && is_a($structure, "StructureFile");  // Wrong argument type for argument 1. Instance of StructureFile expected.');
        $this->structure =& $structure;
        $this->structure->read();
    }

    /**
     * create SQL for MySQL
     *
     * Returns a numeric array of SQL statements.
     * Each element is a single statement.
     * If you want to send the result to a SQL file
     * you should "implode()" the array to a string.
     *
     * @access  public
     * @return  array
     */
    function createMySQL()
    {
        /* this is the result var that will be returned when finished */
        $SQL = array();

        /* this is for statements, which have to come last */
        $sql_last = array();

        $list_of_tables = $this->structure->getTables();

        /*
         *  Create Table
         */
        assert('!isset($table); /* cannot redeclare variable $table */');
        foreach ($list_of_tables as $table)
        {
            assert('is_string($table) && !empty($table);');
            $is_innodb = false;
            $stmt = "CREATE TABLE IF NOT EXISTS `".YANA_DATABASE_PREFIX."${table}` (\n";

            $list_of_columns = $this->structure->getColumns($table);
            $list_of_unique_cols = $this->structure->getUniqueConstraints($table);
            $list_of_foreign_keys = $this->structure->getForeignKeys($table);
            $list_of_indexes = $this->structure->getIndexes($table);

            /*
             *  Create Column
             */
            assert('!isset($column); /* cannot redeclare variable $column */');
            foreach ($list_of_columns as $column)
            {
                $type = $this->structure->getType($table,$column);
                $length = $this->structure->getLength($table,$column);
                $default = $this->structure->getDefault($table,$column);
                $comment = $this->structure->getDescription($table,$column);
                $stmt .= "\t`${column}` ";

                /*
                 *  Add Type
                 */
                if ($this->structure->isForeignKey($table,$column)) {
                    $foreign_table = $this->structure->getTableByForeignKey($table,$column);
                    $foreign_pk    = $this->structure->getPrimaryKey($foreign_table);
                    $type = $this->structure->getType($foreign_table,$foreign_pk);
                    $length = $this->structure->getLength($foreign_table,$foreign_pk);
                }
                switch (strtolower($type))
                {
                    case "time":
                        $type   = "INT";
                        $length = 11;
                    break;

                    case "image": case "file":
                        $type   = "VARCHAR";
                        $length = 128;
                    break;

                    case "ip":
                        $type   = "VARCHAR";
                        $length = 15;
                    break;

                    case "profile":
                        $type   = "VARCHAR";
                        $length = 128;
                    break;

                    case "mail": case "url": case "uri": case "string":
                        if (empty($length)) {
                            $type = 'TEXT';
                        } else {
                            $type = 'VARCHAR';
                        }
                    break;

                    case "select": case "array": case "text":
                        $type  = 'TEXT';
                        $length = 0;
                    break;

                    case "int": case "integer":
                        $type = "INT";
                    break;

                    case "float": case "double":
                    break;

                    case "boolean": case "bool":
                        $type = "BOOL";
                    break;

                    default:
                        if ($length == 0) {
                            $type = 'TEXT';
                        } else {
                            $type = 'VARCHAR';
                        }
                    break;

                } /* end switch */
                $stmt .= "$type";

                /*
                 *  Add Length
                 */
                if (is_int($length) && $length > 0 && $type !== 'text') {
                    $stmt .= "(${length})";
                }

                /*
                 *  Decide wether this column is nullable
                 */
                if ($this->structure->isNullable($table,$column) === true) {
                    $stmt .= " NULL";
                } else {
                    $stmt .= " NOT NULL";
                }

                /*
                 *  Add Default-Value
                 */
                if (!empty($default) && is_scalar($default)) {
                    $stmt .= " DEFAULT '${default}'";
                }

                /*
                 *  Add Autonumber
                 */
                if ($this->structure->isAutonumber($table,$column) === true) {
                    $stmt .= " AUTO_INCREMENT";
                }

                /*
                 *  Add Comment
                 */
                if (!empty($comment)) {
                    $stmt .= " COMMENT '".addcslashes($comment,"'")."'";
                }

                $stmt .= ",\n";

            } /* end foreach column */
            unset($column);


            /*
             *  Create Primary Key
             */
            $stmt .= "\tPRIMARY KEY  (`".$this->structure->getPrimaryKey($table)."`)";

            if (!empty($list_of_unique_cols)) {
                foreach ($list_of_unique_cols as $i => $unique_col)
                {
                    $list_of_unique_cols[$i] = "`$unique_col`";
                }
                $stmt .= ",\n\tUNIQUE ${table}_uq (".implode(', ',$list_of_unique_cols).")";
            }

            $stmt .= "\n";

            /*
             *  Decide for InnoDB or MyISAM
             */

            if (!empty($list_of_foreign_keys)) {
                $is_innodb = true;
            }
            if (!empty($list_of_indexes)) {
                $is_innodb = true;
            }
            if ($is_innodb === false) {
                $stmt .= ") TYPE=MyISAM;";
            } else {
                $stmt .= ") TYPE=InnoDB;";
            }
            assert('is_array($SQL);');
            $SQL[] = $stmt;
            unset($stmt);

            /*
             *  Create Indexes
             */
            if (!empty($list_of_indexes)) {
                assert('!isset($index); /* cannot redeclare variable $index */');
                assert('!isset($i); /* cannot redeclare variable $i */');
                $i = 0;
                foreach ($list_of_indexes as $index)
                {
                    $indexName = "${table}_${i}_idx";
                    assert('is_array($SQL);');
                    $SQL[] = "ALTER TABLE `".YANA_DATABASE_PREFIX."${table}` ADD INDEX `${indexName}` (`${index}`);";
                    $i++;
                }
                unset($index);
            }

            /*
             *  Create Foreign Key Constraints
             */
            if (!empty($list_of_foreign_keys)) {
                foreach ($list_of_foreign_keys as $foreign_key => $foreign_table)
                {
                    $foreign_key = strtolower($foreign_key);
                    $foreign_pk = $this->structure->getPrimaryKey($foreign_table);
                    $constraintName = "${table}_fk_${foreign_table}";
                    if (strlen($constraintName) > 18) {
                        $constraintName = "${foreign_key}_fk";
                    }
                    if (strlen($constraintName) > 18) {
                        $constraintName = substr($constraintName, 0, 18);
                    }
                    assert('is_array($sql_last);');
                    $sql_last[] = "ALTER TABLE `".YANA_DATABASE_PREFIX."${table}` ADD CONSTRAINT `$constraintName` FOREIGN KEY (`${foreign_key}`) REFERENCES `".YANA_DATABASE_PREFIX."${foreign_table}` (`${foreign_pk}`);";
                }
            }

        } /* end foreach table */
        unset($table);

        assert('is_array($sql_last);');
        assert('!isset($stmt); /* cannot redeclare variable $stmt */');
        foreach ($sql_last as $stmt)
        {
            assert('is_array($SQL);');
            $SQL[] = $stmt;
        }
        unset($stmt);

        return $SQL;
    }

    /**
     * create SQL for PostgreSQL
     *
     * Returns a numeric array of SQL statements.
     * Each element is a single statement.
     * If you want to send the result to a SQL file
     * you should "implode()" the array to a string.
     *
     * @access  public
     * @return  array
     */
    function createPostgreSQL()
    {
        /* this is the result var that will be returned when finished */
        $SQL = array();

        /* this is for statements, which have to come last */
        $sql_last = array();

        $list_of_tables = $this->structure->getTables();

        /*
         *  Create Table
         */
        assert('!isset($table); /* cannot redeclare variable $table */');
        foreach ($list_of_tables as $table)
        {
            assert('is_string($table) && !empty($table);');
            $stmt = "CREATE TABLE \"".YANA_DATABASE_PREFIX."${table}\" (\n";

            $list_of_columns      = $this->structure->getColumns($table);
            $list_of_indexes      = $this->structure->getIndexes($table);

            /*
             *  Create Column
             */
            assert('!isset($i); /* cannot redeclare variable $i */');
            for ($i = 0; $i < count($list_of_columns); $i++)
            {
                $column = $list_of_columns[$i];

                $type = $this->structure->getType($table,$column);
                $length = $this->structure->getLength($table,$column);
                $default = $this->structure->getDefault($table,$column);
                $comment = $this->structure->getDescription($table,$column);
                $stmt .= "\t\"${column}\" ";

                /*
                 *  Add Type
                 */
                if ($this->structure->isForeignKey($table,$column)) {
                    $foreign_table = $this->structure->getTableByForeignKey($table,$column);
                    $foreign_pk    = $this->structure->getPrimaryKey($foreign_table);
                    $type = $this->structure->getType($foreign_table,$foreign_pk);
                    $length = $this->structure->getLength($foreign_table,$foreign_pk);
                }
                switch (strtolower($type))
                {
                    case "time":
                        $type   = "integer";
                        $length = 0;
                    break;

                    case "image": case "file":
                        $type   = "varchar";
                        $length = 128;
                    break;

                    case "ip":
                        $type   = "varchar";
                        $length = 15;
                    break;

                    case "profile":
                        $type   = "varchar";
                        $length = 128;
                    break;

                    case "mail": case "url": case "uri": case "string":
                        if (empty($length)) {
                            $type = 'text';
                        } else {
                            $type = 'varchar';
                        }
                    break;

                    case "select": case "array": case "text":
                        $type  = 'text';
                        $length = 0;
                    break;

                    case "int": case "integer":
                        $type = "integer";
                        $length = 0;
                    break;

                    case "float": case "double":
                        $type = "double precision";
                        $length = 0;
                    break;

                    case "boolean": case "bool":
                        $type = "boolean";
                    break;

                    default:
                        if ($length == 0) {
                            $type = 'text';
                        } else {
                            $type = 'varchar';
                        }
                    break;

                } /* end switch */
                $stmt .= "$type";

                /*  PostgreSQL:
                 *  Add Length
                 */
                if (is_int($length) && $length > 0 && $type !== 'text') {
                    $stmt .= "(${length})";
                }

                /*  PostgreSQL:
                 *  Decide wether this column is nullable
                 */
                if ($this->structure->isNullable($table,$column) === false) {
                    $stmt .= " NOT NULL";
                }

                /*  PostgreSQL:
                 *  Add Autonumber
                 */
                if ($this->structure->isAutonumber($table,$column) === true) {
                    array_unshift($SQL,"CREATE SEQUENCE ${table}_${column}_sq INCREMENT 1;");
                    $stmt .= " DEFAULT nextval('${table}_${column}_sq')";
                /*  PostgreSQL:
                 *  Add Default-Value
                 */
                } else if (!empty($default) && is_scalar($default)) {
                    $stmt .= " DEFAULT '${default}'";
                }

                /*  PostgreSQL:
                 *  Add Unique Constraint
                 */
                if ($this->structure->isUnique($table,$column) === true) {
                    $stmt .= " UNIQUE";
                }

                /*  PostgreSQL:
                 *  Add Primary Key
                 */
                if ($this->structure->isPrimaryKey($table,$column) === true) {
                    $stmt .= " PRIMARY KEY";
                }

                /*  PostgreSQL:
                 *  Create Foreign Key Constraints
                 */
                if ($this->structure->isForeignKey($table,$column) === true) {
                    $ftable = $this->structure->getTableByForeignKey($table,$column);
                    assert('is_array($sql_last);');
                    $sql_last[] = "ALTER TABLE \"".YANA_DATABASE_PREFIX."${table}\" ADD FOREIGN KEY (\"${column}\") REFERENCES \"".YANA_DATABASE_PREFIX.$this->structure->getTableByForeignKey($table,$column)."\";";
                }

                /*  PostgreSQL:
                 *  Add Comment
                 */
                if (!empty($comment)) {
                    assert('is_array($sql_last);');
                    $sql_last[] = "COMMENT ON COLUMN \"".YANA_DATABASE_PREFIX."${table}\".\"${column}\" IS '".addcslashes($comment,"'")."';";
                }

                if ($i<count($list_of_columns) - 1) {
                    $stmt .= ",";
                }
                $stmt .= "\n";

            } /* end foreach column */
            unset($i);

            assert('is_array($SQL);');
            $SQL[] = $stmt.");";
            unset($stmt);

            /*  PostgreSQL:
             *  Create Indexes
             */
            if (!empty($list_of_indexes)) {
                assert('!isset($index); /* cannot redeclare variable $index */');
                foreach ($list_of_indexes as $index)
                {
                    assert('is_array($SQL);');
                    $SQL[] = "CREATE INDEX ${table}_${index}_idx ON \"".YANA_DATABASE_PREFIX."${table}\" (\"${index}\");";
                }
                unset($index);
            }

        } /* end foreach table */
        unset($table);

        assert('is_array($sql_last);');
        assert('!isset($stmt); /* cannot redeclare variable $stmt */');
        foreach ($sql_last as $stmt)
        {
            assert('is_array($SQL);');
            $SQL[] = $stmt;
        }
        unset($stmt);

        return $SQL;
    }

    /**
     * create SQL for MS SQL Server
     *
     * Returns a numeric array of SQL statements.
     * Each element is a single statement.
     * If you want to send the result to a SQL file
     * you should "implode()" the array to a string.
     *
     * @access  public
     * @return  array
     */
    function createMSSQL()
    {
        /* this is the result var that will be returned when finished */
        $SQL = array();

        /* this is for statements, which have to come last */
        $sql_last = array();

        $list_of_tables = $this->structure->getTables();

        /*
         *  Create Table
         */
        assert('!isset($table); /* cannot redeclare variable $table */');
        foreach ($list_of_tables as $table)
        {
            assert('is_string($table) && !empty($table);');
            $stmt = "CREATE TABLE [dbo].[".YANA_DATABASE_PREFIX."${table}] (\n";

            $list_of_columns = $this->structure->getColumns($table);
            $list_of_indexes = $this->structure->getIndexes($table);

            /*
             *  Create Column
             */
            assert('!isset($i); /* cannot redeclare variable $i */');
            for ($i = 0; $i < count($list_of_columns); $i++)
            {
                $column = $list_of_columns[$i];

                $type = $this->structure->getType($table,$column);
                $length = $this->structure->getLength($table,$column);
                $default = $this->structure->getDefault($table,$column);
                $comment = $this->structure->getDescription($table,$column);
                $stmt .= "\t\"${column}\" ";

                /*  MS-SQL:
                 *  Add Type
                 */
                if ($this->structure->isForeignKey($table,$column)) {
                    $foreign_table = $this->structure->getTableByForeignKey($table,$column);
                    $foreign_pk    = $this->structure->getPrimaryKey($foreign_table);
                    $type = $this->structure->getType($foreign_table,$foreign_pk);
                    $length = $this->structure->getLength($foreign_table,$foreign_pk);
                }
                switch (strtolower($type))
                {
                    case "time":
                        $type   = "integer";
                        $length = 0;
                    break;

                    case "image": case "file":
                        $type   = "varchar";
                        $length = 128;
                    break;

                    case "ip":
                        $type   = "varchar";
                        $length = 15;
                    break;

                    case "profile":
                        $type   = "varchar";
                        $length = 128;
                    break;

                    case "mail": case "url": case "uri": case "string":
                        if (empty($length)) {
                            $type = 'text';
                        } else {
                            $type = 'varchar';
                        }
                    break;

                    case "select": case "array": case "text":
                        $type  = 'text';
                        $length = 0;
                    break;

                    case "int": case "integer":
                        $type = "integer";
                        $length = 0;
                    break;

                    case "float": case "double":
                        $type = "float";
                        $length = 0;
                    break;

                    case "boolean": case "bool":
                        $type = "bit";
                    break;

                    default:
                        if ($length == 0) {
                            $type = 'text';
                        } else {
                            $type = 'varchar';
                        }
                    break;

                } /* end switch */
                $stmt .= "$type";

                /*  MS-SQL:
                 *  Add Length
                 */
                if (is_int($length) && $length > 0 && $type !== 'text') {
                    $stmt .= "(${length})";
                }

                /*  MS-SQL:
                 *  Decide wether this column is nullable
                 */
                if ($this->structure->isNullable($table,$column) === false) {
                    $stmt .= " NOT NULL";
                } else {
                    $stmt .= " NULL";
                }

                /*  MS-SQL:
                 *  Add Autonumber
                 */
                if ($this->structure->isAutonumber($table,$column) === true) {
                    $stmt .= " IDENTITY (1,1)";

                /*  MS-SQL:
                 *  Add Default-Value
                 */
                } else if (!empty($default) && is_scalar($default)) {
                    $stmt .= " DEFAULT '${default}'";
                }

                /*  MS-SQL:
                 *  Add Unique Constraint
                 */
                if ($this->structure->isUnique($table,$column) === true) {
                    assert('is_array($sql_last);');
                    $sql_last[] = "ALTER TABLE [dbo].[".YANA_DATABASE_PREFIX."${table}] ADD CONSTRAINT ${table}_${column}_uq UNIQUE (${column});";
                }

                /*  MS-SQL:
                 *  Add Primary Key
                 */
                if ($this->structure->isPrimaryKey($table,$column) === true) {
                    $stmt .= " CONSTRAINT ${table}_pk PRIMARY KEY";
                }

                /*  MS-SQL:
                 *  Create Foreign Key Constraints
                 */
                if ($this->structure->isForeignKey($table,$column) === true) {
                    $ftable = $this->structure->getTableByForeignKey($table,$column);
                    assert('is_array($sql_last);');
                    $sql_last[] = "ALTER TABLE [dbo].[".YANA_DATABASE_PREFIX."${table}] ADD CONSTRAINT ${table}_${ftable}_fk FOREIGN KEY (${column}) REFERENCES [dbo].[".YANA_DATABASE_PREFIX."${ftable}]";
                }

                /*  MS-SQL:
                 *  Add Comment
                 */
                if (!empty($comment)) {
                    /* intentionally left blank */
                }

                if ($i<count($list_of_columns) - 1) {
                    $stmt .= ",";
                }
                $stmt .= "\n";

            } /* end foreach column */
            unset($i);

            assert('is_array($SQL);');
            $SQL[] = $stmt.");";
            unset($stmt);

            /*  MS-SQL:
             *  Create Indexes
             */
            if (!empty($list_of_indexes)) {
                assert('!isset($index); /* cannot redeclare variable $index */');
                foreach ($list_of_indexes as $index)
                {
                    assert('is_array($SQL);');
                    $SQL[] = "CREATE NONCLUSTERED INDEX ${table}_${index}_idx ON \"".YANA_DATABASE_PREFIX."${table}\" (\"${index}\");";
                }
                unset($index);
            }

        } /* end foreach table */
        unset($table);

        assert('is_array($sql_last);');
        assert('!isset($stmt); /* cannot redeclare variable $stmt */');
        foreach ($sql_last as $stmt)
        {
            assert('is_array($SQL);');
            $SQL[] = $stmt;
        }
        unset($stmt);

        return $SQL;
    }

    /**
     * create SQL for MS Access
     *
     * Same as DbCreator::createMSSQL()
     *
     * @access  public
     * @return  array
     * @see     DbCreator::createMSSQL()
     */
    function createMSAccess()
    {
        return $this->createMSSQL();
    }

    /**
     * create SQL for IBM DB2
     *
     * Returns a numeric array of SQL statements.
     * Each element is a single statement.
     * If you want to send the result to a SQL file
     * you should "implode()" the array to a string.
     *
     * @access  public
     * @return  array
     */
    function createDB2()
    {
        /* this is the result var that will be returned when finished */
        $SQL = array();

        /* this is for statements, which have to come last */
        $sql_last = array();

        $list_of_tables = $this->structure->getTables();

        global $YANA;
        $file = $YANA->registry->get('reserved_sql_keywords');
        $sql_keywords = file($file->getPath());

        /*
         *  Create Table
         */
        assert('!isset($table); /* cannot redeclare variable $table */');
        foreach ($list_of_tables as $table)
        {
            assert('is_string($table) && !empty($table);');
            $tableName = YANA_DATABASE_PREFIX.$table;
            if (qSearchArray($sql_keywords, $tableName) !== false) {
                $tableName = "\"${tableName}\"";
            }

            $stmt = "CREATE TABLE ${tableName}(\n";

            $list_of_columns = $this->structure->getColumns($table);
            $list_of_foreign_keys = $this->structure->getForeignKeys($table);
            $list_of_indexes = $this->structure->getIndexes($table);

            /*
             *  Create Column
             */
            assert('!isset($i); /* cannot redeclare variable $i */');
            for ($i = 0; $i < count($list_of_columns); $i++)
            {
                $column = $list_of_columns[$i];

                if (qSearchArray($sql_keywords, $column) !== false) {
                    $columnName = "\"${column}\"";
                } else {
                    $columnName = $column;
                }

                $type = $this->structure->getType($table,$column);
                $length = $this->structure->getLength($table,$column);
                $default = $this->structure->getDefault($table,$column);
                $comment = $this->structure->getDescription($table,$column);
                $stmt .= "\t${columnName} ";

                /*  DB2:
                 *  Add Type
                 */
                if ($this->structure->isForeignKey($table,$column)) {
                    $foreign_table = $this->structure->getTableByForeignKey($table,$column);
                    $foreign_pk    = $this->structure->getPrimaryKey($foreign_table);
                    $type = $this->structure->getType($foreign_table,$foreign_pk);
                    $length = $this->structure->getLength($foreign_table,$foreign_pk);
                }
                switch (strtolower($type))
                {
                    case "time":
                        $type   = "INTEGER";
                        $length = 0;
                    break;

                    case "image": case "file":
                        $type   = "VARCHAR";
                        $length = 128;
                    break;

                    case "ip":
                        $type   = "VARCHAR";
                        $length = 15;
                    break;

                    case "profile":
                        $type   = "VARCHAR";
                        $length = 128;
                    break;

                    case "mail": case "url": case "uri": case "string":
                        if (empty($length)) {
                            $type = 'LONG VARCHAR';
                        } else {
                            $type = 'VARCHAR';
                        }
                    break;

                    case "select": case "array": case "text":
                        $type  = 'LONG VARCHAR';
                        $length = 0;
                    break;

                    case "int": case "integer":
                        $type = "INTEGER";
                        $length = 0;
                    break;

                    case "float": case "double":
                        $length = 0;
                    break;

                    case "boolean": case "bool":
                        $type = "SMALLINT";
                        $length = 1;
                    break;

                    default:
                        if ($length == 0) {
                            $type = 'CLOB';
                        } else {
                            $type = 'VARCHAR';
                        }
                    break;

                } /* end switch */
                $stmt .= "$type";

                /*  DB2:
                 *  Add Length
                 */
                if (is_int($length) && $length > 0 && $type !== 'text') {
                    $stmt .= "(${length})";
                }

                /*  DB2:
                 *  Decide wether this column is nullable
                 */
                if ($this->structure->isNullable($table,$column) === false || $this->structure->isAutonumber($table,$column) === true) {
                    $stmt .= " NOT NULL";
                }

                /*  DB2:
                 *  Add Autonumber
                 */
                if ($this->structure->isAutonumber($table,$column) === true) {
                    $stmt .= " GENERATED ALWAYS AS IDENTITY (START WITH 1, INCREMENT BY 1, NO CACHE)";

                /*  DB2:
                 *  Add Default-Value
                 */
                } else if (!empty($default) && is_scalar($default)) {
                    $stmt .= " DEFAULT '${default}'";
                }

                /*  DB2:
                 *  Add Primary Key
                 */
                if ($this->structure->isPrimaryKey($table,$column) === true) {
                    $constraintName = "${table}_pk";
                    if (strlen($constraintName) > 18) {
                        $constraintName = substr($constraintName, strlen($constraintName)-18, strlen($constraintName));
                    }
                    $stmt .= " CONSTRAINT ${constraintName} PRIMARY KEY";
                }

                /*  DB2:
                 *  Add Primary Key
                 */
                if ($this->structure->isUnique($table,$column) === true) {
                    $constraintName = "${table}_${column}_uq";
                    if (strlen($constraintName) > 18) {
                        $constraintName = "${column}_uq";
                    }
                    if (strlen($constraintName) > 18) {
                        $constraintName = substr($constraintName, strlen($constraintName)-18, strlen($constraintName));
                    }
                    $stmt .= " CONSTRAINT ${constraintName} UNIQUE";
                }

                /*  DB2:
                 *  Add Comment
                 */
                if (!empty($comment)) {
                    assert('is_array($sql_last);');
                    $sql_last[] = "COMMENT ON ${tableName} (${columnName} IS '".addcslashes($comment,"'")."');";
                }

                if ($i<count($list_of_columns) - 1) {
                    $stmt .= ",";
                }
                $stmt .= "\n";

            } /* end foreach column */
            unset($i);

            assert('is_array($SQL);');
            $SQL[] = $stmt.");";
            unset($stmt);

            /*  DB2:
             *  Create Indexes
             */
            if (!empty($list_of_indexes)) {
                assert('!isset($index); /* cannot redeclare variable $index */');
                foreach ($list_of_indexes as $index)
                {
                    if (qSearchArray($sql_keywords, $index) !== false) {
                        $indexName = "\"${index}\"";
                    } else {
                        $indexName = $index;
                    }
                    assert('is_array($SQL);');
                    $SQL[] = "CREATE INDEX ${table}_${index}_idx ON ${tableName} (${indexName});";
                } /* end foreach index */
                unset($index);
            }

            /*  DB2:
             *  Create Foreign Key Constraints
             */
            if (!empty($list_of_foreign_keys)) {
                foreach ($list_of_foreign_keys as $foreign_key => $foreign_table)
                {
                    $foreign_key = strtolower($foreign_key);
                    $foreign_pk = $this->structure->getPrimaryKey($foreign_table);
                    $foreign_table_name = YANA_DATABASE_PREFIX.$foreign_table;
                    if (qSearchArray($sql_keywords, $foreign_table_name) !== false) {
                        $foreign_table_name = "\"${foreign_table_name}\"";
                    } else {
                        $foreign_table_name = $foreign_table_name;
                    }
                    if (qSearchArray($sql_keywords, $foreign_key) !== false) {
                        $foreign_key_name = "\"${foreign_key}\"";
                    } else {
                        $foreign_key_name = $foreign_key;
                    }

                    $constraintName = "${table}_fk_${foreign_table}";
                    if (strlen($constraintName) > 18) {
                        $constraintName = "${foreign_key_name}_fk";
                    }
                    if (strlen($constraintName) > 18) {
                        $constraintName = substr($constraintName, 0, 18);
                    }
                    assert('is_array($sql_last);');
                    $sql_last[] = "ALTER TABLE ${tableName} ADD CONSTRAINT ${constraintName} FOREIGN KEY (${foreign_key_name}) REFERENCES ${foreign_table_name};";
                }
            }

        } /* end foreach table */
        unset($table);

        assert('is_array($sql_last);');
        assert('!isset($stmt); /* cannot redeclare variable $stmt */');
        foreach ($sql_last as $stmt)
        {
            assert('is_array($SQL);');
            $SQL[] = $stmt;
        }
        unset($stmt);

        return $SQL;

    }

    /**
     * create SQL for Oracle
     *
     * Returns a numeric array of SQL statements.
     * Each element is a single statement.
     * If you want to send the result to a SQL file
     * you should "implode()" the array to a string.
     *
     * @access  public
     * @return  array
     */
    function createOracleDB()
    {
        /* this is the result var that will be returned when finished */
        $SQL = array();

        /* this is for statements, which have to come last */
        $sql_last = array();

        $list_of_tables = $this->structure->getTables();

        /*
         *  Create Table
         */
        assert('!isset($table); /* cannot redeclare variable $table */');
        foreach ($list_of_tables as $table)
        {
            assert('is_string($table) && !empty($table);');
            $stmt = "CREATE TABLE \"".YANA_DATABASE_PREFIX."${table}\" (\n";

            $list_of_columns      = $this->structure->getColumns($table);
            $list_of_indexes      = $this->structure->getIndexes($table);
            $list_of_foreign_keys = $this->structure->getForeignKeys($table);
            $seq_cnt              = 0;

            /*
             *  Create Column
             */
            assert('!isset($i); /* cannot redeclare variable $i */');
            for ($i = 0; $i < count($list_of_columns); $i++)
            {
                $column = $list_of_columns[$i];

                $type = $this->structure->getType($table,$column);
                $length = $this->structure->getLength($table,$column);
                $default = $this->structure->getDefault($table,$column);
                $comment = $this->structure->getDescription($table,$column);
                $stmt .= "\t\"${column}\" ";

                /*  Oracle:
                 *  Add Type + Length
                 */
                switch (strtolower($type))
                {
                    case "time":
                        $type   = 'INTEGER';
                    break;

                    case "image": case "file":
                        $type   = "VARCHAR2(128 CHAR)";
                    break;

                    case "ip":
                        $type   = 'VARCHAR2(15 CHAR)';
                    break;

                    case "profile":
                        $type   = "VARCHAR2(128 CHAR)";
                    break;

                    case "mail": case "url": case "uri": case "string":
                        if (empty($length)) {
                            $type = 'CLOB';
                        } else {
                            $type = 'VARCHAR2';
                            if (!empty($length)) {
                                $type .= "(${length} CHAR)";
                            }
                        }
                    break;

                    case "select": case "array": case "text":
                        $type  = 'CLOB';
                    break;

                    case "int": case "integer":
                        $type = 'INTEGER';
                    break;

                    case "float": case "double":
                        $type  = 'NUMBER';
                    break;

                    case "boolean": case "bool":
                        $type = 'NUMBER(1)';
                    break;

                    default:
                        if ($length == 0) {
                            $type = 'CLOB';
                        } else {
                            $type = 'VARCHAR2';
                            if (!empty($length)) {
                                $type .= "(${length} CHAR)";
                            }
                        }
                    break;

                } /* end switch */
                $stmt .= "$type";

                /*  Oracle:
                 *  Decide wether this column is nullable
                 */
                if ($this->structure->isNullable($table,$column) === false || $this->structure->isAutonumber($table,$column) === true) {
                    $stmt .= " NOT NULL";
                }

                /*  Oracle:
                 *  Add Autonumber
                 */
                if ($this->structure->isAutonumber($table,$column) === true) {
                    $sql_last[] = "CREATE SEQUENCE \"${table}_sq${seq_cnt}\";";
                    assert('!isset($tmp); /* cannot redeclare variable $tmp */');
                    $tmp  = "CREATE OR REPLACE TRIGGER \"${table}_${column}_inc\"\n";
                    $tmp .= "\tBEFORE INSERT ON \"${table}\"\n";
                    $tmp .= "\tFOR EACH ROW BEGIN\n";
                    $tmp .= "\tSELECT \"${table}_${column}_sq\".nextval INTO :new.\"${column}\" FROM dual;\n";
                    $tmp .= "\tEND;\n;";
                    $sql_last[] = $tmp;
                    $seq_cnt++;
                    unset($tmp);

                /*  Oracle:
                 *  Add Default-Value
                 */
                } else if (!empty($default) && is_scalar($default)) {
                    $stmt .= " DEFAULT '${default}'";
                }

                /*  Oracle:
                 *  Add Unique Constraint
                 */
                if ($this->structure->isUnique($table,$column) === true) {
                    assert('is_array($sql_last);');
                    $sql_last[] = "ALTER TABLE \"".YANA_DATABASE_PREFIX."${table}\" ADD CONSTRAINT \"${table}_${column}_uq\" UNIQUE (\"${column}\");";
                }

                /*  Oracle:
                 *  Add Comment
                 */
                if (!empty($comment)) {
                    assert('is_array($sql_last);');
                    $sql_last[] = "COMMENT ON COLUMN \"".YANA_DATABASE_PREFIX."${table}\".\"${column}\" IS '".addcslashes($comment,"'")."';";
                }

                $stmt .= ",\n";

            } /* end foreach column */
            unset($i,$seq_cnt);

            /*  Oracle:
             *  Add Primary Key
             */
            $stmt .= "\tCONSTRAINT \"${table}_pk\" PRIMARY KEY (\"".$this->structure->getPrimaryKey($table)."\")\n";

            assert('is_array($SQL);');
            $SQL[] = $stmt.");";
            unset($stmt);

            /*  Oracle:
             *  Create Indexes
             */
            if (!empty($list_of_indexes)) {
                assert('!isset($index); /* cannot redeclare variable $index */');
                $i = 0;
                foreach ($list_of_indexes as $index)
                {
                    assert('is_array($SQL);');
                    $SQL[] = "CREATE INDEX \"${table}_idx${i}\" ON \"".YANA_DATABASE_PREFIX."${table}\" (\"${index}\");";
                    $i++;
                }
                unset($i, $index);
            }

            /*
             *  Create Foreign Key Constraints
             */
            if (!empty($list_of_foreign_keys)) {
                $i = 0;
                foreach ($list_of_foreign_keys as $foreign_key => $foreign_table)
                {
                    $foreign_key = strtolower($foreign_key);
                    $foreign_pk = $this->structure->getPrimaryKey($foreign_table);
                    $constraintName = "${table}_fk${i}";
                    assert('is_array($sql_last);');
                    $sql_last[] = "ALTER TABLE \"".YANA_DATABASE_PREFIX."${table}\" ADD CONSTRAINT \"$constraintName\" FOREIGN KEY (\"${foreign_key}\") REFERENCES \"".YANA_DATABASE_PREFIX."${foreign_table}\" (\"${foreign_pk}\");";
                    $i++;
                }
                unset($i,$foreign_key,$foreign_table);
            }

        } /* end foreach table */
        unset($table);

        assert('is_array($sql_last);');
        assert('!isset($stmt); /* cannot redeclare variable $stmt */');
        foreach ($sql_last as $stmt)
        {
            assert('is_array($SQL);');
            $SQL[] = $stmt;
        }
        unset($stmt);

        return $SQL;
    }

    /**
     * self test
     *
     * This function exists only for debugging purposes.
     * Do not uncomment in productive environments.
     *
     * @access  public
     * @static
     * @param   string  $dbName
     *
     * @ignore
     * /
    function test($dbName)
    {
        global $YANA;
        $db = $YANA->connect($dbName);
        $dbc = new DbCreator($db->structure);
        $list = get_class_methods($dbc);
        foreach ($list as $i => $method)
        {
            if (!preg_match('/^create/i',$method)) {
                unset($list[$i]);
            }
        }
        print "<h1>self test</h1>\n";
        print "<ol>\n";
        print '<li><a href="#idx1">Object</a></li>';
        $i = 0;
        foreach ($list as $name)
        {
            print "<li><a href=\"#idx${i}\">${name}</a></li>\n";
            $i++;
        }
        print "</ol>\n";
        print "<h2 id=\"idx1\">Object</h2>\n";
        $result = htmlentities(print_r($dbc,true));
        print '<pre>'.$result.'</pre>';
        $i = 0;
        foreach ($list as $name)
        {
            print "<h2 id=\"idx${i}\">${name}</h2>\n";
            $result = $dbc->$name();
            print '<pre>'.print_r(implode("\n",$result),true).'</pre>';
            $i++;
        }
        exit;
    } /* end of function */

}

?>
