<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * datatype mapping tool
 *
 * This class handles transformation of semantic data types to
 * syntactic data types and vice versa.
 *
 * This class is part of the data access layer.
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 * @ignore
 */
class DbDatatype extends Object
{
    /**#@+
     * @access  private
     * @ignore
     */

    /** @var StructureFile */ var $structure = null;
    /** @var DbTrigger     */ var $exec = null;
    /** @var string        */ var $dbDir = null;

    /**#@-*/

    /**
     * constructor
     *
     * Create a new instance of this class.
     *
     * @param  StructureFile  &$structure
     */
    function DbDatatype(&$structure)
    {
        assert('is_object($structure) && is_a($structure, "StructureFile"); // Wrong argument type for argument 1. Instance of StructureFile expected');
        $this->structure =& $structure;
        $this->exec = new DbTrigger($structure);
    }

    /**
     * prepare data for input to database
     *
     * @access  public
     * @param   mixed   &$value
     * @param   string  $table
     * @param   string  $column
     * @param   string  $row
     * @param   object  $db
     * @return  bool
     */
    function encode(&$value, $table, $column, $row = null, $db = null)
    {
        assert('is_string($table); // Wrong argument type for argument 2. String expected');
        assert('is_string($column); // Wrong argument type for argument 3. String expected');
        assert('is_null($row) || is_scalar($row); // Wrong argument type for argument 4. Scalar value expected');
        assert('is_null($db) || is_object($db); // Wrong argument type for argument 5. Object expected');

        $table  = (string) $table;
        $column = (string) $column;
        $row    = (string) $row;
        $row    = strtolower($row);

        $type = $this->structure->getType($table, $column);
        switch ($type)
        {
            case 'text':
                $value = $this->structure->untaintInput($table, $column, $value, YANA_ESCAPE_USERTEXT);
            break;

            case 'string': case 'select': case 'ip': case 'url': case 'mail': case 'profile':
                $value = $this->structure->untaintInput($table, $column, $value, YANA_ESCAPE_LINEBREAK);
            break;

            /* handle file uploads
             *
             * syntax: _handleUpload($table, $row, $column, $index, true)
             */
            case 'image': case 'file':
                $value = $this->structure->untaintInput($table, $column, $value, YANA_ESCAPE_LINEBREAK);

                /* if row already has an associated file, overwrite the old one, rather than creating another */
                if ($row !== '' && $row !== '*' && $row !== '?' && $db->exists("$table.$row.$column")) {
                    $fileId = DbDatatype::_decodeFileId($db->get("$table.$row.$column"));
                } else {
                    $fileId = DbDatatype::_encodeFileId($table, $column);
                }

                $value = $this->_handleUpload($table, $row, $column, $value, $fileId, $type == 'image');
            break;

            default:
                $value = $this->structure->untaintInput($table, $column, $value, YANA_ESCAPE_NONE);
            break;
        }
    }

    /**
     * prepare a database entry for output
     *
     * @access  public
     * @param   mixed   &$out
     * @param   string  $table
     * @param   string  $column
     * @param   mixed   $value
     * @param   string  $arrayKey
     * @return  bool
     */
    function decode(&$out, $table, $column, $value, $arrayKey = '')
    {
        assert('is_string($table); // Wrong argument type for argument 1. String expected');
        assert('is_string($column); // Wrong argument type for argument 2. String expected');
        assert('is_string($arrayKey); // Wrong argument type for argument 4. String expected');

        /* 1) check constraint */
        if ($this->exec->constraint('SELECT', $table, $column, $value) !== false) {

            switch ($this->structure->getType($table, $column))
            {
                case 'integer': case 'int':
                    if (is_numeric($value)) {
                        $out = (int) $value;
                    } else if ($value === true) {
                        $out = 1;
                    } else if ($value === false) {
                        $out = 0;
                    }
                break;

                case 'float':
                    if (is_numeric($value)) {
                        $out = (float) $value;
                    } else if ($value === true) {
                        $out = 1.0;
                    } else if ($value === false) {
                        $out = 0.0;
                    }
                break;

                case 'bool': case 'boolean':
                    if (is_bool($value)) {
                        $out = $value;
                    } else if ($value == 1) {
                        $out = true;
                    } else if (empty($value)) {
                        $out = false;
                    }
                break;

                case 'array':
                    if (is_array($value)) {
                        $out = $value;
                    } else {
                        $out = SML::getFile(explode("\n", $value), CASE_UPPER);
                    }
                    if ($arrayKey !== "") {
                        assert('is_array($REF); /* unexpected result: $REF should be an array */');
                        $out = Hashtable::get($result, strtolower($arrayKey));
                        if (is_null($result)) {
                            $out = false;
                        }
                    }
                break;

                case 'image': case 'file':
                    if (is_null($this->dbDir) && $this->dbDir !== false) {
                        global $YANA;
                        $dir = $YANA->getVar('DBBLOB');
                        if (is_dir($dir)) {
                            $this->dbDir = $dir;
                        } else {
                            $this->dbDir = false;
                        }
                    }
                    if ($this->dbDir !== false) {
                        $out = $this->dbDir . $value;
                    }
                break;

                default:
                    $out = stripslashes($value);
                break;
            }
            return true;

        } else {
            return false;
        }
    }

    /**
     * handle file uploads
     *
     * This handles all file uploads.
     * Set $isImage to true to upload images.
     *
     * @access  private
     * @param   string  $table
     * @param   string  $row
     * @param   string  $column
     * @param   string  $id
     * @param   string  $fileId
     * @param   bool    $isImage
     * @return  string|null
     * @ignore
     */
    function _handleUpload($table, $row, $column, $id, $fileId, $isImage = false)
    {
        assert('is_string($table);  // Wrong argument type for argument 1. String expected');
        assert('is_string($row);    // Wrong argument type for argument 2. String expected');
        assert('is_string($column); // Wrong argument type for argument 3. String expected');
        assert('is_string($id);     // Wrong argument type for argument 4. String expected');
        assert('is_string($fileId); // Wrong argument type for argument 5. String expected');
        assert('is_bool($isImage);  // Wrong argument type for argument 6. Boolean expected');

        global $YANA;

        /* get directory */
        assert('!isset($dir); // Cannot redeclare var $dir');
        $dir = $YANA->getVar('DBBLOB');
        if ($dir === false) {
            $YANA->report( new Alert("upload_failed") );
            return null;            
        }

        /* 1. file uploads */
        if (!$isImage) {

            /* name of output file */
            $file = "file.${fileId}.gz";
            /* name of input file */
            $maxSize = $this->structure->getLength($table, $column);
            $result = untaintInput($id, "upload", $maxSize);

        /* 2. image uploads */
        } else {

            /* name of output files */
            $file = "image.${fileId}";
            $thumb = "thumb.${fileId}";
            /* name of input file */
            $args = $this->structure->getImageSettings($table, $column);
            if (!is_array($args)) {
                return null;
            }
            $result = Image::uploadFile($id, $dir.$file, null, $args['size'], $args['width'], $args['height'], $args['ratio'], $args['background']);

        }

        unset($fileId);


        if (is_string($result) && is_file($result) && is_readable($result)) {

            /* 1. file uploads */
            if (!$isImage) {

                assert('!isset($FILES); // Cannot redeclare var $FILES');
                $FILES = array_change_key_case($_FILES, CASE_LOWER);
                assert('!isset($name); // Cannot redeclare var $name');
                if (!empty($FILES[$id]['name'])) {
                    $name = $FILES[$id]['name'];
                    $name = preg_replace('/[^\d\w\_\-\.]/', '_', $name);
                } else {
                    $name = "";
                }
                assert('!isset($mimetype); // Cannot redeclare var $mimetype');
                if (!empty($FILES[$id]['type'])) {
                    $mimetype = $FILES[$id]['type'];
                    $mimetype = preg_replace('/\s/', ' ', $mimetype);
                } else {
                    $name = "";
                }
                unset($FILES);

                /* create zip file */
                assert('!isset($gz); /* Cannot redeclare var $gz */');
                $gz = gzopen($dir . $file, 'w9');

                /**
                 * insert Header information
                 *
                 * 1) original filename
                 * 2) original size
                 * 3) original mime-type
                 *
                 * Note: line-break "\n" is used as a delimiter
                 */
                gzwrite($gz, "$name\n");
                gzwrite($gz, filesize($result)."\n");
                gzwrite($gz, "$mimetype\n");

                /* save file contents */
                gzwrite($gz, file_get_contents($result));
                gzclose($gz);
                unset($gz);

                return $file;

            /* 2. image uploads */
            } else {

                /* create thumbnail */
                assert('!isset($width); // Cannot redeclare $width');
                if (isset($args['width']) && $args['width'] < 75) {
                    $width = null;
                } else {
                    $width = 75;
                }
                assert('!isset($height); // Cannot redeclare $height');
                if (isset($args['height']) && $args['height'] < 75) {
                    $height = null;
                } else {
                    $height = 75;
                }
                Image::uploadFile($id, $dir . $thumb, null, null, $width, $height, true, array(128, 128, 128));
                unset($width, $height);
                $result = preg_replace('/^'.toRegExp($dir).'/', '', $result);
                return $result;

            }

        /* handle errors */
        } else {

            $FILES    = array_change_key_case($_FILES, CASE_UPPER);
            $filename = @$FILES[$id]['name'];

            switch ($result)
            {
                case UPLOAD_ERR_SIZE: case UPLOAD_ERR_INI_SIZE: case UPLOAD_ERR_FORM_SIZE:
                    $YANA->report( new Alert("too_large_file", array('FILE' => $filename, 'MAXIMUM' => $args['size'])) );
                    return null;
                break;
                case UPLOAD_ERR_FILE_TYPE:
                    $YANA->report( new Alert(((!$isImage) ? "upload_failed" : "invalid_image"), array('FILE' => $filename)) );
                    return null;
                break;
                case UPLOAD_ERR_INVALID_TARGET:
                    $YANA->report( new Log("Unable to write uploaded file to '${file}'.") );
                    $YANA->report( new Error("500") );
                    return null;
                break;
                case UPLOAD_ERR_NO_FILE:
                    $YANA->report( new Alert(((!$isImage) ? "upload_failed" : "invalid_image")) );
                    return null;
                break;
                default:
                    $YANA->report( new Alert("upload_failed", array('FILE' => $filename)) );
                    return null;
                break;
            } /* end switch */

        } /* end error handling */
    }

    /**
     * create unique id to identify a file
     *
     * @access  private
     * @static
     * @param   string  $table
     * @param   string  $column
     * @return  string
     * @ignore
     */
    function _encodeFileId($table, $column)
    {
        assert('is_string($table);  // Wrong argument type for argument 1. String expected');
        assert('is_string($column); // Wrong argument type for argument 2. String expected');
        return md5(uniqid(strtolower($table).'.'.rand().'.'.$column, true));
    }

    /**
     * extract unique file-id from a database value
     *
     * @access  private
     * @static
     * @param   string  $fileId
     * @return  string
     * @ignore
     */
    function _decodeFileId($fileId)
    {
        assert('is_string($fileId); // Wrong argument type for argument 1. String expected');
        return preg_replace('/^.*(?:image|thumb|file)\.(\w+)\.\w+$/', '${1}', $fileId);
    }

}

?>