<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * Database query builder
 *
 * This class is a query builder that can be used to build SQL statements.
 *
 * Note: this class does NOT untaint input data for you.
 * It also does NOT automatically resolve foreign keys.
 *
 * This class is still experimental.
 * Names and functions might get added, removed or changed without further notice.
 * Use at your own risk!
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 * @since       2.9 RC1
 *
 * @todo        check integration with DbStream - class
 */
class DbQuery extends Object
{
    /**#@+
     * @access  private
     */

    /** @var int    */ var $type    = YANA_DB_UNKNOWN;

    /** @var string */ var $table   = null;
    /** @var string */ var $row     = null;
    /** @var string */ var $column  = null;

    /** @var array  */ var $where   = array();
    /** @var array  */ var $orderBy = null;
    /** @var bool   */ var $desc    = false;

    /** @var object */ var $db      = null;
    /** @var array  */ var $joins   = array();

    /** @var array  */ var $values  = null;

    /**#@-*/

    /**
     * create a new instance
     *
     * This creates and initializes a new
     * instance of this class.
     *
     * @param  object  $database
     */
    function DbQuery($database)
    {
        /* error - data type */
        if (is_string($database)) {

            global $YANA;
            $database = $YANA->connect($database);
            if (is_object($database)) {
                $this->db =& $database;
            }

        } else if (!is_object($database)) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Instance of DbStream', gettype($database)), E_USER_WARNING);

        /* error - invalid base class */
        } else if (!is_a($database, 'DbStream') && !is_subclass_of($database, 'DbStream')) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Instance of DbStream', get_class($database)), E_USER_WARNING);

        } else {

            $this->db =& $database;

        }
    }

    /**
     * select the kind of statement
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   int  $type
     * @return  bool
     */
    function setType($type)
    {
        switch ($type)
        {
            case YANA_DB_UNKNOWN:
            case YANA_DB_SELECT:
            case YANA_DB_UPDATE:
            case YANA_DB_INSERT:
            case YANA_DB_DELETE:
            case YANA_DB_EXISTS:
            case YANA_DB_LENGTH:
                $this->type = $type;
            break;

            default:
                trigger_error("The statement type is unknown.", E_USER_WARNING);
                return false;
            break;
        }
    }

    /**
     * get the currently selected type of statement
     *
     * Returns currently selected constant.
     *
     * @access  public
     * @return  int
     */
    function getType()
    {
        if (is_int($this->type)) {
            return $this->type;
        } else {
            /* error: no type has been selected, yet */
            return YANA_DB_UNKNOWN;
        }
    }

    /**
     * set source table
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $table
     * @return  bool
     */
    function setTable($table)
    {
        if (!is_string($table)) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($table)), E_USER_WARNING);
            return false;

        } else if (!is_null($this->db->structure) && !$this->db->structure->isTable($table)) {

            trigger_error("The table '$table' is unknown.", E_USER_WARNING);
            return false;

        } else {

            $this->table = strtolower($table);
            return true;

        }
    }

    /**
     * get the currently selected table
     *
     * Returns the lower-cased name of the currently
     * selected table, or bool(false) if none has been
     * selected yet.
     *
     * @access  public
     * @return  bool(false)|string
     */
    function getTable()
    {
        if (is_string($this->table)) {
            return $this->table;
        } else {
            /* error: no table has been selected, yet */
            return false;
        }
    }

    /**
     * set source column
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $column
     * @return  bool
     */
    function setColumn($column)
    {
        if (!is_string($column)) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($column)), E_USER_WARNING);
            return false;

        } else if (empty($this->table)) {

            trigger_error("Cannot set column - need to set table first!", E_USER_WARNING);
            return false;

        } else if ($column === '*' || $column === '') {

            $this->column = $column;
            return true;

        } else if (!is_null($this->db->structure) && !$this->db->structure->isColumn($this->table, $column)) {

            trigger_error("The column '$column' is not found in table '".$this->table."'.", E_USER_WARNING);
            return false;

        } else {

            $this->column = strtolower($column);
            return true;

        }
    }

    /**
     * get the currently selected column
     *
     * Returns the lower-cased name of the currently
     * selected column.
     *
     * If none has been selected, '*' is returned.
     *
     * @access  public
     * @return  string
     */
    function getColumn()
    {
        if (is_string($this->column)) {
            return $this->column;
        } else {
            /* error: no column has been selected, yet */
            return '*';
        }
    }

    /**
     * set source row
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * Note: does not check if row exists.
     *
     * Currently you may only request 1 row or all.
     * To search for all rows, use the wildcard '*'.
     *
     * @access  public
     * @param   scalar  $row
     * @return  bool
     */
    function setRow($row)
    {
        if (!is_scalar($row)) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Scalar', gettype($row)), E_USER_WARNING);
            return false;

        } else if (empty($row) || $row === '*') {

            $this->row = '*';
            return true;

        } else {

            $this->row = $row;
            return true;

        }
    }

    /**
     * get the currently selected row
     *
     * Returns the lower-cased name of the currently
     * selected column, or bool(false) if none has been
     * selected yet.
     *
     * @access  public
     * @return  bool(false)|string
     */
    function getRow()
    {
        if (is_string($this->row)) {
            return $this->row;
        } else {
            /* error: no row has been selected, yet */
            return false;
        }
    }

    /**
     * resolve key address to determine table, column and row
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $key
     * @return  bool
     */
    function setKey($key)
    {
        if (!is_string($key)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($key)), E_USER_WARNING);
            return false;
        } else if (!preg_match("/^[\w\d-_]+(\.([\w\d-_]*|\*|\?)){0,}(\.\*)?$/i", $key)) {
            trigger_error("Syntax error. The key '${key}' is not valid.", E_USER_WARNING);
            return false;
        } else {
            $key = preg_replace("/\.(\*)?$/", '', $key);
            $array = explode(".", $key);
        }

        if (count($array) > 3 || count($array) < 1) {

            return false;

        } else {

           if (!$this->setTable($array[0])) {
               return false;
           }

           if (empty($array[1])) {
               $this->setRow('*');
           } else if (!$this->setRow($array[1])) {
               return false;
           }

           if (empty($array[2])) {
               $this->setColumn('*');
           } else if (!$this->setColumn($array[2])) {
               return false;
           }

           return true;

        }

    }

    /**
     * set column to sort the resultset by
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string|array  $orderBy  column name / list of column names
     * @param   bool          $desc     sort descending (true=yes, false=no)
     * @return  bool
     */
    function setOrderBy($orderBy, $desc = false)
    {
        if (!is_string($orderBy) && !is_array($orderBy)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($orderBy)), E_USER_WARNING);
            return false;
        } else if (!is_bool($desc)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Boolean', gettype($desc)), E_USER_WARNING);
            return false;
        }

        /* settype to ARRAY */
        settype($orderBy, "array");

        /* check and lower-case all values */
        foreach($orderBy as $i => $column)
        {
            if (!$this->db->structure->isColumn($this->table, $column)) {

                trigger_error("The column '$column' is not found in table '".$this->table."'.", E_USER_WARNING);
                return false;

            } else {

                $orderBy[$i] = strtolower($column);

            }
        }

        $this->orderBy = $orderBy;
        $this->desc = $desc;
        return true;
    }

    /**
     * get the list of columns the resultset is ordered by
     *
     * Returns the lower-cased column name, or bool(false)
     * if none has been set yet.
     *
     * @access  public
     * @return  bool(false)|array
     */
    function getOrderBy()
    {
        if (is_array($this->orderBy) && !empty($this->orderBy)) {
            return $this->orderBy;
        } else {
            /* error: unexpected value */
            return false;
        }
    }

    /**
     * check if resultset is sorted in descending order
     *
     * Returns bool(true) if sorted descending and bool(false) otherwise.
     *
     * @access  public
     * @return  bool
     */
    function isDescending()
    {
        if (is_bool($this->desc)) {
            return $this->desc;
        } else {
            /* error: unexpected value */
            return false;
        }
    }

    /**
     * set where clause (filter)
     *
     * The syntax is as follows:
     * column1=value1,column2=value2,...,columnN=valueN
     *
     * The alternative is as follows:
     * array(array(0=>column1,1=>value1,2=>operator1),...)
     * Where "operator" can be one of the following:
     * '=', 'LIKE', '<', '>', '!=', '<=', '>='
     *
     * To unset the where clause, call this function without
     * providing a parameter.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string|array  $where
     * @return  bool
     */
    function setWhere($where = null)
    {
        if (is_null($where)) {

            $this->where = array();
            return true;

        } else if (!is_array($where) && !is_string($where)) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($where)), E_USER_WARNING);
            return false;

        }

        /* if scanning a whole table you may search for primary keys - otherwise not */
        if (is_null($this->row) || $this->row === '?' || $this->row === '*') {
            $isTableScan = true;
        } else {
            $isTableScan = false;
        }

        /* convert string to array */
        if (is_string($where)) {
            if (!preg_match("/^((, ?)?([\w\d-_\.])+=([\w\d-_\.\@ \&;])+)*$/i", $where)) {
                global $YANA;
                $YANA->report( new Log("SQL-Statement rejected du to invalid where clause.", $where) );
                trigger_error("The where clause you provided contains syntax errors.\n\t\tMake sure it does only contain alpha-numeric characters and matches the following syntax: fieldname1=value1,fieldname2=value2", E_USER_WARNING);
                return false;
            }

            $where = preg_split("/, ?/", $where);
            /* settype to ARRAY */
            $where = (array) $where;
            assert('!isset($a); /* cannot redeclare variable $a */');
            assert('!isset($b); /* cannot redeclare variable $b */');
            foreach ($where as $a => $b)
            {
                $where[$a] = explode("=", $b);
                if ($this->db->structure->isForeignKey($this->table, $where[$a][0]) || $this->db->structure->isPrimaryKey($this->table, $where[$a][0])) {
                    $where[$a][2] = '=';
                } else {
                    $where[$a][1] = '%'.$where[$a][1].'%';
                    $where[$a][2] = 'like';
                }
                $where[$a][0] = strtolower($where[$a][0]);
            } /* end foreach */
            unset($a, $b);
        }

        assert('is_array($where); // unexpected result: $where should have been converted to an array - but it is not.');

        /* check the array */
        if (count($where) > 0) {
            assert('!isset($i); /* cannot redeclare variable $i */');
            assert('!isset($clause); /* cannot redeclare variable $clause */');
            foreach ($where as $i => $clause)
            {
                if (is_array($clause) && count($clause) >= 2 && count($clause) <= 3 && !empty($clause[0]) && isset($clause[1])) {

                    /* use names that are easier to read */
                    $columnName = $clause[0];
                    $value = $clause[1];
                    if (isset($clause[2])) {
                        $operator = $clause[2];
                    }

                    /* settype to STRING
                     *            STRING
                     *            STRING
                     */
                    $columnName = (string) $columnName;
                    $value      = (string) $value;
                    $operator   = (string) $operator;

                    /* lower case input */
                    $columnName = strtolower($columnName);
                    $operator   = strtolower($operator);

                    /* map aliases */
                    if ($operator === '==') {
                        $operator = '=';
                    }

                    /* catch exceptions */
                    if (!in_array($operator, array('=', 'like', '<', '>', '!=', '<=', '>='))) {
                        global $YANA;
                        $YANA->report( new Log("SQL-Statement rejected du to invalid where clause.", $where) );
                        trigger_error("Invalid where clause. The operator '${operator}' is not supported.", E_USER_WARNING);
                        return false;
                    }
                    if (!$isTableScan && $this->db->structure->isStrict() && $this->db->structure->isPrimaryKey($this->table, $columnName)) {
                        global $YANA;
                        $YANA->report( new Log("SQL-Statement rejected du to invalid where clause.", $where) );
                        trigger_error("Invalid where clause. You are trying to search for a primary key.\n\t\tThis is not allowed, since it might cause results wether to be ambigious or empty.\n\t\tTry setting 'USE_STRICT' to 'false' if you want to turn this message off and continue without checking.", E_USER_WARNING);
                        return false;
                    }
                    if ($this->db->structure->isStrict() && !$this->db->structure->isColumn($this->table, $columnName) && !$this->db->structure->isForeignKey($this->table,$columnName)) {
                        global $YANA;
                        $YANA->report( new Log("SQL-Statement rejected du to invalid where clause.", $where) );
                        trigger_error("Invalid where clause. The column '".$columnName."' does not exist in table '".$this->table."'.\n\t\tPerhaps your structure file is incomplete.\n\t\tTry setting 'USE_STRICT' to 'false' if you want to turn this message off and continue without checking.", E_USER_WARNING);
                        return false;
                    }

                    $where[$i][0] = $columnName;
                    $where[$i][1] = $value;
                    $where[$i][2] = $operator;
                }
            } /* end foreach */
            unset($i, $clause);
        }

        $this->where = $where;
        return true;

    }

    /**
     * get the currently set where clause
     *
     * Returns the current where clause.
     *
     * @access  public
     * @return  array
     */
    function getWhere()
    {
        if (is_array($this->where)) {
            return $this->where;
        } else {
            return array();
        }
    }

    /**
     * join the resultsets for two tables
     *
     * This will join the currently selected table with another.
     *
     * @access  public
     * @param   string $table  name of another table to join table1 with (when omitted will remove all previously set joins from table1)
     * @param   string $key1   name of the foreign key in current table (when omitted the API will look up the key in the structure file)
     * @param   string $key2   name of the key in foreign table that is referenced (may be omitted if it is the primary key)
     * @return  bool
     */
    function setJoin($table = null, $key1 = null, $key2 = null)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        assert('!isset($_types);');
        $_types = array(
        /* argument 1 */ IS_STRING|IS_NULL,
        /* argument 2 */ IS_STRING|IS_NULL,
        /* argument 3 */ IS_STRING|IS_NULL
        );
        $_chk = checkArgumentList(func_get_args(), $_types, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk, $_types);
        }

        /* 1 check input */
        /* 1.1 argument 1 */
        if (is_null($table)) {

            /* if no arguments are provided, reset the list of joined tables */

            $this->joins = array();
            return true;

        } else {
            /* lower-case input */
            $table = strtolower($table);
        }

        /* 1.2 argument 2 */
        if (is_null($key1)) {

            /* if no key is provided, take the first association available */

            assert('!isset($foreignKeys); /* Cannot redeclare variable $foreignKeys */');
            $foreignKeys = $this->db->structure->getForeignKeys($this->table);

            if (empty($foreignKeys)) {
                trigger_error("Cannot join tables '".$this->table."' and '${table}'. No foreign key constraint has been found.", E_USER_NOTICE);
                return false;
            }

            assert('!isset($foreignTable); /* Cannot redeclare variable $foreignTable */');
            assert('!isset($foreignKey); /* Cannot redeclare variable $foreignKey */');
            foreach ($foreignKeys as $foreignKey => $foreignTable)
            {
                if (strcasecmp($table, $foreignTable) === 0) {
                    $key1 = strtolower($foreignKey);
                    $key2 = $this->db->structure->getPrimaryKey($foreignTable);
                    break;
                }
            }
            unset($foreignKey, $foreignKeys, $foreignTable);

        } else  {
            /* lower-case input */
            $key1 = strtolower($key1);
        }
        /* 1.3 argument 3 */
        if (is_null($key2)) {

            /* if the target column is not provided, take the primary key instead */
            $key2 = $this->db->structure->getPrimaryKey($table);

        } else {
            /* lower-case input */
            $key2 = strtolower($key2);
        }

        /* strict input checking */
        if ($this->db->structure->isStrict()) {

            if (!$this->db->structure->isTable($table)) {
                trigger_error("Cannot join tables '".$this->table."' and '${table}'. Table does not exist in structure file.", E_USER_WARNING);
                return false;
            }
            if (!is_null($key1) && !$this->db->structure->isColumn($this->table, $key1)) {
                trigger_error("Cannot join tables '".$this->table."' and '${table}'. Field '${key1}' does not exist in table '".$this->table."'.", E_USER_WARNING);
                return false;
            }
            if (!is_null($key2) && !$this->db->structure->isColumn($table, $key2)) {
                trigger_error("Cannot join tables '".$this->table."' and '".$table."'. Field '${key2}' does not exist in table '${table}'.", E_USER_WARNING);
                return false;
            }

        /* loose input checking */
        } else {

            for ($i = 0; $i < func_num_args(); $i++)
            {
                if (!preg_match("/^[\"\'`]?[\w\d-_]+[\"\'`]?$|^\*$/si", func_get_arg($i))) {
                    trigger_error("Argument ${i} is not a valid identifier.", E_USER_WARNING);
                }
            }

        }

        /* reset old association */
        if (isset($this->joins[$table])) {
            unset($this->joins[$table]);
        }

        /* create new association */
        $this->joins[$table] = array($key1, $key2);
        return true;

    }

    /**
     * get a list of the tables, the query has joined together
     *
     * If $table is not provided, all tables are returned.
     *
     * @access  public
     * @param   string  $table
     * @return  bool(false)|array
     */
    function getJoin($table = null)
    {
        if (is_null($table)) {

            return $this->joins;

        } else if (!is_string($table)) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($table)), E_USER_WARNING);
            return false;

        } else if (!$this->db->structure->isTable($table)) {

            trigger_error("The table '$table' is unknown.", E_USER_WARNING);
            return false;

        } else {

            $table = strtolower($table);

        }

        if (!isset($this->joins[$table])) {

            return false;

        } else {

            return $this->joins[$table];

        }
    }

    /**
     * set value(s) for current query
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @return  bool
     */
    function setValues(&$values)
    {
        if (is_array($values)) {
            $values = Hashtable::changeCase($values, CASE_LOWER);
        }

        if ($this->type === YANA_DB_INSERT) {

            $test = $this->db->structure->checkRow($this->table, $values, true);

        } else if ($this->type === YANA_DB_UPDATE) {

            $test = $this->db->structure->checkRow($this->table, $values, false);

        } else {

            $test = true;

        }

        if ($test) {

            $this->values =& $values;
            return true;

        } else {

            return false;

        }
    }

    /**
     * get the list of values
     *
     * If none are available, NULL (not bool(false)!) is returned.
     *
     * @access  public
     * @return  mixed
     */
    function &getValues()
    {
        return $this->values;
    }

    /**
     * build a SQL-query
     *
     * @access  public
     * @return  string
     */
    function toString()
    {
        /* prepare where clause */
        $where = $this->getWhere();
        if (is_array($where) && count($where) > 0) {

            assert('!isset($_where); // Cannot redeclare $_where');
            $_where = "";
            /* build filter */
            if (count($where) > 0) {
                assert('!isset($clause); // Cannot redeclare $clause');
                foreach ($where as $clause)
                {
                    $_where .= (($_where !== '') ? 'AND ' : '') . $clause[0] . ' ' . $clause[2] . ' ' . $this->db->quote($clause[1]) . ' ';

                }
                unset($clause);
                $_where = 'WHERE ' . $_where;
            }
            $where = $_where;
            unset($_where);

        } else {

            $where = "";

        }

        /* select a statement */
        switch ($this->type)
        {
            case YANA_DB_SELECT:
                $stmt = "SELECT %COLUMN% FROM %TABLE% %WHERE% %ORDERBY% %DESC%;";
            break;

            case YANA_DB_EXISTS:
                $stmt = "SELECT 1 FROM %TABLE% %WHERE%;";
            break;

            case YANA_DB_LENGTH:
                $stmt = "SELECT count(%COLUMN%) FROM %TABLE% %WHERE%;";
            break;

            case YANA_DB_UPDATE:
                $stmt = "UPDATE %TABLE% SET %SET% %WHERE%;";

                /**
                 * replace %SET%
                 *
                 * Note: this is done here, since all other types
                 * of statements do not have this token.
                 */
                assert('!isset($set); // Cannot redeclare $set');
                $set = "";
                if (is_array($this->values)) {
                    assert('!isset($column); // Cannot redeclare $column');
                    assert('!isset($value);  // Cannot redeclare $value');
                    foreach ($this->values as $column => $value)
                    {
                        if ($set !== '') {
                            $set .= ', ';
                        }
                        $set .= $column . ' = ' . $this->db->quote($value);
                    }
                    unset($column, $value);
                }
                $stmt = str_replace('%SET%', $set, $stmt);
                unset($set);

            break;

            case YANA_DB_INSERT:
                $stmt = "INSERT INTO %TABLE% (%KEYS%) VALUES(%VALUES%);";

                /**
                 * replace %KEYS% and %VALUES%
                 *
                 * Note: this is done here, since all other types
                 * of statements do not have these token.
                 */
                if (is_array($this->values) && count($this->values) > 0) {

                    assert('!isset($keys);   // Cannot redeclare $keys');
                    assert('!isset($values); // Cannot redeclare $values');
                    $keys = implode(', ', array_keys($this->values));
                    $values = '';
                    assert('!isset($value); // Cannot redeclare $values');
                    foreach ($this->values as $value)
                    {
                        $values .= (($values !== '') ? ', ' : '' ) . $this->db->quote($value);
                    }
                    unset($value);
                    $stmt = str_replace('%KEYS%', $keys, $stmt);
                    $stmt = str_replace('%VALUES%', $values, $stmt);
                    unset($keys, $values);

                } else {
                    return "";
                }

            break;

            case YANA_DB_DELETE:
                $stmt = "DELETE FROM %TABLE% %WHERE% %ORDERBY% %DESC%;";
            break;

            default:
                return "";
            break;

        }

        /* 1. replace %TABLE% */
        $table = $this->getTable();
        if (is_string($table)) {
            $table = $this->db->quoteId(YANA_DATABASE_PREFIX.$table);

            if ($this->type === YANA_DB_SELECT && !empty($this->joins)) {

                assert('!isset($_where);    /* cannot redeclare variable $_where    */');
                $_where = array();

                assert('!isset($tableName); /* cannot redeclare variable $tableName */');
                assert('!isset($join);      /* cannot redeclare variable $join      */');
                foreach ($this->joins as $tableName => $join)
                {
                    /* collect tables */
                    if (empty($join)) {

                        $table .= "NATURAL JOIN ".$this->db->quoteId(YANA_DATABASE_PREFIX.$tableName)." ";

                    } else {

                        $table .= ', ' . $this->db->quoteId(YANA_DATABASE_PREFIX.$tableName) . ' ';

                    }

                    /* collect tables */
                    if (!empty($join[0]) && !empty($join[1])) {

                          $_where[] = array($this->table . '.' . $join[0], $tableName . '.' . $join[1]);

                    } else if (!empty($join[0])) {

                          $_where[] = array($this->table . '.' . $join[0], $tableName . '.' . $join[0]);

                    } else {

                          /* intenionally left blank */

                    }

                } /* end foreach */
                unset($tableName, $join);

                if (count($_where) > 0) {
                    assert('!isset($clause); // Cannot redeclare $clause');
                    foreach ($_where as $clause)
                    {
                        if ($where === '') {
                            $where = 'WHERE ';
                        } else {
                            $where .= 'AND ';
                        }
                        $where .= YANA_DATABASE_PREFIX.$clause[0] . ' = ' . YANA_DATABASE_PREFIX.$clause[1] . ' ';

                    }
                    unset($clause);
                }
                unset($_where);


            }

            $stmt = str_replace('%TABLE%', $table, $stmt);

        } else {

            return false;

        }


        /* 2. replace %COLUMN% */
        $stmt = str_replace('%COLUMN%', $this->getColumn(), $stmt);


        /* 3. handle ROW */
        assert('!isset($row); // Cannot redeclare $row');
        $row = $this->getRow();
        if ($row !== '*') {
            $where[] = array($this->db->structure->getPrimaryKey($this->table), $row, '=');
        }
        unset($row);


        /* 4. replace %WHERE% */
        $stmt = str_replace('%WHERE%', trim($where), $stmt);
        unset($where);


        /* 5. replace %ORDERBY% */
        assert('!isset($orderBy); // Cannot redeclare $orderBy');
        $orderBy = $this->getOrderBy();
        if (is_array($orderBy)) {
            $stmt = str_replace('%ORDERBY%', 'ORDER BY '.implode(', ', $orderBy), $stmt);
        } else {
            $stmt = str_replace(' %ORDERBY%', '', $stmt);
        }
        unset($orderBy);

        /* 6. replace %DESC% */
        if ($this->isDescending()) {
            $stmt = str_replace('%DESC%', 'desc', $stmt);
        } else {
            $stmt = str_replace(' %DESC%', '', $stmt);
        }

        return $stmt;

    }

    /**
     * self test
     *
     * This function exists only for debugging purposes.
     * Do not uncomment in productive environments.
     *
     * @access  public
     * @static
     *
     * @ignore
     * /
    function test($nr = 1)
    {
        switch ($nr)
        {
            case 1:
                $query = new DbQuery("user");

                // test set/get table
                $query->setTable('UseR');
                if ($query->getTable() !== 'user') {
                    return '"table": test failed';
                }

                // test set/get column
                $query->setColumn('AcTiVe');
                if ($query->getColumn() !== 'active') {
                    return '"column": test failed';
                }
                $query->setColumn('*');
                if ($query->getColumn() !== '*') {
                    return '"column": test failed';
                }

                // test set/get row
                $query->setRow('AbC');
                if ($query->getRow() !== 'AbC') {
                    return '"row": test failed';
                }
                $query->setRow('*');
                if ($query->getRow() !== '*') {
                    return '"row": test failed';
                }

                // test set/get type
                if ($query->getType() !== YANA_DB_UNKNOWN) {
                    return '"type": test failed';
                }
                $query->setType(YANA_DB_SELECT);
                if ($query->getType() !== YANA_DB_SELECT) {
                    return '"type": test failed';
                }

                // test set/get order by / desc
                if ($query->getOrderBy() !== false) {
                    return '"order by": test failed';
                }
                $query->setOrderBy('Active', true);
                if ($query->getOrderBy() !== array('active')) {
                    return '"order by": test failed';
                }
                if ($query->isDescending() !== true) {
                    return '"descending order": test failed';
                }
                $query->setOrderBy(array('ID', 'Active'));
                if ($query->getOrderBy() !== array('id', 'active')) {
                    return '"order by": test failed';
                }
                if ($query->isDescending() !== false) {
                    return '"descending order": test failed';
                }

                // test set/get where
                if ($query->getWhere() !== array()) {
                    return '"where": test failed';
                }
                $query->setWhere(array());
                if ($query->getWhere() !== array()) {
                    return '"where": test failed';
                }
                $query->setWhere('Id=AbC, Active=0');
                if ($query->getWhere() != array(array('id', 'AbC', '='), array('active', '%0%', 'like'))) {
                    return '"where": test failed';
                }
                $query->setWhere(array(array('Id', 'AbC', '='), array('Active', '1', 'lIke')));
                if ($query->getWhere() != array(array('id', 'AbC', '='), array('active', '1', 'like'))) {
                    return '"where": test failed';
                }

                // test set/get join
                if ($query->getJoin() !== array()) {
                    return '"join": test failed';
                }
                $query->setJoin('GrOup', 'Type', 'tYPE');
                if ($query->getJoin() !== array('group' => array('type', 'type'))) {
                    return '"join": test failed';
                }
                $query->setJoin('GrouP', 'Type');
                if ($query->getJoin() !== array('group' => array('type', 'type'))) {
                    return '"join": test failed';
                }
                $query->setJoin();
                if ($query->getJoin() !== array()) {
                    return '"join": test failed';
                }
                $query->setJoin('GroUp');
                if ($query->getJoin() !== array('group' => array('type', 'type'))) {
                    return '"join": test failed';
                }

                // test set/get values
                if ($query->getValues() !== null) {
                    return '"values": test failed';
                }
                $value = true;
                if ($query->setValues($value) !== true) {
                    return '"values": test failed';
                }
                if ($query->getValues() !== true) {
                    return '"values": test failed';
                }
                $query->setType(YANA_DB_UPDATE);
                $value = array('ID' => 'AbC', 'maIl' => 'abc@foo.de', 'acTive' => 0);
                if ($query->setValues($value) !== true) {
                    return '"values": test failed';
                }
                if ($query->getValues() !== array('id' => 'AbC', 'mail' => 'abc@foo.de', 'active' => 0)) {
                    return '"values": test failed';
                }
                if ($query->getValues() !== $value) {
                    return '"values": test failed';
                }

                // test conversion to SELECT stmt
                $query->setType(YANA_DB_SELECT);
                if ($query->toString() != 'SELECT * FROM "user", "group"  WHERE id = "AbC" AND active like "1" AND user.type = group.type ORDER BY id, active;') {
                    return '"toString" - select statement: test failed';
                }

                // test conversion to EXISTS stmt
                $query->setType(YANA_DB_EXISTS);
                if ($query->toString() !== 'SELECT 1 FROM "user" WHERE id = "AbC" AND active like "1";') {
                    return '"toString" - exists statement: test failed';
                }

                // test conversion to LENGTH stmt
                $query->setType(YANA_DB_LENGTH);
                if ($query->toString() !== 'SELECT count(*) FROM "user" WHERE id = "AbC" AND active like "1";') {
                    return '"toString" - length statement: test failed';
                }

                // test conversion to UPDATE stmt
                $query->setType(YANA_DB_UPDATE);
                if ($query->toString() !== 'UPDATE "user" SET id = "AbC", mail = "abc@foo.de", active = "0" WHERE id = "AbC" AND active like "1";') {
                    return '"toString" - update statement: test failed';
                }

                // test conversion to INSERT stmt
                $query->setType(YANA_DB_INSERT);
                if ($query->toString() !== 'INSERT INTO "user" (id, mail, active) VALUES("AbC", "abc@foo.de", "0");') {
                    return '"toString" - insert statement: test failed';
                }

                // test conversion to DELETE stmt
                $query->setType(YANA_DB_DELETE);
                if ($query->toString() !== 'DELETE FROM "user" WHERE id = "AbC" AND active like "1" ORDER BY id, active;') {
                    return '"toString" - delete statement: test failed';
                }

                // test set key
                if ($query->setKey('User.Test.Active.1') !== false) {
                    return '"key": test failed';
                }
                $query->setKey('User.Test.Active');
                if ($query->getTable() !== 'user' || $query->getRow() !== 'Test' || $query->getColumn() !== 'active') {
                    return '"key": test failed';
                }
                $query->setKey('User.*.Active');
                if ($query->getTable() !== 'user' || $query->getRow() !== '*' || $query->getColumn() !== 'active') {
                    return '"key": test failed';
                }
                $query->setKey('User.*');
                if ($query->getTable() !== 'user' || $query->getRow() !== '*' || $query->getColumn() !== '*') {
                    return '"key": test failed';
                }
                $query->setKey('User.Test');
                if ($query->getTable() !== 'user' || $query->getRow() !== 'Test' || $query->getColumn() !== '*') {
                    return '"key": test failed';
                }
                $query->setKey('User');
                if ($query->getTable() !== 'user' || $query->getRow() !== '*' || $query->getColumn() !== '*') {
                    return '"key": test failed';
                }


                return true;

            break;
        }
    } /* end of function */

}

?>
