<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**#@+
 * @ignore
 */

if (!defined('YANA_DB_UNKNOWN')) {
    define('YANA_DB_UNKNOWN', 0);
}
if (!defined('YANA_DB_SELECT')) {
    define('YANA_DB_SELECT', 1);
}
if (!defined('YANA_DB_UPDATE')) {
    define('YANA_DB_UPDATE', 2);
}
if (!defined('YANA_DB_INSERT')) {
    define('YANA_DB_INSERT', 3);
}
if (!defined('YANA_DB_DELETE')) {
    define('YANA_DB_DELETE', 4);
}
if (!defined('YANA_DB_EXISTS')) {
    define('YANA_DB_EXISTS', 5);
}
if (!defined('YANA_DB_LENGTH')) {
    define('YANA_DB_LENGTH', 6);
}
if (!defined('YANA_DB_TABLE')) {
    define('YANA_DB_TABLE', 1);
}
if (!defined('YANA_DB_ROW')) {
    define('YANA_DB_ROW', 2);
}
if (!defined('YANA_DB_CELL')) {
    define('YANA_DB_CELL', 3);
}
if (!defined('YANA_DB_COLUMN')) {
    define('YANA_DB_COLUMN', 4);
}
if (!defined('YANA_DB_NONE')) {
    define('YANA_DB_NONE', 0);
}
if (!defined('YANA_DB_ROW_WILDCARD')) {
    define('YANA_DB_ROW_WILDCARD', 1);
}
if (!defined('YANA_DB_COL_WILDCARD')) {
    define('YANA_DB_COL_WILDCARD', 2);
}

/**#@-*/

/**
 * @ignore
 */
@include_once("DB.php");

/**
 * database API
 *
 * this class is a database abstraction api, that uses pear db
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 */
class DbStream extends Object
{
    /**#@+
     * @ignore
     * @access  private
     */

    /** @var  int        */   var $exists              = YANA_DATABASE_FAILURE;
    /** @var  Object     */   var $database            = null;
    /** @var  DbDatatype */   var $transformer         = null;
    /** @var  DbTrigger  */   var $exec                = null;
    /** @var  string     */   var $lastSelectedTable   = "";
    /** @var  string     */   var $path                = "";
    /** @var  array      */   var $queue               = array();
    /** @var  array      */   var $cache               = array();
    /** @var  array      */   var $joins               = array();
    /** @var  array      */   var $reservedSqlKeywords = null;
    /** @var  SML        */   var $lastModified        = null;

    /**#@-*/

   /**
    * @ignore
    * @access public
    * @var    StructureFile
    */                    var $structure           = null;

    /**
     * create a new instance
     *
     * Each database connection depends on a structure file
     * describing the database. Therefore you have to provide
     * a $filename of such structure file (without path or extension)
     *
     * These files are to be found in config/db/*.config
     *
     * Since version 2.8.7 the argument $file can also be an instance
     * of class StructureFile.
     *
     * @param  string|StructureFile  $file
     * @param  dbServer              $server
     */
    function DbStream($file = "", $server = null)
    {
        /* 1 CHECK INPUT
         */

        /* 1.1 FIRST ARGUMENT
        */
        if (is_object($file)) {

            /* 1.1 a) ERROR CASE: invalid base class
            */
            if (!is_a($file, 'StructureFile')) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Instance of StructureFile', get_class($file)), E_USER_WARNING);
                $file = "";

            /* 1.1 b) SETTING CLASS ATTRIBUTES
            */
            } else {
                $this->path  = $file->getPath();
                $this->structure = $file;
            }

        } else {

            /* 1.1 a) ERROR CASE: invalid data type
            */
            if (!empty($file) && !is_string($file)) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($file)), E_USER_WARNING);
                $this->path  = "";
                $this->structure = null;

            /* 1.1 b) SETTING CLASS ATTRIBUTES
            */
            } else {
                /* settype to STRING */
                $file = (string) $file;
                $this->path = $file;
                $this->structure = new StructureFile($this->getPath());
            }

        }

        /* 1.2 SECOND ARGUMENT
        */

        /* 1.2 a) ERROR CASE: invalid data type
        */
        if (is_null($server)) {
            $server = new dbServer();
        } else if (is_object($server)) {
            if (!is_a($server, 'DbServer')) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Instance of DbServer', get_class($server)), E_USER_WARNING);
                $server = new dbServer();
            }
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Object', gettype($server)), E_USER_WARNING);
            $server = new dbServer();
        }

        /* 1.2 b) SETTING CLASS ATTRIBUTES
        */
        $this->database = $server->get();

        /* 2 LIST OF GLOBALS
         */

        global $YANA;

        /* 3 INIT DATABASE CONNECTION
         */

        if (empty($this->database) || $this->isError($this->database)) {
            /* connection failed - abort */
            $this->exists = YANA_DATABASE_FAILURE;
            $this->handleError($this->database);
        } else {

            /* deactivate autocommit by default */
            $result = $this->database->autoCommit(false);
            if ($this->isError($result)) {
                $this->handleError($result);
            }

            if (is_null($this->structure)) {

                /* If no structure file is available,
                 * this framework has the ability to estimate the structure
                 * of the database "on the fly" at runtime, via the
                 * "DbStream::buildStrucutre()" method.
                 */

                $this->buildStructure();
                $this->exists = YANA_DATABASE_NO_FILE;

            } else {

                $test = $this->structure->read();
                if ($test === false) {
                    /* given file does not exist */
                    unset($this->structure);
                    $this->exists = YANA_DATABASE_NOT_AVAILABLE;
                    trigger_error("Unable to initialize database. Structure file '".$this->getPath()."' not found.", E_USER_WARNING);
                } else {
                    /* all fine, proceed */
                    $this->exists = YANA_DATABASE_OK;
                }

            } /* end if */
        } /* end if */

        /* 4 INIT IO/FILTER
         */
        if (is_object($this->structure)) {
            $this->transformer = new DbDatatype($this->structure);
        }

        /* 5 INIT TRIGGER
         */
        if (is_object($this->structure)) {
            $this->exec = new DbTrigger($this->structure);
        }    
    }

    /**
     * Commit current transaction
     *
     * This writes all changes to the database
     *
     * @access  public
     * @return  bool
     */
    function write()
    {
        if (!$this->_check_writeable()) {
            trigger_error('Operation aborted, not writeable.', E_USER_NOTICE);
            return false;
        }

        if (count($this->queue) == 0) {
            /* Buffer empty */
            return true;
        } else {
            for ($i = 0; $i < count($this->queue); $i++)
            {
                /* SQL statement */
                if (is_array($this->queue[$i]) && isset($this->queue[$i][0])) {
                    $sqlStmt = trim($this->queue[$i][0]);
                } else if (is_string($this->queue[$i])) {
                    $sqlStmt = trim($this->queue[$i]);
                } else {
                    continue;
                }
                /* trigger */
                if (is_array($this->queue[$i]) && isset($this->queue[$i][1]) && is_array($this->queue[$i][1])) {
                    $args = $this->queue[$i][1];
                } else {
                    $args = null;
                }
                /* ignore empty queue elements */
                if (empty($sqlStmt)) {
                    continue;
                /* for security reasons all delete queries will automatically be limited to 1 row at a time */
                } else if (preg_match('/^DELETE /', $sqlStmt)) {
                    $result = $this->database->limitQuery($sqlStmt,0,1);
                } else {
                    $result = $this->database->query($sqlStmt);
                }
                assert('!isset($success); // cannot redeclare variable $success');
                if ($this->isError($result)) {
                    /* fire trigger */
                    if (!is_null($args)) {
                        $success = false;
                        $this->exec->trigger(YANA_TRIGGER_AFTER, $args[0], $args[1], $args[2], $success);
                    }
                    /* rollback on error */
                    $this->handleError($result, 'SQL ERROR: Statement [code]'.$sqlStmt.'[/code]');
                    $result = $this->database->rollback();
                    if ($this->isError($result)) {
                          $this->handleError($result, "SQL ERROR: Unable to rollback changes. Database might contain corrupt data. Please inform webmaster about time of event and anything you might have done to cause this failure.");
                    }
                    return false;
                } else {
                    if (!is_null($args)) {
                        $success = true;
                        $this->exec->trigger(YANA_TRIGGER_AFTER, $args[0], $args[1], $args[2], $success);
                    }
                }
                unset($success);
            } /* end foreach */
            $result = $this->database->commit();
            if ($this->isError($result)) {
                if (!is_null($this->lastModified)) {
                    $this->lastModified->reset();
                }
                $this->handleError($result);
                return false;
            } else {
                if (!is_null($this->lastModified)) {
                    $this->lastModified->write();
                }
                $this->queue = array();
                return true;
            }
        }
    }

    /**
     * get values from the database
     *
     * This returns the values at adress $key
     * starting from $offset and limited to
     * $limit results.
     *
     * The $key parameter has two synopsis.
     * <ul>
     *     <li>
     *         If $key is a string, this parameter is interpreted
     *         as the address of the values you want to retrieve.
     *     </li>
     *     <li>
     *         If $key is an associative array, it overwrites any of the
     *         paramters you provide, without the need to enter default
     *         parameters - this might come in handy in certain situations.
     *     </li>
     * </ul>
     *
     * The array $key can have the following values
     * <ul>
     *     <li>  $key      = the address of the value(s) to retrieve (mandatory) </li>
     *     <li>  $where    = a search string (optional) </li>
     *     <li>  $search   = alias of $where (deprecated) </li>
     *     <li>  $order_by = array of columns to order the resultset by (optional) </li>
     *     <li>  $desc     = boolean, if true results will be ordered in descending, otherwise in ascending order (optional) </li>
     *     <li>  $offset   = integer, the number of the first result to be returned (optional) </li>
     *     <li>  $limit    = integer, maximum number of results to return (optional) </li>
     * </ul>
     *
     * Since version 2.8.5 the parameter $order_by has two synopsis.
     * <ul>
     *     <li>
     *         $order_by is the name of the column in the current table to order the resultset by.
     *     </li>
     *     <li>
     *         $order_by is a numeric array of strings, where each element
     *         is the name of a column in the current table.
     *         The resultset will get ordered by the values of these columns
     *         in the direction in which they are provided.
     *         This feature became available in version 2.8.5
     *     </li>
     * </ul>
     *
     * Since version 2.8.5 the parameter $where has two synopsis.
     * <ul>
     *     <li>
     *         $where is a comma seperated list of column-value pairs using the following syntax:
     *         <pre>column1=value1,column2=value2,...,columnN=valueN</pre>
     *         Note that comparison takes place using the SQL operator 'LIKE', except for
     *         columns that are keys (either a primary key or a foreign key), which are compared
     *         using the SQL operator '='.
     *     </li>
     *     <li>
     *         $where is a two-dimensional numeric array and its entries follow this syntax:
     *         <pre>array(array(0=>column1,1=>value1,2=>operator1),...)</pre>
     *         The "column" name is mandatory and needs to be a string.
     *         The "value" is mandatory and needs to be a scalar value.
     *         The "operator" is optional and defaults to 'LIKE'.
     *         It needs to be a string and can be one of the following:
     *         <pre> '=', 'LIKE', '<', '>', '!=', '<=', '>=', '==' (alias of '=') </pre>
     *     </li>
     * </ul>
     * The second alternative offers more flexibility and should be prefered over the first one.
     *
     * Note: The $desc parameter became available in version 2.8.5
     *
     * @access  public
     * @param   array|string  $key       the address of the value(s) to retrieve
     * @param   array|string  $where     a search string
     * @param   array         $order_by  a list of columns to order the resultset by
     * @param   int           $offset    the number of the first result to be returned
     * @param   int           $limit     maximum number of results to return
     * @param   bool          $desc      if true results will be ordered in descending, otherwise in ascending order
     * @return  mixed
     */
    function get($key, $where = array(), $order_by = array(), $offset = 0, $limit = 0, $desc = false)
    {
        /* 1.1 handle $key array */
        if (is_array($key)) {
            if (isset($key['where'])) {
                $where = $key['where'];
            } else if (isset($key['search'])) {
                $where = $key['search'];
            }
            if (isset($key['order_by'])) {
                $order_by = $key['order_by'];
            }
            if (isset($key['desc'])) {
                $desc = $key['desc'];
            }
            if (isset($key['offset'])) {
                $offset = $key['offset'];
            }
            if (isset($key['limit'])) {
                $limit = $key['limit'];
            }
            if (isset($key['key'])) {
                $key = $key['key'];
            } else {
                return false;
            }
        }
        /* 1.2 check input */
        /* 1.2.1 first argument */
        if (is_string($key)) {
            $key = strtolower($key);
            $table = explode('.', $key);
            $table = $table[0];
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($key)), E_USER_WARNING);
            return false;
        }
        /* 1.2.2 next argument */
        if (!is_string($where) && !is_array($where)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String or array', gettype($where)), E_USER_WARNING);
            return false;
        }
        /* 1.2.3 next argument */
        if (is_string($order_by)) {
            if (!empty($order_by)) {
                $order_by = array($order_by);
            } else {
                $order_by = array();
            }
        }
        if (is_array($order_by)) {
            if (count($order_by) < 1) {
                $order_by = "";
            } else {
                /* for strict structures also test if input is a column */
                if ($this->structure->isStrict()) {
                    assert('!isset($column);');
                    assert('!isset($m);');
                    foreach ($order_by as $column)
                    {
                        if (empty($column)) {
                            continue;
                        }
                        /* if a table name is provided */
                        if (is_string($column) && strpos($column,'.') !== false) {
                            preg_match('/^(.+)\.(.*)$/U',$column,$m);
                            if (!$this->structure->isColumn($m[1],$m[2])) {
                                trigger_error("The 'order_by' parameter is invalid. There is no column '".$m[2]."' in table '".$m[1]."'.",E_USER_WARNING);
                                return false;
                            }
                        /* if no table name is provided, use current table */
                        } else if (!$this->structure->isColumn($table,$column)) {
                            trigger_error("The 'order_by' parameter is invalid. There is no column '${column}' in table '${table}'.",E_USER_WARNING);
                            return false;
                        }
                    } /* end foreach */
                    unset($m, $column);
                } /* end if */
                /* quote ids */
                $order_by = array_map(array($this,'quoteId'),$order_by);
                /* convert array into a string */
                $order_by = 'order by '.implode(',',$order_by);
            } /* end if */
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 3, 'String or array', gettype($order_by)), E_USER_WARNING);
            return false;
        }
        $order_by = strtolower($order_by);
        /* 1.2.4 next argument */
        if (!is_int($offset)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 4, 'Integer', gettype($offset)), E_USER_WARNING);
            return false;
        } else {
            assert('$offset >= 0; /* SQL select statement: the offset '.$offset.' is out of range. */');
        }
        /* 1.2.5 next argument */
        if (!is_int($limit)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 5, 'Integer', gettype($limit)), E_USER_WARNING);
            return false;
        } else {
            assert('$limit >= 0; /* SQL select statement: the limit '.$limit.' is out of range. */');
        }
        /* 1.2.6 last argument */
        if (!is_bool($desc)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 6, 'Boolean', gettype($desc)), E_USER_WARNING);
            return false;
        } else {
            /* have values in descending order if $desc is set to true */
            if ($desc === true) {
                $order_by = trim($order_by.' desc');
            }
        }

        /* 1.3 check availability of database connection */
        if (!$this->_check_available()) {
            return false;
        }

        if ($offset < 0) {
            $offset = 0;
        }
        if ($limit < 0) {
            $limit = 0;
        }
        /* 2 handle request */
        $result = $this->_computeRequest($key, $where, NULL, $offset, $limit, $order_by);

        /* 3 return result */
        return $result;
    }

    /**
     * update or insert row
     *
     * insert $value at position $key
     *
     * If $key already exists, the previous value
     * gets updated, else the value is created
     *
     * This function returns bool(true) on success
     * and bool(false) on error. Note, that this
     * function does not auto-commit. This means,
     * changes to the database will NOT be saved
     * unless you call $DbStream->write().
     *
     * @access  public
     * @param   string  $key
     * @param   mixed   $value
     * @return  bool
     */
    function insert($key, $value)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($key); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ $key = (string) $key;

        $key = strtolower($key);

        if (!$this->_check_writeable()) {
            trigger_error('Operation aborted, not writeable.', E_USER_NOTICE);
            return false;
        }

        /* check whether the row has been modified since last access */
        if ($this->structure->isStrict()) {
            if (isset($_SESSION['transaction_isolation_created'])) {
                if ($this->_lastModified($key) > $_SESSION['transaction_isolation_created']) {
                    global $YANA;
                    if (isset($YANA)) {
                        $YANA->report( new Log("SQL ERROR: The user was trying to save form data, which has been changed since he accessed the corrsponding form. The operation has been aborted, as updating this row would have overwritten changes made by another user.") );
                        $YANA->report( new Alert("form_expired") );
                    } else {
                        $this->handleError("SQL ERROR: Input data has been changed since last access. Updating this row would overwrite existing changes. Insert/Update aborted for security reasons.");
                    }
                    return false;
                }
            }
        }

        if ($this->_computeRequest($key, "", $value)) {
            return Hashtable::set($this->cache, $key, $value);
        } else {
            $this->handleError("SQL ERROR: Insert on database failed for key '${key}'.", preg_replace('/\s/', ' ', SML::encode($value)));
            return false;
        }

    }

    /**
     * remove one row
     *
     * For security reasons all delete queries will automatically be limited to 1 row at a time.
     * While this might be seen as a limitation the far more valueable advantage is,
     * no user is able to destroy a whole table - wether by intention or by accident -
     * in only one query. (At least not via this API.)
     *
     * The function returns bool(true) on success and bool(false) on error.
     *
     * Since version 2.8.5 the parameter $where has two synopsis.
     * <ul>
     *     <li>
     *         $where is a comma seperated list of column-value pairs using the following syntax:
     *         <pre>column1=value1,column2=value2,...,columnN=valueN</pre>
     *         Note that comparison takes place using the SQL operator 'LIKE', except for
     *         columns that are keys (either a primary key or a foreign key), which are compared
     *         using the SQL operator '='.
     *     </li>
     *     <li>
     *         $where is a two-dimensional numeric array and its entries follow this syntax:
     *         <pre>array(array(0=>column1,1=>value1,2=>operator1),...)</pre>
     *         The "column" name is mandatory and needs to be a string.
     *         The "value" is mandatory and needs to be a scalar value.
     *         The "operator" is optional and defaults to 'LIKE'.
     *         It needs to be a string and can be one of the following:
     *         <pre> '=', 'LIKE', '<', '>', '!=', '<=', '>=', '==' (alias of '=') </pre>
     *     </li>
     * </ul>
     * The second alternative offers more flexibility and should be prefered over the first one.
     *
     * @access  public
     * @param   string        $key    the address of the row that should be removed
     * @param   string|array  $where  a search string
     * @return  bool
     */
    function remove($key, $in_where = array())
    {
        /**
         * 1) check input arguments
         */

        /* 1 check input */
        if (!is_string($key)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($key)), E_USER_WARNING);
            return false;
        }
        if (!is_string($in_where) && !is_array($in_where)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String or array', gettype($in_where)), E_USER_WARNING);
            return false;
        }

        /* untaint input */
        $key = strtolower($key);

        if (!$this->_check_writeable()) {
            trigger_error('Operation aborted, not writeable.', E_USER_NOTICE);
            return false;
        }

        $keyArray = explode(".", $key);
        /* settype to ARRAY */
        $keyArray = (array) $keyArray;
        if (!$this->structure->isTable($keyArray[0])) {
            trigger_error("Argument 1 is invalid, '".$keyArray[0]."' is not a table name.", E_USER_WARNING);
            return false;
        }
        $this->lastSelectedTable = $keyArray[0];

        if ((is_string($in_where) && !empty($in_where)) || (is_array($in_where) && count($in_where) > 0)) {
            $where = $this->_computeSearchString($keyArray[0], $in_where, count($keyArray) == 1);
            if ($where === false) {
                trigger_error("Argument 2 is invalid. Where available, see log file for details.", E_USER_WARNING);
                return false;
            }
        } else {
            $where = "";
        }

        if ((count($keyArray) < 2 && empty($where)) || (count($keyArray) > 2) || ($keyArray[1] == '0')) {
            trigger_error("Argument 1 is invalid. Must be a valid address.", E_USER_WARNING);
            return false;
        } else if  ($this->exec->constraint('DELETE', $keyArray[0], $keyArray[1]) === false) {
            trigger_error("Unable to delete row '".$keyArray[1]."' in table '".$keyArray[0]."'. Constraint check failed.", E_USER_WARNING);
            return false;
        }

        /**
         * 2) Security checks
         */
        /* check constraint */
        if ($this->exec->constraint('DELETE', $keyArray[0]) === false) {
            return false;

        /* check profile constraint*/
        } else if ($this->_checkProfile(YANA_DB_DELETE, YANA_DB_ROW, $where, @$dummy, $keyArray[0], $keyArray[1]) !== true) {
            return false;

        /* fire trigger */
        } else {
            assert('!isset($value); // cannot redeclare variable $value');
            $value = null;
            $this->exec->trigger(YANA_TRIGGER_BEFORE, 'DELETE', $keyArray[0], null, $value);
            unset($value);
        }

        /**
         * 3) Build SQL statement
         */

        /* create SQL */
        $sql = array();
        if ($keyArray[1] == "*") {
            /* remove last entry */
            $sql = array();
            $sql[0] = "DELETE FROM ".$this->quoteId(YANA_DATABASE_PREFIX.$keyArray[0]).( (!empty($where)) ? " WHERE $where " : '' )." order by ".$this->structure->getPrimaryKey($keyArray[0])." desc";
            $sql[1] = array('DELETE', $keyArray[0], null, null);
        } else {
            /* remove entry with id $keyArray[1] */
            if (isset($keyArray[1]) && $keyArray[1] != '') {
                if (!empty($where)) {
                    $where .= ' AND ';
                }
                $where .= $this->structure->getPrimaryKey($keyArray[0])." = ".$this->quote($keyArray[1]);
            }
            $sql = array();
            $sql[0] = "DELETE FROM ".$this->quoteId(YANA_DATABASE_PREFIX.$keyArray[0])." WHERE ".( ($where) ? " $where " : '' );
            $sql[1] = array('DELETE', $keyArray[0], null);
        }
        
        /* add SQL statment to queue and wait for commit */
        $this->queue[] = $sql;

        /**
         * 4) Logging
         */

        /* backup entry before deleting it */
        global $YANA;
        if (isset($YANA)) {
            assert('!isset($result); // cannot redeclare variable $result');
            $result = $this->get($key,$in_where,'',0,1);
            if (!empty($result)) {
                if (is_string($result)) {
                    $YANA->report(new Log('Deleting entry from '.$key.':', preg_replace('/\s/', ' ', $result)));
                } else {
                    $YANA->report(new Log('Deleting entry from '.$key.':', $result));
                }
            }
        }

        /**
         * 5) Handling of associated files and images
         *
         * Remove files associated with deleted entry.
         *
         * Note: can not (yet) be reversed. This will
         * apply immediately and NOT wait for commit!
         *
         * @todo improve this behaviour
         *
         * {@internal
         *
         * This might be done as soon as the query
         * builder is integrated, as it will allow
         * statements to be kept as objects and thus
         * required actions associated with them
         * can still be done at a later stage of
         * execution.
         *
         * }}
         */
        $columnList = $this->structure->getFiles($keyArray[0]);
        assert('!isset($dbDir); // Cannot redeclare $dbDir');
        $dbDir = $YANA->getVar('DBBLOB');
        foreach ($columnList as $column)
        {
            $result = $this->database->limitQuery("SELECT $column from ".$this->quoteId(YANA_DATABASE_PREFIX.$keyArray[0])." where ".$this->structure->getPrimaryKey($keyArray[0])." = ".$this->quote($keyArray[1]), 0, 1);
            if ($this->isError($result)) {
                $file = '';
            } else {
                $file = $result->fetchRow(DB_FETCHMODE_ASSOC, 0);
                assert('is_array($file); /* unexpected result: $file */');
                $file = $dbDir . array_pop($file);
            }
            if (is_file($file)) {
                unlink($file);
                $file = str_replace('image.', 'thumb.', $file);
                if (is_file($file)) {
                    unlink($file);
                }
            }
        } /* end foreach */

        /* return true to indicate the request was successfull */
        return true;
    }

    /**
     * join the resultsets for two tables
     *
     * Results in an INNER JOIN $table1, $table2 WHERE $table1.$key1 = $table2.$key2 .
     *
     * Note that if you ommit the parameters $key1 and $key2, the API will try to determine
     * the foreign key and target key itself by looking up the foreign key in the database's
     * structure file. The first foreign key association that matches will be used.
     *
     * Also note that two tables may only be joined via one pair of columns - not two or more.
     * Instead if you may add additional rules to the where clause as you see fit.
     *
     * Note that joins are permanent. So in opposition to what you might have learned from
     * common SQL statements and other APIs, you do not need to repeat joins for each query.
     * Instead, this API "remembers" what it was told and once set your joins will automatically
     * be used each time you query the table until you explicitly remove it.
     *
     * To remove all perviously set joins from a table, use the following function call:
     * <code> $dbStream->join('myTable'); </code>
     * As you can see above, if the second argument ($table2) is ommited, all joins bound
     * to 'myTable' are released.
     *
     * Also note, that the wildcard '*' may be used to refer to the "least recently used" table.
     * This is a shortcut that you may use in your scripts.
     *
     * @access  public
     * @param   string $table1  name of the table to join another one with
     * @param   string $table2  name of another table to join table1 with (when omitted will remove all previously set joins from table1)
     * @param   string $key1    name of the foreign key in table1 that references table2 (when omitted the API will look up the key in the structure file)
     * @param   string $key2    name of the key in table2 that is referenced from table1 (may be omitted if it is the primary key)
     * @return  bool
     */
    function join($table1, $table2 = null, $key1 = null, $key2 = null)
    {
        /* 1 check input */
        /* 1.1 argument 1 */
        if (!is_string($table1)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($table1)), E_USER_WARNING);
            return false;
        } else if (!preg_match("/^[\w\d-_]*$|^\*$/si", $table1)) {
            trigger_error("Argument 1 is invalid for function ".get_class($this)."->".__FUNCTION__."(). Must be a valid table name.", E_USER_WARNING);
            return false;
        } else {
            /* lower-case input */
            $table1 = strtolower($table1);
        }
        /* 1.2 argument 2 */
        if (!is_string($table2) && !is_null($table2)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($table2)), E_USER_WARNING);
            return false;
        } else if (!preg_match("/^[\w\d-_]*$|^\*$/si", $table2)) {
            trigger_error("Argument 2 is invalid for function ".get_class($this)."->".__FUNCTION__."(). Must be a valid table name.", E_USER_WARNING);
            return false;
        } else if (!is_null($table2)) {
            /* lower-case input */
            $table2 = strtolower($table2);
        }
        /* 1.3 argument 3 */
        if (!is_string($key1) && !is_null($key1)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 3, 'String', gettype($key1)), E_USER_WARNING);
            return false;
        } else if (!preg_match("/^[\w\d-_]*$|^\*$/si", $key1)) {
            trigger_error("Argument 3 is invalid for function ".get_class($this)."->".__FUNCTION__."(). Must be a valid column name.", E_USER_WARNING);
            return false;
        } else if (!is_null($key1)) {
            /* lower-case input */
            $key1   = strtolower($key1);
        }
        /* 1.4 argument 4 */
        if (!is_string($key2) && !is_null($key2)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 4, 'String', gettype($key2)), E_USER_WARNING);
            return false;
        } else if (!preg_match("/^[\w\d-_]*$|^\*$/si", $key2)) {
            trigger_error("Argument 4 is invalid for function ".get_class($this)."->".__FUNCTION__."(). Must be a valid column name.", E_USER_WARNING);
            return false;
        } else if (!is_null($key2)) {
            /* lower-case input */
            $key2   = strtolower($key2);
        }

        if (!$this->_check_available()) {
            return false;
        }

        if (empty($table1) || $table1 == "*") {
            if (empty($this->lastSelectedTable)) {
                trigger_error("Wrong parameter count for function ".get_class($this)."->".__FUNCTION__."(). Unable to join tables, since no table has been selected for function join.", E_USER_WARNING);
                return false;
            } else {
                $table1 = $this->lastSelectedTable;
            }
        }

        /* strict input checking */
        if ($this->structure->isStrict()) {

            if (!$this->structure->isTable($table1) || (!is_null($table2) && !$this->structure->isTable($table2))) {
                $this->handleError("Cannot join tables '${table1}' and '${table2}'. Table does not exist in structure file.");
                return false;
            }
            if (!is_null($key1) && !$this->structure->isColumn($table1, $key1)) {
                $this->handleError("Cannot join tables '${table1}' and '${table2}'. Field '${key1}' does not exist in table '${table1}'.");
                return false;
            }
            if (!is_null($table2) && !is_null($key2) && !$this->structure->isColumn($table2, $key2)) {
                $this->handleError("Cannot join tables '${table1}' and '${table2}'. Field '${key2}' does not exist in table '${table2}'.");
                return false;
            }
        }

        /* reset old association */
        if (isset($this->joins[$table1][$table2]['TKEY'])) {
            unset($this->joins[$table1][$table2]['TKEY']);
        }
        if (isset($this->joins[$table1][$table2]['FKEY'])) {
            unset($this->joins[$table1][$table2]['FKEY']);
        }

        /* create new association */
        if (!is_null($table2)) {
            $this->joins[$table1][$table2]['TABLE'] = $table2;
            if (!empty($key1)) {
                $this->joins[$table1][$table2]['FKEY'] = $key1;
                if (!empty($key2)) {
                    $this->joins[$table1][$table2]['TKEY'] = $key2;
                } else if ($this->structure->isForeignKey($table1,$key1)) {
                    $this->joins[$table1][$table2]['TKEY'] = $this->structure->getPrimaryKey($table2);
                }
            } else {
                assert('!isset($foreignKeys); /* Cannot redeclare variable $foreignKeys */');
                $foreignKeys = $this->structure->getForeignKeys($table1);
                assert('!isset($foreignTable); /* Cannot redeclare variable $foreignTable */');
                assert('!isset($foreignKey); /* Cannot redeclare variable $foreignKey */');
                foreach ($foreignKeys as $foreignKey => $foreignTable)
                {
                    if (strcasecmp($table2,$foreignTable) === 0) {
                        $this->joins[$table1][$table2]['FKEY']  = strtolower($foreignKey);
                        $this->joins[$table1][$table2]['TKEY']  = strtolower($this->structure->getPrimaryKey($foreignTable));
                        break;
                    }
                }
                unset($foreignKey,$foreignKeys,$foreignTable);
            }
            return true;
        /* if second argument is omitted, release association */
        } else {
            $this->joins[$table1] = array();
            return true;
        }

    }

    /**
     * optional API bypass
     *
     * Send a sql-statement directly to the PEAR database API, bypassing this API.
     *
     * Note: for security reasons this only sends one single SQL statement at a time.
     * This is done by checking the input for a semicolon followed by anything but
     * whitespace. If such input is found, an E_USER_WARNING is issued and the
     * function will return bool(false).
     *
     * While bypassing the API leaves nearly all of the input checking to you, this
     * is meant to prevent at least a minimum of the common SQL injection attacks.
     * A known attack is to try to terminate a current statement with ';' and afterwards
     * "inject" their own stuff as a second statement. The common attack vector usually
     * is unchecked form data.
     *
     * If you want to send a sequence of statements, call this function multiple times.
     *
     * The function will return bool(false) if the database connection or the
     * PEAR API is not available and otherwise will whatever PEAR sends back as the
     * result of your statement.
     *
     * Note: when database usage is disabled via the administrator's menu,
     * the PEAR-DB API can not be used and this function will return bool(false).
     *
     * The $offset and $limit arguments became available in version 2.8.8
     *
     * @access  public
     * @param   string  $sqlStmt  one SQL statement to execute
     * @param   int     $offset   the row to start from
     * @param   int     $limit    the maximum numbers of rows in the resultset
     * @return  mixed
     */
    function query($sqlStmt, $offset = 0, $limit = 0)
    {
        /* 1 check input type */
        if (!is_string($sqlStmt)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($sqlStmt)), E_USER_WARNING);
            return false;
        } else if (!is_int($offset) || $offset < 0) {
            trigger_error("Invalid argument 2. Must be a positive integer.", E_USER_WARNING);
            return false;
        } else if (!is_int($limit) || $limit < 0) {
            trigger_error("Invalid argument 3. Must be a positive integer.", E_USER_WARNING);
            return false;
        /* 2 check input syntax */
        } else if (is_int(strpos($sqlStmt, ';')) && preg_match("/;.*(?:select|insert|delete|update|create|alter|grant|revoke).*$/is", $sqlStmt)) {
            trigger_error("A semicolon has been found in the current input '${sqlStmt}', indicating multiple queries.\n\t\tAs this might be the result of a hacking attempt it is prohibited for security reasons and the queries won't be executed.", E_USER_WARNING);
            return false;
        /* 3 check if database connection is available */
        } else if (!$this->_check_available()) {
            return false;
        /* 4 send query to database */
        } else {
            if ($offset > 0 || $limit > 0) {
                return $this->database->limitQuery($sqlStmt, $offset, $limit);
            } else {
                return $this->database->query($sqlStmt);
            }
        }
    }

    /**
     * query building
     *
     * This function builds queries from input data.
     *
     * @access  private
     * @param   string        $key       address of the value
     * @param   string|array  $search    values for where clause
     * @param   mixed         $value     values to be inserted / updated
     * @param   int           $offset    positive number
     * @param   int           $limit     positive number
     * @param   string        $order_by  column name (including 'desc')
     * @return  mixed
     *
     * @ignore
     */
    function _computeRequest($key, $search = array(), $value = NULL, $offset = 0, $limit = 0, $order_by = "")
    {
        assert('is_string($key) && !empty($key);');
        assert('is_string($search) || is_array($search);');
        assert('is_int($offset) && $offset >= 0;  /* SQL select statement: the offset '.$offset.' is out of range. */');
        assert('is_int($limit) && $limit >= 0;    /* SQL select statement: the limit '.$limit  .' is out of range. */');
        assert('is_string($order_by);');

        global $YANA;

        /* settype to INTEGER
         *            INTEGER
         */
        $offset = (int) $offset;
        $limit  = (int) $limit;

        if (!$this->_check_available()) {
            trigger_error('Database not available', E_USER_NOTICE);
            return false;
        }

        $key = strtolower($key);
        $where = "";

        /* int */ $returnedType = YANA_DB_UNKNOWN; /* 0 = unknown, 1 = table,  2 = row,    3 = cell, 4 = column */
        /* int */ $queryType    = YANA_DB_UNKNOWN; /* 0 = unknown, 1 = select, 2 = update, 3 = insert */
        /* int */ $useWildcard  = YANA_DB_NONE;    /* 0 = none,    1 = foo.*,  2 = foo.*.foo */
        /* string */ $arrayKey  = '';

        /* 1) check input */

        /* 1.1 check if search is valid */
        if (!is_array($search)) {
            /* settype to STRING */
            $search = (string) $search;
            if (!preg_match("/^((, ?)?([\w\d-_\.])+=([\w\d-_\@\. \&;])+)*$/i", $search)) {
                trigger_error("Syntax error. The input '${search}' is not a valid search string.", E_USER_WARNING);
                return false;
            }
        }
        /* 1.2 check if key is valid */
        /* settype to STRING */
        $key = (string) $key;
        $key = strtolower($key);
        if (!preg_match("/^[\w\d-_]+(\.([\w\d-_]*|\*|\?)){0,}(\.\*)?$/i", $key)) {
            trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
            return false;
        }

        /* 2) prepare input for computation */

        if (preg_match("/\.\*?$/", $key)) {
            $useWildcard = YANA_DB_ROW_WILDCARD;
        } else if (preg_match("/[\w\d-_]+\.\?\.([\w\d-_]+)$/i", $key)) {
            $key = preg_replace("/([\w\d-_]+)\.\?\.([\w\d-_]+)$/i", '${1}.*.${2}', $key);
            $useWildcard = YANA_DB_COL_WILDCARD;
        } else if (preg_match("/[\w\d-_]+\.\*\.([\w\d-_]+)$/i", $key)) {
            $useWildcard = YANA_DB_COL_WILDCARD;
            $returnedType = YANA_DB_COLUMN;
        } else {
            $useWildcard = YANA_DB_NONE;
        }

        $key = preg_replace("/\.(\*)?$/", '', $key);
        $keyArray = explode(".", $key);

        /* 3) resolve foreign keys to get true adress */
        if (count($keyArray) > 3) {
            if ($this->structure->getType($keyArray[0], $keyArray[2]) == 'array') {
                $arrayKey = implode('.', array_slice($keyArray, 3));
                $keyArray = array_slice($keyArray, 0, 3);
            } else {
                while (count($keyArray) > 3 && $this->structure->isForeignKey($keyArray[0],$keyArray[2]))
                {
                    assert('!isset($a); /* cannot redeclare variable $a */');
                    $a = $this->_computeRequest($keyArray[0].".".$keyArray[1].".".$keyArray[2]);
                    if (!$a) {
                        trigger_error('Operation aborted due to invalid foreign key.', E_USER_NOTICE);
                        $this->handleError("SQL ERROR: Unable to resolve foreign key '${key}'. This may mean one of the associated tables contains inconsistent data. Check if foreign key association has been broken.");
                        return false;
                    }
                    assert('!isset($foreignTable); /* cannot redeclare variable $foreignTable */');
                    $foreignTable = $this->structure->getTableByForeignKey($keyArray[0], $keyArray[2]);
                    @ array_shift($keyArray); @ array_shift($keyArray); @ array_shift($keyArray);
                    array_unshift($keyArray, $a);
                    array_unshift($keyArray, $foreignTable);
                    unset($a, $foreignTable);
                }
            }
        } else {
            /* 3.1) inheritance */
            assert('!isset($primaryKey);');
            $primaryKey = strtoupper($this->structure->getPrimaryKey($keyArray[0]));
            if ($this->structure->isForeignKey($keyArray[0],$primaryKey)) {

                $foreignKeys = $this->structure->getForeignKeys($keyArray[0]);
                $foreignTable = $foreignKeys[$primaryKey];
                $this->join($keyArray[0],$foreignTable,$primaryKey,$primaryKey);
            }
        }

        if(count($keyArray) > 3) {
            trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
            return false;
        } else {
            $key = implode('.', $keyArray);
            if ($returnedType == YANA_DB_UNKNOWN) {
                $returnedType = count($keyArray);
            }
        }

        /* rename for better readability */
        assert('!isset($table); // cannot redeclare variable $table');
        $table = $keyArray[0];
        assert('!isset($row); // cannot redeclare variable $row');
        if (isset($keyArray[1])) {
            $row = strtoupper($keyArray[1]);
        } else {
            $row = "";
        }
        assert('!isset($column); // cannot redeclare variable $column');
        if (isset($keyArray[2])) {
            $column = $keyArray[2];
        } else {
            $column = "";
        }
        unset($keyArray);

        /* 4) now that there is a resolved adress: if $value is set, check first if it is valid before going on */

        if ($returnedType > YANA_DB_TABLE) {
            /* Let's check the primary key first */
            $primaryKey = $this->structure->getPrimaryKey($table);
            $row = $this->structure->untaintInput($table, $primaryKey, $row, YANA_ESCAPE_SLASHED);
        }

        if (!is_null($value)) {

            if (!$this->_check_writeable()) {
                trigger_error('Operation aborted, not writeable.', E_USER_NOTICE);
                return false;
            }

            if ($returnedType == YANA_DB_TABLE && ($useWildcard == YANA_DB_ROW_WILDCARD)) {
                $returnedType = YANA_DB_ROW;
                $queryType    = YANA_DB_INSERT;
            }
            if ($returnedType == YANA_DB_COLUMN) {
                switch (true)
                {
                    case !empty($column):
                        $returnedType = YANA_DB_CELL;
                    break;
                    case !empty($row):
                        $returnedType = YANA_DB_ROW;
                    break;
                    default:
                        $returnedType = YANA_DB_TABLE;
                    break;
                }
            }

            /* Decide wether value is updated or inserted. */
            if ($queryType === YANA_DB_UNKNOWN) {
                if ($returnedType == YANA_DB_CELL) {
                    /* Inserting/Updating a cell */
                    if ($this->exists($table.'.'.$row)) {
                        $queryType = YANA_DB_UPDATE;
                    } else {
                        $queryType = YANA_DB_INSERT;
                    }
                } else {
                    /* Inserting/Updating smt. else */
                    if ($this->exists($key)) {
                        $queryType = YANA_DB_UPDATE;
                    } else {
                        $queryType = YANA_DB_INSERT;
                    }
                }
            } /* end if */

            switch ($returnedType)
            {

                /* Inserting/Updating a table */
                case YANA_DB_TABLE:
                    if ($this->structure->isStrict()) {
                        $this->handleError("Inserting/Updating a table is illegal in strict mode. Operation aborted.");
                    }
                    trigger_error('Illegal operation, cannot insert/update a table.', E_USER_NOTICE);
                    return false;
                break;

                /* Updating a row */
                case YANA_DB_ROW:
                    /* settype to ARRAY */
                    $value = (array) $value;
                    $value = array_change_key_case($value, CASE_LOWER);

                    /* autonumber */
                    if ($queryType == YANA_DB_INSERT && $useWildcard == YANA_DB_NONE) {
                        $primaryKey = $this->structure->getPrimaryKey($table);
                        if (!$this->structure->isAutonumber($table, $primaryKey)) {
                            $value[$primaryKey] = $row;
                            unset($primaryKey);
                        }
                    } /* end if */

                    /* untaint raw data */
                    assert('!isset($a); /* cannot redeclare variable $a */');
                    assert('!isset($b); /* cannot redeclare variable $b */');
                    foreach ($value as $a => $b)
                    {
                        if ($queryType == YANA_DB_UPDATE && $this->structure->isReadonly($table, $a)) {
                            $this->handleError("SQL ERROR: Database is readonly. Insert/Update operation aborted.");
                            return false;
                        }

                        /* untaint input */
                        $this->transformer->encode($value[$a], $table, $a, @$row, $this);

                        /* if it is a foreign key, check wether there is a field with that id */
                        $foreignTable = $this->structure->getTableByForeignKey($table, $a);
                        if ($this->structure->isForeignKey($table, $a) && !$this->exists("$foreignTable.$b")) {
                            /* OK - there is none. Now let's check if there is one recently inserted field in cache */
                            if (!isset($this->cache[$foreignTable][$b])) {
                                $this->handleError("SQL Error: The value '${b}' for attribute '${a}' refers to a non-existing entry in table '${foreignTable}'. Insert/Update operation aborted.");
                                return false;
                            }
                        }
                        unset($foreignTable);
                    } /* end foreach */
                    unset($a,$b);

                    /* check if all required data is available */
                    if ($queryType == YANA_DB_INSERT) {
                        if ($this->structure->checkRow($table, $value) === false) {
                            return false;
                        }
                    } else if ($queryType == YANA_DB_UPDATE) {
                        if ($this->structure->checkRow($table, $value, false) === false) {
                            return false;
                        }
                    } /* end if */
                break;

                /* Updating a cell */
                case YANA_DB_CELL:

                    /* untaint raw data */
                    if ($queryType == YANA_DB_UPDATE && $this->structure->isReadonly($table, $column)) {
                        $this->handleError("SQL ERROR: Database is readonly. Insert/Update operation aborted.");
                        return false;
                    }

                    /* untaint value */
                    $this->transformer->encode($value, $table, $column, $row, $this);

                    /* if it is a foreign key, check wether there is a field with that id */
                    if ($this->structure->isForeignKey($table, $column)) {
                        $foreignTable = $this->structure->getTableByForeignKey($table, $column);
                        if (!$this->exists("${foreignTable}.${value}")) {
                            /* OK - there is none. Now let's check if there is one recently inserted field in cache */
                            if (!isset($this->cache[$foreignTable][$value])) {
                                $this->handleError("SQL Error: The value '${value}' for attribute '".$column."' refers to a non-existing entry in table '${foreignTable}'. Insert/Update operation aborted.");
                                return false;
                            } /* end if */
                        } /* end if */
                    } /* end if */

                break;

                /* error */
                default:
                    trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
                    return false;
                break;

            } /* end switch */

        } else {
            /* $value is not set - since there no value to set this will be a simple read operation */
            $queryType = YANA_DB_SELECT;
        } /* end if */

        $this->lastSelectedTable = $table;

        /* 5) build sql-statement */

        switch ($queryType)
        {
            case YANA_DB_SELECT:
                $sqlStmt = "SELECT";
                if (empty($column)) {
                    $sqlStmt .= " * ";
                } else {
                    $sqlStmt .= " ".$column." ";
                }
                $sqlStmt .= "FROM ".$this->quoteId(YANA_DATABASE_PREFIX.$table)." ";
            break;

            case YANA_DB_UPDATE:
                if ($returnedType == YANA_DB_ROW) {
                    if (!is_array($value)) {
                        trigger_error("Syntax error. The argument \$value '${value}' is not valid.", E_USER_WARNING);
                        return false;
                    }
                    $sqlStmt = $this->_buildUpdate($table, $value);
                } else {
                    $sqlStmt = $this->_buildUpdate($table, array($column => $value));
                }
            break;

            case YANA_DB_INSERT:
                if ($returnedType == YANA_DB_ROW) {
                    $temp = $this->structure->getPrimaryKey($table);
                    if (!isset($value[$temp]) && !empty($row)) {
                        $value[$temp] = $row;
                    }
                    $sqlStmt = 'INSERT INTO '.$this->quoteId(YANA_DATABASE_PREFIX.$table).' '.$this->_sqlValues($value).' ';
                    unset($temp);
                } else {
                    $this->handleError("SQL ERROR: Invalid insert operation. You may only insert rows (not fields or tables).");
                    return false;
                }
            break;

            default:
                trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
                return false;
            break;

        }

        /* 6) resolve joins */

        if ($queryType < YANA_DB_INSERT && $queryType > YANA_DB_UNKNOWN && ($useWildcard == YANA_DB_NONE || $useWildcard == YANA_DB_ROW_WILDCARD)) {
            if (!empty($row)) {
                $where = $this->structure->getPrimaryKey($table)." = ".$this->quote(addslashes($row))." ";
            } else {
                $where = "";
            }
        }

        if ($queryType == YANA_DB_SELECT) {
            assert('!isset($JOIN); /* cannot redeclare variable $JOIN */');
            $JOIN  = array();
            if (isset($this->joins[$table])) {
                $JOIN = $this->joins[$table];
            }
            /* settype to ARRAY */
            $JOIN = (array) $JOIN;

            assert('!isset($b); /* cannot redeclare variable $b */');
            foreach ($JOIN as $b)
            {
                if (count($JOIN) == 1 && empty($b['TKEY']) && empty($b['FKEY'])) {
                    $sqlStmt .= "NATURAL JOIN ".$this->quoteId(YANA_DATABASE_PREFIX.$b['TABLE'])." ";
                } else {
                    $sqlStmt .= ", ".$this->quoteId(YANA_DATABASE_PREFIX.$b['TABLE'])." ";
                    if (!empty($b['FKEY']) && !empty($b['TKEY'])) {
                          $where .= ( (!empty($where)) ? "AND " : "" ).YANA_DATABASE_PREFIX.$table.'.'.$b['FKEY'].' = '.YANA_DATABASE_PREFIX.$b['TABLE'].'.'.$b['TKEY'].' ';
                    } else if (!empty($b['FKEY'])) {
                          $where .= ( (!empty($where)) ? "AND " : "" ).YANA_DATABASE_PREFIX.$table.'.'.$b['FKEY'].' = '.YANA_DATABASE_PREFIX.$b['TABLE'].'.'.$b['FKEY'].' ';
                    } else {
                          /* intenionally left blank */
                    }
                }
            } /* end foreach */
            unset($b);
        }

        /* 7) resolve search clause */

        /* 7.a) process search string */
        $where = trim($where);
        if ((is_string($search) && !empty($search)) || (is_array($search) && count($search) > 0)) {
            if (!empty($where)) {
                $where .= ' AND ';
            }
            if (empty($row) || $row == '?' || $row == '*') {
                $where .= $this->_computeSearchString($table, $search, true);
            } else {
                $where .= $this->_computeSearchString($table, $search, false);
            }
        }

        /* 7.b) check for profile constraint */
        if ($this->_checkProfile($queryType, $returnedType, $where, $value, $table, $row, $column) !== true) {
            return false;
        }

        /* 7.c) attach where clause */
        $sqlStmt .= ( (!empty($where)) ? "WHERE $where " : " " ).$order_by;

        /* 8) check constraints and triggers */
        assert('!isset($triggerArgs);');
        /* array */ $triggerArgs = null; /* 0 => OPERATION, 1 => TABLE, 2 => COLUM */
        switch ($queryType)
        {
            /**
             * 1) Operation select has no constraints or triggers
             */
            case YANA_DB_SELECT:
                /* intentionally left blank */
            break;
            /**
             * 2) before update: check constraints and triggers
             */
            case YANA_DB_UPDATE:
                if ($this->exec->constraint('UPDATE', $table, $column, $value) === false) {
                    return false;
                } else {
                    $this->exec->trigger(YANA_TRIGGER_BEFORE, 'UPDATE', $table, $column, $value);
                }
                $triggerArgs = array('UPDATE', $table, $column);
            break;
            /**
             * 3) before insert: check constraints and triggers
             */
            case YANA_DB_INSERT:
                if (@$this->exec->constraint('INSERT', $table, $column, $value) === false) {
                    return false;
                } else {
                    @$this->exec->trigger(YANA_TRIGGER_BEFORE, 'INSERT', $table, $column, $value);
                }
                $triggerArgs = array('INSERT', $table, $column);
            break;
                /**
                 * 4) syntax error
                 */
            default:
                trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
                return false;
            break;
        }

        /* 9) store / evaluate query */
        switch ($queryType)
        {
            /**
             * 1) update and insert are stored and will wait for commit
             */
            case YANA_DB_UPDATE:
            case YANA_DB_INSERT:

                if (is_null($triggerArgs)) {
                    $this->queue[] = $sqlStmt;
                } else {
                    $this->queue[] = array($sqlStmt, $triggerArgs);
                }
                return true;

            break;
            /**
             * 2) issue select query and format output
             */
            case YANA_DB_SELECT:

                if ($offset > 0 || $limit > 0) {
                    $result = $this->database->limitQuery($sqlStmt, $offset, $limit);
                } else {
                    $result = $this->database->query($sqlStmt);
                }
                if ($this->isError($result)) {
                    $this->handleError($result, "SQL WARNING: Query '${sqlStmt}' on database failed");
                } else {
                    $REF = array();
                    $id  = $this->structure->getPrimaryKey($this->getTable());
                    assert('!isset($dataTransformer); // Cannot redeclare var $dataTransformer');

                    for ($i = 0; $i < $result->numRows(); $i++)
                    {
                        $row = $result->fetchRow(DB_FETCHMODE_ASSOC,$i);
                        if (!is_array($row)) {
                            $this->handleError("DB NOTICE: The returned data for statement '${sqlStmt}' must be of type array. Instead database returned the following value '${row}'. The result was considered to be an error.");
                            break;
                        }
                        if ($returnedType === YANA_DB_TABLE) {
                            assert('!isset($refKey);');
                            if (!empty($REF[strtoupper($row[$id])])) {
                                $REF[strtoupper($row[$id])] = array($REF[strtoupper($row[$id])]);
                                $REF[strtoupper($row[$id])][1] = false;
                                $refKey =& $REF[strtoupper($row[$id])][1];
                            } else {
                                $refKey =& $REF[strtoupper($row[$id])];
                            }
                        }
                        assert('!isset($a); /* cannot redeclare variable $a */');
                        assert('!isset($b); /* cannot redeclare variable $b */');
                        foreach ($row as $a => $b)
                        {
                            if (!is_null($b)) {
                                switch ($returnedType)
                                {
                                    case YANA_DB_TABLE:
                                        if (!$this->transformer->decode($refKey[strtoupper($a)], $table, $a, $b)) {
                                            unset($REF[strtoupper($row[$id])]);
                                            break;
                                        }   
                                    break;

                                    case YANA_DB_ROW:
                                        if (!$this->transformer->decode($REF[strtoupper($a)], $table, $a, $b)) {
                                            return array();
                                        }
                                    break;

                                    case YANA_DB_CELL:
                                        if (!$this->transformer->decode($REF, $table, $column, $b, $arrayKey)) {
                                            return array();
                                        }
                                    break;

                                    case YANA_DB_COLUMN:
                                        $this->transformer->decode($REF[$i], $table, $column, $b);
                                    break;

                                    default:
                                        trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
                                        return false;
                                    break;
                                } /* end switch */
                            } /* end if */
                        } /* end foreach */
                        unset($a,$b,$refKey);
                    } /* end for */
                    return $REF;
                } /* end else */
                return false;
            break;
            /**
             * 3) syntax error
             */
            default:
                trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
                return false;
            break;
        } /* end switch $queryType */
    }

    /**
     * get the most recently queried table
     *
     * @access  public
     * @return  string
     */
    function getTable()
    {
        assert('is_string($this->lastSelectedTable);');
        return $this->lastSelectedTable;
    }

    /**
     * get the number of entries inside a table
     *
     * @access  public
     * @param   string        $table   name of a table
     * @param   string|array  $search  optional where clause
     * @return  int
     */
    function length($table = "", $search = array())
    {
        /* 1 check input */
        if (!is_string($table)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($table)), E_USER_WARNING);
            return false;
        }
        if (!is_string($search) && !is_array($search)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String or array', gettype($search)), E_USER_WARNING);
            return false;
        }

        if (!$this->_check_available()) {
            return false;
        }

        /* 2 check table parameter */
        $table = strtolower($table);
        if ($table == "" || $table == "*") {
            if (empty($this->lastSelectedTable)) {
                return false;
            } else {
                $table = $this->lastSelectedTable;
            }
        }
        if (!$this->exists($table)) {
            return 0;
        }

        /* 3 check search parameter */
        if ((is_string($search) && !empty($search)) || (is_array($search) && count($search) > 0)) {
            $search = $this->_computeSearchString($table, $search, true);
            if ($search === false) {
                return false;
            }
        } else {
            $search = "";
        }

        $dummy1 = false;
        $this->_checkProfile(YANA_DB_SELECT, YANA_DB_TABLE, $search, $dummy1, $table);
        unset($dummy1);

        /* 4 query database */
        $result = $this->database->query("SELECT COUNT(*) FROM ".$this->quoteId(YANA_DATABASE_PREFIX.$table). ( (!empty($search)) ? " WHERE ".$search : "" ) );

        /* 5 return result */
        if ($this->isError($result)) {
            return false;
        } else {
            $i = $result->fetchRow(DB_FETCHMODE_ORDERED,0);
            if (is_null($i) || !isset($i[0])) {
                return false;
            } else {
                return (int) $i[0];
            }
        }
    }

    /**
     * check wether a certain table has no entries
     *
     * Note: if no table is provided, the most recently used
     * table will be tested instead.
     *
     * @access  public
     * @param   string   $table  name of a table
     * @return  bool
     */
    function isEmpty($table = "")
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($table); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ $table = (string) $table;

        $table = strtolower($table);

        if ($table == "" || $table == "*") {
            if (!$this->lastSelectedTable) {
                return false;
            } else {
                $table = $this->lastSelectedTable;
            }
        }

        $length = (int) $this->length($table);
        return ($length == 0);
    }

    /**
     * Check wether a certain key exists
     *
     * Returns bool(true), if the adress $key (table, row, column) is defined, and otherwise bool(false).
     * If no argument is provided, the function returns bool(true) if a database connection exists and bool(false) if not.
     *
     * @uses $DbStream->exists('table.5')
     *
     * @access  public
     * @param   string  $key  adress to check
     * @return  bool
     */
    function exists($key = "")
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($key); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ $key = (string) $key;

        $key = preg_replace("/\.(\*)?$/", "", $key);
        $key = strtolower($key);

        if ($key == "" || $this->exists == YANA_DATABASE_FAILURE) {
            if ($this->exists == YANA_DATABASE_FAILURE) {
                return false;
            } else {
                return $this->exists;
            }
        } else {
            $key = strtolower($key);
            $key = explode('.', $key);
            /* int */ $result = 0;
            if (count($key) == 1) {
                return $this->structure->isTable($key[0]);
            } else if (count($key) < 0) {
                return false;
            } else if (count($key) == 2) {
                $result = $this->database->query("SELECT 1 FROM ".$this->quoteId(YANA_DATABASE_PREFIX.$key[0])." WHERE ".$this->structure->getPrimaryKey($key[0])." = ".$this->quote($key[1]));
            } else if (count($key) == 3) {
                if ($key[1] === '*') {
                    $result = $this->database->query("SELECT 1 FROM ".$this->quoteId(YANA_DATABASE_PREFIX.$key[0])." WHERE ".$key[2]." IS NOT NULL");
                } else {
                    $result = $this->database->query("SELECT 1 FROM ".$this->quoteId(YANA_DATABASE_PREFIX.$key[0])." WHERE ".$this->structure->getPrimaryKey($key[0])." = ".$this->quote($key[1])." AND ".$key[2]." IS NOT NULL");
                }
            } else {
                /* syntax error */
                return false;
            }

            if ($this->isError($result)) {
                return false;
            } else {
                if ($result->numRows() > 0) {
                    return true;
                } else {
                    return false;
                }
            }
        }
    }

    /**
     * check wether the current database is readonly
     *
     * @uses $DbStream->isWriteable()
     *
     * @access  public
     * @return  bool
     */
    function isWriteable()
    {
        if ($this->exists() && !$this->structure->isReadonly()) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * resolve search clause
     *
     * Parses the input string and returns the result as a string.
     *
     * @access  private
     * @param   string        $table        name of a table
     * @param   string|array  $search       where clause
     * @param   bool          $isTableScan  if scanning a whole table you may search for primary keys - otherwise not
     * @return  string
     * @ignore
     */
    function _computeSearchString($table, $search, $isTableScan = false)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('preg_match("/^\w+$/s",$table);           // invalid table name, may only contain alpha-numeric characters and can not be empty');
        /* argument 2 */ assert('is_string($search) || is_array($search); // wrong argument type for argument 2, string expected');
        /* argument 3 */ assert('is_bool($isTableScan);                   // wrong argument type for argument 3, boolean expected');

        /* optimistic type casting */
        /* argument 1 */
        /* settype to STRING */
        $table = (string) $table;
        $table = strtolower($table);
        /* argument 2 */
        if (!is_array($search)) {
            /* settype to STRING */
            $search = (string) $search;
        }
        /* argument 3 */
        /* settype to BOOLEAN */
        $isTableScan = (bool) $isTableScan;

        /* convert string to array */
        if (is_string($search)) {
            if (!preg_match("/^((, ?)?([\w\d-_\.])+=([\w\d-_\.\@ \&;])+)*$/i", $search)) {
                $this->handleError("SQL ERROR: Invalid where clause. The search string you provided contains syntax errors. Make sure it does only contain alpha-numeric characters and matches the following syntax: fieldname1=value1,fieldname2=value2");
                return false;
            }

            $search = preg_split("/, ?/", $search);
            /* settype to ARRAY */
            $search = (array) $search;
            assert('!isset($a); /* cannot redeclare variable $a */');
            assert('!isset($b); /* cannot redeclare variable $b */');
            foreach ($search as $a => $b)
            {
                $search[$a] = explode("=", $b);
                if ($this->structure->isForeignKey($table, $search[$a][0]) || $this->structure->isPrimaryKey($table, $search[$a][0])) {
                    $search[$a][2] = '=';
                } else {
                    $search[$a][1] = '%'.$search[$a][1].'%';
                    $search[$a][2] = 'like';
                }
            } /* end foreach */
            unset($a,$b);
        }

        /* resolve array */
        $result = "";
        if (count($search) > 0) {
            assert('!isset($a); /* cannot redeclare variable $a */');
            assert('!isset($b); /* cannot redeclare variable $b */');
            foreach ($search as $a => $b)
            {
                if (is_array($b) && count($b) >= 2 && count($b) <= 3 && !empty($b[0]) && isset($b[1])) {

                    /* use names that are easier to read */
                    $columnName = $b[0];
                    $value      = $b[1];
                    if (isset($b[2])) {
                        $operator   = $b[2];
                    }

                    /* settype to STRING
                     *            STRING
                     *            STRING
                     */
                    $columnName = (string) $columnName;
                    $value      = (string) $value;
                    $operator   = (string) $operator;

                    /* lower case input */
                    $columnName = strtolower($columnName);
                    $operator   = strtolower($operator);

                    /* map aliases */
                    if ($operator === '==') {
                        $operator = '=';
                    }

                    /* catch exceptions */
                    if (!in_array($operator, array('=', 'like', '<', '>', '!=', '<=', '>='))) {
                        $this->handleError("SQL ERROR: Invalid SQL statement. The operator '${operator}' is not supported.");
                        return false;
                    }
                    if (!$isTableScan && $this->structure->isStrict() && $this->structure->isPrimaryKey($table,$columnName)) {
                        $this->handleError("SQL ERROR: Invalid SQL statement. You are trying to search for a primary key. This is not allowed, since it might cause results wether to be ambigious or empty. Try setting 'USE_STRICT' to 'false' if you want to turn this message off and continue without checking.");
                        return false;
                    }
                    if ($this->structure->isStrict() && !$this->structure->isColumn($table,$columnName) && !$this->structure->isForeignKey($table,$columnName)) {
                        $this->handleError("SQL ERROR: Field '".$columnName."' does not exist in table '".$table."'. Perhaps your structure file is incomplete. Try setting 'USE_STRICT' to 'false' if you want to turn this message off and continue without checking.");
                        return false;
                    }

                    /* create result */
                    $result .= ( (!empty($result)) ? "AND " : "" ).$columnName." ".$operator." ".$this->quote(addslashes($value))." ";
                }
            } /* end foreach */
            unset($a,$b);
        }

        return $result;
    }

    /**
     * import SQL from a file
     *
     * The input parameter $sqlFile can wether be filename,
     * or a numeric array of SQL statements.
     *
     * Returns bool(true) on success or bool(false) on error.
     * Note that the statements are executed within a transaction.
     * If the function fails,
     *
     * An error is encountered and an E_USER_NOTICE is issued, if:
     * <ul>
     * <li> the file does not exist or is not readable </li>
     * <li> the $sqlFile parameter is empty </li>
     * <li> the database connection is not available </li>
     * <li> the parameter "readonly" on the database structure file is set to "true" </li>
     * <li> at least one database statement failed (does not issue an E_USER_NOTICE) </li>
     * <li> there are uncommited statements in the queue </li>
     * </ul>
     *
     * @uses $DbStream->importSQL('some_file.sql')
     *
     * @access  public
     * @return  bool
     * @name    DbStream::importsql()
     * @param   string|array  $sqlFile
     */
    function importSQL($sqlFile)
    {
        if (!is_string($sqlFile) && !is_array($sqlFile)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String or array', gettype($sqlFile)), E_USER_WARNING);
            return false;
        }
        if (empty($sqlFile)) {
            trigger_error("Argument \$sqlFile is empty in ".getclass($this)."->".__FUNCTION__.".", E_USER_NOTICE);
            return false;
        }
        if (!empty($this->queue)) {
            trigger_error("Cannot import SQL statements in ".getclass($this)."->".__FUNCTION__.".\n\t\tThere is a pending transaction that needs to be commited before proceeding.", E_USER_NOTICE);
            return false;
        }

        if ($this->_check_writeable() === true) {
            if ($this->structure->isReadonly()) {
                trigger_error("Database is readonly. SQL import aborted.", E_USER_NOTICE);
                return false;
            } else if (is_string($sqlFile) && !is_readable($sqlFile)) {
                trigger_error("The file '${sqlFile}' is not readable.", E_USER_NOTICE);
                return false;
            } else if (is_array($sqlFile)) {
                $this->queue = $sqlFile;
                if ($this->write()) {
                    return true;
                } else {
                    return false;
                }
            } else if (!empty($sqlFile)) {
                $raw_data = @fread(@fopen($sqlFile, 'r'), @filesize($sqlFile));
                /* remove comments and line breaks */
                $raw_data = preg_replace("/\s*\#[^\n]*/i", "", $raw_data);
                $raw_data = preg_replace("/\s*\-\-[^\n]*/i", "", $raw_data);
                $raw_data = preg_replace("/;\s*\n\s*/i", "[NEXT_COMMAND]", $raw_data);
                $raw_data = preg_replace("/\s/", " ", $raw_data);
                /* add items */
                $this->queue = explode("[NEXT_COMMAND]", $raw_data);
                if ($this->write()) {
                    return true;
                } else {
                    return false;
                }
            } else {
                return true;
            }
        } else {
            trigger_error("Database connection is not available. Check your connection settings.", E_USER_NOTICE);
            return false;
        }
    }

    /**
     * get CSV string from a table
     *
     * @uses $DbStream->toString('guestbook')
     *
     * @access  public
     * @param   string  $table  name of a table
     * @return  string
     */
    function toString($table = "")
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($table); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ $table = (string) $table;

        /* lower-case input */
        $table = strtolower($table);

        if ($this->exists === YANA_DATABASE_FAILURE || $this->exists === YANA_DATABASE_NOT_AVAILABLE) {
            return "No Database Connection: Database structure file not available or not loaded";
        }

        if ($table == "" || $table == "*") {
            if (empty($this->lastSelectedTable)) {
                return false;
            } else {
                $table = $this->lastSelectedTable;
            }
        }

        assert('!isset($text); /* cannot redeclare variable $text */');
        $text = "";

        if (!$this->exists($table)) {
            trigger_error("The selected table '${table}' does not exist.", E_USER_WARNING);
            return "";
        } else {

            assert('!isset($a); /* cannot redeclare variable $a */');
            $a = $this->get($table);
            if(is_array($a)) {
                assert('!isset($k); /* cannot redeclare variable $k */');
                assert('!isset($e); /* cannot redeclare variable $e */');
                foreach ($a as $k => $e)
                {
                    assert('!isset($i); /* cannot redeclare variable $i */');
                    for ($i = 0; $i < count($e); $i++)
                    {
                        $b = current($e);
                        /* settype to STRING */
                        $b = (string) $b;
                        $text .= $b;
                        if (next($e)) {
                              $text .= ", ";
                        }
                    } /* end for */
                    unset($i);
                    $text .= "\n";
                } /* end foreach */
                unset($k,$e);
            } /* end if */
            unset($a);
            return $text;
        }
    }

    /**
     * export database structure to a file
     *
     * Note: if the specified file already exists it will get
     * overwritten.
     *
     * Structure files need to have the extension '.config' and are
     * to be stored at the directory 'config/db/*.config'.
     *
     * @uses $DbStream->exportStructure('some_file.config')
     *
     * @access  public
     * @param   string  $filename  relative path to output-file
     * @return  bool
     */
    function exportStructure($filename)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($filename); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ $filename = (string) $filename;

        if (!preg_match("/\.config$/", $filename)) {
            trigger_error("Unexpected file extension. The filename '${filename}' should end with '.config'.", E_USER_WARNING);
            return false;
        }
        $file = new SML($filename, CASE_UPPER);
        if (!$file->exists()) {
            $file->create();
        }
        $file->read();
        if (isset($this->structure) && $file->set($this->structure->getStructure()) && $file->write()) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * handleError
     *
     * @access  private
     * @param   mixed   $error
     * @param   mixed   $additional_info (optional)
     * @ignore
     */
    function handleError($error, $additional_info = "")
    {
        global $YANA;

        if ($this->isError($error)) {
            if (isset($YANA)) {
                $YANA->report( new Log("$additional_info", $error->getMessage()) );
            }
            if (is_string($additional_info)) {
                trigger_error($error->getMessage()." $additional_info", E_USER_WARNING);
            } else {
                trigger_error($error->getMessage(), E_USER_WARNING);
            }
        } else if (is_string($error)) {
            if (isset($YANA)) {
                $YANA->report( new Log("${error}",$additional_info) );
            } else {
                if (is_string($additional_info)) {
                    trigger_error("$error $additional_info", E_USER_WARNING);
                } else {
                    trigger_error("$error", E_USER_WARNING);
                }
            }
        }
    }

    /**
     * buildStructure
     *
     * {@internal
     *
     * WARNING: be aware that this won't work for all DBMS since
     * the type and ammount of information provided might vary.
     * ALSO: be aware that this won't work with PEAR-DB prior
     * to version 1.6.0
     *
     * ... or in other words: avoid this function in any productive
     * environment.
     *
     * It is only there as a "last-resort" in case the structure file
     * is not there as well as for debugging, testing and reverse
     * engineering your own database to a structure file when writing
     * new plugins (and by doing so possibly making your work a bit easier).
     *
     * }}
     *
     * @access  private
     * @ignore
     */
    function buildStructure()
    {
        if (!isset($this->structure)) {
            $this->structure  = new StructureFile($this->getPath());
            $this->structure->read();
        }
        $tables = $this->database->getListOf('tables');
        if (!$this->isError($tables)) {
            assert('!isset($table); /* cannot redeclare variable $table */');
            foreach ($tables as $table)
            {
                $result = $this->database->query('SELECT * FROM '.$this->quoteId(YANA_DATABASE_PREFIX.$table).' WHERE 1=0');
                if ($this->isError($result)) {
                    $this->handleError("DB WARNING: unable to read structure of table '${table}'.");
                } else {
                    $result = $this->database->tableInfo($result, DB_TABLEINFO_ORDER);
                    if (!is_array($result)) {
                        $this->handleError("DB WARNING: unable to read structure of table '${table}'.");
                    } else {
                        $this->structure->addStructure($table, $result);
                    }
                }
            } /* end foreach */
            unset($table);
        } else {
            $this->handleError("DB WARNING: unable to read structure of database.");
        }
    }

    /**
     * read
     *
     * @deprecated  since version 2.8.3
     * @access      private
     * @return      bool
     * @ignore
     */
    function read()
    {
        if ($this->exists != YANA_DATABASE_OK) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * available
     *
     * @access  private
     * @return  bool
     * @ignore
     */
    function available()
    {
        if (($this->exists == YANA_DATABASE_OK && !$this->read()) || empty($this->database)) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * get path and name of current structure file
     *
     * @access  protected
     * @return  string
     * @ignore
     */
    function getPath()
    {
        return $this->path;
    }

    /**
     * _sqlValues
     *
     * @access  protected
     * @param   array    $value
     * @return  string
     * @ignore
     */
    function _sqlValues($value)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_array($value); // wrong argument type for argument 1, array expected');

        /* optimistic type casting */
        /* argument 1 */ $value = (array) $value;

        $k_value = array_keys($value);
        $v_value = array_values($value);

        $newfunc = create_function('&$item,$key,&$obj', '$item = $obj->quote($item);');
        array_walk($v_value, $newfunc, $this);
        $result  = '('.join(',', $k_value).')';
        $result .= ' VALUES('.join(',',$v_value).')';

        assert('is_string($result) && !empty($result);');
        return $result;
    }

    /**
     * isError
     *
     * @access  private
     * @param   mixed   $result
     * @return  bool
     * @ignore
     */
    function isError($result)
    {
        return DB::isError($result);
    }

    /**
     * quote a value
     *
     * Returns the quoted values as a string
     * surrounded by delimiters, depending on
     * the DBMS selected.
     *
     * @access  public
     * @param   mixed  $value
     * @return  string
     * @ignore
     */
    function quote($value)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_scalar($value); // wrong argument type for argument 1, scalar expected');

        /* optimistic type casting */
        if (!is_scalar($value)) {
            trigger_error("Your SQL statement contains an unexpected non-scalar value: '".print_r($value,true)."'", E_USER_NOTICE);
            /* argument 1 */ $value = (string) $value;
        }
        return $this->database->quoteSmart($value);
    }

    /**
     * smart id quoting
     *
     * Returns the quotes Id as a string
     * surrounded by delimiters, depending on
     * the DBMS selected.
     *
     * Implements a blacklist approach to automated quoting.
     * This will only quote such ids, which are a known
     * SQL keyword.
     *
     * @access  public
     * @param   mixed  $value
     * @return  string
     * @ignore
     */
    function quoteId($value)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($value); // Wrong argument type for argument 1. String expected, found \''.gettype($value).'\' instead.');

        /* optimistic type casting */
        /* argument 1 */ $value = (string) $value;

        if (preg_match('/\s/s', $value) || $this->_isSqlKeyword($value) === true) {
            return $this->database->quoteIdentifier($value);
        } else {
            return $value;
        }
    }

    /**
     * _check_available
     *
     * @access  private
     * @return  bool
     * @ignore
     */
    function _check_available()
    {
        if ($this->available() === false) {
            trigger_error("Datebase is not available. In structure file '".$this->getPath()."'.", E_USER_WARNING);
            return false;
        } else {
            return true;
        }
    }

    /**
     * _check_writeable
     *
     * @access  private
     * @return  bool
     * @ignore
     */
    function _check_writeable()
    {
        if ($this->_check_available() === false) {
            return false;
        } else if (!$this->isWriteable()) {
            trigger_error("Datebase is not writeable. Check structure file '".$this->getPath()."' if attribute readonly is set to true.", E_USER_WARNING);
            return false;
        } else {
            return true;
        }
    }

    function _checkProfile($queryType, $returnedType, &$where, &$value, $table, $row = "", $column = "")
    {
        global $YANA;
        $profile = $this->structure->getProfile($table);
        if (!empty($profile)) {
            $profile = strtolower($profile);
            switch ($queryType)
            {
                /**
                 * 1) Auto-attach profile check to where clause
                 */
                case YANA_DB_SELECT:
                    if (!empty($where)) {
                        $where .= ' AND ';
                    }
                    $where .= "${profile} = ".$this->quote($YANA->session->getId());
                    return true;
                break;
                /**
                 * 2) If user does not have permission to write new entries,
                 *    issue an error message and abort.
                 */
                case YANA_DB_INSERT:
                    if ($YANA->session->checkPermission() !== true) {
                        $YANA->report(new Alert("ungenuegende_rechte"));
                        return false;
                    } else {
                        $value[$profile] = $YANA->session->getId();
                        return true;
                    }
                break;
                /**
                 * 3) If user does not have permission to update the entry,
                 *    issue an error message and abort.
                 */
                case YANA_DB_UPDATE:
                    $primaryKey = $this->structure->getPrimaryKey($table);
                    $result = $this->query("SELECT ${profile} from ".$this->quoteId(YANA_DATABASE_PREFIX.$table)." WHERE ${primaryKey} = ".$this->quote($row), 0, 1);
                    if ($this->isError($result)) {
                        trigger_error("Unable to update entry ${table}.${row}.\n\t\t".$result->getMessage(), E_USER_WARNING);
                        return false;
                    } else {
                        $resultRow = $result->fetchRow(DB_FETCHMODE_ASSOC, 0);
                        assert('is_array($resultRow); /* unexpected result: $resultRow */');
                        $profileId = array_pop($resultRow);
                    }
                    if ($YANA->session->checkPermission($profileId) !== true) {
                        $YANA->report(new Alert("ungenuegende_rechte"));
                        return false;
                    } else {
                        switch ($returnedType)
                        {
                            case YANA_DB_ROW:
                                if (isset($value[$profile]) && $value[$profile] != $profileId) {
                                    trigger_error("Security restriction. The profile id of an entry may not be changed.", E_USER_WARNING);
                                    $YANA->report(new Log("Security restriction. The profile id of an entry may not be changed."));
                                    return false;
                                } else {
                                    return true;
                                }
                            break;
                            case YANA_DB_CELL:
                                if (strcasecmp($column, $profile) === 0) {
                                    trigger_error("Security restriction. The profile id of an entry may not be changed.", E_USER_WARNING);
                                    $YANA->report(new Log("Security restriction. The profile id of an entry may not be changed."));
                                    return false;
                                } else {
                                    return true;
                                }
                            break;
                            default:
                                return false;
                            break;
                        }
                    }
                break;
                /**
                 * 4) If user does not have permission to delete the entry,
                 *    issue an error message and abort.
                 */
                case YANA_DB_DELETE:
                    $primaryKey = $this->structure->getPrimaryKey($table);
                    $result = $this->query("SELECT ${profile} from ".$this->quoteId(YANA_DATABASE_PREFIX.$table)." WHERE ${primaryKey} = ".$this->quote($row), 0, 1);
                    if ($this->isError($result)) {
                        trigger_error("Unable to delete entry ${table}.${row}.\n\t\t".$result->getMessage(), E_USER_WARNING);
                        return false;
                    } else {
                        $resultRow = $result->fetchRow(DB_FETCHMODE_ASSOC, 0);
                        assert('is_array($resultRow); /* unexpected result: $resultRow */');
                        $profileId = array_pop($resultRow);
                    }
                    if ($YANA->session->checkPermission($profileId) !== true) {
                        $YANA->report(new Alert("ungenuegende_rechte"));
                        return false;
                    } else {
                        return true;
                    }
                break;
                /**
                 * 5) syntax error
                 */
                default:
                    trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
                    return false;
                break;
            }
        }
        return true;
    }

    /**
     * build an update statement (without where clause)
     *
     * @access  private
     * @param   string  $table
     * @param   array   $values
     * @return  bool
     * @ignore
     */
    function _buildUpdate($table, $values)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($table); // wrong argument type for argument 1, string expected');
        /* argument 2 */ assert('is_array($values); // wrong argument type for argument 2, array expected');

        $sqlStmt = '';
        if (is_string($table) && !empty($table) && is_array($values)) {
            assert('!isset($row); /* cannot redeclare variable $row */');
            assert('!isset($value); /* cannot redeclare variable $value */');
            foreach ($values as $row => $value)
            {
                if (!empty($sqlStmt)) {
                    $sqlStmt .= ',';
                }
                $sqlStmt .= $row.'='.$this->quote($value);
            }
            unset($row,$value);
            $sqlStmt = 'UPDATE '.$this->quoteId(YANA_DATABASE_PREFIX.$table).' SET '.$sqlStmt. ' ';
            return $sqlStmt;
        } else {
            trigger_error('SQL syntax error.', E_USER_WARNING);
            assert('empty($sqlStmt);');
            return $sqlStmt;
        }
    }

    /**
     * returns true if $name is a known SQL keyword and false otherwise
     *
     * implements quick-search
     * + assume that the input is sorted
     * + assume that the input is upper case
     *
     * this algorithm has O(log(n)) running time
     *
     * @access  private
     * @param   string  $name
     * @return  bool
     * @ignore
     */
    function _isSqlKeyword($name)
    {
        if (!is_string($name)) {
            trigger_error("SQL syntax error. Invalid id '".gettype($name)."(".print_r($name,true).")'.", E_USER_WARNING);
            return true;
        } else if (is_numeric($name)) {
            return true;
        }

        if (is_null($this->reservedSqlKeywords)) {
            global $YANA;
            /* Load list of reserved SQL keywords (required for smart id quoting) */
            if (isset($YANA)) {
                $file = $YANA->registry->get('reserved_sql_keywords');
                $this->reservedSqlKeywords = file($file->getPath());
            } else {
                $this->reservedSqlKeywords = array();
            }
            if (!is_array($this->reservedSqlKeywords)) {
                $this->reservedSqlKeywords = array();
            }
        } else if (empty($this->reservedSqlKeywords)) {
            return false;
        }

        $name = strtoupper($name);
        if (qSearchArray($this->reservedSqlKeywords, $name) === false) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * return when the row was last modified
     *
     * Check when the data has been last modified by any other than the current user.
     * Use this to handle conditions, where two or more users request the same form.
     * One of them submits his changes and recieves an affirmation. Now the form of
     * the other user contains invalid data, because meanwhile the data he is currently
     * editing has been changed.
     * If then the other also submits the form, the previously made changes of the other
     * user would be overwritten. We call this a "dirty-write".
     *
     * So instead, this function is to be used to check for any such conditions and
     * present the second user an appropriate error message, telling him that his form
     * data is no longer valid and ensure, the "dirty" data is not written to the
     * database.
     *
     * @access  private
     * @param   string  $key
     * @return  int
     * @ignore
     */
    function _lastModified($key)
    {
        assert('is_string($key);');
        /* settype to STRING */
        $key = (string) $key;
        global $YANA;

        /* 1. convert argument 1 to array */
        $keyArray = explode('.', $key);

        /* 2. load file */
        $file_exists = false;

        if (is_null($this->lastModified)) {
            if (isset($YANA)) {
                $this->lastModified = new SML($YANA->getVar('TEMPDIR').'db_last_modified.tmp', CASE_UPPER);
            } else {
                $this->lastModified = new SML('cache/'.'db_last_modified.tmp', CASE_UPPER);
            }
            $this->lastModified->read();
            if (!$this->lastModified->exists()) {
                $this->lastModified->create();
                $file_exists = false;
            } else {
                $file_exists = true;
            }
        }

        /* 3. get IP to identify user */
        if (isset($_SERVER['REMOTE_ADDR'])) {
            $userId = $_SERVER['REMOTE_ADDR'];
        } else {
            $userId = '*';
        }

        /* 4. get time */
        if ($file_exists === true) {
            if (count($keyArray) >= 2) {
                $lastModified = $this->lastModified->get($keyArray[0].'.'.$keyArray[1]);
            } else if (count($keyArray) === 1) {
                $array = $this->lastModified->get($keyArray[0]);
                if (is_array($array)) {
                    sort($array);
                    $lastModified = array_pop($array);
                } else {
                    $lastModified = 0;
                }
            } else {
                return 0;
            }
        } else {
            $lastModified = 0;
        }

        /* 5. check if changed by different user */
        if (is_array($lastModified)) {
            assert('count($lastModified) === 2; /* unexpected result, $lastModified. Should have 2 values: timestamp and user id */');
            if ($userId === array_pop($lastModified)) {
                $lastModified = 0;
            } else {
                $lastModified = array_pop($lastModified);
            }
        }

        /* 6. update time */
        if (($file_exists === false || $lastModified != time()) && count($keyArray) >= 2) {
            $this->lastModified->insert(array($keyArray[0] => array($keyArray[1] => array(time(), $userId))));
        }

        /* 7. return data */
        if ($lastModified > time()) {
            return 0;
        } else if (!empty($lastModified)) {
            return (int) $lastModified;
        } else {
            return 0;
        }
    }

    /**
     * Reset the object to default values
     *
     * Resets the history for the last selected table,
     * resets the queue of pending SQL statements and
     * resets the database cache.
     *
     * @access  public
     */
    function reset()
    {
        $this->lastSelectedTable   = "";
        $this->queue               = array();
        $this->cache               = array();
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $anotherObject
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class and they both
     * refer to the same structure file and use equal database connections.
     *
     * @access   public
     * @param    object $anotherObject
     * @return   string
     */
    function equals($anotherObject)
    {
        if (is_object($anotherObject) && method_exists($anotherObject, 'getClass') && $anotherObject->getClass() === $this->getClass()) {
            if ($this->getPath() === $anotherObject->getPath() && $this->database == $anotherObject->database) {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

    /**
     * self test
     *
     * This function exists only for debugging purposes.
     * Do not uncomment in productive environments.
     *
     * @access  public
     * @static
     * @param   int  $nr  number of test to take
     * @return  bool
     * @ignore
     * /
    function test($nr = 1)
    {
        switch ($nr)
        {
            case 1:
                global $YANA;
                $db = $YANA->connect('user');

                // get table
                $test = $db->get('user');
                if (!is_array($test) || !isset($test['ADMINISTRATOR']) || !is_array($test['ADMINISTRATOR'])) {
                    return '"get table" test failed';
                }

                // get row
                $test = $db->get('user.administrator');
                if (!is_array($test) || !isset($test['TYPE'])) {
                    return '"get row" test failed';
                }

                // get cell (1)
                $test = $db->get('user.administrator.type');
                if ($test !== 'ADMIN') {
                    return '"get cell 1" test failed';
                }

                // get cell (2)
                $test = $db->get('user.administrator.active');
                if ($test != true) {
                    return '"get cell 2" test failed';
                }

                // resolve foreign key
                $test = $db->get('user.administrator.type.value');
                if ($test !== 100) {
                    return '"resolving foreign key" test failed';
                }

                // get column
                $test = $db->get('user.*.id');
                $test2 = $db->length('user');
                if (!is_array($test) || count($test) !== $test2) {
                    return '"get column" test failed';
                }

                // get last entry
                $test = $db->get('user.?.id');
                if (!is_string($test)) {
                    return '"get last entry" test failed';
                }

                // test foreign key constraint
                $test = $db->insert('user.administrator.type', 'foo'); // supposed to fail
                if ($test !== false) {
                    return '"foreign key" test failed';
                }

                // test buffer
                $test = $db->insert('group.foo', array('name' => 'bar', 'value' => 99));
                if ($test !== true) {
                    return '"buffer test" insert failed';
                }

                $test = $db->insert('user.administrator.type', 'foo'); // supposed to succeed
                if ($test !== true) {
                    return '"buffer key" read of buffer content failed';
                }

                // rollback
                $db->reset();

                // @todo  should anything else be tested?

                return true;
            break;
        }
    } /* end of function */
}

?>
