<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * Database trigger handling
 *
 * {@internal
 * This class provides functions for handling of triggers and constraints.
 * Calls the function StructureFile::getFields().
 * }}
 *
 * This is a utility class that provides a collection of default
 * triggers, that you might find usefull when writing database structure
 * files.
 *
 * A "trigger" here means, a custom fragment of PHP-Code, which
 * is associated with some database event.
 * So the code gets executed every time the database event
 * is triggered.
 *
 * So a trigger gives you the chance to "hook" directly into the
 * database API and ensure that a certain code is always executed.
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 * @since       2.9
 * @ignore
 */
class DbTrigger extends Object
{
    /**#@+
     * @access  private
     */

    /** @var object */ var $structure   = null;

    /**#@-*/

    /**
     * create a new instance
     *
     * This creates and initializes a new
     * instance of this class.
     *
     * @param  object  &$structure
     * @ignore
     */
    function DbTrigger(&$structure)
    {
        assert('is_object($structure) && is_a($structure, "StructureFile"); // Wrong argument type for argument 1. Instance of StructureFile expected');
        $this->structure =& $structure;

    }

    /**
     * constraint
     *
     * Evaluates a constraint and returns bool(true) on success or bool(false) on error or if the constraint fails.
     *
     * @access  public
     * @param   string  $operation
     * @param   string  $table
     * @param   string  $field      (optional)
     * @param   mixed   $value      (optional)
     * @return  bool
     * @ignore
     */
    function constraint($operation, $table, $field = NULL, $value = NULL)
    {
        if (!is_string($operation) || !preg_match('/^select|insert|update|delete$/is', $operation)) {
            trigger_error("Invalid input for argument 1 in ".__CLASS__."->".__FUNCTION__."().\n\t\tExpected a string value 'select', 'insert', 'update' or 'delete' found (".gettype($operation).")'".print_r($operation, true)."' instead.",E_USER_WARNING);
            return false;
        }
        return $this->_eval('constraint', $operation, $table, $field, $value);
    }

    /**
     * trigger
     *
     * Evaluates a trigger and returns bool(true) on success or bool(false) on error.
     *
     * @access  public
     * @param   int     $prefix  currently one of YANA_TRIGGER_BEFORE, YANA_TRIGGER_AFTER
     * @param   string  $operation
     * @param   string  $table
     * @param   string  $field
     * @param   mixed   &$value
     * @return  bool
     * @ignore
     */
    function trigger($prefix, $operation, $table, $field, &$value)
    {
        if (!is_int($prefix)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "1 in ".__CLASS__."->".__FUNCTION__."()", 'Integer', gettype($prefix)), E_USER_WARNING);
            return false;
        }
        if (!is_string($operation) || !preg_match('/^insert|update|delete$/is', $operation)) {
            trigger_error("Invalid input for argument 2 in ".__CLASS__."->".__FUNCTION__."().\n\t\tExpected a string value 'insert', 'update' or 'delete' found (".gettype($operation).")'".print_r($operation, true)."' instead.",E_USER_WARNING);
            return false;
        }
        switch ($prefix)
        {
            case YANA_TRIGGER_BEFORE:
                $operation = 'BEFORE_'. $operation;
            break;
            case YANA_TRIGGER_AFTER:
                $operation = 'AFTER_'. $operation;
            break;
            default:
                trigger_error("Invalid input for argument 1 in ".__CLASS__."->".__FUNCTION__."(). Expected an integer of value '".YANA_TRIGGER_BEFORE."', or '".YANA_TRIGGER_AFTER."' found  ".gettype($prefix)." '".print_r($prefix, true)."' instead.",E_USER_WARNING);
                return false;
            break;
        }

        return $this->_eval('trigger', $operation, $table, $field, $value);

    }

    /**
     * evaluate triggers and constraints
     *
     * Use $fieldname parameter to separate
     * between a constraint check and triggers.
     *
     * @access  private
     * @param   string  $fieldname
     * @param   string  $operation
     * @param   string  $table
     * @param   string  $field
     * @param   mixed   &$value
     * @return  bool
     * @ignore
     */
    function _eval($fieldname, $operation, $table, $field, &$value)
    {
        global $YANA;
        $constraintArgs = '$OPERATION,$PERMISSION,$TABLE,$FIELD,$ID,$VALUE';
        $triggerArgs    = '$OPERATION,$PERMISSION,$TABLE,$FIELD,$ID,&$VALUE';
        $profileId      = $YANA->getVar('ID');

        if (!is_string($fieldname)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "1 in ".__CLASS__."->".__FUNCTION__."()", 'String', gettype($fieldname)), E_USER_WARNING);
            return false;
        } else {
            $fieldname = strtoupper($fieldname);
        }
        if ($fieldname !== 'CONSTRAINT' && $fieldname !== 'TRIGGER') {
            trigger_error("Invalid argument 1 in ".__CLASS__."->".__FUNCTION__."(). String 'CONSTRAINT' or 'TRIGGER' expected, found '".print_r($fieldname,true)."' instead.",E_USER_WARNING);
            return false;
        }
        if (!is_string($operation)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "2 in ".__CLASS__."->".__FUNCTION__."()", 'String', gettype($operation)), E_USER_WARNING);
            return false;
        } else {
            $operation = strtoupper($operation);
        }
        if (!is_string($table)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "3 in ".__CLASS__."->".__FUNCTION__."()", 'String', gettype($table)), E_USER_WARNING);
            return false;
        } else {
            $table = strtoupper($table);
        }
        if (!is_null($field) && !is_string($field)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "4 in ".__CLASS__."->".__FUNCTION__."()", 'String', gettype($field)), E_USER_WARNING);
            return false;
        } else {
            /* settype to STRING */
            $field = (string) $field;
            $field = strtoupper($field);
        }

        /* get list of fields */
        if (empty($field)) {
            $list_of_fields = $this->structure->getFields($fieldname, $operation, $table, array(), true);
        } else {
            $list_of_fields = $this->structure->getFields($fieldname, $operation, $table, array($field), true);
        }
        assert('is_array($list_of_fields);');

        /* 1.1 prepare value */
        if (is_string($value)) {
            $value = stripslashes($value);
        } else if (is_array($value) && (is_null($field) || empty($field))) {
            $value = Hashtable::changeCase($value, CASE_UPPER);
        }

        /* 1.2 prepare permission */
        if (isset($YANA)) {
            $arg_permission = $YANA->getVar("PERMISSION");
        } else {
            $arg_permission = 0;
        }

        /* 2 evaluate element */
        $test = true;
        foreach ($list_of_fields as $column_name => $element)
        {
            if ($element === '') {
                continue;
            } else if (!is_scalar($element)) {
                continue;
            } else {
                if ($fieldname == 'CONSTRAINT') {
                    if (is_bool($element)) {
                        $function = create_function($constraintArgs, "return ".(($element)?'true':'false').";");
                    } else if (trim($element) == '') {
                        $function = create_function($constraintArgs, 'return false;');
                    } else {
                        if (!preg_match(YANA_DB_CONSTRAINT_SYNTAX, $element)) {
                            trigger_error("ERROR: in structure file '".$this->structure->getPath()."', syntax error in constraint: '".trim($element)."'.", E_USER_WARNING);
                            return false;
                        }
                        $function = create_function($constraintArgs, "if ($element) return true; else return false;");
                    }
                } else if (is_string($element)) {
                    $function = create_function($triggerArgs, $element);
                } else {
                    $function = create_function($triggerArgs, 'return true;');
                }
            }
            /* handling row element */
            if (empty($field) && is_string($column_name)) {
                $column_name = strtoupper($column_name);
                if (!isset($value[$column_name])) {
                    $value[$column_name] = null;
                }
                $col_value =& $value[$column_name];
                switch ($operation)
                {
                    case 'SELECT': case 'DELETE':
                        $test = $function($operation, $arg_permission, $table, $column_name, $profileId, $col_value);
                    break;
                    case 'INSERT': case 'UPDATE':
                        if (!is_null($col_value)) {
                            $test = $function($operation, $arg_permission, $table, $column_name, $profileId, $col_value);
                        }
                    break;
                    default:
                    break;
                }
            /* handling database, table and field element */
            } else {
                $test = $function($operation, $arg_permission, $table, $field, $profileId, $value);
            }
            if (is_bool($test) && $test === false) {
                if (isset($YANA)) {
                    $YANA->report( new log("DB WARNING: Operation '${operation}' stopped by '${fieldname}' on table '${table}'".((!empty($field))?" at field '${field}'":"").((!is_null($value))?" with value '".print_r($value,true)."'":"")."."));
                    return false;
                }
            }
        }

        if ($test) {
            return true;
        } else {
            return false;
        }
    }

}

?>