<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * Manipulate a directory
 *
 * This class represents a directory.
 * You may use this to get a list of contents, or remove,
 * or create a directory.
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 */
class DirStream extends InputStream
{

    /**#@+
     * @ignore
     * @access  protected
     */
    /** @var array  */ var $content = array();
    /** @var array  */ var $options = array();
    /** @var string */ var $filter  = "";
    /**
     * @static
     * @var array   */ var $size    = null;
    /**#@-*/

    /**
     * constructor
     *
     * Create a new instance of this class.
     *
     * @param  string  $path
     */
    function DirStream($path)
    {
        parent::InputStream($path);
        /* PHP4 workaround for creating static vars */
        static $size;
        $this->size = &$size;
    }

    /**
     * read contents and put results in cache (filter settings will be applied)
     *
     * @uses    $DirStream->read()
     * @access  public
     * @return  bool
     */
    function read()
    {
        if (!$this->exists()) {
            return false;
        } else {
            $this->content = dirlist($this->getPath(), $this->filter);
            sort($this->content);
            $this->options = $this->attributes;
            return true;
        }
    }


    /**
     * return list of files within the directory
     *
     * NOTE: will only return filenames with the path stripped
     *
     * @uses    $DirStream->get()
     * @access  public
     * @return  array
     */
    function get()
    {
        if (!$this->isEmpty()) {
            return $this->content;
        } else if ($this->read()) {
            return $this->content;
        } else {
            return array();
        }
    }

    /**
     * create this directory
     *
     * Tries to create the directory.
     * Check the developer's cookbook for an example
     * to this function.
     *
     * @uses    $DirStream->create(660)
     *
     * @access  public
     * @param   int  $mode  access mode
     * @return  bool
     * @name    DirStream::create()
     */
    function create($mode = 777)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_NULL|IS_INT, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        if (!is_int($mode) || $this->exists() || !isset($this->path)) {
            return false;
        } else if (!preg_match('/^[1-7]{1}[0-7]{2}$/',"$mode")) {
            trigger_error("Invalid mode (access level) '".$mode."' in ".__CLASS__."->".__FUNCTION__."().\n\t\tThis argument needs to be an integer of 3 or 4 digits between 0-7.\n\t\tNote: when providing 4 digits, the first digit needs to be 0.", E_USER_WARNING);
            return false;
        } else {
            return @mkdir($this->getPath(), "0$mode");
        }
    }

    /**
     * remove this directory
     *
     * By option you may choose to also recursivly remove all files and subdirectories inside.
     * Otherwise the directory will only be removed if it is empty.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @uses    $DirStream->delete(false)
     *
     * @access  public
     * @param   bool  $is_recursive  triggers wether to remove directories even if they are not empty, default = false
     * @return  bool
     */
    function delete($is_recursive = false)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_NULL|IS_BOOL, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        if ($is_recursive === true) {
            $content = dirlist($this->getPath(), '');
            assert('!isset($element); /* cannot redeclare variable $element */');
            foreach ($content as $element)
            {
                $element = '/'.$element;
                if (is_file($this->getPath().$element)) {
                    $file = new secureFileStream($this->getPath().$element);
                    if ($file->delete() === false) {
                        return false;
                    }
                } else if (is_dir($this->getPath().$element)) {
                    $dir = new DirStream($this->getPath().$element);
                    if ($dir->delete(true) === false) {
                        return false;
                    }
                } else {
                    /* intentionally left blank */
                }
            } /* end foreach */
            unset($element);
        }
        if (@rmdir($this->getPath()) === false) {
            trigger_error("Unable to delete directory '".$this->getPath()."'.", E_USER_ERROR);
            exit(1);
            return false;
        } else {
            return true;
        }
    }

    /**
     * return a string representation of this directory
     *
     * Returns a string with the contents of this directory.
     * Entries are seperated by line-breaks.
     *
     * @uses    $DirStream->toString()
     *
     * @access  public
     * @return  string
     */
    function toString()
    {
        if (!$this->exists()) {
            return "Directory ".$this->getPath()." does not exist\n";
        } else if ($this->isEmpty()) {
            return "Directory ".$this->getPath()." is not loaded\n";
        } else {
            return implode("\n", $this->get());
        }
    }

    /**
     * check wether the directory has no contents
     *
     * Returns bool(true) if there are no files that
     * match the current filter and bool(false) if there
     * is at least 1 file that matches.
     *
     * @uses    $DirStream->isEmpty()
     *
     * @access  public
     * @return  bool
     */
    function isEmpty()
    {
        return empty($this->content);
    }

    /**
     * get the number of files inside the directory
     *
     * This returns a positive integer.
     * Note that this functions counts the files in respect
     * to the currently set file filter. So the number
     * of files reported here and the number in total
     * may vary.
     *
     * @uses    $DirStream->length()
     *
     * @access  public
     * @return  int
     */
    function length()
    {
        return count($this->content);
    }

    /**
     * list contents of a directory
     *
     * The argument $filter may contain multiple file extension,
     * use a pipe '|' sign to seperate them.
     * Example: "*.xml|*.html" will find all xml- and html-files
     *
     * @uses $DirStream->dirlist('*.cfg')
     *
     * @access  public
     * @param   string  $filter
     * @return  array
     */
    function dirlist($filter)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_STRING, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        $this->filter = $filter;
        $this->read();
        return $this->content;
    }

    /**
    * getContent
    *
    * @uses    $DirStream->getContent()
    *
    * @access  public
    * @return  array
    * @ignore
    */
    function getContent()
    {
         if ($this->isEmpty()) {
             $this->read();
         }
         $REF = array();
         for ($i = 0; $i < count($this->content); $i++)
         {
             $REF[$i]['VALUE'] = $this->content[$i];
             $REF[$i]['NAME']  = substr($this->content[$i], 0, strrpos($this->content[$i], "."));
         }
         return $REF;
    }

    /**
    * get size of directory
    *
    * Returns the size of $directory in bytes.
    *
    * This function gets the sum of the sizes of all files in
    * a directory.
    *
    * If $directory is not provided or NULL, the current directory is used.
    * If $includeSubDirs is not provided or true, the result will
    * include all subdirectories.
    *
    * If $useCache is not provided or true, the result is cached for
    * the current directory.
    *
    * Returns bool(false) if the directory does not exist and issues an E_USER_WARNING.
    *
    * @uses    $DirStream->getSize()
    * @uses    $DirStream->getSize('foo/bar')
    *
    * @access  public
    * @param   string    $directory         directory name
    * @param   bool      $includeSubDirs    on / off
    * @param   bool      $useCache          on / off
    * @return  int|bool(false)
    * @since   2.8.8
    */
    function getSize($directory = null, $includeSubDirs = true, $useCache = true)
    {
        assert('is_string($directory);    // Wrong argument type for argument 1. String expected.');
        assert('is_bool($includeSubDirs); // Wrong argument type for argument 2. Boolean expected.');
        assert('is_bool($useCache);       // Wrong argument type for argument 3. Boolean expected.');

        /* use default value */
        if (is_null($directory)) {
            if ($this->exists()) {
                $dir = $this->getPath();
            } else {
                trigger_error("The directory '".$this->getPath()."' does not exist.", E_USER_WARNING);
                return false;
            }

        /* directory does not exist */
        } else if (!is_string($directory) || !is_dir($directory)) {
            trigger_error("Argument 1 '".print_r($directory, true)."' is not a directory.", E_USER_WARNING);
            return false;

        /* add slash */
        } else {
            $dir = $directory.'/';
        }

        /* if value is already cached, return the cached size */
        if ($useCache === true && isset($this->size[$dir])) {
            return $this->size[$dir];
        }

        /* else determine the size */
        $d = dir($dir);
        $dirsize = 0;
        while (false !== ($filename = $d->read()))
        {
            if ($filename == '.' || $filename == '..') {
                continue;

            /* accumulate file sizes */
            } else if (is_file($dir.$filename)) {
                $dirsize += filesize($dir.$filename);

            /* only recurse if subdirs are to be included */
            } else if ($includeSubDirs === true) {
                $dirsize += $this->getSize($dir.$filename, true, false);
            }
        }
        $d->close();

        /* cache result for later requests */
        if ($useCache === true) {
           $this->size[$dir] = $dirsize;
        }

        /* return result */
        return $dirsize;
    }

    /**
     * reset statistic data
     *
     * @uses $DirStream->resetStats()
     *
     * @access  protected
     * @ignore
     */
    function resetStats()
    {
        if (!preg_match('/.*\/$/', $this->path)) {
            $this->path .= '/';
        }
        if ($this->exists()) {
            clearstatcache();
            $this->attributes['MTIME'] = filemtime($this->path);
            $this->size                = null;
            clearstatcache();
        }
    }

    /**
     * check if directory exists and is readable
     *
     * Returns bool(true) on success and bool(false) otherwise.
     *
     * @access  public
     * @return  bool
     */
    function exists()
    {
        $name = $this->getPath();
        return is_dir($name) && file_exists($name) && is_readable($name);
    }

}

?>
