<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * simulate a database
 *
 * this class simulates a sql-database on a flat-file
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 */
class FileDb extends DbStream
{

    /**
     * constructor
     *
     * Create a new instance of this class.
     *
     * @param  string  $filename  name of the structure-file
     */
    function FileDb($filename = "")
    {
        /**
         * 1) Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_NULL|IS_STRING, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
        } else {
            unset($_chk);
        }

        /* settype to STRING */
        $filename = (string) $filename;
        $this->database = new FileDb_connection($filename);
        $this->database->autoCommit(false);
        $this->path = $filename;

        /**
         * 2) Init database connection
         */
        if (empty($this->database)) {
            /* connection failed - abort */
            $this->exists = YANA_DATABASE_FAILURE;
        } else if (empty($filename)) {
            /* no structure file available, estimate structure */
            $this->exists = YANA_DATABASE_NO_FILE;
        } else {
            $this->structure = new StructureFile($this->getPath());
            $test = $this->structure->read();
            if ($test === false) {
                /* given file does not exist */
                unset($this->structure);
                $this->exists = YANA_DATABASE_NOT_AVAILABLE;
                trigger_error("Unable to initialize database. Structure file '${filename}' not found.", E_USER_WARNING);
            } else {
                /* all fine, proceed */
                $this->exists = YANA_DATABASE_OK;
                $this->database->setStructure($this->structure);
            }
        }

        /* 3) Init IO/Filter
         */
        if (is_object($this->structure)) {
            $this->transformer = new DbDatatype($this->structure);
        }

        /* 4) Init trigger
         */
        if (is_object($this->structure)) {
            $this->exec = new DbTrigger($this->structure);
        }    
    }

    /**
     * optional API bypass
     *
     * Send a sinle sql-statement directly to FileDB, bypassing the query-builder.
     *
     * The function will return bool(false) if the database connection is not
     * available and otherwise will return whatever FileDB sends back as the
     * result of your statement.
     *
     * Note: FileDB was designed to understand the statements generated
     * by the query-builder and is not intended to be called directly.
     * This means it will understand only SQL statements that comply with the syntax
     * that is provided by the query-builder.
     *
     * @access  public
     * @param   string  $sqlStmt  one SQL statement to execute
     * @param   int     $offset   the row to start from
     * @param   int     $limit    the maximum numbers of rows in the resultset
     * @return  mixed
     * @since   2.8.8
     * @ignore
     */
    function query($sqlStmt, $offset = 0, $limit = 0)
    {
        /* 1 check input type */
        if (!is_string($sqlStmt)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($sqlStmt)), E_USER_WARNING);
            return false;
        } else if (!is_int($offset) || $offset < 0) {
            trigger_error("Invalid argument 2. Must be a positive integer.", E_USER_WARNING);
            return false;
        } else if (!is_int($limit) || $limit < 0) {
            trigger_error("Invalid argument 3. Must be a positive integer.", E_USER_WARNING);
            return false;
        /* 2 check input syntax */
        } else if (is_int(strpos($sqlStmt, ';')) && preg_match("/;.*(?:select|insert|delete|update|create|alter|grant|revoke).*$/is", $sqlStmt)) {
            trigger_error("A semicolon has been found in the current input '${sqlStmt}', indicating multiple queries.\n\t\tAs this might be the result of a hacking attempt it is prohibited for security reasons and the queries won't be executed.", E_USER_WARNING);
            return false;
        /* 3 check if database connection is available */
        } else if (empty($this->database)) {
            return false;
        /* 4 send query to database */
        } else {
            return $this->database->limitQuery($sqlStmt, $offset, $limit);
        }
    }

    /**
     * not implemented importSQL
     *
     * file database is created automatically on first use - nothing to do here!
     *
     * @access  public
     * @name    FileDb::importSQL()
     * @param   string  $sqlFile
     * @return  bool
     *
     * @ignore
     */
    function importSQL($sqlFile)
    {
        /* intentionally left blank */
        return true;
    }

    /**
     * not implemented buildStructure
     *
     * @access  public
     * @ignore
     */
    function buildStructure()
    {
        /* intentionally left blank */
    }

    /**
     * isError
     *
     * @access  private
     * @param   mixed  $result
     * @return  bool
     * @ignore
     */
    function isError($result)
    {
        if (method_exists($result, 'isError') && $result->isError()) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * quote a value
     *
     * Returns the quoted values as a string
     * surrounded by double-quotes.
     *
     * @access  private
     * @param   mixed  $value
     * @return  string
     * @ignore
     */
    function quote($value)
    {
        assert('is_scalar($value);');
        return "\"$value\"";
    }

    /**
     * quote an identifier
     *
     * Returns the quotes Id as a string
     * surrounded by double-quotes.
     *
     * @access  private
     * @param   mixed  $value
     * @return  string
     * @ignore
     */
    function quoteId($value)
    {
        assert('is_scalar($value);');
        return "\"$value\"";
    }

}

?>