<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * @ignore
 */

if (!defined('DB_OK')) {
    /**
     * @ignore
     */
    define('DB_OK', 1);
}
if (!defined('DB_ERROR_UNSUPPORTED')) {
    /**
     * @ignore
     */
    define('DB_ERROR_UNSUPPORTED', 0);
}

/**
 * FileDb_connection
 *
 * Mapper for sql statements to datfile commands.
 * It implements only a required subset of the interface
 * of PEAR-DB as needed by the DbStream class
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 *
 * @ignore
 */
class FileDb_connection extends Object
{

    /**#@+
     * @ignore
     * @access  private
     */

    /** @var Counter       */ var $auto_increment = null;
    /** @var StructureFile */ var $structure      = null;
    /** @var string        */ var $database       = "";
    /** @var string        */ var $table          = "";
    /** @var array         */ var $src            = array();
    /** @var array         */ var $idx            = array();
    /** @var string        */ var $sort           = "";
    /** @var bool          */ var $desc           = false;
    /** @var bool          */ var $auto_commit    = true;
    /** @var string        */ var $BASE_DIR       = 'config/db/';
    /** @var array         */ var $cache          = array();

    /**#@-*/

    /**
     * constructor
     *
     * @param string $database (optional)
     */
    function FileDb_connection($database)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_STRING, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
        } else {
            unset($_chk);
        }

        /* settype to STRING */
        $database = (string) $database;
        global $YANA;

        $database = strtolower(preg_replace('/([\w\d_]+)\.config$/', '${1}', $database));
        $database = preg_replace('/[\.\*]/', '', $database);

        if (isset($YANA)) {
            $this->BASE_DIR = $YANA->getVar('DBDIR');
        }

        if (!empty($database)) {
            if (!is_dir($this->BASE_DIR . $database)) {
                mkdir($this->BASE_DIR . $database, 0600);
            }
            if (is_readable($this->BASE_DIR . $database)) {
                $this->database = $database;
                $this->src[$database] = array();
                $this->idx[$database] = array();
                $this->cache = array();
            }
        }
    }

    /**
     * activate/deactive auto-commit
     *
     * @access  public
     * @param   bool  $mode  on / off
     * @return  int
     */
    function autoCommit($mode = false)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_NULL|IS_BOOL, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        $this->auto_commit = $mode;
        return DB_OK;
    }

    /**
     * rollback current transaction
     *
     * @access  public
     * @return  int
     */
    function rollback()
    {
        $this->cache = array();
        $this->src[$this->database][$this->table]->reset();
        $this->idx[$this->database][$this->table]->rollback();
        return DB_OK;
    }

    /**
     * commit current transaction
     *
     * @access  public
     * @return  bool
     */
    function commit()
    {
        return $this->_write(true);
    }

    /**
     * get list of database objects
     *
     * Returns a list of all 'tables' inside the current database,
     * or a list of all available 'databases' as a numeric array.
     * Returns DB_ERROR_UNSUPPORTED otherwise.
     *
     * @access  public
     * @param   string  $type  valid values are 'tables' and 'databases'
     * @return  array
     */
    function getListOf($type)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_STRING, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        $type = strtolower($type);
        switch($type)
        {
            case 'tables':
                return $this->structure->getTables();
            break;
            case 'databases':
                $databases = dirlist($this->BASE_DIR,'*.config');
                for ($i = 0; $i < count($databases); $i++)
                {
                    $databases[$i] = preg_replace('/\.config$/i', '', $databases[$i]);
                }
                return $databases;
            break;
            default:
                return DB_ERROR_UNSUPPORTED;
            break;
        }
    }

    /**
     * execute a single query
     *
     * @access  public
     * @param   string  $sqlStmt  one SQL statement to execute
     * @param   int     $offset   the row to start from
     * @param   int     $limit    the maximum numbers of rows in the resultset
     * @return  filedb_result
     */
    function limitQuery($sqlStmt, $offset = 0, $limit = 0)
    {

        /* settype to STRING
         *            INTEGER
         *            INTEGER
         */
        $sqlStmt = (string) $sqlStmt;
        $offset  = (int)    $offset;
        $limit   = (int)    $limit;
        
        global $YANA;

        $sqlStmt = str_replace("\n", '\n', $sqlStmt);

        /* this is meant for debugging */
        /* error_log($sqlStmt."\n", 3, 'test.log'); */

        /* caching */
        if (isset($this->cache[$sqlStmt])) {
            return new filedb_result($this->cache[$sqlStmt]);
        }

        if (preg_match('/^select\s1\sfrom\s\"([\w\d_]+)\"\swhere\s(.*?)\s*;?$/si', $sqlStmt, $m)) {
            $from       = $this->_select( @ $m[1]);
            $where      = @ $m[2];

            if ($from === false) {
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            } else {
                if ($this->_isset($from, $where)) {
                    $result = array(1);
                    $this->cache[$sqlStmt] = $result;
                    return new filedb_result($result);
                } else {
                    return new filedb_result(array());
                }
            }

        } else if (preg_match('/^select\scount\(\*\)\sfrom\s\"([\w\d_]+)\"(?:\swhere\s(.*?))?\s*;?$/si', $sqlStmt, $m)) {

            /* parameter 1 */
            $from = $this->_select($m[1]);
            if ($from === false) {
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            }

            /* parameter 2 */
            if (!empty($m[2])) {
                $where = $this->_where($m[2]);
                if ($where === false) {
                    return new filedb_result(null, 'SQL ERROR: Syntax error.');
                }
            } else {
                $where = array();
            }

            /* look up */
            $length = $this->_length($from,$where);

            /* create result */
            if (is_int($length)) {
                $result = array(array($length));
                $this->cache[$sqlStmt] = $result;
                return new filedb_result($result);
            } else {
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            }

        } else if (preg_match('/^select\s(.*?)\sfrom\s((?:\"[\w\d_]+\"(?:\s*,\s*)?){2,}|\"[\w\d_]+\"\s+NATURAL\sJOIN\s+\"[\w\d_]+\")\s*(?:where\s(.+?))?(?:\s*order by\s([\w\d_",]+))?(\s*desc)?\s*;?$/si', $sqlStmt, $m)) {
            /**
             * {@internal
             *
             * Important note:
             * This API should NOT allow cross-joins.
             * The 'where'-clause is supposed to be mandatory here.
             *
             * }}
             */
            $what     = trim(strtoupper(@ $m[1]));
            if ($what === '*') {
                $what = null;
            }
            $this->sort = $this->_order_by(@$m[4]);
            $this->desc = ! @empty(         $m[5]);

            /* BEGIN:
             * resolve natural join
             */
            if (preg_match('/\"([\w\d_]+)\"\s+NATURAL\sJOIN\s+\"([\w\d_]+)\"/',$m[2], $m2)) {
                if (!$this->structure->isTable($m2[1])) {
                    return new filedb_result(null, "SQL error: '".$m2[1]."' is not a table");
                }
                if (!$this->structure->isTable($m2[2])) {
                    return new filedb_result(null, "SQL error: '".$m2[2]."' is not a table");
                }
                $m[2] = $m2[1].",".$m2[2];
                $listA = $this->structure->getColumns($m2[1]);
                $listB = $this->structure->getColumns($m2[2]);
                foreach ($listA as $colA)
                {
                    if (in_array($colA, $listB)) {
                        if (empty($m[3])) {
                            $m[3] = "";
                        } else {
                            $m[3] = " and ";
                        }
                        $m[3] .= $m2[1].'.'.$colA.'='.$m2[2].'.'.$colA;
                    }
                } /* end foreach */
                unset($listA,$listB,$colA);
            }
            /* END:
             * resolve natural join
             */
            $from     = $m[2];
            $from     = str_replace('"', '', $from);
            $from     = preg_split('/,\s?/',$from);

            /* BEGIN:
             * resolve where clause
             */
            $tableA   = null;
            $tableB   = null;
            $columnA  = null;
            $columnB  = null;
            $where    = $this->_where($m[3]);
            $listOfResultSets = array();
            foreach ($where as $clause)
            {
                /* if is join-clause */
                if (is_array($clause[0]) && is_array($clause[1])) {
                    if (empty($tableA)) {
                        $tableA  = $clause[0][0];
                    }
                    $columnA = $clause[0][1];
                    $tableB  = $clause[1][0];
                    $columnB = $clause[1][1];
                    $listOfResultSets[] = $this->_join($tableA, $tableB, $columnA, $columnB, $what, $where);
                } /* end if */
            } /* end foreach */
            /* merge resultsets */
            if (empty($listOfResultSets)) {
                $resultSet = array();
            } else {
                $resultSet = array();
                assert('!isset($result);');
                foreach ($listOfResultSets as $result)
                {
                    if (!empty($result)) {
                        $resultSet = Hashtable::merge($resultSet, $result);
                    }
                }
                unset($result);
            }
            /* END:
             * resolve where clause
             */

            /* sorting and limiting */
            $this->_doSort($resultSet);
            $this->_doLimit($resultSet, $offset, $limit);

            $this->cache[$sqlStmt] = $resultSet;
            return new filedb_result($resultSet);

        } else if (preg_match('/^select\s(.*?)\sfrom\s\"([\w\d_]+)\"(?:\swhere\s(.*?))?(?:\s*order by\s([\w\d_",]+))?(\s*desc)?\s*;?$/si', $sqlStmt, $m)) {

            /* SELECT */
            $what       = trim(strtolower(@ $m[1]));
            $from       = $this->_select( @ $m[2]);
            if (!empty($m[3])) {
                $where  = $this->_where($m[3]);
            } else {
                $where  = array();
            }
            $this->sort = $this->_order_by(@$m[4]);
            $this->desc = ! @empty(         $m[5]);

            if ($where === false || $from === false) {
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            } else {
                $result = $this->_get($what, $where, $offset, $limit);
                $this->cache[$sqlStmt] = $result;
                return new filedb_result($result);
            }

        } else if (preg_match('/^insert into\s\"([\w\d_]+)\"\s*\((.*?)\)\s+VALUES\((.*?)\)\s*;?$/si', $sqlStmt, $m)) {

            /* INSERT */
            $from  = $this->_select(@ $m[1]);
            if ($from === false) {
                return new filedb_result(null, "SQL ERROR: Table '".(@$m[1])."' is unknown.");
            }
            if (!preg_match("/^\".*?(\",\".*?)*\"$/", $m[3])) {
                trigger_error("SQL syntax error. The string '".print_r($m[3], true)."' is not a valid value in the provided SQL-statement.", E_USER_WARNING);
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            } else {
                assert('!isset($keys);');
                assert('!isset($values);');
                $keys = explode(',',@$m[2]);
                $values = explode('","', substr(@ $m[3], 1, -1));
                $set   = $this->_set($keys, $values);
                if (empty($set)) {
                    return new filedb_result(null, 'SQL ERROR: The statement contains illegal values.');
                } else {
                    $set   = $this->_increment($set);
                }
                unset($keys,$values);
            }

            /* get reference to SML file */
            assert('!isset($smlfile); /* Cannot redeclare var $smlfile */');
            $smlfile =& $this->_getSmlFile();
            /* get reference to Index file */
            assert('!isset($idxfile); /* Cannot redeclare var $idxfile */');
            $idxfile =& $this->_getIndexFile();

            $primary_id  = strtoupper($this->structure->getPrimaryKey($this->table));
            $primary_key = $set[$primary_id];
            unset($set[$primary_id]);

            /* create column */
            assert('!isset($column); /* Cannot redeclare variable $column */');
            foreach ($this->structure->getColumns($this->table) as $column)
            {
                /* handle columns of type "array" */
                $column = strtoupper($column);
                if (isset($set[$column])) {
                    if ($this->structure->getType($this->table,$column) == 'array') {
                        $set[$column] = explode('\n', $set[$column]);
                        $set[$column] = SML::getFile($set[$column], CASE_UPPER);
                    } /* end if */
                } /* end if */

                if (isset($set[$column])) {
                    /* check unique constraint
                     *
                     * Unique constraint is breached if:
                     * 1) an entry with the same key already exists
                     */
                    if ($this->structure->isUnique($this->table, $column) === true) {
                        assert('!isset($tmp); /* Cannot redeclare variable $tmp */');
                        $tmp = $idxfile->get($column, $set[$column]);
                        /* error - unique constraint has already been breached by some previous operation
                         *
                         * This may occur, when a unique constraint is added, that has not existed before.
                         */
                        if (is_array($tmp) && count($tmp) > 1) {
                            $YANA->report( new Log("SQL WARNING: The column ${column} has an unique constraint, but multiple rows with the same name have been found. This conflict can not be solved automatically. Please edit and update the affected table.", array('affected rows:' => $tmp)) );
                        } /* end if */
                        /* error - constraint is breached */
                        if (!empty($tmp)) {
                            return new filedb_result(null, "SQL ERROR: Cannot insert entry with column ${column} = ".$set[$column].". The column has an unique constraint and another entry with the same value already exists.");
                        } /* end if */
                        unset($tmp);
                    }
                    /* update indexes
                     *
                     * Index needs to be updated if:
                     * 1) any column which explicitely has an index is updated OR
                     * 2) any column which implicitely requires an index is updated
                     *   (currently that is the case for columns using unique constraints)
                     */
                    if ($this->structure->hasIndex($this->table, $column) === true || $this->structure->isUnique($this->table, $column) === true) {
                        $idxfile->create($column, array($primary_key, $set[$column]));
                    } /* end if */
                } /* end if */
            } /* end foreach */
            unset($column);
            $set = array($primary_id => array($primary_key => $set));
            if (is_array($smlfile->get($primary_id.'.'.$primary_key))) {
                return new filedb_result(null, "SQL ERROR: Cannot insert entry with primary key = ${primary_key}. Another entry with this key already exists.");
            } else {
                if ($smlfile->insert($set)) {
                    return $this->_write();
                } else {
                    return new filedb_result(null, "unable to save changes");
                } /* end if */
            } /* end if */

        } else if (preg_match('/^update\s\"([\w\d_]+)\"\sset\s((?:\s*\w+\s*\=\s*".*?"\s*,?)+)(?:\s+where\s(.*))?\s*;?$/si', $sqlStmt, $m)) {

            /* UPDATE */
            $from  = $this->_select(   @$m[1]);
            if ($from === false) {
                return new filedb_result(null, "SQL ERROR: Table '".(@$m[1])."' is unknown.");
            }
            assert('!isset($nr_of_hits) && !isset($m1);');
            $nr_of_hits = preg_match_all('/\s*(\w+)\s*\=\s*"(.*?)"\s*,?/s', $m[2], $m1);
            if ($nr_of_hits === false || $nr_of_hits < 1) {
                trigger_error("SQL syntax error. The string '".print_r($m[2], true)."' is not a valid value in the provided SQL-statement.", E_USER_WARNING);
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            } else {
                unset($nr_of_hits);
                $set   = $this->_set($m1[1], $m1[2]);
                unset($m1);
            } /* end if */
            if (!empty($m[3])) {
                $where  = $this->_where($m[3]);
            } else {
                $where  = array();
            } /* end if */
            if ($where === false) {
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            } /* end if */

            assert('!isset($primary_key);');
            $primary_key = strtoupper($this->structure->getPrimaryKey($this->table));

            /* get reference to Index file */
            assert('!isset($idxfile); // Cannot redeclare var $idxfile');
            $idxfile =& $this->_getIndexFile();

            $listOfArrays = array();

            /* extract primary key */
            assert('!isset($rule); /* cannot redeclare variable $rule */');
            foreach ($where as $rule)
            {
                if (strcasecmp($rule[0], $primary_key) == 0) {
                    $key = strtoupper($rule[1]);
                    break;
                } /* end if */
            } /* end foreach */
            unset($rule);


            assert('!isset($column); /* cannot redeclare variable $column */');
            foreach ($this->structure->getColumns($this->table) as $column)
            {
                $column = strtoupper($column);
                /* handle columns of type "array" */
                if ($this->structure->getType($this->table, $column) == 'array') {
                    if (isset($set[$column])) {
                        if (!empty($set[$column])) {
                            $set[$column] = str_replace('\n', "\n", $set[$column]);
                            $set[$column] = SML::decode($set[$column], CASE_UPPER);
                        } else {
                            $set[$column] = array();
                        } /* end if */
                        $listOfArrays[] = $column;
                    } /* end if */
                } /* end if */

                /* check unique constraint
                 *
                 * Unique constraint is breached if:
                 * 1) an entry with the same key already exists AND
                 * 2) the existing entry is not the same as the one which is currently being updated
                 */
                if (isset($set[$column]) && $this->structure->isUnique($this->table, $column) === true) {
                    assert('!isset($tmp); /* cannot redeclare variable $tmp */');
                    $tmp = $idxfile->get($column, $set[$column]);
                    /* error - unique constraint has already been breached by some previous operation
                     *
                     * This may occur, when a unique constraint is added, that has not existed before.
                     */
                    if (is_array($tmp) && count($tmp) > 1) {
                        $YANA->report( new Log("SQL WARNING: The column ${column} has an unique constraint, but multiple rows with the same name have been found. This conflict can not be solved automatically. Please edit and update the affected table.", array('affected rows:' => $tmp)) );
                    } /* end if */
                    /* error - constraint is breached */
                    if (!empty($tmp) && $tmp != $key) {
                        return new filedb_result(null, "SQL ERROR: Cannot update entry with column ${column} = ".$set[$column].". The column has an unique constraint and another entry with the same value already exists.");
                    } /* end if */
                    unset($tmp);
                } /* end if */
            } /* end foreach */
            unset($column);

            $smlfile =& $this->_getSmlFile();

            /* if primary key is renamed, the old one has to be replaced */
            if (isset($set[$primary_key]) && $key != $set[$primary_key]) {
                $smlfile->remove("$primary_key.$key");
                $temp = $set[$primary_key];
                $set  = array($temp => $set);
                unset($set[$temp][$primary_key]);
                unset($temp);
            } else if (!empty($key)) {
                $set = array($key => $set);
            } else {
                return new filedb_result(null, "SQL ERROR: Syntax error $sqlStmt");
            } /* end if */
            $set = array($primary_key => $set);

            /* remove old array values */
            foreach ($listOfArrays as $column)
            {
                /* delete old value only if a new value is set */
                $smlfile->insert("$primary_key.$key.$column", array());
            } /* end foreach */

            if ($smlfile->insert($set)) {
                /* after data has been changed, reorganize all indexes */
                $idxfile->create();
                return $this->_write();
            } else {
                return new filedb_result(null, "unable to save changes");
            } /* end if */

        } else if (preg_match('/^delete\s+from\s\"([\w\d_]+)\"\swhere\s(.*)(?:\s*order by\s([\w\d_",]+))?(\s*desc)?\s*;?$/Usi', $sqlStmt, $m)) {
            /* DELETE */

            $from       = $this->_select( @ $m[1]);
            if (!empty($m[2])) {
                $where  = $this->_where($m[2]);
            } else {
                $where  = array();
            } /* end if */
            $this->sort = $this->_order_by(@$m[3]);
            $this->desc = ! @empty(         $m[4]);

            if ($where === false || $from === false) {
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            } /* end if */

            $smlfile =& $this->_getSmlFile();
            $idxfile =& $this->_getIndexFile();

            $i = $this->_get($this->structure->getPrimaryKey($this->table), $where, 0, 1);

            if (empty($i)) {
                /* error */
                return new filedb_result(null, "SQL-ERROR: unable to delete; entry does not exist");
            } else {
                $i = $i[0][strtoupper($this->structure->getPrimaryKey($this->table))];
                if ($smlfile->remove($this->structure->getPrimaryKey($this->table).'.'.$i)) {
                    /* after data has been changed, reorganize all indexes */
                    $idxfile->create();
                    return $this->_write();
                } else {
                    return new filedb_result(null, "unable to save changes");
                } /* end if */
            } /* end if */

        } else {
            /* throw a line if unknown statement encountered */

            global $YANA;

            if (isset($YANA)) {
                $YANA->setVar("__PHP_DBASE_ERROR__", "Ress#id: @74a34b2b616b34f8eb3f5b2fbce7187c");
            }

            trigger_error("Invalid or unknown SQL statement: $sqlStmt.",E_USER_ERROR);
            return new filedb_result(null, "SQL ERROR: Syntax error '$sqlStmt'.");
        } /* end if */
    }

    /**
     * execute a single query
     *
     * alias of limitQuery() with $offset and $limit params stripped
     *
     * @access  public
     * @param   string  $sqlStmt
     * @return  filedb_result
     */
    function query($sqlStmt)
    {
        assert('is_string($sqlStmt);');
        /* settype to STRING */
        $sqlStmt = (string) $sqlStmt;
        return $this->limitQuery($sqlStmt);
    }

    /**
     * setStructure
     *
     * @access  public
     * @param   StructureFile  &$structure
     */
    function setStructure(&$structure)
    {
        /* this is a simple work-around to use PHP5-like type-checking in PHP4 */
        assert('is_a($structure, "StructureFile");');
        $this->structure =& $structure;
    }

    /**
     * _select
     *
     * loads a table and returns table name
     *
     * NOTE: returns boolean(false) on error
     *
     * @access  private
     * @param   string  $select
     * @return  string
     * @ignore
     */
    function _select($select)
    {
        assert('is_string($select);');
        /* settype to STRING */
        $select = (string) $select;
        $select = strtolower(trim($select));

        if (empty($select)) {
            return false;
        } else if (!isset($this->src[$this->database][$select])) {

            /* set current database */
            assert('!isset($database); // Cannot redeclare $database');
            $database = $this->structure->getAssociation($this->table);
            if (is_null($database)) {
                $database = $this->database;
            }
            assert('is_string($database) && !empty($this->database); // Unexpected result: $database');

            /* set current table */
            $this->table = $select;
            if ($this->structure->isStrict() && !$this->structure->isTable($this->table)) {
                trigger_error("There is no table '".$this->table."' in current structure file.", E_USER_WARNING);
                return null;
            }
            assert('is_string($this->table) && !empty($this->table);// Unexpected result: $this->table');

            /* open source file(s) */
            $this->_setSmlFile($database);
            $this->_setIndexFile($database);
            unset($database);

            return $this->table;

        } else {
            $this->table = $select;
            assert('is_string($this->table);');
            return $this->table;
        } /* end if */
    }

    /**
     * _where
     *
     * NOTE: returns boolean(false) on error
     *
     * @access  private
     * @param   string  $where
     * @param   string  $table
     * @return  array|bool(false)
     * @ignore
     */
    function _where($where)
    {
        assert('is_string($where);');
        /* settype to STRING */
        $where = (string) $where;

        $where = explode(' AND ', trim($where));
        if (empty($where)) {
            return array();
        } else {
            assert('!isset($i); /* cannot redeclare variable $i */');
            assert('!isset($val); /* cannot redeclare variable $val */');
            foreach($where as $i => $val)
            {
                if (!empty($val)) {
                    if (preg_match('/^\s*(\S+)\s?(=|LIKE|like|<|>|\!=|<=|>=)\s?\"(.+)\"\s*;?$/', $val, $m)) {
                        $where[$i]    = array();
                        $where[$i][0] = strtoupper($m[1]);
                        $where[$i][1] = toRegExp($m[3]);
                        $m[2] = strtolower($m[2]);
                        if ($m[2] === 'like') {
                            $where[$i][1] = str_replace('%','.*',$where[$i][1]);
                            $where[$i][1] = str_replace('_','.?',$where[$i][1]);
                        }
                        $where[$i][2] = $m[2];
                    /* detect joins */
                    } else if (preg_match('/^\s*(\S+)\.(\S+)\s?(=|LIKE|like|<|>|\!=|<=|>=)\s*(\S+)\.(\S+)\s?;?$/', $val, $m)) {
                        $where[$i]    = array();
                        $where[$i][0] = array(strtoupper($m[1]),strtoupper($m[2]));
                        $where[$i][1] = array(strtoupper($m[4]),strtoupper($m[5]));
                        $where[$i][2] = strtoupper($m[3]);
                    } else {
                        /* syntax error */
                        return false;
                    } /* end if */
                } else {
                    /* intentionally left blank */
                    continue;
                } /* end if */
            } /* end foreach */
            unset($i, $val);
            return $where;
        }
    }

    /**
     * _order_by
     *
     * NOTE: returns boolean(false) on error
     *
     * @access  private
     * @param   string  $order_by
     * @return  array
     * @ignore
     */
    function _order_by($order_by)
    {
        assert('is_string($order_by) || is_null($order_by);');
        if (empty($order_by)) {
            return array();
        } else {
            /* settype to STRING */
            $order_by = (string) $order_by;

            $order_by = explode(',', strtoupper(trim($order_by)));
            $result = array();
            if (empty($order_by)) {
                return array();
            } else {
                assert('!isset($m); /* cannot redeclare variable $m */');
                assert('!isset($val); /* cannot redeclare variable $val */');
                foreach($order_by as $val)
                {
                    /* unquote id */
                    if (!empty($val) && preg_match('/^\s*"(.*)"\s*$/',$val,$m)) {
                        $result[] = $m[1];
                    } else {
                        /* intentionally left blank */
                        continue;
                    } /* end if */
                } /* end foreach */
                unset($val);
                return $result;
            } /* end if */
        } /* end if */
    }

    /**
     * _set
     *
     * @access  private
     * @param   array  $keys
     * @param   array  $values
     * @return  array
     * @ignore
     */
    function _set($keys, $values)
    {
        assert('is_array($keys);');
        assert('is_array($values);');
        if (!is_array($keys) || !is_array($values)) {
            return array();
        }

        assert('count($keys) == count($values);');
        /* array_combine() became available in PHP 5, here's a code to simulate this behaviour for PHP 4 */
        $insert = array();
        assert('!isset($isStrict); /* Cannot redeclare var $isStrict */');
        $isStrict = $this->structure->isStrict();
        assert('!isset($i); /* Cannot redeclare var $i */');
        for ($i = 0; $i < count($keys); $i++)
        {
            if ($isStrict === true && !$this->structure->isColumn($this->table, $keys[$i])) {
                trigger_error("SQL syntax error. The column '".$keys[$i]."' does not exist in table '".$this->table."'.", E_USER_WARNING);
                return array();
            } else {
                $insert[strtoupper($keys[$i])] = $values[$i];
            }
        } /* end foreach */
        unset($i);

        assert('is_array($insert)');
        return $insert;
    }

    /**
     * _increment
     *
     * simulate MySQL's auto-increment feature
     *
     * @access  private
     * @param   array  $res
     * @return  array
     * @ignore
     */
    function _increment($res)
    {
        /* settype to ARRAY */
        $res = (array) $res;
        $primary_key = strtoupper($this->structure->getPrimaryKey($this->table));

        if (is_null($this->auto_increment)) {
         $this->auto_increment = new counter('file_db_auto_increment',YANA_COUNTER_IGNORE_IP);
        }

        /* simulate auto-increment */
        if (empty($res[$primary_key])) {
            $index = $this->auto_increment->count($this->database.'__'.$this->table);
            $res[$primary_key] = $index;
        }

        return $res;
    }

    /**
     * _isset
     *
     * simulate select 1 from ... where ...
     *
     * @access  private
     * @param   string  $tableName
     * @param   string  $where
     * @return  bool
     * @ignore
     */
    function _isset($tableName, $where)
    {
        assert('is_string($tableName);');
        assert('is_string($where);');
        /* simple case - no further restrictions */
        if (preg_match('/^(\S+) IS NOT NULL$/', $where, $m)) {
            $result = $this->_get($m[1]);
            if (count($result) > 0) {
                return true;
            } else {
                return false;
            }
        /* complex case - restriction by where clause */
        } else if (preg_match('/^(.+?)(?: AND (\S+) IS NOT NULL)?$/', $where, $m)) {
            if (!empty($m[1])) {
                $where = $this->_where($m[1]);
                if ($where === false) {
                    /* syntax error */
                    return false;
                }
            } else {
                $where = array();
            }
            if (empty($m[2])) {
                /**
                 * {@internal
                 *
                 * currently used:
                 *
                 * $what = '*';
                 *
                 * instead of:
                 *
                 * $what = $this->getPrimaryKey($this->table);
                 *
                 * Note: due to the current implementation this single operation (which
                 * takes about 0.0005 to 0.0006 seconds on my current machine) is about
                 * 5 to 7% faster if the whole resultset ('*') is taken, instead of
                 * limiting (filtering) the resultset to a certain column (like the
                 * primary key).
                 *
                 * I mention this here because it is quite the opposite of what one might
                 * expect in the first place.
                 *
                 * The current implementation should be revised, in case the underlaying
                 * implementation has changed to reflect on the performance.
                 *
                 * }}
                 */
                $what = '*';
            } else {
                $what = $m[2];
            }
            $result = $this->_get($what,$where);
            if (count($result) > 0) {
                return true;
            } else {
                return false;
            }
        } else {
            /* syntax error */
            return false;
        }

    }

    /**
     * _length
     *
     * simulate select count(*) from ... where ...
     *
     * @access  private
     * @param   string  $tableName
     * @param   array   $where
     * @return  int
     * @ignore
     */
    function _length($tableName, $where)
    {
        assert('is_string($tableName);');
        assert('is_array($where);');
        /* settype to STRING
         *            ARRAY
         */
        $tableName = (string) $tableName;
        $where     = (array)  $where;

        $tableName = strtolower(trim($tableName));
        /* if table does not exist, then there is nothing to get */
        if (!isset($this->src[$this->database][$tableName])) {
            return 0;
        }

        if (empty($where)) {
            $smlfile =& $this->src[$this->database][$tableName];
            $PrimaryKey = $this->structure->getPrimaryKey($tableName);
            /* note: the first index is the primary key - NOT the table name
             */
            return $smlfile->length($PrimaryKey);
        } else {
            $result = $this->_get('*',$where);
            return count($result);
        }
    }

    /**
     * _get
     *
     * @access  private
     * @param   string  $what
     * @param   array   $where
     * @param   int     $offset
     * @param   int     $limit
     * @return  array
     * @ignore
     */
     function _get($what = "*", $where = array(), $offset = 0, $limit = 0)
    {
        assert('is_string($what);');
        assert('is_array($where);');
        assert('is_int($offset) && $offset >= 0;');
        assert('is_int($limit) && $limit >= 0;');
        /* settype to STRING
         *            ARRAY
         *            INTEGER
         *            INTEGER
         */
        $what   = (string) $what;
        $where  = (array)  $where;
        $limit  = (int)    $limit;
        $offset = (int)    $offset;

        /* if table does not exist, then there is nothing to get */
        if (!isset($this->src[$this->database][$this->table])) {
            return array();
        }

        /* initialize vars */
        $result     =  array();
        $smlfile    =& $this->_getSmlFile();
        $primaryKey =  strtoupper($this->structure->getPrimaryKey($this->table));
        $data       =  $smlfile->get($primaryKey);

        /* if there is data */
        if (is_array($data)) {

            /* 1) add primary key column */
            assert('!isset($i); /* cannot redeclare variable $i */');
            assert('!isset($current); /* cannot redeclare variable $current */');
            /* implements a table-scan */
            foreach (array_keys($data) as $i)
            {
                $data[$i][$primaryKey] = $i;
            }

            /* 2) pre-sort data */
            $this->_doSort($data);

            /* 3) apply where clause */
            foreach ($data as $i => $current)
            {
                /* implements the resultset */
                if ($this->_doWhere($current, $where) === true) {
                    if ($what == '*' || empty($what)) {
                        $result[] = $current;
                    } else if (isset($current[strtoupper($what)])) {
                        $result[] = array(strtoupper($what) => $current[strtoupper($what)]);
                    }
                }
            } /* end foreach */
            unset($i, $current);

            /* 4) limit results */
            $this->_doLimit($result, $offset, $limit);

            /* 5) return resultset */
            return $result;

        } else {
            /* result is empty */
            return array();
        }
    }

    /**
     * sort by column
     *
     * @access  private
     * @param   array  &$result
     * @ignore
     */
    function _doSort(&$result)
    {
        /* If no sort criteria is provided, the array will be returned order in which the entries
         * where inserted. If descending order is requested, the order will be reversed, so the
         * latest entries will come first and the oldest last.
         */
        if ($this->desc === true && empty($this->sort)) {
            $result = array_reverse($result);
        /* If a column has been provided to sort entries by, then the resultset will get sorted
         * by it.
         */
        } else if (!empty($this->sort)) {
            uasort($result, array($this, '_sort'));
        }
    }

    /**
     * _sort
     *
     * @access  private
     * @param   array  $a
     * @param   array  $b
     * @return  int
     * @ignore
     */
    function _sort($a, $b, $columns = null)
    {
        if (is_null($columns)) {
            $columns = $this->sort;
        }
        if (!is_array($columns) || count($columns) === 0) {
            return 0;
        } else {
            $sort = array_shift($columns);

            if (!is_array($a) || !is_array($b)) {
                return 0;
            } else if (!isset($a[$sort]) && !isset($b[$sort])) {
                $result = $this->_sort($a, $b, $columns);
            } else if (!isset($a[$sort])) {
                $result = -1;
            } else if (!isset($b[$sort])) {
                $result = 1;
            } else if ($a[$sort] < $b[$sort]) {
                $result = -1;
            } else if ($a[$sort] > $b[$sort]) {
                $result = 1;
            } else {
                if (count($columns) > 0) {
                    $result = $this->_sort($a, $b, $columns);
                } else {
                    $result = 0;
                }
            } /* end if */
            assert('is_int($result) && $result >= -1 && $result <= 1; // unexpected result');

            if ($this->desc) {
                return - $result;
            } else {
                return $result;
            } /* end if */
        } /* end if */
    }

    /**
     * limits and offsets
     *
     * @access  private
     * @param   array  &$result
     * @param   int    $limit
     * @param   int    $offset
     * @ignore
     */
    function _doLimit(&$result, $offset, $limit)
    {
        if ($limit > 0 || $offset > 0) {
            if ($limit <= 0) {
                $limit = count($result);
            }
            if ($offset < 0) {
                $offset = 0;
            }
            $result = array_slice($result, $offset, $limit);
        }
    }

    /**
     * commit transaction
     *
     * @access  private
     * @param   bool  $commit on / off
     * @return  filedb_result
     * @ignore
     */
    function _write($commit = false)
    {
        /* pessimistic scanning */
        assert('is_bool($commit); /* Wrong argument type for argument 1. Boolean expected. */');

        /* optimistic type casting */
        /* settype to BOOLEAN */
        $commit = (bool) $commit;

        $this->cache = array();
        if (!$commit && !$this->auto_commit) {
            /* wait for commit */
            return new filedb_result(array());
        } else {
            /* auto commit */
            $smlfile =& $this->_getSmlFile();
            if (!$smlfile->write()) {
                /* error */
                return new filedb_result(null, 'unable to save changes');
            } else {
                $this->idx[$this->database][$this->table]->commit();
                /* success */
                return new filedb_result(array());
            }
        }
    }

    /**
     * _join
     *
     * This implements joining two tables.
     *
     * @access  private
     * @param   string  $tableA
     * @param   string  $tableB
     * @param   string  $columnA
     * @param   string  $columnB
     * @param   string  $what
     * @param   array   $where
     * @return  filedb_result
     * @ignore
     */
    function _join($tableA, $tableB, $columnA, $columnB, $what = null, $where = array())
    {
        /* pessimistic scanning */
        assert('is_string($tableA);  /* Wrong argument type for argument 1. String expected. */');
        assert('is_string($tableB);  /* Wrong argument type for argument 2. String expected. */');
        assert('is_string($columnA); /* Wrong argument type for argument 3. String expected. */');
        assert('is_string($columnB); /* Wrong argument type for argument 4. String expected. */');

        /* optimistic type casting */
        /* settype to STRING */
        $tableA  = (string) $tableA;
        $tableB  = (string) $tableB;
        $columnA = (string) $columnA;
        $columnB = (string) $columnB;

        /* prepare input */
        $tableA  = strtoupper($tableA);
        $columnA = strtoupper($columnA);
        $tableB  = strtoupper($tableB);
        $columnB = strtoupper($columnB);

        /* this information is needed optimize performance */
        $aHasIndex = $this->structure->hasIndex($tableA, $columnA);
        $bHasIndex = $this->structure->hasIndex($tableB, $columnB);
        $aIsFk     = $this->structure->isForeignKey($tableA, $columnA);
        $bIsPk     = $this->structure->isPrimaryKey($tableB, $columnB);
        $bIsUnique = $this->structure->IsUnique($tableB, $columnB);
        $typeOfA   = $this->structure->getType($tableA, $columnA);
        $typeOfB   = $this->structure->getType($tableB, $columnB);
        $pkA       = strtoupper($this->structure->getPrimaryKey($tableA));
        $pkB       = strtoupper($this->structure->getPrimaryKey($tableB));

        /* notify me if results are not valid */
        assert('is_bool($aHasIndex); /* unexpected result $aHasIndex */');
        assert('is_bool($bHasIndex); /* unexpected result $bHasIndex */');
        assert('is_bool($aIsFk);     /* unexpected result $aIsFk     */');
        assert('is_bool($bIsPk);     /* unexpected result $bIsPk     */');
        assert('is_string($typeOfA); /* unexpected result $typeOfA   */');
        assert('is_string($typeOfB); /* unexpected result $typeOfB   */');
        assert('is_string($pkA);     /* unexpected result $pkA       */');
        assert('is_string($pkB);     /* unexpected result $pkB       */');

        /* strict key checking */
        if ($this->structure->isStrict()) {
            if (!$aIsFk) {
                trigger_error("Joining table '${tableA}' with '${tableB}' might fail.\n\t\tColumn '${columnA}' is not a foreign key in table '${tableA}'.", E_USER_NOTICE);
            }
            if (!$bIsPk && !$bHasIndex) {
                trigger_error("Joining table '${tableA}' with '${tableB}' might fail.\n\t\tColumn '${columnB}' is not a primary key in table '${tableB}'.", E_USER_NOTICE);
            }
            if (!$bIsPk && !$bIsUnique) {
                trigger_error("Joining table '${tableA}' with '${tableB}' might return ambigious results.\n\t\tColumn '${columnB}' is not unique in table '${tableB}'.", E_USER_NOTICE);
            }
            if (strcasecmp($typeOfA, 'select') !== 0 && $typeOfA != $typeOfB) {
                trigger_error("Note: data type of column '${tableA}.${columnA}' (${typeOfA}) does not match column '${tableB}.${columnB}' (${typeOfB}).\n\t\tThis might lead to unexpected results.", E_USER_NOTICE);
            }
        }

        /* the following implements the productive process */

        /* get a cursor on A */
        $this->_select($tableA);
        $SMLA       =& $this->_getSmlFile();
        $cursorA    =  $SMLA->getByReference($pkA);
        if ($aHasIndex) {
            $indexA =& $this->idx[$this->database][$this->table];
        } else {
            $indexA =  null;
        }

        /* get a cursor on B */
        $this->_select($tableB);
        $SMLB    =& $this->_getSmlFile();
        $cursorB =  $SMLB->getByReference($pkB);
        if ($bHasIndex) {
            $indexB =& $this->idx[$this->database][$this->table];
        } else {
            $indexB =  null;
        }

        $resultSet = array();

        /**
         * iterate through tables
         *
         * {@internal
         *
         * Note: there are several cases for table A and B
         *
         * - case 1: $columnB is a primary key
         *           enables direct look-up of values in B
         *           with the best possible performance - this should be the default
         *           O(n) running time
         * - case 2: $columnB has an index
         *           enables indirect look-up of values in B via the index
         *           still good performance (currently we assume $columnB is unique)
         *           O(n) running time
         * - case 3: other
         *           will result in expensive recursive table-scans - this scenario should be avoided!
         *           worst case O(n*m) running time
         * }}
         */
        assert('!isset($row); /* cannot redeclare variable $row */');
        assert('!isset($id);  /* cannot redeclare variable $id */');
        foreach ($cursorA as $id => $row)
        {
            if (isset($row[$columnA]) || !is_scalar($row[$columnA])) {

                /* get value of foreign key */
                $keyA = strtoupper($row[$columnA]);

                /* case 1: $columnB is a primary key */
                if ($bIsPk === true) {
                    if (isset($cursorB[$keyA])) {
                        $joinedValueExists = true;
                        $value             = array($cursorB[$keyA]);
                    } else {
                        $joinedValueExists = false;
                        $value             = null;
                    }
                /* case 2: $columnB has an index */
                } else if ($bHasIndex === true && !is_null($indexB)) {
                    $keyA = $indexB->get($columnB, $keyA);
                    if (is_scalar($keyA) && isset($cursorB[$keyA])) {
                        $joinedValueExists = true;
                        $value             = array($cursorB[$keyA]);
                    } else if (is_array($keyA) && !empty($keyA)) {
                        $joinedValueExists = true;
                        $value = array();
                        foreach ($keyA as $currentKey)
                        {
                            $value[]       =& $cursorB[$currentKey];
                        }
                    } else {
                        $joinedValueExists = false;
                        $value             = null;
                    }
                /* other */
                } else {
                    $value             = array();
                    $joinedValueExists = false;
                    foreach ($cursorB as $idB => $rowB)
                    {
                        if (isset($rowB[$columnB]) && strcmp($rowB[$columnB],$row[$columnA]) === 0) {
                            $joinedValueExists = true;
                            $row[$pkB]         = $idB;
                            $value[]           = $rowB;
                        }
                    }
                    if (empty($value)) {
                        $value         = null;
                    }
                }

                if ($joinedValueExists === true) {

                    /* add primary key to result - otherwise it would be missing */
                    $row[$pkA] = $id;

                    assert('!isset($newResult);');
                    assert('!isset($currentValue);');
                    foreach ($value as $currentValue)
                    {
                        if (!is_null($what)) {
                            if (isset($currentValue[$what])) {
                                $newResult = array($what => $currentValue[$what]);
                            } else if (isset($row[$what])) {
                                $newResult = array($what => $row[$what]);
                            }
                        } else {
                            $newResult = Hashtable::merge($row, $currentValue);
                        }
                        if ($this->_doWhere($newResult, $where) === true) {
                            $resultSet[] = $newResult;
                        }
                    } /* end foreach value */
                    unset($newResult, $currentValue);

                }
            } /* end if */
        } /* end foreach row */
        unset($id,$row);

        /* create output */
        if (!empty($resultSet)) {
            return $resultSet;
        } else {
            return array();
        }

    }

    /**
     * implements where-clause
     *
     * Returns bool(true) if where clause matches $current,
     * returns bool(false) otherwise.
     *
     * @access  private
     * @param   array  &$current
     * @param   array  &$where
     * @return  bool
     * @ignore
     */
    function _doWhere(&$current, &$where)
    {
        $test = true;
        if (!empty($where)) {
            assert('!isset($value); /* cannot redeclare variable $value */');
            foreach ($where as $value)
            {
                /* if all required information is provided */
                if (count($value) === 3 && is_string($value[0]) && is_scalar($value[1])) {
                    assert('!isset($currentValue); /* cannot redeclare variable $currentValue */');
                    if (isset($current[strtoupper($value[0])])) {
                        $currentValue = $current[strtoupper($value[0])];
                    } else {
                        $currentValue = null;
                    }
                    switch ($value[2])
                    {
                        case '=':
                            if ($this->structure->isPrimaryKey($this->table,$value[0]) || $this->structure->isForeignKey($this->table,$value[0])) {
                                if (strcasecmp($currentValue, $value[1]) !== 0) {
                                    $test = false;
                                }
                            } else {
                                if (strcmp($currentValue, $value[1]) !== 0) {
                                    $test = false;
                                }
                            }
                        break;

                        case 'like':
                            if (!preg_match('/^'.$value[1].'$/i',$currentValue)) {
                                $test = false;
                            }
                        break;

                        case '<':
                            if (!($currentValue < $value[1])) {
                                $test = false;
                            }
                        break;

                        case '>':
                            if (!($currentValue > $value[1])) {
                                $test = false;
                            }
                        break;

                        case '!=':
                            if ($this->structure->isPrimaryKey($this->table,$value[0]) || $this->structure->isForeignKey($this->table,$value[0])) {
                                if (strcasecmp($currentValue, $value[1]) === 0) {
                                    $test = false;
                                }
                            } else {
                                if (strcmp($currentValue, $value[1]) === 0) {
                                    $test = false;
                                }
                            }
                        break;

                        case '<=':
                            if (!($currentValue <= $value[1])) {
                                $test = false;
                            }
                        break;

                        case '>=':
                            if (!($currentValue >= $value[1])) {
                                $test = false;
                            }
                        break;

                        default:
                            /* intentionally left blank */
                        break;
                    }
                    unset($currentValue);
                }
            } /* end foreach */
            unset($value);
        } /* end if*/
        return $test;
    }

    /**
     * initialize and return current index file by reference
     *
     * @access  private
     * @return  FileDb_index
     * @ignore
     */
    function &_getIndexFile()
    {
        $idxfile =& $this->idx[$this->database][$this->table];
        if (!is_object($idxfile)) {
            trigger_error("No index-file found.", E_USER_ERROR);
        }
        return $idxfile;
    }

    /**
     * create index file
     *
     * @access  private
     * @param   string  $database
     * @ignore
     */
    function _setIndexFile($database = null)
    {
        $filename = $this->_getFilename($database, 'idx');
        $smlfile =& $this->_getSmlFile();
        $idxfile = new FileDb_index($this->structure, $this->table, $smlfile, $filename);
        $this->idx[$this->database][$this->table] =& $idxfile;
    }

    /**
     * initialize and return current SML file by reference
     *
     * @access  private
     * @return  SML
     * @ignore
     */
    function &_getSmlFile()
    {
        return $this->src[$this->database][$this->table];
    }

    /**
     * create SML file
     *
     * @access  private
     * @param   string  $database
     * @ignore
     */
    function _setSmlFile($database = null)
    {
        $filename = $this->_getFilename($database, 'sml');
        $smlfile = new SML($filename, CASE_UPPER);
        $smlfile->read();
        if (!$smlfile->exists()) {
            $smlfile->create();
            if ($database != $this->database) {
                $filename = $this->_getFilename($this->database, 'sml');
                $smlfile = new SML($filename, CASE_UPPER);
            }
        }

        $this->src[$this->database][$this->table] =& $smlfile;
    }

    /**
     * return filename
     *
     * @access  private
     * @param   string  $database
     * @param   string  $extension
     * @return  string
     * @ignore
     */
    function _getFilename($database, $extension)
    {
        return (string) $this->BASE_DIR . $database . '/' . $this->table . '.' . $extension;
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $anotherObject
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class and they both
     * refer to the same structure file.
     *
     * @access   public
     * @param    object  $anotherObject
     * @return   string
     */
    function equals($anotherObject)
    {
        if (is_object($anotherObject) && method_exists($anotherObject,'getClass') && $anotherObject->getClass() === $this->getClass()) {
            if (!isset($this->structure) || !isset($anotherObject->structure)) {
                return isset($this->structure) === isset($anotherObject->structure);
            } else if ($this->structure->equals($anotherObject->structure)) {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

}

?>
