<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * @ignore
 */

/**
 * FileDb_result
 *
 * This class represents a FileDB resultset.
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 *
 * @ignore
 */
class FileDb_result extends Object
{

    /**#@+
     * @ignore
     * @access  private
     */

    /** @var array */  var $result  = array();
    /** @var string */ var $message = '';

    /**#@-*/

    /**
     * constructor
     *
     * Create a new resultset.
     *
     * @param  mixed   $result   resultset (set "null" for error)
     * @param  string  $message  error message
     */
    function FileDb_result($result, $message = '')
    {
        /* settype to STRING */
        $message = (string) $message;
        if (is_null($result)) {
            $this->result = null;
        } else {
            assert('is_array($result);');
            $this->result = Hashtable::changeCase($result,CASE_LOWER);
        }
        $this->message  = trim($message);
    }

    /**
     * get number of rows in resultset
     *
     * returns the number of rows in the result set
     *
     * @access  public
     * @return  int
     */
    function numRows()
    {
        return count($this->result);
    }

    /**
     * fetch a row from the resultset
     *
     * returns an associatvie array of the row at index $i in the result set
     *
     * @access  public
     * @param   mixed  $dummy  (ignored) this is here for compatibility reasons
     * @param   int    $i      row number
     * @return  array
     */
    function fetchRow($dummy,  $i)
    {
        /* settype to INTEGER */
        $i = (int) $i;
        if (isset($this->result[$i])) {
            return $this->result[$i];
        } else {
            return array();
        }
    }

    /**
     * get error message
     *
     * This returns an error message (if any)
     * if the resultset is in error-state.
     *
     * If there is none, an empty string is
     * returned instead.
     *
     * Use FileDb_result::isError() to check,
     * if the result is in an error state.
     *
     * @access  public
     * @return  string
     */
    function getMessage()
    {
        return $this->message;
    }

    /**
     * check wether the result is an error
     *
     * Returns bool(true) if the request resulted
     * in an error state and bool(false) otherwise.
     *
     * {@internal
     * If something went wrong, the property "result"
     * is not set. This means, $this->result is NULL.
     * }}
     *
     * @access  public
     * @return  bool
     */
    function isError()
    {
        return is_null($this->result);
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $anotherObject
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class and their
     * attributes are equal.
     *
     * @access   public
     * @param    object  $anotherObject
     * @return   string
     */
    function equals($anotherObject)
    {
        if (is_object($anotherObject) && method_exists($anotherObject,'getClass') && $anotherObject->getClass() === $this->getClass()) {
            $attrs = array_keys(get_class_vars($this->getClass()));
            foreach ($attrs as $attribute)
            {
                if (!isset($this->$attribute) && !isset($anotherObject->$attribute)) {
                    continue;
                } else if (!isset($this->$attribute) && isset($anotherObject->$attribute)) {
                    return false;
                } else if (isset($this->$attribute) && !isset($anotherObject->$attribute)) {
                    return false;
                } else if ($this->$attribute !== $anotherObject->$attribute) {
                    return false;
                }
            } /* end foreach */
            return true;
        } else {
            return false;
        }
    }

}

?>