<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * Brush wrapper class
 *
 * This class is meant as an add-on to the framework's
 * Image class. It is intended to handle predefined brushes
 * in PNG format, which reside in the directory
 * "common_files/brush/*.png" to ease work with images.
 *
 * Brush images need to be 2-colored black/white images
 * with color index 0 being black and index 1 being
 * the transparent color and width == height.
 *
 * @access      public
 * @package     yana
 * @subpackage  utilities
 * @since       2.8.7
 * @see         Image
 */
class Brush extends Object
{
    /**#@+
     * @access  private
     * @ignore
     */

    var $brushname = null;
    var $brushdir  = null;
    var $image     = null;

    /**#@-*/

    /**
     * create a new instance of this class
     *
     * The argument $brushname determines wich brush to take the input from.
     * If $brushname is not provided, a 1px square is used.
     *
     * The brush name can be any name of an existing PNG image (without the file extension),
     * that must reside in the "common_files/brush/*.png" directory.
     *
     * Here are some examples:
     * <ul>
     * <li> airbrush </li>
     * <li> small circle </li>
     * <li> circle </li>
     * <li> dot </li>
     * <li> small star </li>
     * <li> star </li>
     * <li> square </li>
     * <li> point (default) </li>
     * </ul>
     *
     * @param  string  $brushname  see list
     */
    function Brush($brushname = null)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_NULL|IS_STRING, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        global $YANA;

        /**
         * Set Brush directory
         */
        /* PHP4 workaround for creating static vars */
        static $brushdir;
        $this->brushdir = &$brushdir;
        if (is_null($this->brushdir)) {
            if (isset($YANA)) {
                $this->brushdir = $YANA->getVar('BRUSHDIR');
            } else {
                $this->brushdir = 'common_files/brush/';
            }
        }

        /**
         * Check if GD-libary is available and able to handle PNG images
         *
         * In case of an error a log entry is created.
         * Note: no error is triggered, since this might result in a broken image,
         * when calling Image::outputToScreen().
         */
        if (!function_exists('imagecreate') || !function_exists('imagecreatefrompng')) {
            if (isset($YANA)) {
                $YANA->report(new Log("Cannot create brush. This server is unable to handle PNG images."));
            }

        /* try to load the file */
        } else {

            /* 1 check input */
            if (is_null($brushname)) {
                $brushname = 'point';
            }
            if (!is_string($brushname)) {
                $brushFile = null;
            } else {
                $brushFile = $this->brushdir . str_replace(' ', '-', $brushname) . '.png';
            }

            /* 2 create image resource */
            if (!is_string($brushFile) || !file_exists($brushFile)) {
                trigger_error("The brush '${brushname}' does not exist.", E_USER_NOTICE);
            } else {
                $this->brushname = $brushname;
                $this->image     = imagecreatefrompng($brushFile);
            }

            /* check if result is valid */
            if (!is_resource($this->image)) {
                trigger_error("The brush ${brushname} is invalid.", E_USER_WARNING);
            } else {
                imagecolortransparent($this->image, 1);
            }
        } /* end of section "create image resource" */
    }

    /**
     * get name of this brush
     *
     * Returns bool(false) on error.
     *
     * @access  public
     * @return  string
     */
    function getName()
    {
        if (!is_string($this->brushname)) {
            return false;
        } else {
            return $this->brushname;
        }
    }

    /**
     * set the directory that contains the brushes
     *
     * This function will set the source directory for brushes.
     * The next time you create a Brush object, the png image
     * will automatically be searched for in the directory
     * you provided here.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * {@internal
     * Note: while this should be a static function, it is not
     * due to a compatibility issue with PHP4, that did not
     * support static class members.
     * }}
     *
     * @access  public
     * @param   string  $directory  new source directory
     * @return  bool
     * @since   2.8.8
     */
    function setSourceDirectory($directory)
    {
        if (!is_string($directory) || !is_dir($directory)) {
            return false;
        } else {
            $this->brushdir = $directory;
            return true;
        }
    }

    /**
     * get brush size
     *
     * Returns the brush's dimension in pixel or bool(false) on error.
     *
     * @access  public
     * @return  int|bool(false)
     */
    function getSize()
    {
        if (!is_resource($this->image)) {
            return false;
        } else if (!function_exists('imagesx')) {
            return false;
        } else {
            return imagesx($this->image);
        }
    }

    /**
     * Resize the brush
     *
     * This resizes the brush.
     *
     * The argument $size is the new size in pixel.
     *
     * Returns bool(false) on error.
     *
     * @access  public
     * @param   int  $size  brush size in pixel
     * @return  bool
     */
    function setSize($size)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_INT, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        if (!is_resource($this->image)) {
            return false;
        } else if (!function_exists('imagepalettecopy')) {
            return false;

        } else {

            $currentSize = $this->getSize();

            /* argument 1 */
            if ($size < 1) {
                trigger_error("Invalid value for argument 1. Size must be greater 0.", E_USER_WARNING);
                return false;
            }

            if ($currentSize === $size) {

                /* if image already has the expected size, then there is nothing to do here */
                return true;

            } else {

                $oldImage    = $this->image;
                $this->image = imagecreate($size, $size);
                imagepalettecopy($this->image, $oldImage);
                imagefill($this->image, 0, 0, 1);
                imagecopyresized($this->image, $oldImage, 0, 0, 0, 0, $size, $size, $currentSize, $currentSize);
                imagecolortransparent($this->image, 1);
                imagedestroy($oldImage);
                return true;

            }
        }
    }

    /**
     * set the color of this brush
     *
     * This function sets the color of the brush
     * to a certain value, where the input is
     * the red, green and blue values of this color.
     *
     * The arguments $r, $g and $b need to be integer
     * values between 0 and 255. For example,
     * $r = 0 is interpreted as "0% red" and $r = 255
     * is interpreted as "100% red".
     *
     * @access  public
     * @param   int    $r        red value
     * @param   int    $g        green value
     * @param   int    $b        blue value
     * @return  string
     */
    function setColor($r, $g, $b)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        assert('!isset($_types);');
        $_types = array(
        /* argument 1 */ IS_INT,
        /* argument 2 */ IS_INT,
        /* argument 3 */ IS_INT
        );
        $_chk = checkArgumentList(func_get_args(), $_types, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk, $_types);
        }

        /*
         * check arguments
         */
        for ($i = 0; $i < 3; $i++)
        {
            $test = func_get_arg($i);
            if ($test < 0 || $test > 255) {
                trigger_error("Invalid value for argument ${i}. Must be between 0 and 255, found '${test}' instead.", E_USER_WARNING);
                return false;
            }
        }
        unset($test, $i);

        /* is image? */
        if (!is_resource($this->image)) {
            return false;
        /* GD library is there? */
        } else if (!function_exists('imagecolorset')) {
            return false;
        /* set color */
        } else {
            imagecolorset($this->image, 0, $r, $g, $b);
        }
    }

    /**
     * get the color of this brush
     *
     * Returns an associative array of the red, green and blue
     * values of the current brush color.
     * Returns bool(false) on error.
     *
     * @access   public
     * @return   array
     */
    function getColor()
    {
        if (!is_resource($this->image)) {
            return false;
        } else if (!function_exists('imagecolorsforindex')) {
            return false;
        } else {
            return imagecolorsforindex($this->image, 0);
        }
    }

    /**
     * get a string representation of this object
     *
     * This function is intended to be called when the object
     * is used in a string context.
     *
     * @access   public
     * @return   string
     */
    function toString()
    {
        $name = $this->getName();
        if (is_string($name)) {
            return $name;
        } else {
            return "untitled brush";
        }
    }

    /**
     * clone this object
     *
     * Creates a copy of this object.
     * You are encouraged to reimplement this for each subclass.
     *
     * @access public
     * @return Object
     */
    function cloneObject()
    {
        $object = parent::cloneObject();

        if ($this->isBroken()) {
            return $object;
        } else {
            $width = $this->getWidth();
            $height = $this->getHeight();

            /* create new image */
            if (function_exists('imagecreatetruecolor')) {
                $copiedImage = imagecreatetruecolor($width, $height);
            } else if (function_exists('imagecreate')) {
                $copiedImage = imagecreate($width, $height);
            } else {
                return false;
            }

            imagecopy($copiedImage, $this->image, 0, 0, 0, 0, $width, $height);
            $object->image = $copiedImage;

            return $object;
        }
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $anotherObject
     * are the booth object which have an image resource that is
     * the same, or $anotherObject is an image resource and
     * identical to the one of this object.
     *
     * Returns bool(false) otherwise.
     *
     * @access  public
     * @param   object  $anotherObject  any var to compare this object with
     * @return  bool
     */
    function equals($anotherObject)
    {
        if (is_resource($anotherObject)) {
            if ($this->image === $anotherObject) {
                return true;
            } else {
                return false;
            }
        }
        if (is_object($anotherObject) && method_exists($anotherObject, 'getResource')) {
            return ( $this->image === $anotherObject->getResource() );
        }
    }

    /**
     * Destructor
     *
     * Automatically free memory for the image if object gets deleted.
     * Note that this is a PHP 5 feature. In PHP 4 you had to call
     * this function by hand.
     *
     * @access public
     * @ignore
     */
    function __destruct()
    {
        if (function_exists('imagedestroy')) {
            if (is_resource($this->image)) {
                imagedestroy($this->image);
            }
        }
    }

}

?>