<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * abstract InputStream
 *
 * This is an abstract super-class for all filesystem wrappers.
 *
 * @abstract
 * @access      public
 * @package     yana
 * @subpackage  io
 */
class InputStream extends Object
{
    /**#@+
     * @access  protected
     * @ignore
     */

    /** @var  string */ var $path       = "";
    /** @var  array */  var $attributes = array();

    /**#@-*/

    /**
     * constructor
     *
     * Create a new instance of this class.
     *
     * @param  string  $filename
     */
    function InputStream($filename)
    {
        assert('is_string($filename); /* Wrong argument type for argument 1. String expected. */');
        /* settype to STRING */
        $filename = (string) $filename;
        $this->path = $filename;
        $this->resetStats();
    }

    /**
     * abstract read
     *
     * Since this is an abstract method, it is to implemented in the sub-classes.
     *
     * @abstract
     * @access  public
     * @return  bool
     *
     * @ignore
     */
    function read()
    {
        /* intentionaly left blank */
        return true;
    }

    /**
     * return stream contents
     *
     * @access  public
     * @return  string
     */
    function get()
    {
        $text = "IO-Stream \"".$this->getPath()."\"\n";
        foreach ($this->attributes as $a => $b)
        {
            $text .= "\t- $a = ".gettype($b)."(".$b.")\n";
        }
        return $text;
    }

    /**
     * abstract toString
     *
     * Since this is an abstract method, it is to implemented in the sub-classes.
     *
     * @abstract
     * @access  public
     * @return  string
     *
     * @ignore
     */
    function toString()
    {
        if (!$this->exists()) {
            return "Path to '".$this->getPath()."' does not exist\n";
        } elseif ($this->isEmpty()) {
            return "Path to '".$this->getPath()."' exists, but it is wether empty or not loaded\n";
        } else {
            return $this->get();
        }
    }

    /**
     * get path to the resource
     *
     * Returns a string with the path and name of the
     * currently open resource (directory or file).
     *
     * {@internal
     *
     * Note: this function has been renamed as of
     * version 2.9 RC3 from "InputStream::getFilename()"
     * for reason of increased plausibility, as input
     * streams may be files AND directories, even though
     * working on files is more common here.
     * The name "getPath()" was choosen to comply with
     * the attribute of the same name in PHP's directory
     * directory class, which is dir::$path.
     *
     * }}
     *
     * @access  public
     * @return  string
     * @name    InputStream::getPath()
     */
    function getPath()
    {
        assert('is_string($this->path); // Unexpected result: $this->path is not a string');
        return ( $this->path );
    }

    /**
     * Alias of InputStream::getPath()
     *
     * @access      public
     * @return      string
     * @deprecated  2.9 RC3
     */
    function getFilename()
    {
        return $this->getPath();
    }

    /**
     * return true, if the input stream resource exists
     *
     * @access  public
     * @return  bool
     */
    function exists()
    {
        return file_exists($this->getPath());
    }

    /**
     * abstract isEmpty
     *
     * Since this is an abstract method, it is to implemented in the sub-classes.
     *
     * @abstract
     * @access  public
     * @return  bool
     *
     * @ignore
     */
    function isEmpty()
    {
        return ( false );
    }

    /**
     * reset file statistics
     *
     * Reset file stats, e.g. after creating a file that did not exist.
     * While this method is never used in THIS class, it will propably
     * be essential for its sub classes.
     *
     * @access  protected
     *
     * @ignore
     */
    function resetStats()
    {
        if (file_exists($this->path) && is_readable($this->path)) {
            clearstatcache();
            $this->attributes['FILESIZE'] = filesize($this->path);
            $this->attributes['MTIME'] = filemtime($this->path);
            clearstatcache();
        }
    }


    /**
     * get time when file was last modified
     *
     * Returns the file MTIME value (from cached value).
     * The result is an UNIX timestamp (UTC),
     * or bool(false) if an error occurs.
     *
     * @access  public
     * @return  int
     * @since   2.8.5
     */
    function getLastModified()
    {
        if (isset($this->attributes['MTIME'])) {
            return $this->attributes['MTIME'];
        } else {
            return false;
        }
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $anotherObject
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class and they both
     * refer to the same filesystem resource.
     *
     * @access  public
     * @param   object  $anotherObject
     * @return  bool
     */
    function equals($anotherObject)
    {
        if (is_object($anotherObject) && method_exists($anotherObject,'getClass') && $anotherObject->getClass() === $this->getClass()) {
            return $this->getPath() === $anotherObject->getPath();
        } else {
            return false;
        }
    }

}

?>