<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * Singleton Language
 *
 * This class may be used to dynamically load additional
 * language files at runtime.
 *
 * @access      public
 * @package     yana
 * @subpackage  core
 */
class Language extends Singleton
{

    /**#@+
     * @ignore
     * @access  private
     */

    /** @var String    */ var $directory   = "";
    /** @var dirStream */ var $dirStream   = null;
    /** @var dirStream */ var $defaultDir  = null;
    /** @var array     */ var $file_loaded = array();

    /**#@-*/

    /**
     * Singleton Constructor
     *
     * @param  string     $directory
     * @param  dirStream  $dirStream
     * @param  string     $defaultDir
     *
     * @ignore
     */
    function Language($directory, $dirStream, $defaultDir)
    {
        $this->checkSingletonConstraint();
        assert('is_string($directory) && is_dir($directory);');
        assert('is_string($defaultDir) && is_dir($defaultDir);');
        /* settype to STRING */
        $directory  = (string) $directory;
        $defaultDir = (string) $defaultDir;

        $this->directory  = $directory;
        $this->dirStream  =& $dirStream;
        $this->dirStream->read();
        $this->defaultDir = $defaultDir;
    }

    /**
     * read language strings from a file
     *
     * The file should be taken from 'languages/*.config',
     * without path and file extension.
     *
     * The file contents will automatically be stored in the global
     * registry and can be accessed using $YANA->getVar('LANGUAGE.SOME_VALUE').
     *
     * This function issues an E_USER_NOTICE if the file does not exist.
     * It returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $filename
     * @return  bool
     */
    function read($filename)
    {
        assert('is_string($filename);');
        /* settype to STRING */
        $filename = (string) $filename;
        global $YANA;

        $fileId = strtoupper($filename);

        if (!preg_match("/^[\w_-\d]+$/i", $filename)) {
            trigger_error("The provided language-file id '".$filename."' contains illegal characters. Be aware that only alphanumeric (a-z,0-9,-,_) characters are allowed.", E_USER_NOTICE);
            return false;
        }
        if (isset($this->file_loaded[$fileId]) && $this->file_loaded[$fileId] === true) {
            return true;
        }

        /* preload defaults */
        $defaultFile = $this->defaultDir.$filename.".config";
        if (file_exists($defaultFile)) {
            $defaultLanguage = SML::getFile($defaultFile, CASE_UPPER);
            $this->file_loaded[$fileId] = true;
        } else {
            $defaultLanguage = array();
            $this->file_loaded[$fileId] = false;
        }
        assert('is_array($defaultLanguage);');

        /* override defaults where available */
        $selectedFile = $this->directory.$filename.".config";
        if ($selectedFile !== $defaultFile) {
            if (file_exists($selectedFile)){
                $selectedLanguage = SML::getFile($selectedFile, CASE_UPPER);
                $this->file_loaded[$fileId] = true;
                assert('is_array($selectedLanguage);'); ;
                assert('is_array($defaultLanguage);');
            } else {
                /* No specific language file found, falling back to defaults. */
                trigger_error("No language-file '".$selectedFile."' found for id '".$filename."'. Falling back to default language.", E_USER_NOTICE);
            }
        } else {
            $selectedLanguage = array();
        }

        $LANGUAGE = Hashtable::merge($defaultLanguage, $selectedLanguage);
        unset($selectedLanguage,$defaultLanguage);
        assert('is_array($LANGUAGE);');

        if (!empty($LANGUAGE)) {
            $YANA->merge("*", $LANGUAGE);
            return true;
        } else {
            return false;
        }
    }

    /**
     * getContent
     *
     * @access  public
     * @return  array  numeric array of strings
     *
     * @ignore
     */
    function getContent()
    {
        $content = $this->dirStream->getContent();
        assert('is_array($content);');
        if (is_array($content)) {
            return $content;
        } else {
            return array();
        }
    }

    /**
     * setDirectory
     *
     * @access  public
     * @return  bool
     *
     * @ignore
     */
    function setDirectory($directory)
    {
        global $YANA;
        if (is_string($directory) && preg_match('/^[\w\d]+$/s', $directory) && is_dir($this->dirStream->getPath().$directory)) {
            $this->directory = $this->dirStream->getPath().$directory.'/';
            $this->read('default');
            return true;
        } else {
            trigger_error("Invalid language directory '".print_r($directory,true)."'.", E_USER_NOTICE);
            return false;
        }
    }

    /**
     * getContent
     *
     * @access  public
     * @return  string
     *
     * @ignore
     */
    function getDirectory()
    {
        assert('is_string($this->directory);');
        return $this->directory;
    }

    /**
     * getInfo
     *
     * @access  public
     * @param   string  $filename  relative path to config file
     * @return  array
     *
     * @ignore
     */
    function getInfo($filename)
    {
        assert('is_string($filename);');
        /* settype to STRING */
        $filename = (string) $filename;

        $file = ($this->dirStream->getPath()).$filename;
        if (is_file($file)) {
            $REF = SML::getFile($file, CASE_UPPER);
            assert('is_array($REF["LANGUAGE_INFO"]);');
            if (is_array($REF['LANGUAGE_INFO'])) {
                $REF['LANGUAGE_INFO']['LAST_CHANGE'] = filemtime($file);
                clearstatcache();
                return $REF["LANGUAGE_INFO"];
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

}

?>
