<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * Registry
 *
 * This class implements a registry.
 * You can operate on it via keys, identifying
 * scalar values or arrays inside the registry.
 *
 * Example for usage of the $key parameter:
 * <pre>
 * Array {
 *     ID1 => Array {
 *         ID2 => 'value'
 *     }
 * }
 * </pre>
 * To get the string 'value' from the array above
 * use $key = 'ID1.ID2'
 *
 * The wildcard '*' may be used to refer to the array
 * as a whole.
 *
 * @access     public
 * @package    yana
 * @subpackage vdrive
 */
class Registry extends VDrive
{

    /**
     * retrieves var from registry
     *
     * This returns the var identified by $key.
     *
     * Note that you may use the following magic keys
     * FOR THIS FUNCTION ONLY:
     *
     * <ul>
     *    <li> <BOOLEAN> returns an one-dimensional associative array of all values converted to boolean  </li>
     *    <li> <STRING>  returns an one-dimensional associative array of all values converted to string, arrays are mapped to bool(true) </li>
     *    <li> <ARRAY>   returns an array of all arrays, scalar values are mapped to an empty array </li>
     * </ul>
     *
     * Returns bool(false) on error;
     *
     * @access  public
     * @param   string  $key  (optional)
     * @return  mixed
     */
    function getVar($key = "*")
    {
        /* argument 1 */ assert('is_string($key);   // wrong argument type for argument 1, string expected');
        /* settype to STRING */
        $key = (string) $key;

        assert('is_array($this->options);');
        if (empty($key) || $key == "*") {
            return $this->options;
        } else if (strspn($key,'<')==1) {
            switch (strtoupper($key)) {

                case '<BOOLEAN>':
                    if (is_array($this->options)) {
                        assert('!isset($a); /* cannot redeclare variable $a */');
                        foreach (array_keys($this->options) as $a)
                        {
                            $RESULT[$a] = true;
                        }
                        unset($a);
                    } else {
                        $RESULT = true;
                    }
                    return $RESULT;
                break;

                case '<STRING>':
                    if (is_array($this->options)) {
                        assert('!isset($a); /* cannot redeclare variable $a */');
                        assert('!isset($b); /* cannot redeclare variable $b */');
                        foreach ($this->options as $a => $b)
                        {
                            if (!is_array($b)) {
                                $RESULT[$a] = $b;
                            } else {
                                $RESULT[$a] = true;
                            }
                        }
                        unset($a,$b);
                    } else {
                      $RESULT = $this->options;
                    }
                    return $RESULT;
                break;

                case '<ARRAY>':
                    if (is_array($this->options)) {
                        assert('!isset($a); /* cannot redeclare variable $a */');
                        assert('!isset($b); /* cannot redeclare variable $b */');
                        foreach ($this->options as $a => $b)
                        {
                            if (!is_array($b)) {
                                $RESULT[$a] = array();
                            } else {
                                $RESULT[$a] = $b;
                            }
                        }
                        unset($a,$b);
                    } else {
                      $RESULT = array();
                    }
                    return $RESULT;
                break;

                default:
                    return false;
                break;

            } /* end switch */
      } else {
          $result =& Hashtable::get($this->options, $key);
          if (is_null($result)) {
              return false;
          } else {
              return $result;
          }
      } /* end if */

    } /* end getVar */

    /**
     * sets var on registry by Reference
     *
     * Sets the element identified by $key
     * to $value by passing it's reference.
     * If the value does not exist it gets inserted.
     * If a previous value existed the value gets updated.
     *
     * This function returns bool(false) if $key = '*'
     * and $value is not an array - which is: trying
     * overwrite the complete registry with a non-array value.
     * It returns bool(true) otherwise.
     *
     * @access  public
     * @param   string  $key
     * @param   mixed   &$value
     * @return  bool
     * @since   2.8.5
     */
    function setVarByReference($key, &$value)
    {
        /* argument 1 */ assert('is_string($key);   // wrong argument type for argument 1, string expected');
        /* settype to STRING */
        $key = (string) $key;

        if ($key == "" || $key == "*") {
            $this->options =& $value;
            return true;
        } else {
            return Hashtable::setByReference($this->options, $key, &$value);
        }

    }

    /**
     * sets var on registry
     *
     * Sets the element identified by $key
     * to $value.
     * If the value does not exist it gets inserted.
     * If a previous value existed the value gets updated.
     *
     * This function returns bool(false) if $key = '*'
     * and $value is not an array - which is: trying
     * overwrite the complete registry with a non-array value.
     * It returns bool(true) otherwise.
     *
     * @access  public
     * @param   string  $key
     * @param   mixed   $value
     * @return  bool
     */
    function setVar($key, $value)
    {
        return $this->setVarByReference($key, $value);
    }

    /**
     * merges the value at adresse $key with the provided array data
     *
     * Merges the element identified by $key
     * with the array $value.
     *
     * Returns bool(false) on error.
     *
     * @access  public
     * @param   string  $key
     * @param   array   $array
     * @return  bool
     */
    function merge($key, $array)
    {
        /* argument 1 */ assert('is_string($key);   // wrong argument type for argument 1, string expected');
        /* argument 2 */ assert('is_array($array);  // wrong argument type for argument 2, array expected');

        /* settype to STRING
         *            ARRAY
         */
        $key   = (string) $key;
        $array = (array)  $array;

        if ($key == "" || $key == "*") {
            $this->options = Hashtable::merge($this->options, $array);
            return true;
        } else {
            $options =& Hashtable::get($this->options, $key);
            if (is_null($options)) {
                return Hashtable::set($this->options, $key, $array);
            } else {
                $options = Hashtable::merge($options, $array);
                return true;
            }
        }
    }

    /**
     * removes var from registry
     *
     * Unsets the element identified by $key in the
     * registry. Returns bool(false) if the element
     * does not exist or the key is invalid.
     * Returns bool(true) otherwise.
     *
     * @access  public
     * @param   string  $key
     * @return  bool
     */
    function unsetVar($key)
    {
        assert('is_string($key);');
        /* settype to STRING */
        $key = (string) $key;

        if ($key == "" || $key == "*") {
            $this->options = $this->attributes;
            return true;
        } else {
            return Hashtable::remove($this->options, $key);
        }

    }

    /**
     * sets the type of a var on registry
     *
     * Set the data type of the element identified by $key
     * to $type.
     *
     * Returns bool(false) if the element is NULL or does not exist,
     * or the $type parameter is invalid. Returns bool(true) otherwise.
     *
     * @access  public
     * @param   string  $key
     * @param   string  $type
     * @return  bool
     */
    function setType($key, $type)
    {
        /* argument 1 */ assert('is_string($key);   // wrong argument type for argument 1, string expected');
        /* argument 2 */ assert('is_string($type);  // wrong argument type for argument 2, string expected');
        /* settype to STRING */
        $key  = (string) $key;
        $type = (string) $type;
        return Hashtable::setType($this->options, $key, $type);
    }

}

?>