<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * checked reading of file sources
 *
 * This class adds functionality to compute the checksum
 * of a resource and failsafe reading of resources.
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 */
class SecureInputStream extends InputStream
{

    /**#@+
     * @ignore
     * @access  protected
     */

    /** @var string */ var $md5     = null;
    /** @var array  */ var $content = null;
    /** @var array  */ var $options = array();

    /**#@-*/

    /**
     * read file contents
     *
     * Returns bool(true) on success and
     * bool(false) on error.
     *
     * @access  public
     * @return  bool
     */
    function read()
    {
        if (!$this->exists()) {
            return false;
        } else {
            $this->content = file($this->getPath());
            $this->options = $this->attributes;
            /* calculate hash */
            $this->md5     = md5(implode('', $this->content));
            return true;
        }
    }

    /**
     * read file contents
     *
     * Automatically restarts reading if the file-resource
     * is temporarily not available after waiting for 0.5 seconds.
     *
     * The process is aborted if it fails 3 times.
     *
     * @access  public
     * @return  bool
     */
    function failSafeRead()
    {
        for ($i = 0; $i < 3; $i++)
        {
            if ($this->read()) {
              return true;
              break;
            } else {
              sleep(0.5);
            }
        }
        return false;
    }

    /**
     * get file contents
     *
     * @access  public
     * @return  string
     */
    function get()
    {
        if (!$this->isEmpty()) {
            return implode("", $this->content);
        } else {
            return "";
        }
    }

    /**
     * return file contents as string
     *
     * @access  public
     * @return  string
     */
    function getFileContent()
    {
        if (is_string($this->content)) {
            return (string) $this->content;
        } else if (is_array($this->content)) {
            return implode("", $this->content);
        } else {
            return "";
        }
    }

    /**
     * alias of get()
     *
     * @access  public
     * @return  string
     */
    function toString()
    {
        if (!$this->exists()) {
            return "File ".$this->getPath()." does not exist\n";
        } elseif ($this->isEmpty()) {
            return "File ".$this->getPath()." is wether empty or not loaded\n";
        } else {
            return $this->get();
        }
    }

    /**
     * returns bool(true) if the source is empty or not loaded
     *
     * @access  public
     * @return  bool
     */
    function isEmpty()
    {
        if (is_null($this->content)) {
            return true;
        } else if (is_array($this->content)) {
            return count($this->content) == 0;
        } else if (is_scalar($this->content)) {
            return empty($this->content);
        } else {
            /* the source is invalid */
            return true;
        }
    }

    /**
     * return crc32 checksum for this file
     *
     * The filename parameter became available
     * in version 2.8.5
     *
     * This function has two synopsis. You may decide
     * either to call it statically or based on a current instance.
     * <ul>
     *  <li> $md5 = SecureInputStream::getMd5(string $filename) </li>
     *  <li> $md5 = $file->getMd5(); </li>
     * </ul>
     * Both will return a MD5 hash. The first one for the file with the
     * name you provided, the second for the file currently represented
     * by the object $file (where $file is an instance of
     * SecureInputStream or a derived sub-class).
     *
     * Note: This function has been renamed in version 2.9 RC3 from
     * "SecureInputStream::checksum", to better comply with the
     * framework's naming convention
     *
     * If you prefer a hash value over a checksum, you may want to have
     * a look at {@link SecureInputStream::getMd5()} instead.
     *
     * @access  public
     * @param   string  $filename
     * @return  int
     * @name    SecureInputStream::getCrc32()
     * @see     SecureInputStream::getMd5()
     */
    function getCrc32($filename = "")
    {
        if (!is_string($filename)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($filename)), E_USER_WARNING);
            return false;
        } else if (!empty($filename)) {
            if (!is_file($filename)) {
                trigger_error("Unable to calculate checksum. The file '${filename}' does not exist.", E_USER_NOTICE);
                return false;
            }
        } else {
            $filename = $this->getPath();
        }

        $source = implode("",file($filename));

        if (empty($source)) {
            return 0;
        } elseif (is_string($source)) {
            return crc32($source);
        } else {
            /* the source is invalid */
            return 0;
        }
    }

    /**
     * return md5 hash for this file
     *
     * This function calculates and returns the MD5
     * hash-string for a file.
     *
     * Usefull to check whether or not a
     * file has been changed since last access.
     *
     * This function has two synopsis. You may decide
     * either to call it statically or based on a current instance.
     * <ul>
     *  <li> $md5 = SecureInputStream::getMd5(string $filename) </li>
     *  <li> $md5 = $file->getMd5(); </li>
     * </ul>
     * Both will return a MD5 hash. The first one for the file with the
     * name you provided, the second for the file currently represented
     * by the object $file (where $file is an instance of
     * SecureInputStream or a derived sub-class).
     *
     * Note: results are cached for the current file.
     *
     * Note: This function has been renamed in version 2.9 RC3 from
     * "SecureInputStream::md5Checksum", as the MD5 algorithm does
     * not really return a checksum, but a hash value.
     *
     * If you prefer a checksum over a hash value, you may want to have
     * a look at {@link SecureInputStream::getCrc32()} instead.
     *
     * @access  public
     * @return  string
     * @since   2.8.5
     * @name    SecureInputStream::getMd5()
     * @see     SecureInputStream::getCrc32()
     */
    function getMd5($filename = "")
    {
        /* input checking*/
        if (!is_string($filename)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($filename)), E_USER_WARNING);
            return false;
        }

        if (!empty($filename)) {
            if (!is_file($filename)) {
                trigger_error("Unable to calculate MD5 hash. The file '${filename}' does not exist.", E_USER_NOTICE);
                return false;
            }
        } else {
            if (isset($this->md5)) {
                assert('is_string($this->md5);');
                return $this->md5;
            } else {
                $filename = $this->getPath();
                assert('is_file($filename);');
            }
        }
        $md5 = md5_file($filename);

        if (!is_string($md5)) {
            return false;
        } else {
            return $md5;
        }
    }

    /**
     * reload file contents
     *
     * You accidently did something wrong with the file?
     * Calling this will reload the file from disk while
     * reseting its current state.
     *
     * @access  public
     */
    function reset()
    {
        $this->read();
    }

}

?>
