<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * decorator SmartTemplate
 *
 * This implements a decorator class for the smarty
 * template engine. The use of the "decorator" pattern
 * actually means it "implements an API" on top
 * of the original.
 *
 * Note that this does not touch or even change the
 * engine itself.
 *
 * This class provides two things: most obviously it
 * provides some features, that smarty does not have and
 * in addition it does some more type checking and
 * automates the initialization process.
 *
 * @access      public
 * @package     yana
 * @subpackage  core
 */
class SmartTemplate extends Object
{

    /**#@+
     * @ignore
     * @access  private
     */

    /** @var string */  var $path   = "";
    /** @var bool */    var $exists = false;
    /** @var smarty */  var $smarty = null;

    /**#@-*/

    /**
     * create an instance
     *
     * You may enter a filename of a template you want to use.
     *
     * @param  string  $filename  (optional)
     */
    function SmartTemplate($filename = "")
    {
        /* settype to STRING */
        $filename = (string) $filename;
        global $YANA;

        /* Config Smarty */
        $this->smarty                  = new Smarty();
        $this->smarty->left_delimiter  = YANA_LEFT_DELIMITER;
        $this->smarty->right_delimiter = YANA_RIGHT_DELIMITER;
        if (isset($YANA)) {
            DisplayUtility::loadSmilies($YANA->getVar("PROFILE.SMILEYDIR"));
            $this->smarty->template_dir = '.';
            $this->smarty->config_dir   = $YANA->getVar('SKINCONFIGDIR');
            $this->smarty->compile_dir  = $YANA->getVar('TEMPDIR');
        } else {
            $this->smarty->template_dir = '.';
            $this->smarty->config_dir   = 'skins/.config/';
            $this->smarty->compile_dir  = 'cache/';
        }
        $this->smarty->php_handling     = SMARTY_PHP_REMOVE;
        $this->smarty->security         = true;
        $this->smarty->caching          = 0;
        if (!defined('YANA_ERROR_REPORTING') || !defined('YANA_ERROR_OFF')) {
            $this->smarty->compile_check = true;
        } else {
            $this->smarty->compile_check = YANA_ERROR_REPORTING !== YANA_ERROR_OFF;
        }
        $this->smarty->use_sub_dirs     = false;
        $this->smarty->register_modifier('embeddedTags',        array('DisplayUtility', 'embeddedTags'));
        $this->smarty->register_modifier('replaceToken',        array('Layout',         'replaceToken'));
        $this->smarty->register_modifier('css',                 array('DisplayUtility', 'css'));
        $this->smarty->register_modifier('date',                array('DisplayUtility', 'date'));
        $this->smarty->register_modifier('entities',            array('DisplayUtility', 'entities'));
        $this->smarty->register_modifier('href',                array('DisplayUtility', 'href'));
        $this->smarty->register_modifier('scanForAt',           array('DisplayUtility', 'scanForAt'));
        $this->smarty->register_modifier('smilies',             array('DisplayUtility', 'smilies'));
        $this->smarty->register_modifier('url',                 array('DisplayUtility', 'url'));
        $this->smarty->register_function('printArray',          array('DisplayUtility', 'printArray'));
        $this->smarty->register_function('printUnorderedList',  array('DisplayUtility', 'printUnorderedList'));
        $this->smarty->register_function('import',              array('DisplayUtility', 'guiImport'));
        $this->smarty->register_function('smilies' ,            array('DisplayUtility', 'guiSmilies'));
        $this->smarty->register_function('embeddedTags',        array('DisplayUtility', 'guiEmbeddedTags'));
        $this->smarty->register_function('create',              array('GuiCreator',     'create'));
        $this->smarty->register_function('sml_load',            array('DisplayUtility', 'guiSmlLoad'));
        $this->smarty->register_block('loop',                   array('DisplayUtility', 'loopArray'));
        $this->smarty->default_modifiers =                      array('replaceToken',   'scanForAt');
        $this->smarty->register_prefilter(                      array('DisplayUtility', 'processLanguageInstructions'));
        $this->smarty->register_prefilter(                      array('DisplayUtility', 'htmlPreProcessor'));
        $this->smarty->register_postfilter(                     array('DisplayUtility', 'htmlPostProcessor'));
        $this->smarty->register_outputfilter(                   array('DisplayUtility', 'outputFilter'));
        $this->setPath($filename);
    }

    /**
     * toString
     *
     * @access public
     * @return string
     *
     * @ignore
     */
    function toString()
    {
        if (!$this->exists()) {
            return "The template file '".$this->getPath()."' does not exist\n";
        } else {
            return "Template class '".get_class($this)."' using template file '".$this->getPath()."'.\n";
        }
    }

    /**
     * fetch a template or template var
     *
     * There are three ways to call this function:
     *
     * If you call $smartTemplate->get($varName)
     * it will get the template var $varName
     * and return it.
     *
     * If you call $smartTemplate->get("*")
     * with the wildcard '*' or an empty string ''
     * it will return an associative array
     * containing all template vars.
     *
     * If you call $smartTemplate->get()
     * without a parameter it will fetch
     * the current template and return it
     * as a string.
     *
     * If the $key parameter is NULL,
     * the second parameter can be used to change
     * the way how variables are imported from the
     * global registry to the template.
     * If set to false, existing template vars will not
     * be replaced by vars of the same name in the
     * registry. Otherwise template vars will get
     * overwritten if the global registry has another value.
     *
     * @access  public
     * @param   string  $key        variable-name
     * @param   bool    $overwrite  true = on / false = off
     * @return  mixed
     */
    function get($key = null, $overwrite = true)
    {
        global $YANA;

        /* 1) fetch template */
        if (is_null($key)) {
            if (isset($YANA)) {
                /* import vars from global registry */
                if ($overwrite === true) {
                    /* import and overwrite */
                    $this->smarty->assign($YANA->getVar('*'));
                } else if ($overwrite === false) {
                    /* import only non-existing vars */
                    $vars = $YANA->getVar('*');
                    $vars = Hashtable::changeCase($vars, CASE_UPPER);
                    $vars = Hashtable::merge($vars, $this->smarty->get_template_vars());
                    $this->smarty->assign($vars);
                } else {

                    trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Boolean', gettype($overwrite)), E_USER_WARNING);
                    return false;
                }
            }
            return $this->smarty->fetch($this->getPath());

        /* 2) get all template vars */
        } else if (empty($key) || $key === '*') {

            return $this->smarty->get_template_vars();

        /* 3) get one template var, identified by $key */
        } else if (is_scalar($key)) {

            $resource = $this->smarty->get_template_vars();
            assert('is_array($resource); /* unexpected result: $resource should be an array */');
            return Hashtable::get($resource, $key);

        /* 4) error: invalid input */
        } else {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Scalar value', gettype($key)), E_USER_WARNING);
            return false;

        }

    }

    /**
     * output the template
     *
     * This function will fetch and print
     * the current template to the screen
     *
     * @access  public
     */
    function write()
    {
        if (ob_get_length() !== false) {
            ob_flush();
        }
        echo $this->get();
    }

    /**
     * bypass template class
     *
     * This function is used to unbox
     * the smarty instance inside the
     * object. It may be used to bypass
     * the template class in cases
     * where direct access to the smarty
     * template engine is necessary.
     *
     * @access  public
     * @return  Smarty
     * @since   2.8.9
     */
    function &getSmarty()
    {
        return $this->smarty;
    }

    /**
     * assign a variable by value
     *
     * This assigns the $var to the name $varName.
     *
     * Unlike Smarty's "assign()" this function takes an
     * additional value for $varName:
     *
     * You may use the wildcard '*' to:
     *
     * <ol>
     *     <li>    $smartTemplate('*', string $var)    assign a string to the global token %INSERT_CONTENT_HERE%    </li>
     *     <li>    $smartTemplate('*', array  $var)    merge an associative array with the template vars    </li>
     * </ol>
     *
     * @access  public
     * @param   string  $varName  address
     * @param   mixed   $var      some new value
     */
    function insert($varName, $var)
    {
        /* 1) error: invalid input */
        if (!is_scalar($varName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Scalar value', gettype($varName)), E_USER_WARNING);
            return false;

        /* 2) assign to standard out */
        } else if ($varName == '*' && is_string($var)) {
            $this->smarty->assign('INSERT_CONTENT_HERE', $var);

        /* 3) assign to global namespace */
        } else if ($varName == '*') {
            $this->smarty->assign($var);

        /* 4) assigned var is an object */
        } else if (is_object($var)) {
            
            $this->smarty->assign_by_ref($varName, $var);

        /* 5) assign to var identified by $varName */
        } else {
            $this->smarty->assign($varName, $var);
        }

    }

    /**
     * assign a variable by reference
     *
     * This assigns the $var to the name $varName.
     *
     * Unlike Smarty's "assign()" this function takes an
     * additional value for $varName:
     *
     * You may use the wildcard '*' to:
     *
     * <ol>
     *     <li>    $smartTemplate('*', string $var)    assign a string to the global token %INSERT_CONTENT_HERE%    </li>
     *     <li>    $smartTemplate('*', array  $var)    merge an associative array with the template vars    </li>
     * </ol>
     *
     * @access  public
     * @param   string  $varName  address
     * @param   mixed   &$var     some new value
     */
    function insertByReference($varName, &$var)
    {
        /* 1) error: invalid input */
        if (!is_scalar($varName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Scalar value', gettype($varName)), E_USER_WARNING);

        /* 2) assign to global namespace */
        } else if ($varName === '*') {

            /* 2.a) assign contents to global namespace */
            if (is_array($var)) {

                assert('!isset($key); /* cannot redeclare variable $key */');
                foreach (array_keys($var) as $key)
                {
                    $this->smarty->assign_by_ref($key, $var[$key]);
                }
                unset($key);

            /* 2.b) assign to standard output */
            } else if (is_string($var)) {

                $this->smarty->assign_by_ref('INSERT_CONTENT_HERE', $var);

            /* 2.c) error: invalid input */
            } else {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String or array', gettype($var)), E_USER_WARNING);
            }

        /* 3) assign to var identified by $varName */
        } else {
            $this->smarty->assign_by_ref($varName, $var);
        }

    }

    /**
     * insert a file
     *
     * Load a file and assign it's content as a string
     * to the provided template var.
     *
     * You may use the wildcard '*' to assign the file content
     * to the global token %INSERT_CONTENT_HERE%
     *
     * @access  public
     * @param   string  $varName
     * @param   string  $filename
     * @return  bool
     */
    function insertFile($varName, $filename)
    {
        /* input check */
        if (!is_scalar($varName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Scalar value', gettype($varName)), E_USER_WARNING);
            return false;
        }
        if (!is_string($filename)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($filename)), E_USER_WARNING);
            return false;
        }
        /* settype to STRING */
        $varName  = (string) $varName;

        if (empty($varName) || $varName == "*") {
            $varName = 'INSERT_CONTENT_HERE';
        }

        if (file_exists($filename) && is_readable($filename)) {
            $t = new smartTemplate($filename);
            $t->insert('FILE_IS_INCLUDE', "true");
            $include = $t->get();
            $this->smarty->assign($varName, $include);
            return true;
        } else {
            return false;
        }

    }

    /**
     * set filename of current template
     *
     * You may set another filename of a template to fetch.
     *
     * Template files may not have a reserved extension like "htaccess", "php", "config" or the like.
     * Files should be adressed from the root, where "index.php" is stored.
     * If you can't access a file, the file does not exist or is not readable, an error is thrown.
     *
     * @access  public
     * @param   string  $filename  name of the template file
     * @return  bool
     */
    function setPath($filename)
    {
        /* input check */
        if (!is_string($filename)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($filename)), E_USER_WARNING);
            return false;
        }

        /* Reset file binding */
        $this->smarty->clear_assign('MTIME');
        $this->smarty->clear_assign('ATIME');
        $this->smarty->clear_assign('CTIME');
        $this->exists = false;
        $this->path = "";

        if (empty($filename)) {

            return false;

        } else if (preg_match("/.*\.(register|config|cfg|lock|dat|htaccess|php|inc|conf)/Ui", $filename)) {

            $this->smarty->trigger_error("ERROR 403: Access denied for file '$filename'.");
            return false;

        } elseif (!file_exists($filename) || !is_readable($filename)) {

            $this->smarty->trigger_error("ERROR 404: File '$filename' not found.");
            return false;

        } else {

            $this->path = $filename;

            /* Append Template Data */
            $this->smarty->assign('MTIME', filemtime($filename));
            $this->smarty->assign('ATIME', fileatime($filename));
            $this->smarty->assign('CTIME', filectime($filename));
            $this->smarty->assign_by_ref('FILENAME', $this->path);
            preg_match("/^.*\//", $filename, $basedir);
            $this->smarty->assign('BASEDIR', $basedir[0]);
            clearstatcache();

            $this->exists = true;
            return true;

        }
    }

    /**
     * get path and name of current template
     *
     * Returns a string containing the path and filename
     * of the current template relative to the framework's
     * home directory.
     *
     * @access  public
     * @return  string
     */
    function getPath()
    {
        assert('is_string($this->path);');
        return $this->path;
    }

    /**
     * exists
     *
     * Returns bool(true) if template exists and bool(false)
     * otherwise.
     *
     * @access  public
     * @return  bool
     *
     * @ignore
     */
    function exists()
    {
        assert('is_bool($this->exists);');
        return $this->exists;
    }

    /**
     * clearCache
     *
     * Deletes all temporary files in the 'cache/' directory.
     *
     * @access  public
     * @static
     * @ignore
     */
    function clearCache()
    {
        global $YANA;

        $dir   = $YANA->getVar('TEMPDIR');
        $files = dirlist($dir, '.php|.cache|.tmp');
        for ($i = 0; $i < count($files); $i++)
        {
            $current_file = $dir.$files[$i];
            /* If file can't be deleted due to active write-protection (e.g. when running under Windows), check wether this can be fixed. */
            if (!is_writeable($current_file)) {
                chmod($current_file,0666);
            }
            if (is_writeable($current_file)) {
                unlink($current_file);
            } else {
                trigger_error("Unable to delete file '${current_file}', because the file is not writeable.", E_USER_WARNING);
            }
        }
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $anotherObject
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class and they both
     * refer to the same filesystem resource.
     *
     * @access   public
     * @param    object  $anotherObject
     * @return   string
     */
    function equals($anotherObject)
    {
        if (is_object($anotherObject) && method_exists($anotherObject, 'getClass') && $anotherObject->getClass() === $this->getClass()) {
            return $this->getPath() === $anotherObject->getPath();
        } else {
            return false;
        }
    }

}
?>
