<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**#@+
 * @ignore
 */
if (!defined('YANA_DB_AUTO')) {
    define('YANA_DB_AUTO', 'auto');
}
if (!defined('YANA_DB_CONSTRAINT_SYNTAX')) {
    define('YANA_DB_CONSTRAINT_SYNTAX', "/^\s*(?:(?:(?:-| |\!)?\\\$[\w\d_]+ ?|true|false|null|-?\d+|\&\&?|(?:preg_match|ereg|eregi)\((?:'[^']*'|\"[^\"]*\"),\s*\\\$[\w\d_]+\)|[\&\|\!\~\-\*\/\%\+\<\>]|\[\"[^\"\]\[]+\"\]|\[\'[^\'\]\[]+\'\]|\"[^\"]*\"|\'[^\']*\'|\d+(?:\.\d*)?|(?:\=|\!|\<|\>)\={1,2})(?:\s+|$))*\s*$/i");
}
/**#@-*/

/**
 * database schema file
 *
 * This wrapper class represents the structure of a database
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 */
class StructureFile extends SML
{
    /**#@+
     * @ignore
     * @access  private
     */

    /** @var  array  */   var $cached_fields = array();

    /**#@-*/


    /**
     * constructor
     *
     * Create a new instance of this class.
     *
     * @param  string $filename
     */
    function StructureFile($filename)
    {
        /* call superclass constructor */
        $constructor = get_parent_class(__CLASS__);
        parent::$constructor($filename, CASE_UPPER);

        /* specific extensions */
        assert('is_string($filename);');
        $this->path = StructureFile::_getFilename($filename);
        $this->resetStats();
    }


    /**
     * read and initialize the file
     *
     * Always call this first before anything else.
     * Returns the file content on success and bool(false) on error.
     *
     * @access  public
     * @return  array|bool
     * @since   2.8.8
     */
    function read()
    {
        if (parent::read() !== false) {
            foreach ($this->content['TABLES'] as $k => $v)
            {
                $this->content['TABLES'][$k]['PRIMARY_KEY'] = strtolower($v['PRIMARY_KEY']);
                if (isset($v['FOREIGN_KEYS'])) {
                    assert('!isset($k1); /* cannot redeclare variable $k1 */');
                    assert('!isset($v1); /* cannot redeclare variable $v1 */');
                    foreach ($v['FOREIGN_KEYS'] as $k1 => $v1)
                    {
                        $this->content['TABLES'][$k]['FOREIGN_KEYS'][$k1] = strtolower($v1);
                    }
                    unset($k1,$v1);
                }
            }
            if (isset($this->content['INCLUDE'])) {
                if (is_string($this->content['INCLUDE'])) {
                    $test = $this->addFile($this->content['INCLUDE']);
                    if ($test === false) {
                        trigger_error("Error importing file '".$this->content['INCLUDE']."'.", E_USER_WARNING);
                    }
                } else if (is_array($this->content['INCLUDE'])) {
                    foreach ($this->content['INCLUDE'] as $filename)
                    {
                        if (is_string($filename)) {
                            $test = $this->addFile($filename);
                            if ($test === false) {
                                trigger_error("Error importing file '${filename}'.", E_USER_WARNING);
                            }
                        } else {
                            trigger_error("Invalid field type in ".__CLASS__."->".__FUNCTION__."().\n\t\tContents of array 'INCLUDE' are supposed to be strings. Found '".gettype($filename)."' instead.", E_USER_NOTICE);
                        }
                    } /* end foreach */
                } else {
                    trigger_error("Invalid field type in ".__CLASS__."->".__FUNCTION__."().\n\t\tField 'INCLUDE' is supposed to be array or string. Found '".gettype($filename)."' instead.", E_USER_NOTICE);
                }
            } /* end if*/
            return $this->content;
        } else {
            return false;
        }
    }

    /**
     * get the compiled structure of the database
     *
     * @access  public
     * @return  mixed
     */
    function getStructure()
    {
        return $this->content;
    }

    /**
     * get the file source
     *
     * Returns the text of the source file containing
     * the database structure as a string or bool(false)
     * on error.
     *
     * @access  public
     * @return  mixed
     */
    function getSource()
    {
        $source = file($this->getPath());
        if (!is_array($source)) {
            return false;
        } else {
            $source = implode("",$source);
            assert('is_string($source);');
            if (empty($source)) {
                return false;
            } else {
                return $source;
            }
        }
    }

    /**
     * check whether a table exists in the current structure
     *
     * Returns bool(true) if a table with the given name is listed
     * in the current structure file and bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access  public
     * @param   string $table
     * @return  bool
     */
    function isTable($table)
    {
        return is_array($this->_getTable($table));
    }

    /**
     * add a new table
     *
     * Returns bool(true) on success and bool(false) otherwise.
     *
     * @access  public
     * @param   string  $table
     * @return  bool
     * @since   2.9
     */
    function addTable($table)
    {
        if (preg_match('/^[\w\d-_]+$/s', $table) && !isset($this->content['TABLES'][strtoupper($table)])) {
            if (!isset($this->content['TABLES'])) {
                $this->content['TABLES'] = array();
            }
            $this->content['TABLES'][strtoupper($table)] = array('CONTENT' => array());
            return true;
        } else {
            return false;
        }
    }

    /**
     * check whether a column exists in the current structure
     *
     * Returns bool(true) if a table named $table is listed
     * in the current structure file and it has a column named $column
     * in its list of contents. Returns bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     */
    function isColumn($table, $column)
    {
        return is_array($this->_getColumn($table, $column));
    }

    /**
     * add a new column
     *
     * Returns bool(true) on success and bool(false) otherwise.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     * @since   2.9
     */
    function addColumn($table, $column)
    {
        $tbl =& $this->_getTable($table);
        if (!is_array($tbl)) {
            /* error: table not found */
            return false;
        } else if (preg_match('/^[\w\d-_]+$/s', $column) && !isset($tbl['CONTENT'][strtoupper($column)])) {
            $tbl['CONTENT'][strtoupper($column)] = array();
            return true;
        } else {
            return false;
        }
    }

    /**
     * check whether a column allows NULL values
     *
     * Returns bool(true) if a table named $table is listed
     * in the current structure file and it has a column named $column
     * that allows undefined values (NULL). Returns bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     * @since   2.8.6
     * @name    StructureFile::isNullable()
     * @see     StructureFile::setNullable()
     */
    function isNullable($table, $column)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else if (!isset($col['REQUIRED'])) {
            return true;
        } else if ($col['REQUIRED'] === false) {
            return true;
        } else if (strcasecmp($col['REQUIRED'], 'false') === 0) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * choose wether a column should be nullable
     *
     * Sets the "required" property of the column.
     * If the argument $isNullable is bool(true), then
     * "required" is set to false. Otherwise it is
     * set to true.
     *
     * If the table or column is not defined, the function
     * returns bool(false).
     * Otherwise it returns bool(true).
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @param   bool    $isNullable
     * @return  bool
     * @since   2.9
     * @name    StructureFile::setNullable()
     * @see     StructureFile::isNullable()
     */
    function setNullable($table, $column, $isNullable)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else {
            if ($isNullable) {
                $col['REQUIRED'] = false;
            } else {
                $col['REQUIRED'] = true;
            }
            return true;
        }
    }

    /**
     * make a column use auto-number / auto-filled values
     *
     * Sets the property "required" of the selected column to
     * the value "auto".
     *
     * This enables the "autofill" feature, which is available for
     * columns of several types. On columns of type "integer" it
     * mimics MySQL's "auto increment" feature.
     * On columns of type "ip" it enters the visitor's remote
     * address (IP) automatically.
     * For type "time" it enters the current server time.
     *
     * However: you should note, that the user input takes
     * precedence over the autofill feature, which defines a default value.
     *
     * Note: this function does not check wether "autofill" does "make sense"
     * for the type of column you selected. It also does not clear the
     * "default" property of the column, if any.
     *
     * Returns bool(false) if the table or column does not exist.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @param   bool    $isNullable
     * @return  bool
     * @since   2.9
     * @name    StructureFile::setAuto()
     * @see     StructureFile::isAuto()
     */
    function setAuto($table, $column)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else {
            $col['REQUIRED'] = YANA_DB_AUTO;
            return true;
        }
    }

    /**
     * check whether a column uses the "autofill" feature
     *
     * Returns bool(true) if a table named $table is listed
     * in the current structure file and it has a column named $column
     * that uses the autofill feature. Returns bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * Autofill is activated by setting the property "REQUIRED"
     * to "AUTO".
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     * @since   2.9
     * @name    StructureFile::isAuto()
     * @see     StructureFile::isAutonumber()
     */
    function isAuto($table, $column)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else if (!isset($col['REQUIRED'])) {
            /* property is undefined */
            return false;
        } else {
            if (strcasecmp($col['REQUIRED'], YANA_DB_AUTO) === 0) {
                return true;
            } else {
                return false;
            }
        }
    }

    /**
     * check whether a column is autonumber/autoincrement
     *
     * Returns bool(true) if a table named $table is listed
     * in the current structure file and it has a column named $column
     * that is an autonumber colummn. Returns bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     * @since   2.8.6
     * @name    StructureFile::isAutonumber()
     */
    function isAutonumber($table, $column)
    {
        $type = $this->getType($table, $column);
        $type = strtolower($type);
        if ($type !== 'int' && $type !== 'integer') {
            return false;
        } else {
            return $this->isAuto($table, $column);
        }
    }

    /**
     * check whether a column is indexed in the current structure
     *
     * Returns bool(true) if a table named $table is listed
     * in the current structure file and it has a column named $column,
     * which has an index. Returns bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     * @since   2.8.5
     * @name    StructureFile::hasIndex()
     * @see     StructureFile::setIndex()
     */
    function hasIndex($table, $column)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else if (!isset($col['INDEX'])) {
            /* index property is undefined */
            return false;
        } else {
            if (is_string($col['INDEX'])) {
                if (strcasecmp($col['INDEX'], 'true') === 0) {
                    return true;
                } else {
                    return false;
                }
            } else if ($col['INDEX'] == true) {
                return true;
            } else {
                return false;
            }
        }
    }

    /**
     * add/remove an index on a column
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @param   bool    $hasIndex
     * @return  bool
     * @since   2.9
     * @name    StructureFile::setIndex()
     * @see     StructureFile::hasIndex()
     */
    function setIndex($table, $column, $hasIndex)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else {
            if ($hasIndex) {
                $col['INDEX'] = true;
            } else {
                $col['INDEX'] = false;
            }
            return true;
        }
    }

    /**
     * check whether the table has a column containing a profile id
     *
     * Returns the name of a column that is supposed to contain the
     * profile id (if any).
     *
     * If no table named $table is listed in the current structure file
     * or $table does not have a column using a profile id, or the
     * specified column does not exist, the function returns bool(false).
     *
     * @access  public
     * @param   string  $table
     * @return  string|bool(false)
     * @since   2.9
     * @name    StructureFile::getProfile()
     * @see     StructureFile::setProfile()
     */
    function getProfile($table)
    {

        $tbl =& $this->_getTable($table);

        if (!is_array($tbl)) {
            /* error: table not found */
            return false;
        } else if (!isset($tbl['PROFILE_KEY'])) {
            return false;
        } else if ($this->isStrict() && !$this->isColumn($table, $tbl['PROFILE_KEY'])) {
            trigger_error("Error in structure file '".$this->getPath."'.\n\t\tTable '$table' refers to a non-existing column '".$tbl['PROFILE_KEY']."' as profile id.", E_USER_WARNING);
            return false;
        } else {
            return $tbl['PROFILE_KEY'];
        }
    }

    /**
     * add/remove a profile reference on a column
     *
     * To unset, leave $column blank.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     * @since   2.9
     * @name    StructureFile::setProfile()
     * @see     StructureFile::getProfile()
     */
    function setProfile($table, $column = null)
    {
        $tbl =& $this->_getTable($table);

        if (!is_array($tbl)) {
            /* error: table not found */
            return false;
        } else if (is_null($column)) {
            if (isset($tbl['PROFILE_KEY'])) {
                unset($tbl['PROFILE_KEY']);
            }
            return true;
        } else if (!$this->isColumn($table, $column)) {
            /* error: column not found */
            return false;
        } else {
            /* settype to STRING */
            $column = (string) $column;
            $column = strtolower($column);
            $tbl['PROFILE_KEY'] = $column;
            return true;
        }
    }

    /**
     * check whether a foreign key exists in the current structure
     *
     * Returns bool(true) if a table named $table is listed
     * in the current structure file and it has a column named $column
     * in its list of foreign keys. Returns bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     */
    function isForeignKey($table, $column)
    {
        /* settype to STRING */
        $column = (string) $column;
        $column = strtoupper($column);

        $tbl =& $this->_getTable($table);

        if (!is_array($tbl)) {
            /* error: table not found */
            return false;
        } else if (!isset($tbl['FOREIGN_KEYS'][$column])) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * add a foreign key constraint
     *
     * To unset, leave $ftable blank.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @param   string  $ftable
     * @return  bool
     */
    function setForeignKey($table, $column, $ftable = null)
    {
        /* settype to STRING */
        $column = (string) $column;
        $column = strtoupper($column);

        $tbl =& $this->_getTable($table);

        if (!is_array($tbl)) {
            /* error: table not found */
            return false;
        } else if (!is_null($ftable) && !$this->isTable($ftable)) {
            /* error: table not found */
            return false;
        } else if (!$this->isColumn($table, $column)) {
            /* error: column not found */
            return false;
        } else if (is_null($ftable)) {
            if (isset($tbl['FOREIGN_KEYS'][$column])) {
                unset($tbl['FOREIGN_KEYS'][$column]);
            }
            return true;
        } else {
            if (!isset($tbl['FOREIGN_KEYS']) || !is_array($tbl['FOREIGN_KEYS'])) {
                $tbl['FOREIGN_KEYS'] = array();
            }
            /* settype to STRING */
            $ftable = (string) $ftable;
            $ftable = strtolower($ftable);
            $tbl['FOREIGN_KEYS'][$column] = $ftable;
            return true;
        }
    }

    /**
     * check whether a primary key exists in the current structure
     *
     * Returns bool(true) if a table named $table is listed
     * in the current structure file and its primary key is named $column.
     * Returns bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     */
    function isPrimaryKey($table, $column)
    {
        /* settype to STRING */
        $column = (string) $column;
        return (strcasecmp($this->getPrimaryKey($table), $column) === 0);
    }

    /**
     * check whether a column has a unique constraint
     *
     * Returns bool(true) if a table named $table is listed
     * in the current structure file, it has a column named $column
     * in its list of contents and this column has a unique constraint.
     * Returns bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     * @since   2.8.5
     * @name    StructureFile::isUnique()
     * @see     StructureFile::setUnique()
     */
    function isUnique($table, $column)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else if (!isset($col['UNIQUE'])) {
            /* property is undefined */
            return false;
        } else if ($col['UNIQUE'] === true || $col['UNIQUE'] === 'true') {
            return true;
        } else {
            return false;
        }
    }

    /**
     * add/remove a unique constraint on a column
     *
     * Returns bool(true) on success and bool(false) otherwise.
     *
     * Note: you don't need to set a "unique" constraint on a
     * primary key. A "primary key" column always requires its
     * values to be unique.
     *
     * Primary keys implicitely have a unique constraint.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @param   bool    $isUnique
     * @return  bool
     * @since   2.9
     * @name    StructureFile::setUnique()
     * @see     StructureFile::isUnique()
     */
    function setUnique($table, $column, $isUnique)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else {
            if ($isUnique) {
                $col['UNIQUE'] = true;
            } else {
                $col['UNIQUE'] = false;
            }
            return true;
        }
    }

    /**
     * return a list of foreign keys defined on a table
     *
     * Returns an associative array of foreign keys.
     * If $table is not a specified in the current structure file,
     * then NULL is returned instead.
     * If the table has no foreign keys, an empty array is returned.
     *
     * Note that this operation is not case sensitive.
     *
     * @access  public
     * @param   string  $table
     * @return  array
     */
    function getForeignKeys($table)
    {
        $tbl =& $this->_getTable($table);

        if (!is_array($tbl)) {
            trigger_error("Unable to get foreign keys. There is no table '${table}' in current structure file.", E_USER_WARNING);
            /* error: table not found */
            return false;
        } else if (!isset($tbl['FOREIGN_KEYS'])) {
            return array();
        } else {
            return $tbl['FOREIGN_KEYS'];
        }
    }

    /**
     * get the primary key of a table
     *
     * Returns the name of the primary key column of $table as a
     * lower-cased string.
     * Returns NULL and issues an E_USER_WARNING if $table is not a listed table in the current
     * structure file.
     * Returns NULL and issues an E_USER_WARNING if there is no primary key for $table.
     *
     * @access  public
     * @param   string  $table
     * @return  string
     */
    function getPrimaryKey($table)
    {
        $tbl =& $this->_getTable($table);

        if (!is_array($tbl)) {
            trigger_error("Unable to get primary key. There is no table '${table}' in current structure file.", E_USER_WARNING);
            /* error: table not found */
            return null;
        } else if (!isset($tbl['PRIMARY_KEY'])) {
            trigger_error("Table '${table}' has no primary key declaration.", E_USER_WARNING);
            return null;
        } else {
            return strtolower($tbl['PRIMARY_KEY']);
        }
    }

    /**
     * set the primary key of a table
     *
     * Select $column as the primary key of $table.
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     */
    function setPrimaryKey($table, $column)
    {
        /* settype to STRING */
        $column = (string) $column;
        $column = strtolower($column);

        $tbl =& $this->_getTable($table);

        if (!is_array($tbl)) {
            /* error: table not found */
            return false;
        } else if (!$this->isColumn($table, $column)) {
            /* error: column not found */
            return false;
        } else {
            $tbl['PRIMARY_KEY'] = $column;
            return true;
        }
    }

    /**
     * get the user description of a column
     *
     * Returns the description text (=comment) of $column in $table as a string or bool(false) if
     * none exists.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  string
     * @since   2.8.6
     * @name    StructureFile::getDescription()
     * @see     StructureFile::setDescription()
     */
    function getDescription($table, $column)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else if (!isset($col['DESCRIPTION'])) {
            return false;
        } else {
            return (string) $col['DESCRIPTION'];
        }
    }

    /**
     * set the description property of a column
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * A description is the "label" of a column. A human-readable
     * small phrase that should tell readers what it is.
     *
     * Note: though you may take any length of text you want,
     * you are best adviced to use really short phrases like
     * 'name', 'address' or 'phone' as a description.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @param   string  $description
     * @return  bool
     * @since   2.9
     * @name    StructureFile::setDescription()
     * @see     StructureFile::getDescription()
     */
    function setDescription($table, $column, $description)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else {
            $col['DESCRIPTION'] = (string) $description;
            return true;
        }
    }

    /**
     * get the maximum length of a column as specified in the structure
     *
     * Returns the 'length' attribute of $column in $table.
     * Returns int(0) if there is no $table, or the table has no
     * column named $column, or the column does not have a 'length'
     * attribute.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  int
     * @name    StructureFile::getLength()
     * @see     StructureFile::setLength()
     */
    function getLength($table, $column)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else {
            if (isset($col['LENGTH'])) {
                $length = $col['LENGTH'];
                if (!is_numeric($length)) {
                    trigger_error("Property 'length' has an invalid value '${length}' in table '${table}' column '${column}'.", E_USER_NOTICE);
                    $length = 0;
                } else {
                    /* settype to INTEGER */
                    $length = (int) $length;
                }
            } else {
                /* property is undefined, defaults to 0 */
                $length = 0;
            }
            return $length;
        }
    }

    /**
     * set the maximum length property of a column
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * The argument $length must be a positive integer.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @param   int     $length
     * @return  bool
     * @since   2.9
     * @name    StructureFile::setLength()
     * @see     StructureFile::getLength()
     */
    function setLength($table, $column, $length)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else if (is_int($length) && $length > 0) {
            $col['LENGTH'] = $length;
            return true;
        } else {
            trigger_error("Invalid length argument '${length}'. Positive integer expected.", E_USER_WARNING);
            /* error: invalid argument */
            return false;
        }
    }

    /**
     * get the data type of a field as specified in the structure
     *
     * Returns the 'type' attribute of $column in $table as a lower-cased string.
     * Returns string("") if there is no $table, or the table has no
     * column named $column, or the column does not have a 'type'
     * attribute.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  string
     * @name    StructureFile::getType()
     * @see     StructureFile::setType()
     */
    function getType($table, $column)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return '';
        } else if (!isset($col['TYPE'])) {
            /* property is undefined */
            return '';
        } else {
            return strtolower($col['TYPE']);
        }
    }

    /**
     * set the type of a field as specified in the structure
     *
     * Note: this function does not check if the provided
     * string is a valid type name.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @param   midex   $value
     * @return  bool
     * @since   2.9 RC3
     * @name    StructureFile::setType()
     * @see     StructureFile::getType()
     */
    function setType($table, $column, $value)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else {
            $col['TYPE'] = $value;
            return true;
        }
    }

    /**
     * get the properties of a field of type 'image'
     *
     * Returns an array of the following values:
     * <code>
     * array (
     *    'size'       : int,  // maximum size in bytes
     *    'width'      : int,  // horizontal dimension in px
     *    'height'     : int,  // vertical dimension in px
     *    'ratio'      : bool, // keep aspect-ratio (true=yes, false=no)
     *    'background' : array(red, green, blue) // color of canvas
     * )
     * </code>
     * If one of the values above does not exist,
     * the field is set to 'null'.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  array
     * @since   2.8.9
     * @name    StructureFile::getImageSettings()
     * @see     StructureFile::setImageSettings()
     */
    function getImageSettings($table, $column)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        }

        if (strcasecmp($this->getType($table, $column), 'image') === 0) {
            $result = array();
            /**
             * here: maximum size in bytes
             *
             * Note: the size after (!) the conversion might
             * be different from the size of the uploaded file.
             */
            if (isset($col['LENGTH'])) {
                $result['size'] = (int) $col['LENGTH'];
            } else {
                $result['size'] = null;
            }
            if (isset($col['WIDTH'])) {
                $result['width'] = (int) $col['WIDTH'];
            } else {
                $result['width'] = null;
            }
            if (isset($col['HEIGHT'])) {
                $result['height'] = (int) $col['HEIGHT'];
            } else {
                $result['height'] = null;
            }
            if (isset($col['RATIO'])) {
                $result['ratio'] = (bool) $col['RATIO'];
            } else {
                $result['ratio'] = null;
            }
            if (isset($col['BACKGROUND'])) {
                $result['background'] = $col['BACKGROUND'];
                if (!is_array($result['background']) || count($result['background']) !== 3) {
                    $result['background'] = null;
                } else {
                   settype($result['background'][0], 'int');
                   settype($result['background'][1], 'int');
                   settype($result['background'][2], 'int');
                }
            } else {
                $result['background'] = null;
            }
            return $result;
        } else {
            return false;
        }
    }

    /**
     * set the properties of a field of type 'image'
     *
     * The argument $settings is an array of the following values:
     * <code>
     * array (
     *    'size'       : int,  // maximum size in bytes
     *    'width'      : int,  // horizontal dimension in px
     *    'height'     : int,  // vertical dimension in px
     *    'ratio'      : bool, // keep aspect-ratio (true=yes, false=no)
     *    'background' : array(red, green, blue) // color of canvas
     * )
     * </code>
     * If one of the values above does not exist,
     * the field is set to 'null'.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @param   array   $settings
     * @return  array
     * @since   2.9
     * @name    StructureFile::setImageSettings()
     * @see     StructureFile::getImageSettings()
     */
    function setImageSettings($table, $column, $settings)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        }

        if (strcasecmp($this->getType($table, $column), 'image') === 0) {
            if (isset($settings['size']) && $settings['size'] > 0) {
                $col['LENGTH'] = (int) $settings['size'];
            } else {
                $col['LENGTH'] = null;
            }
            if (isset($settings['width']) && $settings['width'] > 0) {
                $col['WIDTH'] = (int) $settings['width'];
            } else {
                $col['WIDTH'] = null;
            }
            if (isset($settings['height']) && $settings['height'] > 0) {
                $col['HEIGHT'] = (int) $settings['height'];
            } else {
                $col['HEIGHT'] = null;
            }
            if (isset($settings['ratio'])) {
                $col['RATIO'] = (bool) $settings['ratio'];
            } else {
                $col['RATIO'] = null;
            }
            if (isset($settings['background']) && is_array($settings['background']) && count($settings['background']) === 3) {
                $col['BACKGROUND'] = array();
                $col['BACKGROUND'][] = (int) array_shift($settings['background']);
                $col['BACKGROUND'][] = (int) array_shift($settings['background']);
                $col['BACKGROUND'][] = (int) array_shift($settings['background']);
            } else {
                $col['BACKGROUND'] = null;
            }
            return true;
        } else {
            return false;
        }
    }

    /**
     * get a list of all columns that match a certain type
     *
     * Returns false if the table does not exist.
     * Otherwise it returns a list of column names.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $type
     * @return  array|false
     * @since   2.9 RC3
     */
    function getColumnsByType($table, $type)
    {
        /* settype to STRING */
        $table = (string) $table;
        $type  = (string) $type;

        /* get list of all columns */
        $columnList = $this->getColumns($table);
        if (is_array($columnList)) {
            $result = array();
            foreach ($columnList as $column)
            {
                $columnType = $this->getType($table, $column);
                if (is_string($columnType) && strcasecmp($columnType, $type) === 0) {
                    $result[] = $column;
                }
            } /* end foreach */
            return $result;
        } else {
            /* error: table not found */
            return false;
        }
    }

    /**
     * get a list of all columns that contain blobs
     *
     * Returns false if the table does not exist.
     * Otherwise it returns a list of column names.
     *
     * @access  public
     * @param   string  $table
     * @return  array|false
     * @since   2.8.9
     */
    function getFiles($table)
    {
        /* settype to STRING */
        $table  = (string) $table;
        /* get list of all columns */
        $columnList = $this->getColumns($table);
        if (is_array($columnList)) {
            $result = array();
            foreach ($columnList as $column)
            {
                $type = $this->getType($table, $column);
                if (is_string($type) && ($type === 'image' || $type === 'file')) {
                    $result[] = $column;
                }
            } /* end foreach */
            return $result;
        } else {
            /* error: table not found */
            return false;
        }
    }

    /**
     * get the default value of a field as specified in the structure
     *
     * Returns constant NULL (not bool(false)) if value is not specified.
     * Use is_null($result) to test for existance, don't use empty($result).
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  mixed
     */
    function getDefault($table, $column)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return null;
        } else if (!isset($col['DEFAULT'])) {
            /* property is undefined */
            return null;
        } else if (is_array($col['DEFAULT']) && $this->getType($table, $column) !== 'select') {
            $type = $this->getType($table, $column);
            if ($type !== 'string') {
                return false;
            } else {
                foreach ($col['DEFAULT'] as $default)
                {
                    $default = Layout::replaceToken($default);
                    if (empty($default)) {
                        continue;
                    } else if (strpos($default, YANA_UNI_DELIMITER) === 0 && strrpos($default, YANA_UNI_DELIMITER) >= strlen($default) -1) {
                        continue;
                    } else {
                        return untaintInput($default, "string", 0, YANA_ESCAPE_TOKEN);
                    }
                } /* end foreach */
            }
            return null;
        } else {
            return $col['DEFAULT'];
        }
    }

    /**
     * set the default value of a field as specified in the structure
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @param   midex   $value
     * @return  bool
     * @since   2.8.9
     */
   function setDefault($table, $column, $value)
    {
        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error: table or column not found */
            return false;
        } else {
            $col['DEFAULT'] = $value;
            return true;
        }
    }

    /**
     * get the names of all columns in a table
     *
     * Returns a numeric array of all columns in $table.
     * Issues an E_USER_NOTICE and returns an empty array, if $table does not exist
     * in current structure file.
     *
     * @access  public
     * @param   string  $table
     * @return  array
     */
    function getColumns($table)
    {
        $tbl =& $this->_getTable($table);

        if (!is_array($tbl)) {
            trigger_error("The table '${table}' does not exist.", E_USER_NOTICE);
            /* error: table not found */
            return array();
        } else {
            return array_keys(array_change_key_case($tbl['CONTENT'], CASE_LOWER));
        }
    }

    /**
     * get the name of the table, a foreign key points to
     *
     * @access  public
     * @param   string  $table
     * @param   string  $foreignKey
     * @return  string
     */
    function getTableByForeignKey($table, $foreignKey)
    {
        /* settype to STRING */
        $foreignKey = (string) $foreignKey;
        $foreignKey = strtoupper($foreignKey);

        $tbl =& $this->_getTable($table);

        if (!is_array($tbl)) {
            /* error: table not found */
            return "";
        } else {
            return strtolower(@$tbl['FOREIGN_KEYS'][$foreignKey]);
        }
    }

    /**
     * check whether the structure defines the "USE_STRICT" setting as bool(true)
     *
     * @access  public
     * @return  bool
     * @name    StructureFile::isStrict()
     * @see     StructureFile::setStrict()
     */
    function isStrict()
    {
        return ($this->content['USE_STRICT'] === true || $this->content['USE_STRICT'] === 'true');
    }

    /**
     * select whether the structure should use the "strict" directive
     *
     * Set the property "USE_STRICT" of the database file to
     * the argument $isStrict;
     *
     * @access  public
     * @param   bool  $isStrict
     * @since   2.9
     * @name    StructureFile::setStrict()
     * @see     StructureFile::isStrict()
     */
    function setStrict($isStrict)
    {
        if ($isStrict) {
            $this->content['USE_STRICT'] = true;
        } else {
            $this->content['USE_STRICT'] = false;
        }
    }

    /**
     * get a list of all tables in the current database
     *
     * Returns a numeric array of all tables in the current structure file.
     * Issues an E_USER_NOTICE and returns an empty array, if the list of
     * tables is empty.
     *
     * @access  public
     * @return  array
     */
    function getTables()
    {
        if (!isset($this->content['TABLES'])) {
            trigger_error("The list of tables is empty.",E_USER_NOTICE);
            return array();
        } else {
            return array_keys(array_change_key_case($this->content['TABLES'], CASE_LOWER));
        }
    }

    /**
     * get a list of all indexed columns in a table
     *
     * Returns a numeric array of all columns in $table, that are marked with the
     * option INDEX => true (which means, all columns that have an index).
     *
     * @access  public
     * @param   string  $table
     * @return  array
     * @since   2.8.5
     */
    function getIndexes($table)
    {
        if (!isset($this->content['TABLES'])) {
            trigger_error("The list of tables is empty.", E_USER_NOTICE);
            return array();
        } else if (!$this->isTable($table)) {
            /* table does not exist */
            trigger_error("The table '${table}' does not exist.", E_USER_NOTICE);
            return array();
        } else {
            $result = array();
            $table = strtoupper($table);
            foreach ($this->getColumns($table) as $columnName)
            {
                if ($this->hasIndex($table,$columnName) === true) {
                    array_push($result,$columnName);
                }
            } /* end foreach */
            return $result;
        }
    }

    /**
     * get a list of all unique columns in table
     *
     * Returns a numeric array of all columns in $table, that are marked with the
     * option UNQIUE => true (which means, all columns that have an unique constraint).
     *
     * @access  public
     * @param   string  $table
     * @return  array
     * @since   2.8.5
     */
    function getUniqueConstraints($table)
    {
        if (!isset($this->content['TABLES'])) {
            trigger_error("The list of tables is empty.", E_USER_NOTICE);
            return array();
        } else if (!$this->isTable($table)) {
            /* table does not exist */
            trigger_error("The table '${table}' does not exist.", E_USER_NOTICE);
            return array();
        } else {
            $result = array();
            $table = strtoupper($table);
            foreach ($this->getColumns($table) as $columnName)
            {
                if ($this->isUnique($table, $columnName) === true) {
                    $result[] = $columnName;
                }
            } /* end foreach */
            return $result;
        }
    }

    /**
     * get the database name associated with a table
     *
     * returns the associated database name as string or null of no association is found
     *
     * @access  protected
     * @return  string
     * @ignore
     */
    function getAssociation($table)
    {
        /* settype to STRING */
        $table = (string) $table;
        $table = strtoupper($table);
        if (!isset($this->content['ASSOCIATIONS']) || !is_array($this->content['ASSOCIATIONS']) || !isset($this->content['ASSOCIATIONS'][$table]) || !is_string($this->content['ASSOCIATIONS'][$table])) {
            return null;
        } else {
            return $this->content['ASSOCIATIONS'][$table];
        }
    }

    /**
     * get all constraints for an address
     *
     * Retrieves all "constraint" entries that apply to the given
     * operation on the dataset and returns the results as an numeric
     * array.
     *
     * @access  public
     * @since   version 2.8.4
     * @param   string $operation one of select, insert, update, delete
     * @param   string $table
     * @param   array  $columns (optional)
     * @return  array
     * @name    StructureFile::getConstraint()
     * @see     StructureFile::setConstraint()
     */
    function getConstraint($operation, $table, $columns = array())
    {
        return $this->getFields('constraint', $operation, $table, $columns);
    }

    /**
     * set constraint
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * Note: This function will check some syntax of your code.
     * However, it can't ensure that your codes makes sense.
     * So keep in mind that it is your job in the first place
     * to ensure the constraint is valid!
     *
     * BE WARNED: As always - do NOT use this function with any
     * unchecked user input.
     *
     * @access  public
     * @param   string  $constraint  PHP-Code
     * @param   string  $operation   one of select, insert, update, delete
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     * @since   2.9
     * @name    StructureFile::setConstraint()
     * @see     StructureFile::getConstraint()
     */
    function setConstraint($constraint, $operation, $table = null, $column = null)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        assert('!isset($_types);');
        $_types = array(
        /* argument 1 */ IS_STRING,
        /* argument 2 */ IS_STRING,
        /* argument 3 */ IS_NULL|IS_STRING,
        /* argument 4 */ IS_NULL|IS_STRING
        );
        $_chk = checkArgumentList(func_get_args(), $_types, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk, $_types);
        }

        $operation = strtoupper($operation);

        if (!preg_match(YANA_DB_CONSTRAINT_SYNTAX, $constraint)) {
            trigger_error("Syntax error in constraint: '".trim($constraint)."'.", E_USER_WARNING);
            return false;
        }

        if (!is_null($column)) {
            $col =& $this->_getColumn($table, $column);
            if (!is_array($col)) {
                /* error: table or column not found */
                return false;
            } else {
                if (!isset($col['CONSTRAINT']) || !is_array($col['CONSTRAINT'])) {
                    $col['CONSTRAINT'] = array();
                }
                $col['CONSTRAINT'][$operation] = $constraint;
                return true;
            }
        } else if (!is_null($table)) {
            $tbl =& $this->_getTable($table);
            if (!is_array($tbl)) {
                /* error: table not found */
                return false;
            } else {
                if (!isset($tbl['CONSTRAINT']) || !is_array($tbl['CONSTRAINT'])) {
                    $tbl['CONSTRAINT'] = array();
                }
                $tbl['CONSTRAINT'][$operation] = $constraint;
                return true;
            }
        } else {
            if (!isset($this->content['CONSTRAINT']) || !is_array($this->content['CONSTRAINT'])) {
                $this->content['CONSTRAINT'] = array();
            }
            $this->content['CONSTRAINT'][$operation] = $constraint;
            return true;
        }

    }

    /**
     * get all triggers for an address
     *
     * Retrieves all "trigger" entries that apply to the given
     * operation on the dataset and returns the results as an numeric
     * array.
     *
     * There are two constants to use for argument $prefix:
     * The first, YANA_TRIGGER_BEFORE, refers to triggers that fire BEFORE the action
     * $operation is carried out.
     * And YANA_TRIGGER_AFTER, refers to triggers that fire AFTER the action
     * $operation has been carried out.
     *
     * @access  public
     * @param   int     $prefix     currently one of YANA_TRIGGER_BEFORE, YANA_TRIGGER_AFTER
     * @param   string  $operation  one of insert, update, delete
     * @param   string  $table
     * @param   array   $columns
     * @return  array
     * @since   2.8.5
     * @name    StructureFile::getTrigger()
     * @see     StructureFile::setTrigger()
     */
    function getTrigger($prefix, $operation, $table, $columns = array())
    {
        if (!is_string($operation) || !preg_match('/^select|insert|update|delete$/is', $operation)) {
            trigger_error("Invalid input for argument 2 in ".__CLASS__."->".__FUNCTION__."().\n\t\tExpected a string value 'insert', 'update' or 'delete' found (".gettype($operation).")'".print_r($operation, true)."' instead.",E_USER_WARNING);
            return false;
        } else {
            switch ($prefix)
            {
                case YANA_TRIGGER_BEFORE:
                    $operation = 'BEFORE_'.$operation;
                break;
                case YANA_TRIGGER_AFTER:
                    $operation = 'AFTER_'.$operation;
                break;
                default:
                    trigger_error("Invalid input for argument 1 in ".__CLASS__."->".__FUNCTION__."().\n\t\tExpected an integer of value '".YANA_TRIGGER_BEFORE."', or '".YANA_TRIGGER_AFTER."' found (".gettype($prefix).")'".print_r($prefix, true)."' instead.",E_USER_WARNING);
                    return false;
                break;
            }
            return $this->getFields('trigger', $operation, $table, $columns);
        }
    }

    /**
     * set constraint
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * Note: This function will check some syntax of your code.
     * However, it can't ensure that your codes makes sense.
     * So keep in mind that it is your job in the first place
     * to ensure the constraint is valid!
     *
     * BE WARNED: As always - do NOT use this function with any
     * unchecked user input.
     *
     * @access  public
     * @param   string  $trigger    PHP-Code
     * @param   int     $prefix     currently one of YANA_TRIGGER_BEFORE, YANA_TRIGGER_AFTER
     * @param   string  $operation  one of select, insert, update, delete
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     * @since   2.9
     * @name    StructureFile::setTrigger()
     * @see     StructureFile::getTrigger()
     */
    function setTrigger($trigger, $prefix, $operation, $table = null, $column = null)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        assert('!isset($_types);');
        $_types = array(
        /* argument 1 */ IS_STRING,
        /* argument 2 */ IS_INT,
        /* argument 3 */ IS_STRING,
        /* argument 4 */ IS_NULL|IS_STRING,
        /* argument 5 */ IS_NULL|IS_STRING
        );
        $_chk = checkArgumentList(func_get_args(), $_types, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk, $_types);
        }

        $operation = strtoupper($operation);

        switch ($prefix)
        {
            case YANA_TRIGGER_BEFORE:
                $operation = 'BEFORE_'.$operation;
            break;
            case YANA_TRIGGER_AFTER:
                $operation = 'AFTER_'.$operation;
            break;
            default:
                trigger_error("Invalid input for argument 2 in ".__CLASS__."->".__FUNCTION__."().\n\t\tExpected an integer of value '".YANA_TRIGGER_BEFORE."', or '".YANA_TRIGGER_AFTER."' found (".gettype($prefix).")'".print_r($prefix, true)."' instead.", E_USER_WARNING);
                return false;
            break;
        }

        if (!is_null($column)) {
            $col =& $this->_getColumn($table, $column);
            if (!is_array($col)) {
                /* error: table or column not found */
                return false;
            } else {
                if (!isset($col['TRIGGER']) || !is_array($col['TRIGGER'])) {
                    $col['TRIGGER'] = array();
                }
                $col['TRIGGER'][$operation] = $trigger;
                return true;
            }
        } else if (!is_null($table)) {
            $tbl =& $this->_getTable($table);
            if (!is_array($tbl)) {
                /* error: table not found */
                return false;
            } else {
                if (!isset($tbl['TRIGGER']) || !is_array($tbl['TRIGGER'])) {
                    $tbl['TRIGGER'] = array();
                }
                $tbl['TRIGGER'][$operation] = $trigger;
                return true;
            }
        } else {
            if (!isset($this->content['TRIGGER']) || !is_array($this->content['TRIGGER'])) {
                $this->content['TRIGGER'] = array();
            }
            $this->content['TRIGGER'][$operation] = $trigger;
            return true;
        }

    }

    /**
     * check whether the "READONLY" flag is set to bool(true)
     *
     * Returns bool(true) if:
     * <ol>
     *    <li>database is set to 'readonly=true', or</li>
     *    <li>$table is set to 'readonly=true', or</li>
     *    <li>$column in $table is set to 'readonly=true'</li>
     * </ol>
     *
     * Returns bool(false) otherwise.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     *
     * @see     StructureFile::isVisible()
     * @see     StructureFile::isEditable()
     * @name    StructureFile::isReadonly()
     */
    function isReadonly($table = "", $column = "")
    {
        /* 1) the database is set to readonly */
        if (!empty($this->content['READONLY'])) {
            return true;
        }

        /* 2) the column is set to readonly */
        if (!empty($table) && !empty($column)) {
            $col =& $this->_getColumn($table, $column);

            /* error: table or column not found */
            if (!is_array($col)) {
                return false;

            /* property is true */
            } else if (!empty($col['READONLY'])) {
                return true;

            /* property is undefined */
            } else {
                /* intentionally left blank */
            }
        }

        /* 3) the table is set to readonly */
        if (!empty($table)) {
            $tbl =& $this->_getTable($table);

            /* error: table not found */
            if (!is_array($tbl)) {
                return false;

            /* property is true */
            } else if (!empty($tbl['READONLY'])) {
                return true;

            /* property is undefined */
            } else {
                /* intentionally left blank */
            }
        }

        /* no readonly flag has been set */
        return false;
    }

    /**
     * set the "readonly" property
     *
     * Returns bool(true) on success and bool(false) otherwise.
     *
     * @access  public
     * @param   bool    $isReadonly
     * @param   string  $table
     * @param   string  $column
     * @return  bool
     *
     * @see     StructureFile::isReadonly()
     * @name    StructureFile::setReadonly()
     */
    function setReadonly($isReadonly, $table = null, $column = null)
    {
        if (!is_null($column)) {
            $col =& $this->_getColumn($table, $column);
            if (!is_array($col)) {
                /* error: table or column not found */
                return false;
            } else {
                $readonly =& $col['READONLY'];
            }
        } else if (!is_null($table)) {
            $tbl =& $this->_getTable($table);
            if (!is_array($tbl)) {
                /* error: table not found */
                return false;
            } else {
                $readonly =& $tbl['READONLY'];
            }
        } else {
            $readonly =& $this->content['READONLY'];
        }

        if ($isReadonly) {
            $readonly = true;
        } else {
            $readonly = false;
        }
    }

    /**
     * check whether the column should be visible
     *
     * Returns bool(true) if:
     * <ol>
     *    <li>the table and column exist AND</li>
     *    <li>the column setting DISPLAY.HIDDEN is not set, or false AND</li>
     *    <li>the column setting DISPLAY.HIDDEN.$action is not set OR</li>
     *    <li>the column setting DISPLAY.HIDDEN.$action exists and is set to false.</li>
     * </ol>
     * Returns bool(false) otherwise.
     *
     * If the argument $action is not provided, the last two do not apply.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @param   string  $action
     * @return  bool
     * @since   2.9.0 RC1
     * @see     StructureFile::isEditable()
     * @see     StructureFile::isReadonly()
     * @name    StructureFile::isVisible()
     */
    function isVisible($table, $column, $action = "")
    {
        /* settype to STRING */
        $action = (string) $action;
        $action = strtoupper($action);

        $col =& $this->_getColumn($table, $column);

        /* 1) error - table / column do not exist */
        if (!is_array($col)) {
            return false;

        /* 2) display.hidden.$action property is set */
        } else if (!empty($action) && isset($col['DISPLAY']['HIDDEN'][$action])) {
            if ($col['DISPLAY']['HIDDEN'][$action] === false) {
                return true;
            } else {
                return false;
            }

        /* 3) display.hidden property is set */
        } else if (isset($col['DISPLAY']['HIDDEN']) && is_bool($col['DISPLAY']['HIDDEN'])) {
            if ($col['DISPLAY']['HIDDEN'] === false) {
                return true;
            } else {
                return false;
            }

        /* 4) display.hidden is undefined, defaults to bool(true) */
        } else {
            return true;
        }
    }

    /**
     * select whether the column should be visible
     *
     * This sets the property "display.hidden" of the column
     * to bool(true), if $isVisible is bool(false) and vice versa.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * Note: if you don't provide the attribute $action then
     * settings will apply to all actions. Otherwise you need
     * to set the display property for all actions separately.
     *
     * @access  public
     * @param   bool    $isVisible
     * @param   string  $table
     * @param   string  $column
     * @param   string  $action
     * @return  bool
     * @since   2.9
     * @see     StructureFile::isVisible()
     * @name    StructureFile::setVisible()
     */
    function setVisible($isVisible, $table, $column, $action = "")
    {
        /* settype to STRING */
        $action = (string) $action;
        $action = strtoupper($action);

        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error table / column do not exist */
            return false;
        } else {
            if (!isset($col['DISPLAY'])) {
                $col['DISPLAY'] = array();
            }
            if (!isset($col['DISPLAY']['HIDDEN'])) {
                $col['DISPLAY']['HIDDEN'] = array();
            }
            $prop =& $col['DISPLAY']['HIDDEN'];
            if (!empty($action)) {
                if (!is_array($prop)) {
                    $prop = array();
                }
                if ($isVisible) {
                    $prop[$action] = false;
                } else {
                    $prop[$action] = true;
                }
                return true;
            } else {
                if ($isVisible) {
                    $prop = false;
                } else {
                    $prop = true;
                }
                return true;
            }
        } /* end if */
    }

    /**
     * check whether the column should be editable
     *
     * Returns bool(true) if:
     * <ol>
     *    <li>the table and column exist AND</li>
     *    <li>the column's readonly flag is set to true (see {@link StructureFile::isReadonly}) AND</li>
     *    <li>the column is not visible (see {@link StructureFile::isVisible}) AND</li>
     *    <li>the column setting DISPLAY.READONLY is not set, or false AND</li>
     *    <li>the column setting DISPLAY.READONLY.$action is not set OR</li>
     *    <li>the column setting DISPLAY.READONLY.$action exists and is set to false.</li>
     * </ol>
     * Returns bool(false) otherwise.
     *
     * If the argument $action is not provided, the last two do not apply.
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @param   string  $action
     * @return  bool
     *
     * @see     StructureFile::isVisible()
     * @see     StructureFile::isReadonly()
     * @since   2.9.0 RC1
     * @name    StructureFile::isEditable()
     */
    function isEditable($table, $column, $action = "")
    {
        /* settype to STRING */
        $action = (string) $action;
        $action = strtoupper($action);

        $col =& $this->_getColumn($table, $column);

        /* 1) error - table / column do not exist */
        if (!is_array($col)) {
            return false;

        /* 2) the readonly flag has been set */
        } else if ($this->isReadonly($table, $column)) {
            return false;

        /* 3) the column is not visible */
        } else if (!$this->isVisible($table, $column, $action)) {
            return false;

        /* 4) display.readonly.$action property is set */
        } else if (!empty($action) && isset($col['DISPLAY']['READONLY'][$action])) {
            if ($col['DISPLAY']['READONLY'][$action] === false) {
                return true;
            } else {
                return false;
            }

        /* 5) display.readonly property is set */
        } else if (isset($col['DISPLAY']['READONLY']) && is_bool($col['DISPLAY']['READONLY'])) {
            if ($col['DISPLAY']['READONLY'] === false) {
                return true;
            } else {
                return false;
            }

        /* 6) display.readonly is undefined, defaults to bool(true) */
        } else {
            return true;
        }
    }

    /**
     * select whether the column should be editable
     *
     * This sets the "display.readonly" property of the column
     * to bool(false), if $isEditable is bool(true) and vice versa.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * Note: if you don't provide the attribute $action then
     * settings will apply to all actions. Otherwise you need
     * to set the display property for all actions separately.
     *
     * @access  public
     * @param   bool    $isEditable
     * @param   string  $table
     * @param   string  $column
     * @param   string  $action
     * @return  bool
     * @since   2.9
     * @see     StructureFile::isVisible()
     * @name    StructureFile::setVisible()
     */
    function setEditable($isEditable, $table, $column, $action = "")
    {
        /* settype to STRING */
        $action = (string) $action;
        $action = strtoupper($action);

        $col =& $this->_getColumn($table, $column);

        if (!is_array($col)) {
            /* error table / column do not exist */
            return false;
        } else {
            if (!isset($col['DISPLAY'])) {
                $col['DISPLAY'] = array();
            }
            if (!isset($col['DISPLAY']['READONLY'])) {
                $col['DISPLAY']['READONLY'] = array();
            }
            $prop =& $col['DISPLAY']['READONLY'];
            if (!empty($action)) {
                if (!is_array($prop)) {
                    $prop = array();
                }
                if ($isEditable) {
                    $prop[$action] = false;
                } else {
                    $prop[$action] = true;
                }
                return true;
            } else {
                if ($isVisible) {
                    $prop = false;
                } else {
                    $prop = true;
                }
                return true;
            }
        } /* end if */
    }

    /**
     * validate a row against the schema
     *
     * Returns bool(true) if $row is valid and
     * bool(false) otherwise.
     *
     * @access  public
     * @param   string  $table
     * @param   mixed   &$row
     * @param   bool    $is_insert
     * @return  bool
     */
    function checkRow($table, &$row, $is_insert = true)
    {
        if (!$this->isTable($table)) {
            /* error: table not found */
            return false;
        }

        foreach ($this->getColumns($table) as $column)
        {
            $column = strtolower($column);
            if ($this->isNullable($table, $column)) {
                /* value is optional - ignore */
                continue;
            } else if (($is_insert && !isset($row[$column])) || (isset($row[$column]) && (is_null($row[$column]) || $row[$column] === ""))) {
                $default = $this->getDefault($table, $column);
                $type    = $this->getType($table, $column);
                $isAuto  = $this->isAuto($table, $column);
                if ($isAuto && $type == 'time') {
                    $row[$column] = time();
                } else if ($isAuto && $type == 'ip') {
                    if (isset($_SERVER['REMOTE_ADDR'])) {
                        $row[$column] = $_SERVER['REMOTE_ADDR'];
                    } else {
                        $row[$column] = '0.0.0.0';
                    }
                } else if ($isAuto && $type == 'profile') {
                    global $YANA;
                    $row[$column] = $YANA->session->getId();
                } else if ($isAuto && is_null($default)) {
                    $row[$column] = false;
                } else if (!is_null($default)) {
                    if (is_array($default)) {
                        $row[$column] = $this->untaintInput($table, $column, array_shift(array_keys($default)), YANA_ESCAPE_NONE);
                    } else if (is_scalar($default)) {
                        $row[$column] = $this->untaintInput($table, $column, $default, YANA_ESCAPE_NONE);
                    } else {
                        trigger_error("SYNTAX ERROR: Invalid default value for column '${column}'. The structure file is not valid.", E_USER_ERROR);
                    }
                    if ($this->getType($table, $column) === 'string') {
                        $row[$column] = Layout::replaceToken($row[$column]);
                    }
                } else {
                    trigger_error("SQL ERROR: A required attribute '${column}' has not been provided. Insert/Update operation aborted.", E_USER_NOTICE);
                    return false;
                } /* end if */
            } else {
                /* this input is valid - move to next */
            } /* end if */
        } /* end for */
        return true;
    }

    /**
     * get all fields for an address
     *
     * Retrieves all "$fieldname" entries that apply to the given
     * operation on the dataset and returns the results as an numeric
     * array.
     *
     * If the optional parameter $assoc is set to true, the result
     * will be an associative array.
     *
     * @access  public
     * @param   string  $fieldname  e.g. 'constraint', or 'trigger'
     * @param   string  $operation  one of select, insert, update, delete
     * @param   string  $table
     * @param   array   $columns    (optional)
     * @param   bool    $as_assoc   (optional)
     * @return  array
     * @ignore
     */
    function getFields($fieldname, $operation, $table, $columns = array(), $as_assoc = false)
    {
        /* settype to STRING
         *            STRING
         *            STRING
         *            ARRAY
         *            BOOLEAN
         */
        $fieldname = (string) $fieldname;
        $operation = (string) $operation;
        $table     = (string) $table;
        $columns   = (array)  $columns;
        $as_assoc  = (bool)   $as_assoc;

        $fieldname = strtoupper($fieldname);
        $operation = strtoupper($operation);
        $table = strtoupper($table);

        $cache =& $this->cached_fields[(int)$as_assoc][$fieldname][$operation][$table];

        /* if this result has already been cached than there is nothing to do here */
        if (isset($cache)) {
            if (count($columns) === 0) {
                return $cache;
            } else {
                if (!is_array($cache) || count($cache) === 0) {
                    return array();
                }
                $list_of_results = array();
                if (isset($cache[0])) {
                    $list_of_results[0] =& $cache[0];
                }
                if (isset($cache[1])) {
                    $list_of_results[1] =& $cache[1];
                }
                $columns = Hashtable::changeCase($columns, CASE_UPPER);;
                $match = array_intersect(array_keys($cache),$columns);

                foreach ($match as $key => $definition)
                {
                    if ($as_assoc) {
                        $list_of_results[$key] =& $cache[$definition];
                    } else {
                        $list_of_results[count($list_of_results)] =& $cache[$definition];
                    }
                }
                return $list_of_results;
            }
        /* if no cached result is available it has to be resolved */
        } else {

            /* 1. places where fields are allowed */
            $db_table_col = array();

            /* 1.1 Database-wide definitions */
            $db_table_col[0] =& $this->content;

            /* 1.2 table-wide definitions */
            if (!$this->isTable($table)) {
                /* table does not exist */
                trigger_error("The table '${table}' does not exist.",E_USER_NOTICE);
                return array();
            } else {
                $db_table_col[1] =& $this->content['TABLES'][$table];
            }

            /* 1.3 column specific definitions */
            if (!is_array($columns) || count($columns) === 0) {
                $columns = array_keys($this->content['TABLES'][$table]['CONTENT']);
            }

            assert('is_array($columns); // the table '.$table.' has no contents');
            if (is_array($columns)) {
                foreach ($columns as $column)
                {
                    $column = strtoupper($column);
                    if (!$this->isColumn($table, $column)) {
                        /* column does not exist */
                        trigger_error("The column '${column}' does not exist in table '${table}.'", E_USER_NOTICE);
                        return array();
                    } else {
                        $db_table_col[$column] =& $this->content['TABLES'][$table]['CONTENT'][$column];
                    }
                } /* end foreach */
            } /* end if */

            /* results are put here */
            $list_of_results = array();

            /* 2. find definitions */
            foreach ($db_table_col as $key => $value)
            {
                if (isset($value[$fieldname])) {
                    $definition =& $value[$fieldname];
                    /* fieldname.operation = string */
                    if (is_array($definition) && isset($definition[$operation]) && is_scalar($definition[$operation])) {
                        if ($as_assoc) {
                            $list_of_results[$key] =& $definition[$operation];
                        } else {
                            $list_of_results[count($list_of_results)] =& $definition[$operation];
                        }
                    /* fieldname = string */
                    } else if (is_scalar($definition)) {
                        if ($as_assoc) {
                            $list_of_results[$key] =& $definition;
                        } else {
                            $list_of_results[count($list_of_results)] =& $definition;
                        }
                    /* none */
                    } else {
                        /* no definition present */
                    }
                }
            } /* end foreach */

            $cache = $list_of_results;

            return $list_of_results;
        }

    }

    /**
     * untaint user input data with the help of the schema
     *
     * @access  public
     * @param   string  $table
     * @param   string  $column
     * @param   mixed   $value
     * @param   int     $escape
     * @return  mixed
     */
    function untaintInput($table, $column, $value, $escape = YANA_ESCAPE_NONE)
    {
        return untaintInput($value, (string) $this->getType($table, $column), (int) $this->getLength($table,$column), $escape);
    }

    /**
     * include structure file
     *
     * This function merges the contents of another file with the current structure.
     *
     * Note that you can reach the same result by using the 'INCLUDE' tag in the
     * structure file. See the developer's cookbook for an example.
     *
     * @access  public
     * @param   string $filename
     * @return  bool
     */
    function addFile($filename)
    {
        $filename = StructureFile::_getFilename($filename);
        if (!is_file($filename) || !preg_match('/^(?:[\w\d\_\-\/]*\/)*([\w\d\_\-]+)\.config$/', $filename, $m)) {
            trigger_error("Illegal structure file '${filename}'.", E_USER_ERROR);
            return false;
        } else {
            assert('!isset($databaseName); // cannot redeclare $databaseName');
            assert('!isset($file); // cannot redeclare $file');
            assert('!isset($test); // cannot redeclare $test');
            $databaseName = $m[1];
            assert('is_string($databaseName) && !empty($databaseName); // Invalid database name');
            $file = new StructureFile($filename);
            $test = $file->read();
            if ($test === false) {
                return false;
            } else {
                assert('!isset($current_list); // cannot redeclare $current_list');
                assert('!isset($import_list); // cannot redeclare $import_list');
                assert('!isset($intersection); // cannot redeclare $intersection');
                $current_list = array_keys($this->content['TABLES']);
                $import_list  = array_keys($file->content['TABLES']);
                $intersection = array_intersect($current_list,$import_list);
                if (is_array($intersection) && count($intersection) > 0) {
                    trigger_error("Unable to import structure file '${filename}' into structure file '".$this->getPath()."'.\n\t\tAmbigious table declarations found for: ".implode(', ',$intersection), E_USER_WARNING);
                    return false;
                } else {
                    assert('!isset($table); // cannot redeclare $table');
                    if (!isset($this->content['ASSOCIATIONS'])) {
                        $this->content['ASSOCIATIONS'] = array();
                    }
                    foreach ($import_list as $table)
                    {
                        if (isset($this->content['ASSOCIATIONS'][$table])) {
                            trigger_error("For table '${table}' replacing illegal table association '".$this->content['ASSOCIATIONS'][$table]."' with '${databaseName}'.",E_USER_NOTICE);
                        }
                        $this->content['ASSOCIATIONS'][$table] = $databaseName;
                        $this->content['TABLES'][$table] = $file->content['TABLES'][$table];
                    }
                    return true;
                } /* end if */
            } /* end if */
        } /* end if */
    }

    /**
     * addStructure
     *
     * Try to extract some information on the structure of
     * a database from the information provided by PEAR-DB's
     * tableInfo() function.
     *
     * see DbStream::buildStructure() for additional
     * information on the input
     *
     * @access  public
     * @param   string  $table  current table name
     * @param   array   $array  value returned by getTableInfo() -function
     * @see     DbStream::buildStructure()
     * @ignore
     */
    function addStructure($table, $array)
    {
        assert('is_array($array);');
        if (!is_array($array)) {
            return false;
        } else {
            $array = Hashtable::changeCase($array,CASE_UPPER);
            $config   = array("USE_STRICT" => "true", "READONLY" => "false", "TABLES" => array());
            for ($i = 0; $i < $array['NUM_FIELDS']; $i++)
            {
                $name  = $array[$i]['NAME'];
                $flags = explode(" ", $array[$i]['FLAGS']);
                $config['TABLES'][$table]['CONTENT'][$name] = array();
                switch ($array[$i]['TYPE'])
                {
                    case 'blob':
                        $array[$i]['TYPE'] = "text";
                    break;

                    case 'tinyint': case 'bigint': case 'int':
                        $array[$i]['TYPE'] = "integer";
                    break;

                    case 'bool':
                        $array[$i]['TYPE'] = "boolean";
                    break;

                    case 'double': case 'numeric':
                        $array[$i]['TYPE'] = "float";
                    break;

                    case 'timestamp':
                        $array[$i]['TYPE'] = "time";
                        $array[$i]['REQUIRED'] = YANA_DB_AUTO;
                    break;

                    case 'datetime': case 'string': case 'varchar':
                        $array[$i]['TYPE'] = "string";
                    break;

                    /* more ? */
                    default:
                        /* convert unrecognized type to string */
                        $array[$i]['TYPE'] = "string";
                    break;

                }
                if (!empty($array[$i]['TYPE'])) {
                    $config['TABLES'][$table]['CONTENT'][$name]['TYPE'] = $array[$i]['TYPE'];
                }
                if (!empty($array[$i]['LEN'])) {
                    $config['TABLES'][$table]['CONTENT'][$name]['LENGTH'] = $array[$i]['LEN'];
                }
                $config['TABLES'][$table]['CONTENT'][$name]['REQUIRED'] = in_array('not_null', $flags);
                if (in_array('primary_key', $flags)) {
                     $config['TABLES'][$table]['PRIMARY_KEY'] = $name;
                     $config['TABLES'][$table]['CONTENT'][$name]['HIDDEN']= "true";
                }
            }
            if (isset($this->content)) {
                $this->content = Hashtable::merge($this->content, $config);
            } else {
                $this->content = $config;
            }
        }
    }

    /**
     * _getFilename
     *
     * check input $filename
     *
     * @static
     * @access  private
     * @param   string  $filename
     * @return  string
     * @ignore
     */
    function _getFilename(&$filename)
    {
        /* settype to STRING */
        $filename = (string) $filename;
        if (preg_match('/^([\w\d_]+)$/', $filename)) {
            if (isset($YANA)) {
                $filename = $YANA->getVar('DBDIR').$filename.".config";
            } else {
                $filename = "config/db/".$filename.".config";
            }
        }
        return $filename;
    }

    /**
     * _getTable
     *
     * Check if table exists and if so
     * return a reference to it.
     *
     * Returns bool(false) on error.
     *
     * @static
     * @access  private
     * @param   string  &$table
     * @return  bool(false)|array
     * @since   2.9
     * @ignore
     */
    function &_getTable(&$table)
    {
        /* settype to STRING */
        $table  = (string) $table;
        $table  = strtoupper($table);
        if (isset($this->content['TABLES'][$table])) {
            return $this->content['TABLES'][$table];
        } else {
            $col = false;
            return $col;
        }

    }

    /**
     * _getColumn
     *
     * Check if column exists and if so
     * return a reference to it.
     *
     * Returns bool(false) on error.
     *
     * @static
     * @access  private
     * @param   string  &$table
     * @param   string  &$column
     * @return  bool(false)|array
     * @since   2.9
     * @ignore
     */
    function &_getColumn(&$table, &$column)
    {
        /* settype to STRING */
        $column = (string) $column;
        $column = strtoupper($column);

        $tbl =& $this->_getTable($table);

        if (!is_array($tbl)) {
            $col = false;
            return $col;
        } else if (isset($tbl['CONTENT'][$column])) {
            return $tbl['CONTENT'][$column];
        } else {
            $col = false;
            return $col;
        } /* end if */
    }

    /**
     * self test
     *
     * This function exists only for debugging purposes.
     * Do not uncomment in productive environments.
     *
     * @access  public
     * @static
     * @param   int  $nr  number of test to take
     * @return  bool|string
     * @ignore
     * /
    function test($nr = 1)
    {
        switch ($nr)
        {
            case 1:
                $db = new StructureFile('temp.sml');

                // "USE_STRICT"
                $db->setStrict(true);
                if ($db->isStrict() !== true) {
                    return 'set/get "strict" property failed';
                }

                // "READONLY"
                $db->setReadonly(true);
                if ($db->isReadonly() !== true) {
                    return 'set/get "readonly" property on database failed';
                }
                $db->setReadonly(false);
                if ($db->isReadonly() !== false) {
                    return 'set/get "readonly" property on database failed';
                }

                // add "TABLE"
                if ($db->addTable('TeSt') !== true) {
                    return 'add "table" property failed';
                }

                // get "TABLES"
                if ($db->getTables() !== array('test')) {
                    return 'list "tables" failed';
                }

                // check "is table"
                if ($db->isTable('Test') !== true) {
                    return '"is table" test failed';
                }

                // "READONLY"
                $db->setReadonly(true, 'teST');
                if ($db->isReadonly('teST') !== true) {
                    return 'set/get "readonly" property on table failed';
                }
                $db->setReadonly(false, 'teSt');
                if ($db->isReadonly('tEsT') !== false) {
                    return 'set/get "readonly" property on table failed';
                }

                // add "COLUMN"
                if ($db->addColumn('teSt', 'iD') !== true) {
                    return 'add "column" property failed';
                }

                // get "COLUMNS"
                if ($db->getColumns('Test') !== array('id')) {
                    return 'list "columns" failed';
                }

                // check "is column"
                if ($db->isColumn('Test', 'Id') !== true) {
                    return '"is column" test failed';
                }

                // "TYPE"
                $db->setType('Test', 'Id', 'Integer');
                if ($db->getType('TesT', 'iD') !== 'integer') {
                    return 'set/get "type" property failed';
                }

                // "DEFAULT"
                $db->setDefault('Test', 'Id', 1);
                if ($db->getDefault('TesT', 'iD') !== 1) {
                    return 'set/get "default" property failed';
                }

                // "DESCRIPTION"
                $db->setDescription('Test', 'Id', 'Foo');
                if ($db->getDescription('TesT', 'iD') !== 'Foo') {
                    return 'set/get "description" property failed';
                }

                // "LENGTH"
                $db->setLength('Test', 'Id', 8);
                if ($db->getLength('TesT', 'iD') !== 8) {
                    return 'set/get "length" property failed';
                }

                // get columns by type
                if ($db->getColumnsByType('Test', 'integer') !== array('id')) {
                    return 'list "columns by type" failed';
                }

                // "UNIQUE"
                $db->setUnique('Test', 'Id', true);
                if ($db->isUnique('TesT', 'iD') !== true) {
                    return 'set/get "unique" property failed';
                }

                // get columns by unique constraint
                if ($db->getUniqueConstraints('Test') !== array('id')) {
                    return 'list "columns by unique constraint" failed';
                }

                // "REQUIRED"
                $db->setNullable('Test', 'Id', false);
                if ($db->isNullable('TesT', 'iD') !== false) {
                    return 'set/get "required" property failed';
                }

                // "REQUIRED" = AUTO
                $db->setAuto('Test', 'Id', true);
                if ($db->isAutonumber('TesT', 'iD') !== true) {
                    return 'set/get "required" property to "auto" failed';
                }

                // "READONLY"
                $db->setReadonly(true, 'Test', 'Id');
                if ($db->isReadonly('TesT', 'iD') !== true) {
                    return 'set/get "readonly" property failed';
                }
                $db->setReadonly(false, 'Test', 'Id');

                // "INDEX"
                $db->setIndex('Test', 'Id', true);
                if ($db->hasIndex('TesT', 'iD') !== true) {
                    return 'set/get "index" property failed';
                }

                // get columns which have an index
                if ($db->getIndexes('Test') !== array('id')) {
                    return 'list "columns with index" failed';
                }

                $db->setIndex('Test', 'Id', false);
                if ($db->hasIndex('TesT', 'iD') !== false) {
                    return 'set/get "index" property failed';
                }

                // "DISPLAY.READONLY"
                $db->setEditable(true, 'Test', 'Id', 'eDit');
                if ($db->isEditable('TesT', 'iD', 'eDit') !== true) {
                    return 'set/get "display.readonly" property failed';
                }

                // "DISPLAY.HIDDEN"
                $db->setVisible(true, 'Test', 'Id');
                if ($db->isVisible('TesT', 'iD') !== true) {
                    return 'set/get "display.hidden" property failed';
                }
                $db->setVisible(false, 'Test', 'Id');
                if ($db->isVisible('TesT', 'iD') !== false) {
                    return 'set/get "display.hidden" property failed';
                }
                $db->setVisible(true, 'Test', 'Id', 'New');
                if ($db->isVisible('TesT', 'iD', 'neW') !== true) {
                    return 'set/get "display.hidden.new" property failed';
                }

                // "PRIMARY_KEY"
                $db->setPrimaryKey('Test', 'Id', true);
                if ($db->isPrimaryKey('TesT', 'iD') !== true) {
                    return 'set/get "primary key" property failed';
                }
                if ($db->getPrimaryKey('TesT') !== 'id') {
                    return 'set/get "primary key" property failed';
                }

                // "PROFILE_KEY"
                $db->setProfile('Test', 'Id');
                if ($db->getProfile('TesT') !== 'id') {
                    return 'set/get "primary key" property failed';
                }
                if ($db->getPrimaryKey('TesT') !== 'id') {
                    return 'set/get "primary key" property failed';
                }

                // get Files
                $db->addColumn('teSt', 'my_File');
                $db->setType('Test', 'my_fIle', 'file');
                if ($db->getFiles('TesT', 'iD') !== array('my_file')) {
                    return 'list "columns of type image or file" failed';
                }

                // get/set Foreign keys
                $db->addTable('ftab');
                $db->addColumn('ftab', 'fid');
                $db->setPrimaryKey('ftab', 'fid');

                if ($db->setForeignKey('teSt', 'id', 'fTab') !== true) {
                    return 'set "foreign key" failed';
                }
                if ($db->getForeignKeys('teSt') !== array('ID' => 'ftab')) {
                    return 'get "foreign key" failed';
                }
                if ($db->isForeignKey('teSt', 'id') !== true) {
                    return 'is "foreign key" failed';
                }
                if ($db->getTableByForeignKey('teSt', 'id') !== 'ftab') {
                    return 'get "table by foreign key" failed';
                }

                // image settings
                $db->addColumn('test', 'iMG');
                $db->setType('tEsT', 'imG', 'image');
                $set = array();
                $set['size'] = 50000;
                $set['width'] = 300;
                $set['height'] = 200;
                $set['ratio'] = true;
                $set['background'] = array(255, 255, 255);
                if ($db->setImageSettings('test', 'IMG', $set) !== true) {
                    return 'set "image settings" failed';
                }
                if ($db->getImageSettings('teSt', 'Img') !== $set) {
                    return 'set "image settings" failed';
                }

                // "CONSTRAINT"
                $constraint = '$PERMISSION > 100 && $VALUE > 0';
                if ($db->setConstraint($constraint, 'insert', 'test', 'id') !== true) {
                    return 'set "constraint" failed';
                }
                if ($db->getConstraint('insert', 'test') !== array($constraint)) {
                    return 'get "constraint" failed';
                }

                // "TRIGGER"
                if ($db->setTrigger($constraint, YANA_TRIGGER_BEFORE, 'insert', 'test', 'id') !== true) {
                    return 'set "trigger" failed';
                }
                if ($db->getTrigger(YANA_TRIGGER_BEFORE, 'insert', 'test') !== array($constraint)) {
                    return 'get "trigger" failed';
                }

                // Use this for testing purposes
                // var_dump($db->getStructure());

                // @todo  should anything else be tested?

                return true;
            break;
        }
    } /* end of function */
}

?>
