<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * @ignore
 */

/**
 * Singleton View
 *
 * A manager class to automate searching and loading of templates, belonging to view layer.
 *
 * @access      public
 * @package     yana
 * @subpackage  core
 *
 * @ignore
 */
class View extends Singleton
{
    /**#@+
     * @ignore
     * @access  private
     */

    /** @var String */       var $baseView = null;
    /** @var String */       var $selected = null;

    /**#@-*/

    /**
     * readonly
     * @access  public
     * @var     smartTemplate */ var $document = null;

    /**
    * Singleton Constructor
    *
    * @access  public
    */
    function View()
    {
        $this->checkSingletonConstraint();
        $this->document =& new smartTemplate();
    }

    /**
    * select template by template id
    *
    * Will select the inserted template $viewName
    * for output.
    *
    * Returns bool(true) on success or bool(false) on error.
    *
    * @access  public
    * @param   string  $viewName
    * @return  bool
    */
    function selectView($viewName)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_STRING, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        global $YANA;
        $view = $YANA->skin->getFile($viewName);

        if (!empty($view)) {
            $this->selected = $view;
            return true;
        } else {
            trigger_error("Invalid view name '$viewName' supplied in ".__CLASS__."->".__FUNCTION__."(). View is not registered.", E_USER_WARNING);
            return false;
        }
    }

    /**
    * select template by filename
    *
    * Will set the inserted template to file $viewName
    * for output.
    *
    * Returns bool(true) on success or bool(false) on error.
    *
    * @access  public
    * @param   string  $viewName
    * @return  bool
    */
    function setView($viewName)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_STRING, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        if (file_exists($viewName)) {
            $this->selected = $viewName;
            return true;
        } else {
            trigger_error("Invalid filename '$viewName' supplied in ".__CLASS__."->".__FUNCTION__."(). File does not exist.", E_USER_WARNING);
            return false;
        }
    }

    /**
    * get filename of selected template
    *
    * @access  public
    * @return  string|bool(false)
    */
    function getView()
    {
        assert('is_string($this->selected);');
        if ($this->selected) {
            return $this->selected;
        } else {
            return false;
        }
    }

    /**
    * unselect template file
    *
    * @access  public
    */
    function unsetView()
    {
        $this->selected = "";
    }

    /**
    * select template by template id
    *
    * Will select the base template $viewName
    * for output.
    *
    * Returns bool(true) on success or bool(false) on error.
    *
    * @access  public
    * @param   string  $viewName
    * @return  bool
    */
    function selectBaseView($viewName)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_STRING, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        global $YANA;
        $baseView = $YANA->skin->getFile($viewName);

        if (!empty($baseView)) {
            $this->baseView = $baseView;
            return true;
        } else {
            trigger_error("Invalid view name '$viewName' supplied in ".__CLASS__."->".__FUNCTION__."(). View is not registered.", E_USER_WARNING);
            return false;
        }
    }

    /**
    * select template by filename
    *
    * Will set the base template to file $viewName
    * for output.
    *
    * Returns bool(true) on success or bool(false) on error.
    *
    * @access  public
    * @param   string  $viewName
    * @return  bool
    */
    function setBaseView($viewName)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_STRING, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        if (file_exists($viewName)) {
            $this->baseView = $viewName;
            return true;
        } else {
            trigger_error("Invalid filename '$viewName' supplied in ".__CLASS__."->".__FUNCTION__."(). File does not exist.", E_USER_WARNING);
            return false;
        }
    }

    /**
    * unselect template file
    *
    * @access  public
    */
    function unsetBaseView()
    {
        $this->baseView = "";
    }

    /**
    * getBaseView
    *
    * @access  public
    * @return  string
    */
    function getBaseView()
    {
        assert('is_string($this->baseView);');
        if ($this->baseView) {
            return $this->baseView;
        } else {
            return false;
        }
    }

    /**
    * prepare document for output
    *
    * @access  public
    * @return  bool
    */
    function buildPage()
    {
         if ($this->baseView) {
             $this->document->setPath($this->baseView);
             if ($this->selected) {
                 $this->document->insertFile('*',$this->selected);
             }
             return true;
         } else {
             return false;
         }
    }

    /**
    * output template to screen
    *
    * @access  public
    * @return  bool
    */
    function outputToScreen()
    {
        if (!is_null($this->document) && $this->document->exists()) {
            $this->document->write();
            return true;
        } else {
            return false;
        }
    }

    /**
    * copy options from registry to output
    *
    * @access  public
    * @param   string  $key  the address of the data in registry
    * @return  bool
    */
    function setOptions($key)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        $_chk = checkArgumentList(func_get_args(), IS_STRING, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk);
        }

        if (!is_null($this->document) && $this->document->exists()) {
            global $YANA;
            $this->document->insert('*', $YANA->getVar($key));
            return true;
        } else {
            return false;
        }
    }

    /**
    * output a message to browser and terminate the program
    *
    * NOTE: This function will terminate program execution
    *
    * @access  public
    * @param   string  $type
    * @param   string  $errcode
    * @param   string  $event
    */
    function message($type, $errcode, $event = "")
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        assert('!isset($_types);');
        $_types = array(
        /* argument 1 */ IS_STRING,
        /* argument 2 */ IS_STRING,
        /* argument 3 */ IS_STRING|IS_NULL
        );
        $_chk = checkArgumentList(func_get_args(), $_types, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk, $_types);
        }

        if (!$this->selectBaseView('message')) {
            trigger_error("Invalid filename supplied. Unable to create page.", E_USER_ERROR);
            exit;
        }

        /* argument 3 */
        if (is_null($event)) {
            $event = 'null';
        } else {
            $event = strtolower($event);
        }

        $this->unsetView();
        $this->buildPage();

        global $YANA;
        $YANA->registry->mount('language_message_file');

        $YANA->setVar('ACTION', $event);
        $YANA->setVar('MESSAGE_TYPE', strtolower($type));

        if (!empty($errcode)) {
            $this->setOptions("TEXT.$errcode");
        }

        assert('!isset($a); /* cannot redeclare variable $a */');
        assert('!isset($b); /* cannot redeclare variable $b */');
        foreach ($_GET as $a => $b)
        {
            if (is_null($this->document->get(strtoupper($a)))) {
                $this->document->insert(strtoupper($a), $b);
            }
        }
        unset($a,$b);
        assert('!isset($a); /* cannot redeclare variable $a */');
        assert('!isset($b); /* cannot redeclare variable $b */');
        foreach ($_POST as $a => $b)
        {
            if (is_null($this->document->get(strtoupper($a)))) {
                $this->document->insert(strtoupper($a), $b);
            }
        }
        unset($a,$b);
        $this->setOptions("*");
        $this->outputToScreen();
        exit(0);
    }
}

?>