<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * Singleton Yana
 *
 * This is a primary controller and application loader for the YANA-Framework.
 *
 * @access      public
 * @package     yana
 * @subpackage  core
 */
class Yana extends Singleton
{
    /**#@+
     * @ignore
     * @access  private
     */

    /** @var array  */  var $config  = array();
    /** @var int    */  var $mode    = 0;
    /** @var View   */  var $view    = null;
    /** @var array  */  var $log     = array();
    /** @var array  */  var $request = array();
    /** @var string */  var $id      = "";

    /**#@-*/
    /**#@+
     * This variable is readonly.
     * @access  public
     */

    /**
     * to communicate with plugins
     * @var PluginManager
     */ var $plugins  = null;
    /**
     * to load language strings
     * @var Language
     */ var $language = null;
    /**
     * to load skins and templates
     * @var Skin
     */ var $skin     = null;
    /**
     * to read and write data to the global registry
     * @var Registry
     */ var $registry = null;
    /**
     * to read and write user data and permissions
     * @var SessionManager
     * @since 2.9
     */ var $session = null;

    /**#@-*/

    /**
    * Singleton Constructor
    *
    * This function creates a new instance of the framework.
    * Note that you may only operate one instance at a time.
    *
    * Examples:
    *
    * If you want to use a web interface:
    * <code>
    * global $YANA;
    * $YANA = new yana("config/system.config", $_REQUEST);
    * </code>
    *
    * If you want to use a command line interface:
    * <code>
    * global $YANA;
    * $YANA = new yana("config/system.config", $_SERVER['argv]);
    * </code>
    *
    * @uses    new yana("config/system.config", $_REQUEST)
    * @param   string  $filename  path to system.config
    * @param   array   $ARGS      associative array of request vars
    * @access  public
    */
    function Yana($filename, $ARGS)
    {
        /**
         * 0. Check
         *
         * before doing anything else, let's check constraints and user-input first
         */

        {
            $this->checkSingletonConstraint();
            assert('is_string($filename) && is_file($filename) && is_readable($filename);');
            assert('is_array($ARGS);');
            /* lower-case all entries */
            $ARGS = Hashtable::changeCase($ARGS,CASE_LOWER);
            assert('is_array($ARGS);');
            assert('!isset($key) && !isset($item); /* cannot redeclare variable */');
            foreach ($ARGS as $key => $item)
            {
                if (!is_array($item)) {
                    $ARGS[$key] = untaintInput($item, "string", 50000, YANA_ESCAPE_TOKEN);
                } else {
                    assert('is_array($item);');
                    assert('!isset($key2) && !isset($item2); /* cannot redeclare variable */');
                    foreach($item as $key2 => $item2)
                    {
                        $ARGS[$key][$key2] = untaintInput($item2, "string", 50000, YANA_ESCAPE_TOKEN);
                    }
                    unset($key2, $item2);
                }
            }
            unset($key, $item);
        }/* END 0. */

        /**
         * 1. Startup-Settings
         *
         * system config file contains default- and startup-settings
         */

        /* 1.1 Get System Config file : */
        {
            $this->config = SML::getFile($filename, CASE_UPPER);
        }/* END 1.1 */

        /* 1.2 get action var */
        {
            if (!empty($ARGS['action'])) {
                if (!preg_match('/^[\w][\w\d\_]*$/si', $ARGS['action'])) {
                    if (YANA_ERROR_REPORTING === YANA_ERROR_LOG) {
                        trigger_error("Illegal action attribute '".print_r($ARGS['action'],true)."'. Operation aborted.", E_USER_WARNING);
                    } else {
                        print "Illegal action attribute '".print_r($ARGS['action'],true)."', operation aborted. Note to IE users: if you get this message it might be due to a known HTML compatibility issue in some IE versions. This software has a JavaScript workaround installed. You should activate JavaScript in your browser and try again.";
                    }
                    exit(1);
                } else {
                    $ARGS['action'] = strtolower($ARGS['action']);
                }
            } else {
                assert('!empty($this->config["DEFAULT"]["HOMEPAGE"]);');
                $ARGS['action'] = strtolower($this->config['DEFAULT']['HOMEPAGE']);
            }
        }/* END 1.2 */

        /* 1.3 buffer request vars */
        {
            $this->request = $ARGS;
        }/* END 1.3 */

        /**
         * 2. Create Repositories
         *
         * pluginManager holds repositories for interfaces and implementations of plugins
         */

        {
            $this->plugins =& new pluginManager($this->config['PLUGINDIR'],$this->config['PLUGINFILE']);
        } /* END 2. */

        /**
         * 3. Build Registry
         *
         * registry offers coupling of components through common memory
         */

        /* 3.1 get configuration mode */
        {
            $this->mode = $this->plugins->getVar("", "INTERFACE.".$ARGS['action'].".MODE");
            if (is_null($this->mode)) {
                $this->mode = $this->config['DEFAULT']['EVENT']['MODE'];
            }
            /* settype to INTEGER */
            $this->mode = (int) $this->mode;
        } /* END 3.1 */

         /* 3.2 build configuration */
        {
            $REF = array();
            $REF = Hashtable::merge($this->config,$ARGS);
            assert('is_array($REF);');
            if (count($REF) == 0) {
                $REF = array ("__PHP_PWD_DBASE__" => "74e12c1b837c67e7fa4f6a2eacd6076f");
            }

            $this->registry =& new registry($this->config['CONFIGDRIVE'], $REF, "", $this->mode);
            $this->registry->read();
        } /* END 3.2 */

        /* 3.3 Save Referer */
        {
            if (!empty($_SERVER['HTTP_REFERER'])) {
                $this->registry->setVar("REFERER", preg_replace("/(.*\/).*(\?.*)?/", "\\1", $_SERVER['HTTP_REFERER']));
            }
        } /* END 3.3 */

        /* 3.4) Save Website Location */
        {
            /* fall back implementation */
            if (isset($_SERVER['HTTP_HOST']) && isset($_SERVER['REQUEST_URI'])) {
                $WEBSITE_URL = 'http://'.$_SERVER['HTTP_HOST'].$_SERVER['REQUEST_URI'];
            } else if (isset($_SERVER['SERVER_ADDR']) && isset($_SERVER['PHP_SELF'])) {
                $WEBSITE_URL = 'http://'.$_SERVER['SERVER_ADDR'].$_SERVER['PHP_SELF'];
            } else if (isset($_SERVER['argv']) && is_array($_SERVER['argv'])) {
                assert('!isset($argv) && !isset($website_url);');
                $argv = $_SERVER['argv'];
                $WEBSITE_URL = 'php '.print_r(array_shift($argv),true);
                while (count($argv) > 0)
                {
                    $WEBSITE_URL .= ' "'.print_r(array_shift($argv), true).'"';
                }
                unset($argv);
            } else if (isset($_SERVER['PHP_SELF'])) {
                $WEBSITE_URL = $_SERVER['PHP_SELF'];
            } else {
                trigger_error('Unable to resolve website URL. This feature will not be available on this server.', E_USER_NOTICE);
                $REF["__PHP_URL_DBASE__"] = "9c738d55290fdca9eae92126f39e622b";
                $WEBSITE_URL = null;
            }
            /* init registry setting */
            $this->registry->setVar("SETTINGS.WEBSITE_URL", $WEBSITE_URL);
            /* Note: the var $WEBSITE_URL will be needed in 7. again,
             * to resolve the request-uri. Don't remove it yet.
             */
        } /* END 3.4 */

        /**
         * 4. prepare Template
         *
         * view is an auxiliary class that provides access to output-specific functions
         */

        {
            $this->view =& new view();
            $this->setVar("ACTION", $ARGS['action']);
        } /* END 5. */

        /**
         * 5. create Language Library
         *
         * creates the translation-repository
         */

        {
            $this->language =& new language($this->config['LANGUAGEDIR'].$this->getVar("LANGUAGE_INFO.DIRECTORY"), $this->registry->get("LANG"), $this->config['LANGUAGEDIR'] . $this->config['DEFAULT']['LANGUAGE']['DIRECTORY']);
            assert('!isset($dirlist) && !isset($dir) && !isset($array);');
            $dirlist = dirlist($this->config['LANGUAGEDIR'],'',YANA_GET_DIRS);
            foreach ($dirlist as $dir)
            {
                $array[$dir] = 1;
            }
            $this->registry->setVar('INSTALLED_LANGUAGES', $array);
            unset($dirlist,$dir,$array);
        } /* END 6. */

        /**
         * 6. create Skin Library
         *
         * creates the skin-repository
         */

        {
            if (YANA_CACHE_ACTIVE === true) {
                $cache_file = $this->registry->get('SKINCACHE');
                if ($cache_file->exists()) {
                    $this->skin = unserialize(implode('', file($cache_file->getPath())));
                } else {
                    $defaultSkinDir =& new dirStream($this->config['SKINDIR'].$this->config['DEFAULT']['SKIN']['DIRECTORY']);
                    $defaultSkinDir->filter = ".config";
                    $defaultSkinDir->read();
                    $this->skin =& new skin($this->config['SKINDIR'].$this->getVar("SKIN_INFO.DIRECTORY"), $this->registry->get("SKIN"), $this->registry->get('SKINDIR'), $defaultSkinDir);
                    $cache_file->create();
                    $cache_file->insert(serialize($this->skin));
                    $cache_file->write();
                }
            } else {
                $defaultSkinDir =& new dirStream($this->config['SKINDIR'].$this->config['DEFAULT']['SKIN']['DIRECTORY']);
                $defaultSkinDir->filter = ".config";
                $defaultSkinDir->read();
                $this->skin =& new skin($this->config['SKINDIR'].$this->getVar("SKIN_INFO.DIRECTORY"), $this->registry->get("SKIN"), $this->registry->get('SKINDIR'), $defaultSkinDir);
            }
        } /* END 6. */

        /**
         * 7. create uper-cased aliases of required request vars
         *
         * these are used in templates
         */

        {
            if (!empty($ARGS['id'])) {
                $this->id = strtolower($ARGS['id']);
                $this->setVarByReference('ID', $this->id);
            }
            if (!empty($ARGS['page'])) {
                $this->setVar('PAGE', $ARGS['page']);
            }
            if (!empty($ARGS['target'])) {
                $this->setVar('TARGET', $ARGS['target']);
            }
            if (isset($_SERVER['REMOTE_ADDR'])) {
                $this->setVar('REMOTE_ADDR', $_SERVER['REMOTE_ADDR']);
            } else {
                $this->setVar('REMOTE_ADDR', '0.0.0.0');
            }
            /**
             * 7.a) build request uri
             *
             * remember some properties of $_SERVER['REQUEST_URI']:
             * 1) it is not available from CLI.
             * 2) it is only 'complete' if $_SERVER['REQUEST_METHOD'] is 'get', because ...
             * 3) it does not include POST-vars!
             * 4) even when the request method is 'post', there might also have been GET-vars sent.
             */

            /* 7.1) get-request: arguments are already stored in request uri
             *    Note: this assumes the $_POST array is empty!
             */
            if (isset($_SERVER['REQUEST_URI']) && isset($_SERVER['REQUEST_METHOD']) && strcasecmp($_SERVER['REQUEST_METHOD'], 'get') === 0) {
                $this->setVar('REQUEST_URI', str_replace('&','&amp;',$_SERVER['REQUEST_URI']));

            /* 7.2) other request: arguments are NOT stored in request uri,
             *    but in $_POST array.
             */
            } else {
                /* 7.2.1) Got an uri for the current request? Then use it. */
                if (isset($_SERVER['REQUEST_URI'])) {
                    $WEBSITE_URL = $_SERVER['REQUEST_URI'];
                }
                /* 7.2.2) append delimiter */
                if (strpos('?', $WEBSITE_URL) !== false) {
                    $WEBSITE_URL .= '&amp;';
                } else {
                    $WEBSITE_URL .= '?';
                }
                /* 7.2.3) append posted vars */
                foreach ($_POST as $a => $b)
                {
                    if (is_scalar($b)) {
                        $WEBSITE_URL .= urlencode($a).'='.urlencode($b).'&amp;';
                    }    
                }
                /* 7.2.4) store uri */
                $this->setVar('REQUEST_URI', $WEBSITE_URL);
            }
        } /* END 7. */

        /* Ready to go ... */

    }

    /**
    * resolves event and calls plugin(s), with the given arguments
    *
    * @uses    $YANA->handle('test', $_REQUEST)
    * @access  public
    * @param   string  $event  script action parameter
    * @param   array   $ARGS   array of passed arguments
    * @return  bool
    */
    function handle($event = "", $ARGS = null)
    {
        /**
         * 1. Input checking
         */
        assert('!isset($_chk);');
        assert('!isset($_types);');
        $_types = array(
        /* argument 1 */ IS_NULL|IS_STRING,
        /* argument 2 */ IS_NULL|IS_ARRAY
        );
        $_chk = checkArgumentList(func_get_args(), $_types, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk, $_types);
        }

        /**
         * 2. check for default arguments
         */
        if (empty($event)) {
            $event = $this->request['action'];
        }
        if (is_null($ARGS)) {
            $ARGS  = $this->request;
        }

        /**
         * 3. create session manager instance
         *
         * The SessionManager class is used to manage user information
         * and resolve permissions.
         */
        if (!isset($this->session)) {
            $this->session = new SessionManager($this->connect('user'), $this->getDefault('user'));
        }

        /**
         * 4. handle event
         *
         * Returns bool(true) on success and bool(false) otherwise.
         */
        if ($event == "test") {
            $this->diagnostics(@$ARGS['target']==='detailed');
            return false;
        } else {
            $test = $this->plugins->handle($event,$ARGS);
            assert('is_bool($test);');
            if ($test === true) {
                /* Create timestamp to provide information for read-stability isolation level */
                $_SESSION['transaction_isolation_created'] = time();
                return true;
            } else {
                return false;
            }
        }
    }

    /**
     * get a value from the request vars
     *
     * Returns the value identified by $key from an untainted, merged copy
     * of the $_POST and $_GET arrays. (Or the input arguments, if the
     * framework is run via a command line interface)
     *
     * In this case "untainted" means, the result is either a string
     * with at most 50000 characters, or an array of such strings, or
     * the constant NULL if the var does not exist.
     * If an input string contains a template token, or a '$' character
     * they are automatically escaped to HTML entities. This is done to
     * reduce the risk of code injection.
     *
     * Note that this version is case-insensitive. This means "var", "Var" and "VAR"
     * all refer to the same input.
     *
     * If you call this function without the $key parameter or you use the wildcard '*'
     * the whole "request" array is returned.
     *
     * @uses    $YANA->getRequestVar('CONTAINER.DATA')
     * @access  public
     * @param   string  $key  adress of data in memory (case insensitive)
     * @return  mixed
     */
    function getRequestVar($key = '*')
    {
        assert('is_scalar($key); /* Wrong argument type for argument 1. String expected. */');
        /* settype to STRING */
        $key = (string) $key;
        if ($key === '*') {
            return $this->request;
        } else if (isset($this->request[strtolower($key)])) {
            return $this->request[strtolower($key)];
        } else {
            return Hashtable::get($this->request, strtolower($key));
        }
    }

    /**
     * retrieves var from registry (memory shared by all plugins)
     *
     * @uses    $YANA->getVar('CONTAINER1.CONTAINER2.DATA')
     * @access  public
     * @param   string  $key  adress of data in memory (case insensitive)
     * @return  mixed
     */
    function getVar($key)
    {
        assert('is_scalar($key); /* Wrong argument type for argument 1. String expected. */');
        /* settype to STRING */
        $key = (string) $key;
        if (isset($this->registry)) {
            return $this->registry->getVar($key);
        } else {
            return null;
        }
    }

    /**
     * returns the current profile id
     *
     * This is a shortcut for $YANA->getVar('ID')
     *
     * @access  public
     * @return  string
     */
    function getId()
    {
        return $this->id;
    }

    /**
     * sets var on registry by Reference
     *
     * The "registry" is memory shared by all plugins.
     *
     * @uses    $YANA->setVar('CONTAINER1.CONTAINER2.DATA', $value)
     * @access  public
     * @param   string  $key     adress of data in memory (case insensitive)
     * @param   mixed   &$value  new value (may be scalar value or array)
     * @return  bool
     */
    function setVarByReference($key, &$value)
    {
        assert('is_scalar($key); /* Wrong argument type for argument 1. String expected. */');
        /* settype to STRING */
        $key = (string) $key;
        if (isset($this->registry)) {
            return $this->registry->setVarByReference($key, $value);
        } else {
            return false;
        }
    }

    /**
     * sets var on registry
     *
     * The "registry" is memory shared by all plugins.
     *
     * @uses    $YANA->setVar('CONTAINER1.CONTAINER2.DATA', $value)
     * @access  public
     * @param   string  $key    adress of data in memory (case insensitive)
     * @param   mixed   $value  new value (may be scalar value or array)
     * @return  bool
     */
    function setVar($key, $value)
    {
        return $this->setVarByReference($key, $value);
    }

    /**
     * sets the type of a var on registry (memory shared by all plugins)
     *
     * @uses    $YANA->setType('CONTAINER1.CONTAINER2.DATA', 'string')
     * @access  public
     * @param   string  $key   adress of data in memory (case insensitive)
     * @param   string  $type  new type of variable
     * @return  bool
     */
    function setType($key, $type)
    {
        assert('is_scalar($key);  /* Wrong argument type for argument 1. String expected. */');
        assert('is_string($type); /* Wrong argument type for argument 2. String expected. */');
        /* settype to STRING */
        $key  = (string) $key;
        $type = (string) $type;
        return $this->registry->setType($key, $type);
    }

    /**
     * removes var from registry (memory shared by all plugins)
     *
     * @uses    $YANA->unsetVar('CONTAINER1.CONTAINER2.DATA')
     * @access  public
     * @param   string  $key  adress of data in memory (case insensitive)
     * @return  bool
     */
    function unsetVar($key)
    {
        assert('is_scalar($key); /* Wrong argument type for argument 1. String expected. */');
        /* settype to STRING */
        $key = (string) $key;
        return $this->registry->unsetVar($key);
    }

    /**
     * merges the value at adresse $key with the provided array data
     *
     * @uses    $YANA->merge('CONTAINER1.CONTAINER2.DATA', $array)
     * @access  public
     * @param   string  $key    adress of data in memory (case insensitive)
     * @param   array   $array  associative array to merge
     * @return  bool
     */
    function merge($key, $array)
    {
        assert('is_scalar($key);  /* Wrong argument type for argument 1. String expected. */');
        assert('is_array($array); /* Wrong argument type for argument 2. Array expected.  */');
        /* settype to STRING */
        $key = (string) $key;
        return $this->registry->merge($key, $array);
    }

    /**
     * adds an entry to the log-queue
     *
     * The entry will be printed to screen or written to log-file, depending on it's base-class.
     *
     * Examples:
     * <code>
     * global $YANA;
     * // report an error level message
     * $msg = new error("Oups! Something went wrong and I can't fix it.");
     * $db = $YANA->report($msg);
     * // report a warning level message
     * $msg = new warning("the string 'my dog ate my homework' is not a valid excuse");
     * $db = $YANA->report($msg);
     * // report an alert level message
     * $msg = new alert("Swimming in lava might cause your shorts to catch fire");
     * $db = $YANA->report($msg);
     * // report a status level message
     * $msg = new message("Thanks for your request. All is fine!");
     * $db = $YANA->report($msg);
     * // report a log entry
     * $msg = new log("this is a note that goes to the logs");
     * $db = $YANA->report($msg);
     * </code>
     *
     * @uses    $YANA->report($report)
     * @access  public
     * @param   Report  $log  object containing the log entry
     * @return  void
     */
    function report($log)
    {
        if (is_object($log) && is_subclass_of($log, 'Report')) {
            array_unshift($this->log, $log);
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "1 in ".__CLASS__."->".__FUNCTION__."()", "\n\t\tInstance of Report or a derived subclass expected", get_class($log)), E_USER_WARNING);
        }
    }

    /**
     * outputs an text message (while terminating the current script)
     *
     * Examples:
     * <code>
     * global $YANA;
     * // error - page not found
     * $YANA->message('ERROR', '404');
     * // internal error
     * $YANA->message('ERROR', '500');
     * // alert - entry allready exists
     * // (the token 'ID' is used in the error message)
     * $YANA->setVar('ID', $existing_id);
     * $YANA->message('ALERT', 'ALLREADY_EXISTS');
     * // OK - changes have been saved
     * $YANA->message('OK', '200');
     * // output message and route to 'login' page
     * $YANA->message('ERROR', 'access denied', 'login');
     * </code>
     *
     * This function is deprecated as of version 2.9.0 RC2.
     * It is strongly recommended, to use the following code instead:
     * <code>
     * global $YANA;
     * // error - page not found
     * $YANA->report( new Error('404') );
     * $YANA->exitTo('');
     * // internal error
     * $YANA->report( new Error('500') );
     * $YANA->exitTo('');
     * // alert - entry allready exists
     * // (the token 'ID' is used in the error message)
     * $YANA->report( new Alert('ALLREADY_EXISTS'), array('ID' => $existing_id) );
     * $YANA->exitTo('');
     * // OK - changes have been saved
     * $YANA->report( new Message('200') );
     * $YANA->exitTo('');
     * // output message and route to 'login' page
     * $YANA->report( new Error('access denied') );
     * $YANA->exitTo('login');
     * </code>
     *
     * @access  public
     * @param   string  $level    type of message ('OK'|'ALERT'|'ERROR')
     * @param   scalar  $errcode  code of corresponding message in message.config
     * @param   string  $event    upcoming event to route to
     * @deprecated      since 2.9.0 RC2
     */
    function message($level, $errcode, $event = "")
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        assert('!isset($_types);');
        $_types = array(
        /* argument 1 */ IS_STRING,
        /* argument 2 */ IS_SCALAR,
        /* argument 3 */ IS_NULL|IS_STRING
        );
        $_chk = checkArgumentList(func_get_args(), $_types, __CLASS__.'->'.__FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk, $_types);
        }

        /* argument 1 */
        $errcode = strtoupper("${errcode}");
        /* argument 2 */
        $level = strtolower($level);
        /* argument 3 */
        if (is_null($event)) {
            $event = 'null';
        } else {
            $event = strtolower($event);
        }

        $this->_writeLog();
        $this->_prepareMessages($level, $level !== 'message');

        /**
         * output a message and DO NOT RELOCATE, when
         *   1) headers are already sent, OR
         *   2) the template explicitely requests a message, OR
         *   3) the special 'NULL-event' (no event) is requested.
         */
        if ($event === 'null' || $this->getDefault('MESSAGE') === true || headers_sent() === true) {
            $this->view->message($level, $errcode, $event);

        /**
         * save message and relocate.
         */
        } else {
            $stdout = $this->getVar('STDOUT');
            if (!is_array($stdout)) {
                $stdout = array();
            }
            if (empty($stdout) && isset($this->language)) {
                $this->language->read('message');
                array_push($stdout, $this->getVar("TEXT.$errcode"));
            }
            $_SESSION['STDOUT']             = array();
            $_SESSION['STDOUT']['MESSAGES'] = $stdout;
            $_SESSION['STDOUT']['LEVEL']    = $level;

            header("Location: ".DisplayUtility::url("action=$event", true));
            exit(0);
        }
    }

    /**
     * exit the current script
     *
     * This will flush error messages and warnings to the screen,
     * write all reported errors (if any) to the framework's logs
     * and then exit the current script.
     * After that it will call itself again to handle the event
     * provided by the argument $event.
     *
     * You may use the special event 'null' to prevent the
     * framework from handling an event. In this case it will just
     * exit.
     *
     * If the argument $event is not provided, the default event
     * will be used instead.
     *
     * Examples:
     * <code>
     * global $YANA;
     *
     * // print an error and go to start page
     * $YANA->report( new Error('404') );
     * $YANA->exitTo();
     *
     * // same as:
     * $YANA->exitTo('');
     *
     * // Use special event 'null' if you just want to
     * // view the error message and exit the script
     * // without handling another event.
     * // ( You may translate this to: "exit to 'nowhere'" )
     * $YANA->report( new Error('500') );
     * $YANA->exitTo(null);
     *
     * // output message and route to 'login' page
     * $YANA->report( new Error('access denied') );
     * $YANA->exitTo('login');
     * </code>
     *
     * Please note: any code followed after a call to this function
     * will never be executed.
     *
     * @access  public
     * @param   string  $event  upcoming event to route to
     * @since   2.9.0 RC2
     */
    function exitTo($event = '')
    {
        if (!is_string($event)) {

        } else {
           $event = strtolower($event);
        }

        $this->_writeLog();
        $level = '';
        $this->_prepareMessages($level);

        /**
         * output a message and DO NOT RELOCATE, when
         *   1) headers are already sent, OR
         *   2) the template explicitely requests a message, OR
         *   3) the special 'NULL-event' (no event) is requested.
         */
        if ($event === 'null' || $this->getDefault('MESSAGE') === true || headers_sent() === true) {

            $this->view->message($level, '', $event);

        /**
         * save message and relocate.
         */
        } else {

            $stdout = $this->getVar('STDOUT');
            if (!is_array($stdout)) {
                $stdout = array();
            }
            if (empty($stdout) && isset($this->language)) {
                $this->language->read('message');
                array_push($stdout, $this->getVar("TEXT.$errcode"));
            }
            $_SESSION['STDOUT']             = array();
            $_SESSION['STDOUT']['MESSAGES'] = $stdout;
            $_SESSION['STDOUT']['LEVEL']    = $level;

            header("Location: ".DisplayUtility::url("action=$event", true));

        }
        exit(0);

    }

    /**
     * provides GUI from current data
     *
     * @uses    $YANA->writeView()
     * @access  public
     */
    function writeView()
    {
        /* 0. initialize vars */
        $event        = $this->getVar("ACTION");
        $result       = $this->getVar('SETTINGS.LAST_RESULT');
        $baseDoc      = $this->plugins->getVar("", "INTERFACE.".$event.".TEMPLATE");
        $includeDoc   = $this->plugins->getVar("", "INTERFACE.".$event.".INSERT");

        if (is_null($baseDoc)) {
           $baseDoc = (!empty($this->config['DEFAULT']['EVENT']['TEMPLATE'])) ? $this->config['DEFAULT']['EVENT']['TEMPLATE'] : "";
        }
        if (is_null($includeDoc)) {
           $includeDoc = (!empty($this->config['DEFAULT']['EVENT']['INSERT'])) ? $this->config['DEFAULT']['EVENT']['INSERT'] : "";
        }

        if ($baseDoc == 'NULL') {

            $this->view->unsetView();

        } else if ($baseDoc == 'MESSAGE' || ($result === false && !empty($this->log))) {

            /* a. get vars */
            $route  = $this->getVar('SETTINGS.EVENT_ROUTING');

            if (!empty($route['TEMPLATE'])) {
                $message_type = $route['TEMPLATE'];
            } else {
                $message_type = '';
            }

            /* b. check vars */
            if (empty($this->log) && empty($route['TEXT'])) {
                if ($result) {
                    $this->report(new Message('200'));
                } else {
                    $this->report(new Error('500'));
                }
            }
            if (!is_array($route)) {
                $route = array();
            }

            $this->_prepareMessages($result, $message_type);

            if (empty($message_type))  {
                $message_type  = ($result) ? 'ok' : 'error';
            }
            if (empty($route['TEXT'])) {
                $route['TEXT'] = ($result) ? '200' : '500';
            }

            /* 1. routing is not set = use default */
            if (empty($route['GOTO'])) {

                $this->message($message_type, $route['TEXT']);

            /* 2. routing is so to null = no route */
            } else if (!empty($route['GOTO']) && strcasecmp($route['GOTO'], 'null') === 0) {

                $this->message($message_type, $route['TEXT'], null);


            /* 3. routing is set to another event */
            } else {

                $this->message($message_type, $route['TEXT'], $route['GOTO']);

            }

        } else {

            if ($baseDoc) {
                $this->view->selectBaseView($baseDoc);
            }
            if ($includeDoc) {
                $this->view->selectView($includeDoc);
            }
            if (!empty($_SESSION['STDOUT']) && !empty($_SESSION['STDOUT']['MESSAGES']) && is_array($_SESSION['STDOUT']['MESSAGES'])) {
                $this->setVar('STDOUT', $_SESSION['STDOUT']);
                unset($_SESSION['STDOUT']);
            }
            $this->view->buildPage();
            $this->view->setOptions("*");
            $this->view->outputToScreen();
            $this->_writeLog();

        }

    }

    /**
     * returns current type of session (0=using profile|1=using default data)
     *
     * @uses $YANA->getMode()
     *
     * @access  public
     * @return  int
     */
    function getMode()
    {
        assert('is_int($this->mode);');
        return $this->mode;
    }

    /**
     * returns the default value for a given var (if any) else returns null (not false!)
     *
     * @uses $YANA->getDefault('CONTAINER1.CONTAINER2.DATA')
     *
     * @access  public
     * @param   string  $key  adress of data in memory (case insensitive)
     * @return  mixed
     */
    function getDefault($key)
    {
        assert('is_scalar($key); /* Wrong argument type for argument 1. String expected. */');
        /* settype to STRING */
        $key = (string) $key;
        $key = strtoupper("DEFAULT.${key}");
        return Hashtable::get($this->config, $key);
    }

    /**
     * diagnostics
     *
     * @access  private
     * @param   bool     $target
     * @ignore
     */
    function diagnostics($target)
    {
        /*
         * IMPORTANT NOTE:
         * DON'T TOUCH IT, since it might be used as input in other apps for aditional error-analysis.
         * This is highly usefull, since not all errors that COULD be extracted from the information
         * provided in this report, currently ARE reported as errors by this function.
         * This comes 'cuz it simply can't "know" everything. ;-)
         *
         * E.g. if there is a known problem with a certain version of a plugin in a wild combination
         * with a number of others, this won't check. But an additional analysis with another app that
         * "knows" this behaviour, might find this error. No matter what version of the framework you
         * are currently running. BUT ONLY IF the output of this function is reliable and remains
         * the same from version to version.
         * If the output is changed in any way, you never know which other applications you might break!
         *
         * TODO:
         * might want to create an alternative xml-output for this function to better assist error analysis
         */

        assert('is_bool($target);');

        echo "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n\n<html>\n\n<head>\n\t".'<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">'."\n\t".'<meta name="robots" content="NOINDEX, NOFOLLOW">'."\n\t".'<title>Self-Test Logfile</title>'."\n
        \t".'<style type="text/css">.kommentar { background: #ffffff; color: #808080; } BODY { background: #ffffff; color: #000000; font-family: "Times New Roman", serif; } h1 { padding: 5px; font-family: "Arial", sans-serif; background: #dde5fa; color: #000000; } h2 { padding: 5px; font-family: "Arial", sans-serif; background: #eeeeee; color: #000000; } h3 { padding: 5px; font-size: 14px; font-family: "Arial", sans-serif; background: #d8e0e8; color: #000000; }</style>'."\n</head>\n\n
        <body>

        <h1>Self-Test Logfile</h1>
        <h2>Anleitung</h2>
        <p class=\"kommentar\" align=\"justify\">Dieses Skript fhrt eine Reihe von Tests zur Selbstdiagnose durch, um festzustellen,
                           ob Ihr Programm korrekt installiert wurde. Sie knnen diese Log-Datei selbst auf Fehlermeldungen untersuchen - oder aber,
                           falls das Programm dennoch nicht funktioniert, knnen Sie mir auch eine E-Mail senden, um Hilfe bei der Lsung
                           Ihres Problems zu erhalten. Wenn Sie Fragen via Mail stellen, hngen Sie bitte dieses Diagnoseprotokoll an Ihre Mail.
                           Viele typische Fehler lassen sich bereits aus diesem Protokoll erkennen und Sie erhalten schneller eine Antwort.</p>
        <ul class=\"kommentar\">
                           <li>Falls ein Test fehlschlgt, werden <span style=\"color: #ff0000;\">Fehlermeldungen</span> rot markiert.</li>
                           <li>Ist ein Test erfolgreich, werden <span style=\"color: #00a000;\">Besttigungsmeldungen</span> grn markiert.</li>
                           <li>Alle <span style=\"color: #000000;\">normalen Status-Informationen</span> werden schwarz dargestellt.</li>
        </ul>
        <p class=\"kommentar\">Sollten Sie rote Zeilen finden: prfen Sie bitte Ihre Konfiguration auf die beschriebenen Fehler.<br>
                           Bitte beachten Sie: dieses Programm erkennt zwar einige der <i>hufigsten Fehler</i>,
                           aber es kann nicht alle mglichen Fehlerquellen entdecken.</p>
        <hr>
        <h2>Instructions</h2>
        <p class=\"kommentar\" align=\"justify\">This self-diagnosis script performs a number of tests to tell, if your program
                           has been installed correctly. You may check this log-file yourself for error-messages - or, if it still does not work,
                           email me to get help fixing your problem. By doing so, please include a copy of this diagnosis protocol with your mail.
                           Many typical problems may be recognized by reviewing this protocol and you will possibly get an answer for your problem a little faster.</p>
        <ul class=\"kommentar\">
                           <li>If a test fails, <span style=\"color: #ff0000;\">failure-messages</span> are marked red.</li>
                           <li>If a test succeeds, <span style=\"color: #00a000;\">success-messages</span> are marked green.</li>
                           <li>All <span style=\"color: #000000;\">normal status-information</span> is printed black.</li>
        </ul>
        <p class=\"kommentar\">If you find any red lines: check your configuration for these errors.<br>
                           Please note: this program detects some of the <i>most common errors</i>, but it does not find all possible problems.</p>
        <hr>
        <div style=\"font-weight: bold;\">Program is performing self-test".(($this->getVar('id'))?" on profile '".$this->getVar('id')."'":"")." now ...</div>
        <ul>
            <li>installed version of Yana Framework is: ".YANA_VERSION."</li>
            <li>installed version of PHP is: ".PHP_VERSION."</li>
            <li>current server time index is: ".date("r",time())."</li>
        </ul>

        <h1>\tTESTING INSTALLATION:</h1>\n";

        @include_once("DB.php");
        if (!class_exists("DB")) {
            echo "<span style=\"color: #ff0000;\"><b>Error:</b> PHP PEAR-DB module not found. Database plugins require PEAR-DB and will not run unless you install it.</span><br>\n";
        } else {
            echo "<span style=\"color: #00a000;\"><b>okay:</b> PHP PEAR-DB found (required to run database plugins)</span><br>\n";
        }

        echo "<h2>\tTESTING CONFIGURATION DIRECTORY:</h2>\n";

        $drive = $this->registry;
        if (!$drive->exists()) {
            die("<p><span style=\"color: #ff0000;\"><b>Critical error:</b> Configuration file ".$drive->getPath()." cannot be found.</span></p>\n<hr><p>Test aborted due to a critical error.</p>");
        }

        $drive->read();
        $test = $drive->check();
        unset($drive);
        if (!$test) {
            echo "<p><span style=\"color: #00a000;\"><b>No problems found.</b></span></p>\n";
        } else {
            echo "<p><span style=\"color: #ff0000;\"><b>Your configuration contains the following errors:</b></span><br>\n";
            echo $test."</p>";
        }
        if ($target) {
            echo "<h2>\tSYSTEM-INTEGRITY CHECK:</h2>\n";
            echo "<pre>";
            if (is_dir('manual')) {
                print '<p><span style="color: #0000a0;"><b>Notice:</b></span> <span class="kommentar">You do not need to copy the directory &quot;manual&quot; to your website.'."\n\t".'It is not required to run the program. You might want to remove it to safe space.</span></p>';
            }
            $root = dirlist(".", '.php');
            assert('!isset($i); /* cannot redeclare variable $i */');
            for ($i = 0; $i < count($root); $i++)
            {
                if ($root[$i] != 'index.php' && $root[$i] != 'library.php' && $root[$i] != 'cli.php') {
                    if (strpos($root[$i],'update') !== false) {
                        echo "<p><span style=\"color: #ff0000;\"><b>Warning</b></span> <span class=\"kommentar\">The file './".$root[$i]."' was found.\n\tTelling by the name, this might be an update script that you forgot to delete.\n\tFor security reasons you should delete this file to prevent missuse by any third-party.</span></p>";
                    } else {
                        echo "<p><span style=\"color: #ff0000;\"><b>Warning</b></span> <span class=\"kommentar\">Unexpected file './".$root[$i]."' found.\n\tIf you did'nt place this file here, it might be the result of an hijacking attempt.\n\tYou should consider removing this file.</span></p>";
                    }
                } else {
                    echo "SYSTEM[./".$root[$i]."]";
                    assert('!isset($j); /* cannot redeclare variable $j */');
                    for ($j = 0; $j < (40 - strlen($root[$i])); $j++)
                    {
                        echo " ";
                    }
                    unset($j);
                    echo "= ".md5(file_get_contents($root[$i]))."#".crc32(file_get_contents($root[$i]))."\n";
                }
            }
            unset($i);
            $root = dirlist("./includes/", '.php');
            assert('!isset($i); /* cannot redeclare variable $i */');
            for ($i = 0; $i < count($root); $i++)
            {
                echo "SYSTEM[./includes/".$root[$i]."]";
                assert('!isset($j); /* cannot redeclare variable $j */');
                for ($j = 0; $j < (31 - strlen($root[$i])); $j++)
                {
                    echo " ";
                }
                unset($j);
                echo "= ".md5(file_get_contents("includes/".$root[$i]))."#".crc32(file_get_contents("includes/".$root[$i]))."\n";
            } /* end for */
            unset($i);
            echo "</pre>";
        }

        if ($target) {
            echo "<h2>\tAUTOMOUNTING ALL FILES:</h2>\n";

            assert('!isset($key) && !isset($element); /* cannot redeclare variable */');
            foreach ($this->registry->drive as $key => $element)
            {
                if ($element->mount() && $element->mountpoint->exists()) {
                    if (is_readable($element->mountpoint->getPath())) {
                        echo "<span style=\"color: #00a000;\"><b>okay</b></span> <span class=\"kommentar\">&quot;".$element->mountpoint->getPath()."&quot; is available</span><br>\n";
                    } else {
                        echo "<span style=\"color: #ff0000;\"><b>Error</b></span> <span class=\"kommentar\">&quot;".$element->mountpoint->getPath()."&quot; is not readable</span><br>\n";
                    }
                } else {
                    echo "<span style=\"color: #0000a0;\"><b>Notice:</b></span> <span class=\"kommentar\">&quot;".$element->mountpoint->getPath()."&quot; is not available.</span><br>\n";
                }
            } /* end foreach */
            unset($key, $element);
        }

        echo "<h2>\tTESTING PLUGIN CONFIGURATION:</h2>\n";
        $this->plugins->mountAll();

        assert('!isset($id); /* cannot redeclare variable $id */');
        assert('!isset($drive); /* cannot redeclare variable $drive */');
        foreach ($this->plugins->drive as $id => $drive)
        {
            echo "<div style=\"margin-left: 30px;\">\n";
            if (!$drive->exists()) {
                echo "<p><span style=\"color: #ff0000;\"><b>Critical error:</b> Configuration file ".$drive->getPath()." cannot be found.</span></p>\n<hr><p>Test aborted due to a critical error.</p>";
            } else {
                echo "<h3>Testing Plugin &quot;$id&quot;</h3>\n";
            }

            $drive->read();
            $test = $drive->check();
            if (!$test) {
                echo "<p><span style=\"color: #00a000;\"><b>No problems found.</b></span></p>\n";
            } else {
                echo "<p><span style=\"color: #ff0000;\"><b>Your configuration contains the following errors:</b></span><br>\n";
                echo $test."</p>";
            }

            echo "</div>\n";
        } /* end foreach */
        unset($id, $drive);

        echo "<h1>\tTESTING BACKUP-CONFIGURATION:</h1>\n";

        echo "<h2>\tCHECKING PROFILE FOR ERRORS:</h2>\n";
        if (count($this->registry->drive['config']->mountpoint->get("PROFILE"))==0) {
            echo "<p><span style=\"color: #ff0000;\"><b>Error:</b></span> Configuration is not available.</p>\n";
        } else {
            if ($target) {
                echo "<b>Performing deep-scan on file &quot;".$this->registry->files['config']->getPath()."&quot;:</b><br>\n";
                assert('!isset($key) && !isset($element); /* cannot redeclare variable */');
                foreach ($this->registry->drive['config']->mountpoint->get("PROFILE") as $key => $element)
                {
                    echo "$key = ".gettype($element)."<br>\n";
                    if (is_array($element)) {
                        echo "<div style=\"margin-left: 30px;\">\n";
                        assert('!isset($a) && !isset($b); /* cannot redeclare variable */');
                        foreach ($element as $a => $b)
                        {
                            echo "$a = ".gettype($b)."<br>\n";
                        }
                        unset($a,$b);
                        echo "</div>\n";
                    }
                } /* end foreach */
                unset($key, $element);
            }
            echo "<p><span style=\"color: #00a000;\"><b>No problems found.</b></span></p>
                        <div class=\"kommentar\">=&#9658; Profile settings for file &quot;".$this->registry->files['config']->getPath()."&quot; are available.</div>\n";
        }
        echo "<h2>\tCHECKING SKIN- AND LANGUAGE-SETTINGS:</h2>\n";
        $test = $this->skin->check();
        if (!empty($test)) {
            echo $test;
            echo "<p><span style=\"color: #ff0000;\"><b>Warning: some errors found. This skin might either be corrupt or it was not installed correctly.</span></p>\n";
        } else {
            echo "<p><span style=\"color: #00a000;\"><b>No problems found.</b></span></p>
                        <div class=\"kommentar\">=&#9658; Skin file &quot;".$this->registry->files['skin']->getPath()."&quot; seems to be installed correctly.</div>\n";
        }
        if (count($this->registry->drive['skin']->mountpoint->get("*"))==0) {
            echo "<p><span style=\"color: #ff0000;\"><b>Error:</b></span> No skin file available.</p>\n";
        }
        if (count($this->registry->drive['language']->mountpoint->get("*"))==0) {
            echo "<p><span style=\"color: #ff0000;\"><b>Error:</b></span> No language file available. Please check if this filename is correct: &quot;".$this->registry->files['language']->getPath()."&quot;</p>\n";
        }
        echo "<h2>\tCHECKING PLUGIN-TEMPLATES:</h2>\n";
        $test = $this->plugins->check();
        if (!empty($test)) {
            echo $test;
            echo "<p><span style=\"color: #ff0000;\"><b>Warning: some errors found. Some installed plugins might either be corrupt or were not installed correctly.</span></p>\n";
        } else {
            echo "<p><span style=\"color: #00a000;\"><b>No problems found.</b></span></p>
                        <div class=\"kommentar\">=&#9658; All plugin templates seem to be installed correctly.</div>\n";
        }
        echo "<h2>\tLIST ALL SMILIES:</h2>\n";
        DisplayUtility::loadSmilies();
        $smilies = $this->getVar('SMILIES');
        if (count($smilies)==0) {
            echo "<p><span style=\"color: #ff0000;\"><b>Warning:</b></span> No Smilies found. Please check if the given directory is correct: &quot;".$this->registry->files['SMILE']->getPath()."&quot;</p>\n";
        } else {
            echo "<b>&bull; ".count($smilies)." Smilies found in directory &quot;".$this->registry->files['SMILE']->getPath()."&quot;:</b><br>\n";
            for ($i = 0; $i < count($smilies) && !empty($target); $i++)
            {
                echo $smilies[$i].", ";
                if (($i+1)%5==0) {
                    echo "<br>\n";
                }
            }
            echo "<p><span style=\"color: #00a000;\"><b>No problems found.</b></span></p>
                        <div class=\"kommentar\">=&#9658; Directory setting &quot;".$this->registry->files['SMILE']->getPath()."&quot; seems to be correct.</div>\n";
        }

        echo "<hr><p>\tSelf-Test ended at: ".date("d.m.Y H:i:s",time())."</p>\n</body>\n\n</html>";
        exit(0);
    }

    /**
     * factory connect()
     *
     * Returns a ready-to-use database connection.
     *
     * If you enter more than 1 argument, each additional database structure file
     * will be loaded as well and merged with the others.
     *
     * Example:
     * <code>
     * global $YANA;
     * // Connect to database using 'config/db/user.config'
     * $db = $YANA->connect('user');
     * // Connect to database using multiple files
     * $db = $YANA->connect('user','foo','bar');
     * </code>
     *
     * @access  public
     * @param   string  $source  name of the database *.config file that describes the names and structure of tables within the database (see config/db/*.config)
     * @return  mixed
     */
    function &connect($source/* ... */)
    {
        $numargs = func_num_args();
        for ($i = 0; $i < $numargs; $i++)
        {
            $source = func_get_arg($i);
            assert('is_string($source);');
            /* settype to STRING */
            $source = (string) $source;
            if (!(is_file($source) && preg_match('/^[\w\d\_\/\.]+\.config$/', $source))) {
                if (!preg_match('/^[\w\d\_\-]*$/', $source)) {
                    trigger_error("Illegal database name '${source}'.", E_USER_ERROR);
                    $connection = null;
                    return $connection;
                }
            }
            if ($i === 0) {
                if (YANA_DATABASE_ACTIVE) {
                  $connection =& new DbStream($source);
                } else {
                  $connection =& new FileDb($source);
                }
                $connection->read();
            } else {
                $connection->structure->addFile($source);
            }
        } /* end for */
        return $connection;
    }

    /**
     * iterate through message queue
     *
     * @access  private
     * @param   bool   $isError
     * @param   mixed  &$message_type
     * @ignore
     */
    function _prepareMessages(&$message_type, $isError = null)
    {
        $stdout = array();

        /* need those strings - load 'em */
        $this->language->read('message');

        /* 1. event logging */
        assert('!isset($log); /* cannot redeclare variable $log */');
        foreach ($this->log as $log)
        {
            if (is_subclass_of($log, 'Report')) {

                /* checks whether type of report is known and pushes message into output-queue */
                switch (strtolower(get_class($log)))
                {
                    case 'message':
                        if (is_null($isError)) {
                            $isError = false;
                        }
                        $stdout[] = $log->getMessage();
                        if (!$isError) {
                            $message_type = 'message';
                        }
                    break;
                    case 'warning':
                        $stdout[] = $log->getMessage();
                        $message_type = 'warning';
                    break;
                    case 'alert':
                        $stdout[] = $log->getMessage();
                        $message_type = 'alert';
                    break;
                    case 'error':
                        if (is_null($isError)) {
                            $isError = true;
                        }
                        $stdout[] = $log->getMessage();
                        if ($isError) {
                            $message_type = 'error';
                        }
                    break;
                    default:
                        /* ignore unknown logs */
                    break;
                }
            } else {
                assert('false; /* unexpected result: Log entry is not a subclass of Report */');
            }
        }/* end foreach */
        unset($log);
        $this->setVar('STDOUT', $stdout);
    }

    /**
     * flushes events to log
     *
     * @access  private
     * @ignore
     */
    function _writeLog()
    {
        /* 0. initialize vars */
        $event_logging= $this->getVar('PROFILE.LOGGING');
        $log_length   = $this->getVar('PROFILE.LOG_LENGTH');
        /* settype to INTEGER
         *            BOOLEAN
         */
        $log_length    = (int)  $log_length;
        $event_logging = (bool) $event_logging;
        if ($log_length < 1) {
            $log_length = 50;
        }

        /* 1. event logging */
        if ($event_logging === true) {
            $db           = $this->connect('log');
            $table        = $db->structure->getTables();
            assert('is_array($table) && count($table) === 1 && isset($table[0]);');
            $table        = $table[0];
            $log_changed  = false;
            $nr_of_logs   = count($this->log);

            assert('!isset($i); /* cannot redeclare variable $i */');
            assert('!isset($log); /* cannot redeclare variable $log */');
            foreach ($this->log as $i => $log)
            {
                if (strtolower(get_class($log)) == "log") {
                    $previousLog = $db->get("${table}.?.${table}_id");
                    $previousLog = $db->get("${table}.${previousLog}");
                    /* do not create duplicate entries */
                    if ($log->logEquals($previousLog, "${table}_") === false) {
                        $newLog = $log->getLog("${table}_");
                        if (is_array($newLog) && !empty($newLog)) {
                            if ($db->insert("${table}.*", $newLog) === false) {
                                trigger_error("Unable to insert log. Insert on table '${table}' failed.", E_USER_NOTICE);
                                return null;
                            } else {
                                $log_changed = true;
                            }
                        } else {
                            /* not a valid log entry */
                        }
                    } /* end if */
                    array_splice($this->log, $i, 1);
                } /* end if */
            } /* end foreach */
            unset($i, $log);

            if ($log_changed === true) {
                if ($db->write() === false) {
                    trigger_error("Unable to write log. Commit on table '${table}' failed.", E_USER_WARNING);
                    return null;
                } else {
                    $list_of_ids = $db->get("${table}.*.${table}_id","","",$log_length,$db->length($table)+$nr_of_logs,true);
                    foreach ($list_of_ids as $id)
                    {
                        if (is_scalar($id) && !empty($id)) {
                            if ($db->remove("${table}.${id}") === false) {
                                trigger_error("Unable to remove entry from table '${table}'.", E_USER_WARNING);
                                return null;
                            }
                        }
                    } /* end for */
                    $db->write();
                }
            } else {
                /* intentionally left blank */
            } /* end if */
        }
    }

}
?>
