<?php
/**
 * Fall back functions
 *
 * This file is to collect fall-backs for PHP 4, in order to enhance backwards compatibility.
 *
 * These will apply automatically.
 * If the original function does not exist in the installed version of PHP,
 * the fall-back function listed here will fill the gap so your code will not be broken.
 *
 * Of course these implementations won't reach the same performance as the original functions -
 * but they will work.
 *
 * @package     yana
 * @subpackage  utilities
 * @license     http://creativecommons.org/licenses/by-nc-sa/2.5/
 * @ignore
 */

/**#@+
 * CONSTANTS
 *
 * @ignore
 */

/* the following applies to (among others)
 *
 * ARRAY_CHANGE_KEY_CASE()
 *
 * introduced in PHP 4.2.0
 */
if (!defined('CASE_LOWER')) {
    define('CASE_LOWER', 0);
}
if (!defined('CASE_UPPER')) {
    define('CASE_UPPER', 1);
}

/* the following applies to
 *
 * $_FILES['foo']['error']
 *
 * introduced in PHP 4.3.0
 */
if (!defined('UPLOAD_ERR_OK')) {
    define('UPLOAD_ERR_OK', 0);
}
if (!defined('UPLOAD_ERR_INI_SIZE')) {
    define('UPLOAD_ERR_INI_SIZE', 1);
}
if (!defined('UPLOAD_ERR_FORM_SIZE')) {
    define('UPLOAD_ERR_FORM_SIZE', 2);
}
if (!defined('UPLOAD_ERR_PARTIAL')) {
    define('UPLOAD_ERR_PARTIAL', 3);
}
if (!defined('UPLOAD_ERR_NO_FILE')) {
    define('UPLOAD_ERR_NO_FILE', 4);
}
if (!defined('UPLOAD_ERR_NO_TMP_DIR')) {
    define('UPLOAD_ERR_NO_TMP_DIR', 5);
}

/* introduced in PHP 4.3.10, 5.0.3 */
if (!defined('UPLOAD_ERR_NO_TMP_DIR')) {
    define('UPLOAD_ERR_NO_TMP_DIR', 6);
}

/* introduced in PHP 5.1.0 */
if (!defined('UPLOAD_ERR_CANT_WRITE')) {
    define('UPLOAD_ERR_CANT_WRITE', 7);
}

/* yana framework only */
if (!defined('UPLOAD_ERR_SIZE')) {
    define('UPLOAD_ERR_SIZE', -1);
}
if (!defined('UPLOAD_ERR_FILE_TYPE')) {
    define('UPLOAD_ERR_FILE_TYPE', -2);
}
if (!defined('UPLOAD_ERR_INVALID_TARGET')) {
    define('UPLOAD_ERR_INVALID_TARGET', -4);
}
if (!defined('UPLOAD_ERR_OTHER')) {
    define('UPLOAD_ERR_OTHER', -5);
}

/* The following applies to (among others)
 *
 * FILE_PUT_CONTENTS()
 *
 * introduced in PHP 5
 */
if (!defined('FILE_APPEND')) {
    define('FILE_APPEND', 8);
}
if (!defined('FILE_USE_INCLUDE_PATH')) {
    define('FILE_USE_INCLUDE_PATH', 1);
}

/* introduced in PHP 4.0.1 / 5.1.0 */
if (!defined('LOCK_SH')) {
    define('LOCK_SH', 1);
}
if (!defined('LOCK_EX')) {
    define('LOCK_EX', 2);
}
if (!defined('LOCK_UN')) {
    define('LOCK_UN', 3);
}
if (!defined('LOCK_NB')) {
    define('LOCK_NB', 4);
}

/**#@-*/

/* FUNCTIONS */

if (!function_exists('file_get_contents')) {
    /**
     * file_get_contents() for PHP < 4.3.0
     *
     * The function file_get_contents() was introduced in PHP 4.3.0.
     * It returns the contents of a file as a string or bool(false)
     * on error.
     *
     * @since   2.8.7
     * @param   string   $filename
     * @param   boolean  $use_include_path
     * @return  bool(false)|string
     */
    function file_get_contents($filename, $use_include_path = false)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        assert('!isset($_types);');
        $_types = array(
        /* argument 1 */ IS_STRING,
        /* argument 2 */ IS_NULL|IS_BOOL
        );
        $_chk = checkArgumentList(func_get_args(), $_types, __FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk, $_types);
        }

        /* Input handling */
        if ($use_include_path) {
            $use_include_path = 1;
        } else {
            $use_include_path = 0;
        }
        $fileContents = file($filename, $use_include_path);
        if ($fileContents === false) {
            return false;
        } else {
            return implode('', $fileContents);
        }
    }
} /* end if */

if (!function_exists('file_put_contents')) {
    /**
     * file_put_contents() for PHP < 5.0
     *
     * The function file_put_contents() was introduced in PHP 5
     * It writes the contents of $data to a file and returns the
     * number of bytes written to the file or bool(false) on error.
     *
     * Example:
     * <code>
     * file_put_contents('test.dat', array(1,2,3,4), FILE_APPEND | LOCK_EX);
     * </code>
     *
     * @since   2.8.8
     * @param   string   $filename
     * @param   mixed    $data
     * @param   int      $flags
     * @param   resource $context
     * @param   boolean  $use_include_path
     * @return  bool(false)|int
     */
    function file_put_contents($filename, $data , $flags = null, $context = null)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        assert('!isset($_types);');
        $_types = array(
        /* argument 1 */ IS_STRING,
        /* argument 2 */ IS_SCALAR|IS_ARRAY,
        /* argument 3 */ IS_NULL|IS_INT,
        /* argument 4 */ IS_NULL|IS_RESOURCE
        );
        $_chk = checkArgumentList(func_get_args(), $_types, __FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk, $_types);
        }

        /* Input handling */
        if (is_array($data)) {
            $data = implode("", $data);
        }
        if (is_null($context)) {
            if ($flags & FILE_APPEND) {
                $handle = fopen($filename, "a");
            } else {
                $handle = fopen($filename, "w+");
            }
        } else {
            $handle = $context;
        }
        if ($handle === false) {
            return false;
        } else {
            if ($flags & LOCK_EX) {
                flock($handle, 2);
            }
            $result = fwrite($handle, $data);
            if ($flags & LOCK_EX) {
                flock($handle, 3);
            }
            fclose($handle);
        }
        if ($result === false) {
            return false;
        } else {
            return $result;
        }
    }
} /* end if */

if (!function_exists('stripos')) {
    /**
     * stripos() for PHP < 5.0
     *
     * The function stripos() was introduced in PHP 5
     * It returns the position of $needle in the string $haystack.
     *
     * It is case-insensitive.
     *
     * The name can be read as follows:
     * STRing_Ignore-case_POSition (strripos)
     *
     * @since   2.8.8
     * @param   string   $haystack
     * @param   string   $needle
     * @param   int      $offset
     * @return  bool(false)|int
     */
    function stripos($haystack, $needle, $offset = 0)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        assert('!isset($_types);');
        $_types = array(
        /* argument 1 */ IS_STRING,
        /* argument 2 */ IS_STRING,
        /* argument 3 */ IS_NULL|IS_INT
        );
        $_chk = checkArgumentList(func_get_args(), $_types, __FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk, $_types);
        }

        $haystack = strtolower($haystack);
        $needle   = strtolower($needle);
        if ($offset !== 0) {
            return  strpos($haystack, $needle, $offset);
        } else {
            return  strpos($haystack, $needle);
        }
    }
} /* end if */

if (!function_exists('strripos')) {
    /**
     * strripos() for PHP < 5.0
     *
     * The function strripos() was introduced in PHP 5
     * It returns the position of $needle in the string $haystack
     * starting from the end of the string.
     *
     * It is case-insensitive.
     *
     * The name can be read as follows:
     * STRing_Right_Ignore-case_POSition (strripos)
     *
     * @since   2.8.8
     * @param   string   $haystack
     * @param   string   $needle
     * @param   int      $offset
     * @return  bool(false)|int
     */
    function strripos($haystack, $needle, $offset = 0)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        assert('!isset($_types);');
        $_types = array(
        /* argument 1 */ IS_STRING,
        /* argument 2 */ IS_STRING,
        /* argument 3 */ IS_NULL|IS_INT
        );
        $_chk = checkArgumentList(func_get_args(), $_types, __FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk, $_types);
        }

        $haystack = strtolower($haystack);
        $needle   = strtolower($needle);
        if ($offset !== 0) {
            return  strpos($haystack, $needle, $offset);
        } else {
            return  strpos($haystack, $needle);
        }
    }
} /* end if */

if (!function_exists('substr_compare')) {
    /**
     * substr_compare() for PHP < 5.0
     *
     * The function substr_compare() was introduced in PHP 5
     * It does the same as strcmp(), but can be used on substrings.
     *
     * @since   2.8.8
     * @param   string   $main_str
     * @param   string   $str
     * @param   int      $offset
     * @param   int      $length
     * @param   bool     $case_insensitivity
     * @return  bool(false)|int
     */
    function substr_compare($main_str, $str, $offset, $length = 0, $case_insensitivity = false)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        assert('!isset($_types);');
        $_types = array(
        /* argument 1 */ IS_STRING,
        /* argument 2 */ IS_STRING,
        /* argument 3 */ IS_INT,
        /* argument 4 */ IS_NULL|IS_INT,
        /* argument 5 */ IS_NULL|IS_BOOL
        );
        $_chk = checkArgumentList(func_get_args(), $_types, __FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk, $_types);
        }

        $main_str = substr($main_str, $offset, $length);
        $str      = substr($str,      $offset, $length);

        if ($case_insensitivity === true) {
            return strcasecmp($main_str, $str);
        } else {
            return strcasecmp($main_str, $str);
        }

    }
} /* end if */

if (!function_exists('strpbrk')) {
    /**
     * strpbrk() for PHP < 5.0
     *
     * The function strpbrk() was introduced in PHP 5
     * It searches the string for a list of characters
     * and returns the substring starting at the first
     * match, or bool(false) if no match is found.
     *
     * @since   2.8.8
     * @param   string   $haystack
     * @param   string   $char_list
     * @return  bool(false)|string
     */
    function strpbrk($haystack, $char_list)
    {
        /**
         * Input checking
         */
        assert('!isset($_chk);');
        assert('!isset($_types);');
        $_types = array(
        /* argument 1 */ IS_STRING,
        /* argument 2 */ IS_STRING
        );
        $_chk = checkArgumentList(func_get_args(), $_types, __FUNCTION__);
        if ($_chk !== true) {
            trigger_error($_chk, E_USER_WARNING);
            return false;
        } else {
            unset($_chk, $_types);
        }

        if (empty($char_list) || empty($haystack)) {
            return false;
        }

        $test = preg_match('/['.preg_quote($char_list).'].*$/s', $haystack, $match);

        if (!empty($match[0])) {
            return $match[0];
        } else {
            return false;
        }

    }
} /* end if */

if (!function_exists('md5_file')) {
    /**
     * md5_file() for PHP < 4.2
     *
     * The function md5_file() was introduced in PHP 4.2
     * It returns the MD5 hash for a file.
     *
     * @since   2.8.8
     * @param   string   $filename
     * @param   bool     $raw_output
     * @return  bool(false)|string
     */
    function md5_file($filename, $raw_output = false)
    {
        assert('is_string($filename); /* Wrong argument type for argument 1. String expected */');
        assert('is_bool($raw_output); /* Wrong argument type for argument 2. Boolean expected */');
        /* settype to STRING
         *            BOOLEAN
         */
        $filename   = (string) $filename;
        $raw_output = (bool)   $raw_output;

        if (!is_file($filename)) {
            return false;
        } else {
            $fileContents = file_get_contents($filename);
            return md5($fileContents, $raw_output);
        }

    }
} /* end if */

if (!function_exists('scandir')) {
    /**
     * scandir() for PHP < 5
     *
     * The function scandir() was introduced in PHP 5
     * It returns a list of files and directories inside
     * a $directory.
     *
     * @since   2.8.8
     * @param   string    $directory      directory name
     * @param   int       $sorting_order  0 = ascending / 1 = descending
     * @param   resource  $context        currently not supported
     * @return  bool(false)|array
     * @see     dirlist()
     */
    function scandir($directory, $sorting_order = 0, $context = null)
    {
        if (!is_string($directory)) {
            return false;
        } else if (!is_dir($directory)) {
            trigger_error("The directory '${directory}' does not exist.", E_USER_WARNING);
            return false;
        }

        $contents = dirlist($directory);
        if (!is_array($contents)) {
            return false;
        } else {
            if ($sorting_order === 1) {
                rsort($contents);
            } else {
                sort($contents);
            }
            return $contents;
        }
    }
} /* end if */

if (!function_exists('array_change_key_case')) {
    /**
     * array_change_key_case() for PHP < 4.2
     *
     * The function array_change_key_case() was introduced in PHP 4.2
     * It returns a the input array with the keys in upper-case or
     * lower-case depending on argument $mode.
     * If $mode is not set it defaults to CASE_LOWER.
     *
     * For a recursive version see {@link Hashtable::changeCase()}.
     *
     * @since   2.8.8
     * @param   array    $array    some array
     * @param   int      $mode     CASE_LOWER = lower-case / CASE_UPPER = upper-case
     * @return  array
     * @see     Hashtable::changeCase()
     */
    function array_change_key_case($array, $mode = CASE_LOWER)
    {
        assert('is_array($array); /* Wrong argument type for argument 1. Array expected. */');
        assert('$mode === CASE_LOWER || $mode === CASE_UPPER; /* Invalid argument 2. Expected one of: CASE_LOWER, CASE_UPPER. */');

       if (is_array($array)) {
           foreach ($array as $i => $value)
           {
               unset($array[$i]);
               if (is_string($i)) {
                   if ($mode === CASE_UPPER) {
                       $i = strtoupper($i);
                   } else {
                       $i = strtolower($i);
                   }
               }
               $array[$i] = $value;
           }
       }
       return $array;
    }
}

if (!function_exists('array_combine')) {
    /**
     * array_combine() for PHP < 5
     *
     * The function array_combine() was introduced in PHP 5
     * It takes a list of keys and a list of values and combines both
     * to a new array.
     *
     * @since   2.8.8
     * @param   array    $keys      list of keys
     * @param   array    $values    list of values
     * @return  array
     */
    function array_combine($keys, $values)
    {
        assert('is_array($keys);                 /* Wrong argument type for argument 1. Array expected. */');
        assert('is_array($values);               /* Wrong argument type for argument 2. Array expected. */');
        assert('count($keys) === count($values); /* Key-array and value-array do not have equal size.   */');

        if (is_array($keys) && is_array($values) && !empty($keys) && count($keys) === count($values)) {
            $combinedArray = array();
            while (!empty($keys))
            {
                $combinedArray[array_shift($keys)] = array_shift($values);
            } /* end foreach */
            return $combinedArray;
        } else {
            return false;
        }
    }
}

if (!function_exists('array_product')) {
    /**
     * array_product() for PHP < 5.1
     *
     * The function array_product() was introduced in PHP 5.1.0RC1
     * It returns the product of the numbers inside the array.
     *
     * If a non-numeric value is encountered, it is converted to an integer
     * using the "intval()" function.
     *
     * Note: unlike the original PHP-function this implementation
     * will issue an E_USER_NOTICE if a non-numeric value is found.
     *
     * This function returns NULL on error.
     *
     * @since   2.8.8
     * @param   array    $array    list of numbers
     * @return  number
     */
    function array_product($array)
    {
        if (is_array($array)) {
            if (empty($array)) {
                return 0;
            }
            $product = 1;
            foreach ($array as $element)
            {
                if (is_numeric($element)) {
                    $product *= $element;
                } else {
                    $product *= intval($element);
                    trigger_error("The input-array contains a non-numeric value of type '".gettype($element)."'.\n\t\tThe result is undefined.", E_USER_NOTICE);
                }
            } /* end foreach */
            return $product;
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "1 in ".__FUNCTION__."()", 'Array', gettype($array)), E_USER_WARNING);
            return null;
        }
    }
}

if (!function_exists('property_exists')) {
    /**
     * array_product() for PHP < 5.1.0RC1
     *
     * The function property_exists() was introduced in PHP 5.1.0RC1
     * It returns true if the object or class $class has a property
     * named $property and false otherwise
     *
     * @since   2.8.8
     * @param   mixed    $class     class name or object
     * @param   string   $property  name of the property
     * @return  bool
     */
    function property_exists($class, $property)
    {
        if (!is_string($property)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "2 in ".__FUNCTION__."()", 'String', gettype($property)), E_USER_WARNING);
            return false;
        }

        if (is_object($class)) {
            $class = get_class($class);
        }

        if (is_string($class)) {
            $array = get_class_vars($class);
            if (!is_array($array)) {
                return false;
            } else {
                $array = array_keys($array);
            }
            if (in_array($property, $array)) {
                return true;
            } else {
                return false;
            }
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "1 in ".__FUNCTION__."()", 'String or object', gettype($class)), E_USER_WARNING);
            return false;
        }
    }
}

if (!function_exists('floatval')) {
    /**
     * floatval() for PHP < 4.2.0
     *
     * Alias of doubleval()
     *
     * @since   2.8.8
     * @param   mixed    $var
     * @return  float
     */
    function floatval($var)
    {
        return doubleval($var);
    }
}

$lcn_note = base64_decode("PHAgY2xhc3M9IktPTU1FTlRBUiIgc3R5bGU9InRleHQtYWxpZ246Y2VudGVyIj5yZWdpc3RlcmVkIFZJUC12ZXJzaW9uOiA8c3BhbiBjbGFzcz0iQkVTQ0hSSUZUVU5HIj4mcXVvdDslT1dORVIlJnF1b3Q7PC9zcGFuPjwvcD4=");

?>
