<?php
/**
 * @author     Thomas Meyer
 * @license    http://creativecommons.org/licenses/by-nc-sa/2.5/
 * @package    yana
 * @subpackage plugins
 */

/**
 * automated spam protection
 *
 * @access     public
 * @package    yana
 * @subpackage plugins
 *
 * @ignore
 */
class plugin_antispam extends Plugin
{

    /**
     * Default event handler
     *
     * returns bool(true) on success and bool(false) on error
     *
     * {@internal
     *
     * The corresponding form fields are created in DisplayUtility::htmlPostProcessor();
     *
     * }}
     *
     * @access  public
     * @return  bool
     * @param   string  $event  name of the called event in lower-case
     * @param   array   &$ARGS  array of arguments passed to the function
     */
    function _default($event, &$ARGS)
    {
        global $YANA;
        
        $eventType = strtolower($YANA->plugins->getEventType($event));

        if ($eventType === 'write') {

            /* check input data */
            assert('is_string($event);');
            assert('is_array($ARGS);');
            settype($event, "string");
            if (!is_array($ARGS)) {
                settype($ARGS, "array");
            }
            $event = strtolower($event);
    
            /* global variables */
            global $YANA;
            $permission = $YANA->getVar('PERMISSION');
            $type       = strtolower($YANA->plugins->getEventType($event));
            if (is_int($permission) && $permission > 0 || $type !== 'write') {
                /**
                 * {@internal
                 *
                 * Let's say: a robot would'nt have been able to log in,
                 * because it would have had to defeat the login form
                 * which is also protected by this implementation.
                 * In addition if it DID conquer the login form already,
                 * then it is quite obvious it would conquer them all.
                 *
                 * }}
                 */
            } else {
                /* 0 create header information for review by administrator
                 *
                 * Note: remote address is logged automatically - no need to add this here
                 */
                $header_data = array();
                if (isset($_SERVER['HTTP_REFERER'])) {
                    $header_data['REFERER'] = $_SERVER['HTTP_REFERER'];
                }
                if (isset($_SERVER['REQUEST_METHOD'])) {
                    $header_data['REQUEST_METHOD'] = $_SERVER['REQUEST_METHOD'];
                }
                if (isset($_SERVER['HTTP_USER_AGENT'])) {
                    $header_data['USER_AGENT'] = $_SERVER['HTTP_USER_AGENT'];
                }
                if (isset($_SERVER['REMOTE_PORT'])) {
                    $header_data['PORT'] = $_SERVER['REMOTE_PORT'];
                }
                /* 1 check the time the user required to fill out the form */
                $time =& $_SESSION['transaction_isolation_created'];
                /* if the time is not set or = 0, then we assume the form has never been displayed and we just encountered a direct request */
                if (empty($time)) {

                    $YANA->report(new Log('SPAM: blocked entry because no timestamp present (possibly the form was never displayed).',$header_data));
                    $YANA->report( new Alert('spamprotect'));
                    $YANA->exitTo('');
                    return false;

                /* assume a user would need at least 5 seconds and at most 30 minutes to fill the form */
                } else if (time() - $time < 5) {

                    $YANA->report(new Log('SPAM: blocked entry because a previous entry has been issued within the last 5 seconds.',$header_data));
                    $YANA->report( new Alert('spamprotect_time'));
                    $YANA->exitTo('');
                    return false;

                } else if (time() - $time > 1800) {

                    $YANA->report(new Log('SPAM: blocked entry because maximum time of life (30 minutes) for the form has been exceeded.',$header_data));
                    $YANA->report( new Alert('spamprotect_timeout'));
                    $YANA->exitTo('');
                    return false;

                /* if an invisible field has been filled, then we assume that the request was not send via a web browser */
                } else if (!empty($ARGS['yana_url'])) {

                    $YANA->report(new Log('SPAM: blocked entry because a field that is not visible to human visitors has been filled.',$header_data));
                    $YANA->report( new Alert('spamprotect'));
                    $YANA->exitTo('');
                    return false;

                /* The following was added in version 2.8.6
                 * This checks the header sent by the user for suspicious combinations of data.
                 */
                } else if (!defined('STDIN') && isset($_SERVER['REQUEST_METHOD']) && empty($_COOKIE) && !empty($_SERVER['HTTP_REFERER']) && empty($_SERVER['HTTP_USER_AGENT'])) {

                    $YANA->report(new Log("SPAM: blocked entry because of suspicious header. 'Referer' is set while 'user agent' is missing.",$header_data));
                    $YANA->report( new Alert('spamprotect'));
                    $YANA->exitTo('');
                    return false;

                }
            }
            /* check if input has a valid form id */
            if ($type === 'write' && $YANA->getVar('DISABLE_FORM_ID') !== true && (!is_int($permission) || $permission <= 0)) {
                if (!isset($_SESSION['yana_form_id']) || $_SESSION['yana_form_id'] === 'expired' || empty($ARGS['yana_form_id']) || strcasecmp($ARGS['yana_form_id'], $_SESSION['yana_form_id']) !== 0) {
                    $YANA->report(new Log('SPAM: blocked entry because no valid form Id has been found.'));
                    if ($_SESSION['yana_form_id'] === 'expired') {
                        $YANA->report( new Alert('nospam'));
                        $YANA->exitTo('');
                    } else {
                        $YANA->report( new Alert('ungueltige_eingabe'));
                        $YANA->exitTo('');
                    }
                    $_SESSION['yana_form_id'] = 'expired';
                    return false;
                } else {
                    $_SESSION['yana_form_id'] = 'expired';
                }
            } else {
                $_SESSION['yana_form_id'] = 'expired';
            }
    
            /* clean up the garbage */
            if (isset($ARGS['yana_form_id'])) {
                unset($ARGS['yana_form_id']);
            }
            return true;

        } else {

            $smarty = $YANA->view->document->getSmarty();
            $smarty->register_outputfilter(array(__CLASS__, '_outputFilter'));
            return true;

        }
    }
    
    /**
     * smarty outputfilter outputfilter
     *
     * @access  public
     * @static
     * @param   string  $source
     * @return  string
     */
    function _outputFilter($source)
    {
        global $YANA;
        /* Create form id */
        if ($YANA->getVar('DISABLE_FORM_ID') !== true) {
            $yana_form_id = md5(uniqid(rand(), true));
            if (strpos($source, "</form>") !== false) {
                /* insert form id */
                $source   = str_replace("</form>", "<span class=\"yana_button\"><input type=\"text\" name=\"yana_form_id\" value=\"$yana_form_id\"></span>\n</form>", $source);
                $_SESSION['yana_form_id'] = $yana_form_id;
                $YANA->setVar('DISABLE_FORM_ID', true);
            }
        }
        return $source;
    }

}

//  2007, Thomas Meyer www.All-Community.de
// Copyright Hinweis nicht entfernen!
?>