<?php
/**
 * @author     Thomas Meyer
 * @license    http://creativecommons.org/licenses/by-nc-sa/2.5/
 * @package    yana
 * @subpackage plugins
 */

/**
 * database administration tool
 *
 * @access     public
 * @package    yana
 * @subpackage plugins
 */
class plugin_db_admin extends Plugin
{

    /**
     * @access  private
     * @var     secureFileStream
     * @ignore
     */
    var $file;

    /**
     * Constructor
     *
     * @access  public
     * @param   string  $name a name for this plugin
     * @ignore
     */
    function plugin_db_admin($name)
    {
        settype($name,"string");
        $this->name = $name;
        global $YANA;
        $this->file = $YANA->plugins->get($this->name,"DBCONFIGFILE");
    }

    /**
     * Event handler
     *
     * parameters taken: none
     *
     * @access  public
     * @param   array  $ARGS  array of params passed to the function
     * @return  bool
     */
    function get_db_configuration($ARGS)
    {
        unset($ARGS);

        global $YANA;

        if (!class_exists("DB")) {
            $YANA->report( new error('PEARDB_FAILURE') );
            return false;
        }

        $YANA->setVar("DATABASE_ACTIVE",     YANA_DATABASE_ACTIVE);
        $YANA->setVar("DATABASE_DBMS",       YANA_DATABASE_DBMS);
        $YANA->setVar("DATABASE_HOST",       YANA_DATABASE_HOST);
        $YANA->setVar("DATABASE_PORT",       YANA_DATABASE_PORT);
        $YANA->setVar("DATABASE_USER",       YANA_DATABASE_USER);
        $YANA->setVar("DATABASE_PASSWORD",   YANA_DATABASE_PASSWORD);
        $YANA->setVar("DATABASE_PREFIX",     YANA_DATABASE_PREFIX);
        $YANA->setVar("DATABASE_NAME",       YANA_DATABASE_NAME);
        $YANA->setVar("YANA_DATABASE_ACTIVE",YANA_DATABASE_ACTIVE);
        $DATABASE_LIST = dirlist($YANA->getVar("DBDIR"), '*.config');
        foreach ($DATABASE_LIST as $key => $item)
        {
            $DATABASE_LIST[$key] = preg_replace("/\.\w*$/", "", $item);
        }
        $YANA->setVar("DATABASE_LIST", $DATABASE_LIST);
        return true;
    }

    /**
     * Event handler
     *
     * parameters taken:
     *
     * <ul>
     * <li> string database_dbms </li>
     * <li> array database_list  </li>
     * </ul>
     *
     * @access  public
     * @param   array  $ARGS array of params passed to the function
     * @return  bool
     */
    function db_install($ARGS)
    {
        settype($ARGS, "array");
        global $YANA;

        if (isset($ARGS['silent']) && $ARGS['silent'] === true) {
            $isSilent = true;
        } else {
            $isSilent = false;
        }

        if (!class_exists("DB")) {
            $isSilent || $YANA->report( new Error('PEARDB_FAILURE') );
            return false;
        }

        if (!isset($ARGS['database_dbms']) || !is_string($ARGS['database_dbms']) || !isset($ARGS['database_list']) || !is_array($ARGS['database_list']) || count($ARGS['database_list']) === 0) {
            /* invalid option - the choosen dbms is unknown */
            $isSilent || $YANA->report( new Alert('ungueltige_eingabe'));
            return false;
        } else {
            $dbms   = strtoupper($ARGS['database_dbms']);
            $dbList = $ARGS['database_list'];
        }

        /* Mapping the DBMS to the SQL export function in class DbCreator */
        switch ($dbms)
        {
            case 'DBASE':
            case 'FBSQL':
            case 'IBASE':
            case 'IFX':
            case 'SYBASE':
                $method_name  = null;
            break;
            case 'DB2':
                $method_name  = 'createDB2';
            break;
            case 'ACCESS':
                $method_name  = 'createMSAccess';
            break;
            case 'MSSQL':
                $method_name  = 'createMSSQL';
            break;
            case 'MYSQL': case 'MYSQLI':
                $dbms         = 'MYSQL';
                $method_name  = 'createMySQL';
            break;
            case 'OCI8':
                $method_name  = 'createOracleDB';
            break;
            case 'PGSQL':
                $method_name  = 'createPostgreSQL';
            break;
            default:
                $YANA->report( new Alert('ungueltige_eingabe'));
                return false;
            break;
        }

        /* Mapping the DBMS to it's installation directory */
        $installDirectory = $YANA->registry->get($dbms);
        if ($installDirectory === false) {
            /* invalid option - the choosen dbms is unknown */
            $YANA->report( new Log("Unable to install database. The choosen DBMS '${dbms}' is unknown."));
            $isSilent || $YANA->report( new Alert('ungueltige_eingabe'));
            return false;
        }

        /* we assume the class DbCreator has the desired method. This will be tested later! */

        if (!$installDirectory->exists() && is_null($method_name)) {
            $isSilent || $YANA->report( new Error("There is no installation file for this dbms available.") );
            return false;
        }

        /* get the list of available installation files */
        $installDirectory = $installDirectory->getPath();

        foreach ($dbList as $item)
        {
            /* check the input */
            if (!is_string($item)) {
                $isSilent || $YANA->report( new Alert('ungueltige_eingabe'));
                return false;
            } else {
                $item = strtolower($item);
            }

            $installFile   = $installDirectory.$item.'.sql';
            $structureFile = new StructureFile($item);
            $database      = new DbStream($structureFile);
            $database->read();

            /* If no SQL file for the current $item does exist,
             * we need to call the appropriate DbCreator method
             * instead.
             */
            if (!is_readable($installFile)) {
                $dbCreator = new DbCreator($structureFile);

                /* If the DbCreator class does not support the desired function.
                 */
                if (!method_exists($dbCreator, $method_name)) {

                    if (!file_exists($installFile)) {
                        $isSilent || $YANA->report( new Error("Unable to install database. There is no installation file available for database '$item'.") );
                    } else {
                        $isSilent || $YANA->report( new Error("Unable to install database '$item'. Cannot read sql file '$installFile'.") );
                    }
                    return false;

                /* Else, create and execute the SQL statements.
                 */
                } else {

                    /* create ... */
                    $sqlStmts = $dbCreator->$method_name();

                    /* ... execute */
                    if (!$database->importSQL($sqlStmts)) {

                        /* if execution fails */
                        $isSilent || $YANA->report(new Error("SQL installation fault. Unable to install database '$item'."));
                        return false;

                    } else {

                        /* else (execution finished successfully) */
                        $YANA->report(new Log("SQL import was successful.", $sqlStmts));

                    }
                }

            /* If a SQL file is available, always prefer using the SQL file.
             */
            } else if (!$database->importSQL($installFile)) {

                /* If import of SQL file has failed */
                $isSilent || $YANA->report(new Error("SQL installation fault. Unable to install database '$item'."));
                return false;

            /* If the SQL file has been imported successfully
             */
            } else {

                $YANA->report(new Log("SQL file '$installFile' has been imported."));

            }
        } /* end foreach */

        return true;
    }

    /**
     * Event handler
     *
     * parameters taken:
     *
     * <ul>
     * <li> string database_dbms </li>
     * <li> array database_list  </li>
     * </ul>
     *
     *
     * @todo   sort table definitions by foreign keys, so referenced tables are synchronized before the references
     *
     * @access  public
     * @param   array  $ARGS array of params passed to the function
     * @return  bool
     */
    function db_sync($ARGS)
    {
        settype($ARGS, "array");
        global $YANA;

        if (isset($ARGS['silent']) && $ARGS['silent'] === true) {
            $isSilent = true;
        } else {
            $isSilent = false;
        }

        if (!class_exists("DB")) {
            $isSilent || $YANA->report( new error('PEARDB_FAILURE') );
            return false;
        }

        if (!isset($ARGS['database_dbms']) || !isset($ARGS['database_list']) || !is_array($ARGS['database_list']) || count($ARGS['database_list']) === 0) {
            return false;
        }

        foreach ($ARGS['database_list'] as $item)
        {
            /* read database content */
            $database = new DbStream($item);
            $database->read();
            $file_database = new filedb($item);
            $file_database->read();
            $structure        = array();
            $database_content = array();
            $file_content     = array();

            foreach ($file_database->structure->getTables() as $table)
            {
                /* treat columns of types 'image' and 'file' as string */
                $files = $file_database->structure->getFiles($table);
                foreach ($files as $colName)
                {
                    $file_database->structure->setType($table, $colName, 'string');
                    $database->structure->setType($table, $colName, 'string');
                }

                /* get primary key */
                $primary_key = $file_database->structure->getPrimaryKey($table);

                if ($database->exists($table)) {
                    $database_keys = $database->get("$table.*.$primary_key");
                } else {
                    $database_keys = array();
                }
                if (!is_array($database_keys)) {
                    $database_keys = array();
                }
                if ($file_database->exists($table)) {
                    $file_keys     = $file_database->get("$table.*.$primary_key");
                } else {
                    $file_keys     = array();
                }
                if (!is_array($file_keys)) {
                    $file_keys     = array();
                }

                $difference = array_diff($database_keys, $file_keys);
                $i = 0;
                /* synchronize: database -> file */
                foreach ($difference as $key)
                {
                    if (!$file_database->insert("$table.$key", $database->get("$table.$key"))) {
                        $isSilent || trigger_error("Unable to copy value $table.$key from database to FileDB", E_USER_NOTICE);
                        return false;
                    } else if ($i > 20) {
                        if ($file_database->write()) {
                            $i = 0;
                        } else {
                            $isSilent || trigger_error("Failed to commit changes to FileDB.", E_USER_NOTICE);
                            return false;
                        }
                    } else {
                        $i++;
                    }
                }

                if ($i > 0 && !$file_database->write()) {
                    $isSilent || trigger_error("Failed to commit changes to FileDB.", E_USER_NOTICE);
                    return false;
                }

                $difference = array_diff($file_keys,$database_keys);
                $i = 0;
                /* synchronize: file -> database */
                foreach ($difference as $key)
                {
                    if (!$database->insert("$table.$key", $file_database->get("$table.$key"))) {
                        $isSilent || trigger_error("Unable to copy value $table.$key from FileDB to database", E_USER_NOTICE);
                        return false;
                    } else if ($i > 20) {
                        if ($database->write()) {
                            $i = 0;
                        } else {
                            $isSilent || trigger_error("Failed to commit changes to Database.", E_USER_NOTICE);
                            return false;
                        }
                    } else {
                        $i++;
                    }
                }
                if ($i > 0 && !$database->write()) {
                    $isSilent || trigger_error("Failed to commit changes to Database.", E_USER_NOTICE);
                    return false;
                }

            }

        } /* end foreach */

        return true;
    }

    /**
     * Event handler
     *
     * parameters taken:
     *
     * <ul>
     * <li> string database_active   </li>
     * <li> string database_dbms     </li>
     * <li> string database_host     </li>
     * <li> string database_port     </li>
     * <li> string database_password </li>
     * <li> string database_name     </li>
     * <li> string database_prefix   </li>
     * </ul>
     *
     * @access public
     * @param array $ARGS array of params passed to the function
     * @return boolean
     */
    function set_db_configuration($ARGS)
    {
        settype($ARGS, "array");
        global $YANA;

        /* update database settings */
        if (isset($ARGS['database_dbms'])) {

            /**
             * 1) Test if connection is available
             */
            if ($ARGS['database_active'] === 'true') {
                /* check input data */
                $dsn = array('DBMS'     => $ARGS['database_dbms'],
                             'HOST'     => stripslashes($ARGS['database_host']),
                             'PORT'     => $ARGS['database_port'],
                             'USERNAME' => stripslashes($ARGS['database_user']),
                             'PASSWORD' => stripslashes($ARGS['database_password']),
                             'DATABASE' => stripslashes($ARGS['database_name']));
                $test = dbServer::test($dsn);
                if ($test !== true) {
                    $YANA->report(new Error('Unable to establish connection to database server. Check your input please!'));
                    return false;
                } else {
                    /* all fine - proceed */
                }
            } else {
                /**
                 * A connection to FileDB should always be available,
                 * so there is no need to check that.
                 * Instead we assume this test is always "true".
                 */
            }

            /**
             * 2) Contact database server
             *
             * Install and synchronize the database with the current content of FileDb.
             *
             * This only needs to be done when a user activates / deactivates database support.
             * If the "database_active" setting is unchanged, then this may be skipped
             */
            if (isset($ARGS['database_active'])) {
                if (YANA_DATABASE_ACTIVE && $ARGS['database_active'] === 'true') {
                    /* continue without changes */
                } else if (!YANA_DATABASE_ACTIVE && $ARGS['database_active'] === 'false') {
                    /* continue without changes */
                } else {
                    $ARGS['silent'] = true;
                    if ($ARGS['database_active'] === 'true') {
                        $this->db_install($ARGS);
                    } else {
                        /* continue (FileDB does not need a database installation) */
                    }
                    if (!$this->db_sync($ARGS)) {
                        $YANA->report(new Error('500'));
                        $YANA->report(new Log('Unable to install tables of plugin "user" with the choosen dbms. Operation aborted.'));
                        return false;
                    } else {
                        /* success - proceed to next step */
                    } /* end if */
                } /* end if */
            } else {
                /* continue without changes */
            } /* end if */

            /**
             * 3) Write changes to file
             */
            $text = "<?php

if (!defined('YANA_DATABASE_ACTIVE'))   define('YANA_DATABASE_ACTIVE', ".    ( (isset($ARGS['database_active']))   ? $ARGS['database_active']   : YANA_DATABASE_ACTIVE )   .");
if (!defined('YANA_DATABASE_DBMS'))     define('YANA_DATABASE_DBMS', \"".    ( (isset($ARGS['database_dbms']))     ? $ARGS['database_dbms']     : YANA_DATABASE_DBMS )     ."\");
if (!defined('YANA_DATABASE_HOST'))     define('YANA_DATABASE_HOST', \"".    ( (isset($ARGS['database_host']))     ? $ARGS['database_host']     : YANA_DATABASE_HOST )     ."\");
if (!defined('YANA_DATABASE_PORT'))     define('YANA_DATABASE_PORT',\"".     ( (isset($ARGS['database_port']))     ? $ARGS['database_port']     : YANA_DATABASE_PORT )     ."\");
if (!defined('YANA_DATABASE_USER'))     define('YANA_DATABASE_USER',\"".     ( (isset($ARGS['database_user']))     ? $ARGS['database_user']     : YANA_DATABASE_USER )     ."\");
if (!defined('YANA_DATABASE_PASSWORD')) define('YANA_DATABASE_PASSWORD',\"". ( (isset($ARGS['database_password'])) ? $ARGS['database_password'] : YANA_DATABASE_PASSWORD ) ."\");
if (!defined('YANA_DATABASE_PREFIX'))   define('YANA_DATABASE_PREFIX',\"".   ( (isset($ARGS['database_prefix']))   ? $ARGS['database_prefix']   : YANA_DATABASE_PREFIX )   ."\");
if (!defined('YANA_DATABASE_NAME'))     define('YANA_DATABASE_NAME',\"".     ( (isset($ARGS['database_name']))     ? $ARGS['database_name']     : YANA_DATABASE_NAME )     ."\");

?>";
            if (!$this->file->exists()) {
                $this->file->create();
            }
            $this->file->read();
            $this->file->remove();
            $this->file->insert($text);
            $this->file->failSafeWrite();

        } else {
            $YANA->report(new Error('UNGUELTIGE_EINGABE'));
            return false;
        } /* end if */

        return true;
    }

}

//  2007, Thomas Meyer www.All-Community.de
// Copyright Hinweis nicht entfernen!
?>