<?php
/**
 * default library
 *
 * @author     Thomas Meyer
 * @license    http://creativecommons.org/licenses/by-nc-sa/2.5/
 * @package    yana
 * @subpackage plugins
 */

/**
 * entity RSS feed
 *
 * This class represents a RSS feed.
 * It offers functionality to publish new feeds.
 *
 * Examples of usage:
 * <ol>
 *  <li> Creating and printing a RSS feed inside a plugin:
 *       <code>
 *  function rss_foobar($args)
 *  {
 *       // create rss feed
 *       $rss = new RSS();
 *       // provide some description for it
 *       $rss->title = 'foo';
 *       $rss->description = 'this is my RSS feed';
 *       // create 1st item
 *       $item = new RSSitem();
 *       $item->title = '1st entry';
 *       $item->link = 'http://some.url';
 *       $item->description = 'some text';
 *       $rss->addItem($item);
 *       // create 2nd item
 *       $item = new RSSitem();
 *       $item->title = '2nd entry';
 *       $item->link = 'http://some.other.url';
 *       $item->description = 'some text';
 *       $rss->addItem($item);
 *       // print rss feed
 *       print utf8_encode($rss->toString());
 *       // when you're done, terminate the program
 *       exit(0);
 *  }
 *  </code>
 *  </li>
 *  <li> Showing the RSS logo + link from inside a plugin:
 *       <code>$YANA->setVar('RSS_ACTION', 'rss_foobar')</code>
 *       Where 'rss_foobar' is the value for the "action" parameter of the
 *       URL. You don't need to call it 'rss_foobar' - just call it whatever
 *       you like.
 *       You are to implement the code for action 'rss_foobar' yourself.
 *       To find out how to do that, see the example above.
 *  </li>
 * </ol>
 *
 * @access     public
 * @package    yana
 * @subpackage plugins
 */
class RSS extends Object
{
    /**#@+
     * this field is mandatory
     *
     * @access  public
     */

    /**
     * name of the channel, e.g. title of the website the channel refers to
     * @var string */ var $title       = "";
    /**
     * a link that points to the originating website of the feed
     * @var string */ var $link        = "";
    /**
     * a text to describe context and purpose of this channel
     * @var string */ var $description = "";

    /**#@-*/
    /**#@+
     * this field is optional
     *
     * @access  public
     */

    /**
     * language the channel is written in
     * @var string */ var $language       = "";
    /**
     * copyright notice of your choice
     * @var string */ var $copyright      = "";
    /**
     * mail of person responsible for editorial content
     * @var string */ var $managingEditor = "";
    /** 
     * mail of person responsible for technical issues
     * @var string */ var $webMaster      = "";
    /**
     * "time to live" - a number of minutes that indicates how long a channel can be cached before refreshing from the source
     * @var int    */ var $ttl            = 0;
    /**
     * *.jpg, *.gif, or *.png image to display with the feed
     * @var string */ var $image          = "";
    /**
     * one or more categories that the channel belongs to
     * @var array  */ var $category       = array();
    /**
     * a CSS stylesheet that formats the channel
     * @var string */ var $css            = "";
    /**
     * a XSL transformation that formats the channel
     * @var string */ var $xslt           = "";
    /**
     * list of one or more items of the channel
     * @var array  */ var $item          = array();

    /**#@-*/
    /**#@+
     * @ignore
     * @access  private
     */

    /**
     * date when the channel was published (auto-generated)
     * @var string */ var $pubDate       = "";
    /**
     * pubDate of most recent item (auto-generated)
     * @var string */ var $lastBuildDate = "";
    /**
     * a string to identify the class / program that created the file
     * @var string */ var $generator     = "";

    /**#@-*/
    
    /**
     * constructor
     *
     * creates and initializes a new instance of this class
     *
     */
    function RSS()
    {
        global $YANA;
        $this->title     = '%PROGRAM.TITLE%';
        $this->generator = "Yana Framework ".YANA_VERSION;
        $this->item      = array();
        if (isset($_SESSION['language'])) {
            $this->language = $_SESSION['language'];
        } else if (isset($YANA)) {
            $this->language = str_replace('/','',$YANA->getVar("LANGUAGE_INFO.DIRECTORY"));
        }
        if (isset($YANA)) {
            $this->link = $YANA->getVar("SETTINGS.WEBSITE_URL");
        }
    }

    /**
     * get string value
     *
     * This returns this object as a string in RSS 2.0 syntax.
     *
     * Returns bool(false) and issues an E_USER_NOTICE on error.
     *
     * @access public
     * @return bool(false)|string
     *
     * @name   RSS::toString()
     */
    function toString()
    {
        global $YANA;

        /* check mandatory fields */
        if (!is_string($this->title) || empty($this->title)) {
            trigger_error("Unable to create RSS. Mandatory field 'title' missing.", E_USER_NOTICE);
            return false;
        }
        if (!is_string($this->link) || empty($this->link)) {
            trigger_error("Unable to create RSS. Mandatory field 'link' missing.", E_USER_NOTICE);
            return false;
        }
        if (!is_string($this->description) || empty($this->description)) {
            trigger_error("Unable to create RSS. Mandatory field 'description' missing.", E_USER_NOTICE);
            return false;
        }

        /* check optional fields */
        if (!is_string($this->language) || !preg_match('/^(?:|\w{2}(?:-\w{2})?)$/s',$this->language)) {
            trigger_error("Unable to create RSS. Invalid 'language' value.\n\t\tUse following syntax: 'xx-xx', e.g. 'en', 'en-us', 'de', 'de-de', 'de-at' aso.", E_USER_NOTICE);
            return false;
        }
        if (!is_string($this->copyright)) {
            trigger_error("Unable to create RSS. Invalid 'copyright' value.", E_USER_NOTICE);
            return false;
        }
        if (!is_string($this->image)) {
            trigger_error("Unable to create RSS. Invalid 'image' value.", E_USER_NOTICE);
            return false;
        }
        if (!is_int($this->ttl)) {
            trigger_error("Unable to create RSS. Invalid 'ttl' value.", E_USER_NOTICE);
            return false;
        }
        if (!is_string($this->managingEditor)) {
            trigger_error("Unable to create RSS item. Field 'managingEditor' is invalid.", E_USER_NOTICE);
            return array();
        } else {
            $tmp = $this->managingEditor;
            $this->managingEditor = untaintInput($this->managingEditor, 'mail');
            if ($tmp != $this->managingEditor) {
                trigger_error("RSS item has been auto-converted.\n\t\tThe input value '".$tmp."' did not seem to be valid.\n\t\tThe value has been changed to '".$this->managingEditor."'.", E_USER_NOTICE);
            }
        }
        if (!is_string($this->webMaster)) {
            trigger_error("Unable to create RSS item. Field 'webMaster' is invalid.", E_USER_NOTICE);
            return array();
        } else {
            $tmp = $this->webMaster;
            $this->webMaster = untaintInput($this->webMaster, 'mail');
            if ($tmp != $this->webMaster) {
                trigger_error("RSS item has been auto-converted.\n\t\tThe input value '".$tmp."' did not seem to be valid.\n\t\tThe value has been changed to '".$this->webMaster."'.", E_USER_NOTICE);
            }
        }
        if (!is_string($this->css)) {
            trigger_error("Unable to create RSS. Invalid 'css' value.", E_USER_NOTICE);
            return false;
        }
        if (!is_string($this->xslt)) {
            trigger_error("Unable to create RSS. Invalid 'xslt' value.", E_USER_NOTICE);
            return false;
        }
        if (!is_array($this->item)) {
            trigger_error("Unable to create RSS. Invalid 'item' value.", E_USER_NOTICE);
            return false;
        }

        /* auto-fill fields, that can be generated */
        $this->pubDate = date('r',time());
        if (count($this->item) > 0) {
            $time = 0;
            foreach ($this->item as $item)
            {
                if (!empty($item['pubDate'])) {
                    $tmp = strtotime($item['pubDate']);
                    if (is_int($tmp) && $tmp > $time) {
                        $time = $tmp;
                    }
                }
            } /* end foreach */
            if ($time > 0) {
                $this->lastBuildDate = date('r',$time);
            } else {
                $this->lastBuildDate = "";
            }
        }

        /* publish data */
        $src = $YANA->skin->getFile('RSS_TEMPLATE');
        assert('is_file($src);');
        $tpl = new SmartTemplate($src);
        $tpl->insert('rss',get_object_vars($this));
        return $tpl->get();
    }

    /**
     * add RSS feed item to this channel
     *
     * The $item has to be a valid RSSitem object.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * Note: this function does not check for duplicate guid's
     *
     * @access public
     * @param  RSSitem  $item
     * @return bool
     *
     * @name   RSS::toString()
     */
    function addItem($item)
    {
        if (!is_object($item) || (is_subclass_of($item,'RSSitem') === false && strcasecmp(get_class($item),'RSSitem') !== 0)) {
            trigger_error("Wrong argument type for argument 1. Instance of RSSitem expected, found '".get_class($item)."' instead.", E_USER_WARNING);
            return false;
        } else {
            if (!is_array($this->item)) {
                $this->item = array();
            }
            $item = $item->get();
            if (!is_array($item) || count($item) === 0) {
                return false;
            } else {
                array_push($this->item, $item);
                assert('is_array($this->item);');
                return true;
            }
        }
    }

}

/**
 * entity RSS item
 *
 * This class represents an item of an RSS feed.
 *
 * Note: all attributes are optional, but at least one of title or description must be set.
 *
 * @access     public
 * @package    yana
 * @subpackage plugins
 */
class RSSitem extends Object
{
    /**#@+
     * this field is optional
     *
     * @access  public
     */

    /**
     * title of the item
     * @var string */ var $title       = "";
    /**
     * URL of the item
     * @var string */ var $link        = "";
    /**
     * item synopsis, note that this may contain HTML
     * @var string */ var $description = "";
    /**
     * Email address of the author
     * @var string */ var $author      = "";
    /**
     * include item in one or more categories
     * @var array  */ var $category    = array();
    /**
     * URL of a page for comments
     * @var string */ var $comments    = "";
    /**
     * indicates when the item was published (auto-generated if not provided)
     * @var string */ var $pubDate     = "";

    /**#@-*/
    /**#@+
     * this field is optional
     *
     * @access  private
     */

    /**
     * unique identity (always auto-generated)
     * @var string */ var $guid        = "";

    /**#@-*/

    /**
     * constructor
     *
     * creates and initializes a new instance of this class
     *
     */
    function RSSitem()
    {
        $this->pubDate = date('r',time());
    }
    
    /**
     * return item as associative array
     *
     * This checks the values of all attributes, creates
     * a unique id for this item, and returns
     * an associative array of all object vars.
     *
     * Returns an empty array and issues an E_USER_NOTICE on error.
     *
     * @access  public
     * @return  array
     */
    function get()
    {
        /* check optional fields */
        if (!is_string($this->title)) {
            trigger_error("Unable to create RSS item. Field 'title' is invalid.", E_USER_NOTICE);
            return array();
        }
        if (!is_string($this->link)) {
            trigger_error("Unable to create RSS item. Field 'link' is invalid.", E_USER_NOTICE);
            return array();
        }
        if (!is_string($this->description)) {
            trigger_error("Unable to create RSS item. Field 'description' is invalid.", E_USER_NOTICE);
            return array();
        }
        if (!is_string($this->author)) {
            trigger_error("Unable to create RSS item. Field 'author' is invalid.", E_USER_NOTICE);
            return array();
        } else {
            $tmp = $this->author;
            $this->author = untaintInput($this->author, 'mail');
            if ($tmp != $this->author) {
                trigger_error("RSS item has been auto-converted.\n\t\tThe input value '".$tmp."' did not seem to be valid.\n\t\tThe value has been changed to '".$this->author."'.", E_USER_NOTICE);
            }
        }
        if (!is_array($this->category)) {
            trigger_error("Unable to create RSS item. Field 'category' is invalid.", E_USER_NOTICE);
            return array();
        }
        if (!is_string($this->comments)) {
            trigger_error("Unable to create RSS item. Field 'comments' is invalid.", E_USER_NOTICE);
            return array();
        } else {
            $tmp = $this->comments;
            $this->comments = untaintInput($this->comments, 'url');
            if ($tmp != $this->comments) {
                trigger_error("RSS item has been auto-converted.\n\t\tThe input value '".$tmp."' did not seem to be valid.\n\t\tThe value has been changed to '".$this->comments."'.", E_USER_NOTICE);
            }
        }
        if (!is_string($this->guid)) {
            trigger_error("Unable to create RSS item. Field 'guid' is invalid.", E_USER_NOTICE);
            return array();
        }
        if (!is_string($this->pubDate)) {
            trigger_error("Unable to create RSS item. Field 'pubDate' is invalid.", E_USER_NOTICE);
            return array();
        }
        if (empty($this->title) && empty($this->description)) {
            trigger_error("Unable to create RSS item. Need at least 'title' or 'description' to be present.", E_USER_NOTICE);
            return array();
        }

        /* auto-fill fields, that can be generated */
        $this->guid = md5($this->title.$this->description);

        /* return result */
        return get_object_vars($this);
    }

}

//  2006, Thomas Meyer www.All-Community.de
// Copyright Hinweis nicht entfernen!
?>